/* Get all interface multicast addresses via sysctl */

#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>
#include <errno.h>
#include <signal.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/sysctl.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <net/route.h>

#include <netinet/in.h>

#include <netdb.h>
#include <arpa/inet.h>

#include <string.h>
#include <stdlib.h>
#include <syslog.h>

#ifdef if_type
#define ifma_msghdr	ifa_msghdr
#define ifmam_msglen	ifam_msglen
#define ifmam_addrs	ifam_addrs
#define ifmam_index	ifam_index
#endif

int mib[] = { CTL_NET, PF_ROUTE, 0, /*af*/0, NET_RT_IFMADDR, /*index*/0 };
int size;
caddr_t buf;
struct ifma_msghdr *ifmam;
struct sockaddr *sa, *ifp, *ifa, *gate;
struct timeval tv;

main() {
	caddr_t cp;

	if (sysctl(mib, 6, NULL, &size, NULL, 0) < 0) {
		perror("sysctl0");
		exit(1);
	}
#ifdef DEBUG
	printf("size = %d\n", size);
#endif
	if ((buf = malloc(size)) == NULL) {
		fprintf(stderr, "malloc failed\n");
		exit(1);
	}
	if (sysctl(mib, 6, buf, &size, NULL, 0) < 0) {
		perror("sysctl");
		exit(1);
	}
	if (gettimeofday(&tv, NULL) < 0) {
		perror("gettimeofday");
		exit(1);
	}

	for (cp = buf; cp < buf + size; cp += ifmam->ifmam_msglen) {
		ifmam = (struct ifma_msghdr *)cp;
		ifmaddr();
	}
}

#define ROUNDUP(a) \
        ((a) > 0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))
#define ADVANCE(x, n) (x += ROUNDUP((n)->sa_len))

ifmaddr() {
	caddr_t cp;
	int len, i;

	cp = (caddr_t)(ifmam + 1);
	len = ifmam->ifmam_msglen;
	ifp = ifa = gate = NULL;

	for (i = 0; i < RTAX_MAX; i++)
		if (ifmam->ifmam_addrs & (1 << i)) {
			sa = (struct sockaddr *)cp;
			switch (i) {
			    case RTAX_GATEWAY:
				gate = sa;
				break;
			    case RTAX_IFP:
				ifp = sa;
				break;
			    case RTAX_IFA:
				ifa = sa;
				break;
			}
			ADVANCE(cp, sa);
		}
	if (ifp)
		print_ifp();
	else {
		fprintf(stderr, "no name for %d\n", ifmam->ifmam_index);
		return;
	}
	if (ifa)
		print_ifa();
	else {
		fprintf(stderr, "address?\n");
		return;
	}
	if (gate)
		print_gate();
	else
		printf("\n");

	if (cp != (caddr_t)ifmam + len)
		printf("delta = %d\n", cp - ((caddr_t)ifmam + len));
}

print_ifp() {
	struct sockaddr_dl *sdl;
	int index;

	sdl = (struct sockaddr_dl *)ifp;
	index = (int)ifmam->ifmam_index;
	printf("%*s(#%d)", sdl->sdl_nlen, sdl->sdl_data, index);
}

print_gate() {
	struct sockaddr_dl *sdl;
	u_char *cp;
	u_char lo[6], hi[6];
	int i;

	sdl = (struct sockaddr_dl *)gate;
	cp = (u_char *)(sdl->sdl_data + sdl->sdl_nlen);
	switch (sdl->sdl_alen) {
	    case 0:
#ifndef if_type
		printf(" []");
#endif
		printf("\n");
		return;

	    case 6:
	    ieee:
		bcopy(cp, lo, 6);
#ifndef if_type
		printf(" [");
#endif
		printf("%x:%x:%x:%x:%x:%x",
		       lo[0], lo[1], lo[2], lo[3], lo[4], lo[5]);
#ifndef if_type
		printf("]");
#endif
		printf("\n");
		return;

	    case 12:
		bcopy(cp, lo, 6);
		bcopy(cp + 6, hi, 6);
		if (bcmp(lo, hi, 6) == 0)
			goto ieee;
#ifndef if_type
		printf(" [");
#endif
		printf("%x:%x:%x:%x:%x:%x - %x:%x:%x:%x:%x:%x",
		       lo[0], lo[1], lo[2], lo[3], lo[4], lo[5],
		       hi[0], hi[1], hi[2], hi[3], hi[4], hi[5]);
#ifndef if_type
		printf("]");
#endif
		printf("\n");
		return;

	    default:
#ifndef if_type
		printf(" [");
#endif
		for (i = sdl->sdl_alen - 1; i >= 0; i--)
			printf("%x%s", *cp++, i ? ":" : "");
#ifndef if_type
		printf("]");
#endif
		printf("\n");
	}
}

print_ifa() {
	u_char *cp;
	char buf[64];
	int i;

	switch (ifa->sa_family) {
	    case AF_INET:
#define SIN(x)	((struct sockaddr_in *)x)
		printf(" IP:\t%s", inet_ntoa(SIN(ifa)->sin_addr));
		return;

	    case AF_INET6:
#define SIN6(x)	((struct sockaddr_in6 *)x)
		printf(" IPv6:\t%s",
		       inet_ntop(AF_INET6, &SIN6(ifa)->sin6_addr,
				 buf, sizeof(buf)));
		return;

	    case AF_LINK:
#define SDL(x)	((struct sockaddr_dl *)x)
		printf(" MAC:\t");
#ifdef if_type
		gate = ifa;
#else
		if (SDL(ifa)->sdl_alen == 0) {
			printf("-");
			return;
		}
		cp = (u_char *)(SDL(ifa)->sdl_data + SDL(ifa)->sdl_nlen);
		for (i = SDL(ifa)->sdl_alen - 1; i >= 0; i--)
			printf("%x%s", *cp++, i ? ":" : "");
#endif
		return;

	    default:
		printf(" %d?:\t", ifa->sa_family);
		cp = (u_char *)ifa->sa_data;
		for (i = ifa->sa_len; i > 0; i--)
			printf("%02x ", *cp++);
		printf("\n");
	}
}
