/* SAT test program */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <net/if_atm.h>
#include <netinet/in.h>
#include <netnatm/natm.h>
#include <stdio.h>
#include <time.h>
#include <errno.h>

char *ifname = "en0";
char *satname = "sat0";
int vpi = 0;
int vci = 0;
int tx = 0;
int join = 0;
int settx = 0;

main(argc, argv)
	int argc;
	char *argv[];
{
	struct sockaddr_natm snatm;
	struct ifreq ifr;
	struct vctxreq vtr;
	u_char buf[1024*16];
	struct timeval tv;
	struct tm *t;
	time_t clk;
	int s, i, cc, ch;
	extern char *optarg;
	extern int optind;

	while ((ch = getopt(argc, argv, "i:s:p:v:t:j:")) != -1)
		switch(ch) {
		    case 'i':
			ifname = optarg;
			break;
		    case 's':
			satname = optarg;
			break;
		    case 'p':
			vpi = atoi(optarg);
			break;
		    case 'v':
			vci = atoi(optarg);
			break;
		    case 't':
			tx = atoi(optarg);
			settx++;
			break;
		    case 'j':
			join = atoi(optarg);
			settx++;
			break;
		    default:
			fprintf(stderr,
	"usage: sat -i if -s sat -p vpi -v vci -t txspeed -j join_vci\n");
			exit(-1);
		}
	argc -= optind;
	if (argc) {
		fprintf(stderr, "%d extra arguments\n", argc);
		exit(-1);
	}
	if (vci == 0) {
		fprintf(stderr, "arguments are mandatory\n");
		exit(-1);
	}

	if ((s = socket(AF_NATM, SOCK_STREAM, PROTO_NATMAAL5)) < 0) {
		perror("socket");
		exit(1);
	}

	bzero(&snatm, sizeof(snatm));
	snatm.snatm_len = sizeof(snatm);
	snatm.snatm_family = AF_NATM;
	snprintf(snatm.snatm_if, IFNAMSIZ, "%s", ifname);
	snatm.snatm_vpi = vpi;
	snatm.snatm_vci = vci;

	if (connect(s, (struct sockaddr *)&snatm, sizeof(snatm)) < 0) {
		perror("connect");
		exit(1);
	}

	bzero(&ifr, sizeof(ifr));
	snprintf(ifr.ifr_name, IFNAMSIZ, "%s", satname);

	if (ioctl(s, SIOCATMASIF, (caddr_t)&ifr) < 0) {
		perror("ioctl");
		exit(1);
	}

	bzero(&vtr, sizeof(vtr));
	snprintf(vtr.vc_ifname, IFNAMSIZ, "%s", satname);
	vtr.vc_jvci = join;
	vtr.vc_pcr = tx;

	if (settx &&
	    ioctl(s, SIOCTXSPEED, (caddr_t)&vtr) < 0) {
		perror("set tx speed");
		exit(1);
	}

	snatm.snatm_if[IFNAMSIZ - 1] = '\0';
	printf("Waiting for frames on %s from %d/%d\n",
	       snatm.snatm_if, snatm.snatm_vpi, snatm.snatm_vci);

	for (;;) {
		cc = read(s, (char *)buf, sizeof(buf));
		if (cc < 0) {
			perror("read");
			exit(1);
		}
		if (cc == 0) {
			fprintf(stderr, "closed ?\n");
			exit(1);
		}
		(void) gettimeofday(&tv, NULL);
		clk = (time_t)tv.tv_sec;
		t = localtime(&clk);
		printf("%02d:%02d:%02d.%06u ",
		       t->tm_hour,
		       t->tm_min,
		       t->tm_sec,
		       (u_int)tv.tv_usec);
		printf("%d bytes:", cc);
		for (i = 0; i < cc; i++) {
			if ((i & 15) == 0)
				printf("\n\t");
			else if ((i & 1) == 0)
				printf(" ");
			printf("%02x", buf[i]);
		}
		printf("\n");
	}
}
