/*
 * DAD support function
 *	given a struct if_laddrreq argument, do a SIOCALIFADDR with DAD
 * also manage LL multicast setting
 */
/* ND part */

#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <net/if.h>
#include <net/if_dl.h>
#ifndef if_type
#include <net/if_var.h>
#endif
#include <net/route.h>
#include <netinet/in.h>
#include <netinet/ip_var.h>
#include <netinet/in6_var.h>
#include <netinet/ip6.h>
#include <netinet/ip6_var.h>
#include <netinet/icmp6.h>

#include <err.h>
#include <errno.h>
#include <poll.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define MAX_RTR_SOLICITATION_DELAY	1	/* seconds */
#define RETRANS_TIMER			1	/* seconds */
#define DupAddrDetectTransmits		2
#define MAXPACKET	576	/* max packet size */

int nd_waitbefore_dad = 0;
int dad_to_send = DupAddrDetectTransmits;

static int ntosend = 0;		/* packet to send counter */
static int ninflow = 0;		/* packet on the cable counter */


static int recv_nd __P((int, int, int, int, struct in6_addr *, int *));
static int in_cksum __P((u_short *, int));

int	dad_ipv6 __P((int, u_int, struct if_laddrreq *));
int	dad_add_multi __P((int, int));

#ifdef DEBUG
#include <arpa/inet.h>
char pbuf[64];
#define PRINT6(x)	inet_ntop(AF_INET6, (x), pbuf, sizeof (pbuf))
#endif

/*
 * Try to set a address, performing dad test
 * Input: u: AF_INET6 socket; ifp: interface index #; req: if_laddrreq
 * Output: ret<0 : err (errno); ret>=0: done
 */
int
dad_ipv6(u, ifp, req)
	int u;
	u_int ifp;
	struct if_laddrreq *req;
{
	int flags, val, si = -1, cc, ret, delay, try = 0;
#define NSLEN	(sizeof (struct ip6_hdr) + sizeof(struct nd_neighbor_solicit))
	u_char msgndsol[NSLEN];
	struct ip6_hdr *ip = (struct ip6_hdr *)msgndsol;
	struct nd_neighbor_solicit *nsp = (struct nd_neighbor_solicit *)(ip+1);
	int lgsol = NSLEN;
	struct sockaddr_in6 dst;
	struct ipv6_mreq mreq;
	struct icmp6_filter filter;
	struct timeval tv;
	struct in6_addr *addr;
	struct if_laddrreq reqget;

	flags = req->flags;

	/* get interface ID part */
	if (req->flags & (IFLR_PREFIX|IFLR_NOSET)) {
		/* ask the kernel to complete the if_laddrreq */
		req->flags |= IFLR_NOSET;
		if ((ret = ioctl(u, SIOCALIFADDR, req)) < 0)
			warn("dad ioctl (fill PREFIX)");
		if ((ret < 0) || (flags & IFLR_NOSET))
			goto errdone;
		req->flags = flags & ~IFLR_PREFIX;
	}
	addr = &((struct sockaddr_in6 *)(&req->addr))->sin6_addr;

	reqget = *req;
	reqget.flags |= IFLR_PREFIX;
	reqget.prefixlen = 128;
	if ((ret = ioctl(u, SIOCGLIFADDR, &reqget)) == 0)
		goto modify;	/* already exists: no dad */
	if (errno != EADDRNOTAVAIL) {
		warn("dad ioctl(SIOCGLIFADDR)");
		goto errdone;
	}

	/* Open ICMPv6 "si" socket */
	if ((si = socket(PF_INET6, SOCK_RAW, IPPROTO_ICMPV6)) < 0) {
		warn("dad socket");
		goto errdone;
	}

#define sockopt(opt, x, msg) \
	if (setsockopt(si, IPPROTO_IPV6, (opt), &(x), sizeof(x)) < 0) { \
		warn("dad setsockopt(" msg ")"); \
		goto errdone; }

	val = 1;
	sockopt(IP_HDRINCL, val, "IP_HDRINCL");
	sockopt(IPV6_MULTICAST_LOOP, val, "IPV6_MULTICAST_LOOP");
	sockopt(IPV6_RECVPKTINFO, val, "IPV6_RECVPKTINFO");

	/* needed, even with IP_HDRINCL ! */
	val = ND_DEFAULT_HOPLIM;
	sockopt(IPV6_MULTICAST_HOPS, val, "IPV6_MULTICAST_HOPS");

	val = ifp;
	sockopt(IPV6_MULTICAST_IF, val, "IPV6_MULTICAST_IF");

	/* only interested in ND_NEIGHBOR_SOLICIT and ND_NEIGHBOR_ADVERT */
	ICMP6_FILTER_SETBLOCKALL(&filter);
	ICMP6_FILTER_SETPASS(ND_NEIGHBOR_SOLICIT, &filter);
	ICMP6_FILTER_SETPASS(ND_NEIGHBOR_ADVERT, &filter);
	sockopt(ICMP6_FILTER, filter, "ICMP6_FILTER");

	memset(&mreq, 0, sizeof (mreq));
	mreq.ipv6mr_interface = ifp;
	mreq.ipv6mr_multiaddr.s6_addr[0] = 0xff;
	mreq.ipv6mr_multiaddr.s6_addr[1] = 2;
	mreq.ipv6mr_multiaddr.s6_addr[15] = 1;
	sockopt(IPV6_JOIN_GROUP, mreq, "IPV6_JOIN_GROUP all nodes");

	/* solicited-node multicast address */
	mreq.ipv6mr_multiaddr.s6_addr[11] = 1;
	mreq.ipv6mr_multiaddr.s6_addr[12] = 0xff;
#define LENSOL 3
#define ADSOL	(sizeof(struct in6_addr) - LENSOL)
	memcpy(&mreq.ipv6mr_multiaddr.s6_addr[ADSOL],
	       &addr->s6_addr[ADSOL],
	       LENSOL);
	sockopt(IPV6_JOIN_GROUP, mreq, "IPV6_JOIN_GROUP sollicited");

	/* ICMPv6 Neighbor Solicitation packet */
	memset(msgndsol, 0, lgsol);
	nsp->nd_ns_type = ND_NEIGHBOR_SOLICIT;
	COPY_ADDR6(*addr, nsp->nd_ns_target);

	COPY_ADDR6(mreq.ipv6mr_multiaddr, ip->ip6_dst);

	((struct ip6ovck *)ip)->ih6_pr = IPPROTO_ICMPV6;
	((struct ip6ovck *)ip)->ih6_len =
		htons(lgsol - sizeof(struct ip6_hdr));

	/* other fields are 0, including ip6_src */

	nsp->nd_ns_cksum = in_cksum((u_short *)ip, lgsol);

	/* finish IPv6 header */
	ip->ip6_flow = IPV6_VERSION | ND_DEFAULT_CLASS;
	ip->ip6_plen = htons(lgsol - sizeof(struct ip6_hdr));
	ip->ip6_nxt = IPPROTO_ICMPV6;
	ip->ip6_hlim = ND_DEFAULT_HOPLIM;

	memset(&dst, 0, sizeof dst);
	dst.sin6_family = AF_INET6;
#ifdef SIN6_LEN
	dst.sin6_len = sizeof dst;
#endif
	COPY_ADDR6(ip->ip6_dst, dst.sin6_addr);

	/*
	 * Perform duplicate address detection.
	 */
	ntosend = dad_to_send;
	ninflow = 0;
	if (nd_waitbefore_dad == 0)
		goto start;

	nd_waitbefore_dad = 0;	/* always reset */
	gettimeofday(&tv, NULL);
	srandom((int)tv.tv_sec+tv.tv_usec);
	delay = MAX_RTR_SOLICITATION_DELAY * 1000;
	delay -= random() % 1000;

	for (;;) {
		if ((val = recv_nd(si, ifp, delay, 0, addr, &ninflow)) != 0) {
			if (val > 0) {
				warnx("Duplicate address detected");
				errno = EADDRNOTAVAIL;
			} else
				warnx("dad test");
			goto errdone;
		}

	start:
		delay = RETRANS_TIMER * 1000;
		if (--ntosend < 0)
			break;
	retry:
		cc = sendto(si, (char *)msgndsol, lgsol, 0,
			    (struct sockaddr *)&dst, sizeof dst);
#ifdef DEBUG
		fprintf(stderr, "send %d to %s#%d ",
				lgsol, PRINT6(&dst.sin6_addr), ifp);
		fprintf(stderr, "type %d tgt %s ",
				nsp->nd_ns_type, PRINT6(&nsp->nd_ns_target));
		fprintf(stderr, "src %s ", PRINT6(&ip->ip6_src));
		fprintf(stderr, "dst %s -> %d\n", PRINT6(&ip->ip6_dst), cc);
#endif
		if (cc < 0 && try == 0 && errno == EHOSTUNREACH) {
			(void) dad_add_multi(ifp, 0);
			try = 1;
			goto retry;
		}
		if (cc < 0 || cc != lgsol)  {
			if (cc < 0)
				warn("dad sendto NDsol");
			else {
				warnx("dad NDsol: wrote %d chars, ret=%d",
				      lgsol, cc);
				errno = EIO;
			}
			goto errdone;
		}
		ninflow++;
	}
	ret = 1;
    modify:
	cc = ioctl(u, SIOCALIFADDR, req);
	if (cc < 0) {
		warn("dad ioctl (SIOCALIFADDR)");
		/* ignore route install problems */
		if (errno != EEXIST)
			ret = -1;
	}

	/* last dad test, somebody may have just started */
	if (ret > 0) {
		if (recv_nd(si, ifp, 0, 0, addr, &ninflow) > 0) { /* Conflict */
			/* I already won, try to kill the other guy */
			sendto(si, (char *)msgndsol, lgsol, 0,
				(struct sockaddr *)&dst, sizeof dst);
		}
	}

    errdone:
	req->flags = flags;
	close(si);
	return(ret);
}

/*
 * Wait timeout ms and receive an ND packet
 * ret : 0: done, -1: error (in errno) 1: dad conflict with addr
 */
static int
recv_nd(fd, ifp, timeout, retintr, addr, ninflow)
	int fd, ifp, timeout, retintr, *ninflow;
	struct in6_addr *addr;
{
	struct pollfd fds[1];
	int cc, wait;
	struct msghdr msg;
	struct iovec iov;
	u_char packet[MAXPACKET];
	struct {
		struct cmsghdr c_cm;
		union {
			struct in6_pktinfo c_pi;
			u_char c_uspace[64];
		} c_space;
	} c_cmsg;
#define pi	c_cmsg.c_space.c_pi
#define cm	c_cmsg.c_cm
	struct ip6_hdr *ip = (struct ip6_hdr *)packet;
	struct nd_neighbor_solicit *nsp; /* same as nd_neighbor_advert */
	struct timeval prev, now;

	fds[0].fd = fd;
	fds[0].events = POLLIN;
	fds[0].revents = 0;

	gettimeofday(&prev, NULL);
	goto start;

	for(;;) {
		if (timeout != INFTIM) {
			gettimeofday(&now, NULL);
			timersub(&now, &prev, &prev);
			wait = prev.tv_sec;
			if (wait < 0)
				wait = 0;
			wait = wait * 1000;
			wait += prev.tv_usec / 1000;
			prev = now;
			timeout -= wait;
			if (timeout < 0)
				timeout = 0;
		}
	start:
#ifdef DEBUG
		fprintf(stderr, "start poll wait %d\n", timeout);
#endif
		cc = poll(fds, sizeof fds / sizeof (struct pollfd), timeout);
#ifdef DEBUG
		fprintf(stderr, "ret poll cc %d errno %d\n", cc, errno);
#endif
		if (cc < 0 && errno == EINTR && retintr == 0)
			continue;
		if (cc <= 0)
			return cc;
		bzero(&msg, sizeof(struct msghdr));
		msg.msg_iov = &iov;
		msg.msg_iov->iov_base = (char *)packet;
		msg.msg_iov->iov_len = MAXPACKET;
		msg.msg_iovlen = 1;
		msg.msg_control = (caddr_t)&c_cmsg;
		msg.msg_controllen = sizeof(c_cmsg);
		cc = recvmsg(fd, &msg, MSG_DONTWAIT);
#ifdef DEBUG
		fprintf(stderr, "recv %d bytes flg %x ctl %d errno %d\n",
				cc, msg.msg_flags, msg.msg_controllen, errno);
#endif
		if (cc < 0 && errno == EINTR && retintr == 0)
			continue;
		if (cc < 0)
			return cc;
		if ((cc < sizeof(*ip) + sizeof(struct icmp6_hdr)) ||
		    (msg.msg_flags & (MSG_TRUNC|MSG_CTRUNC)))
			continue;
		if ((msg.msg_controllen == 0) ||
		    (cm.cmsg_len != (sizeof(struct cmsghdr) +
				    sizeof(pi))) ||
		    (cm.cmsg_level != IPPROTO_IPV6) ||
		    (cm.cmsg_type != IPV6_PKTINFO) ||
		    (pi.ipi6_ifindex != ifp))
			continue;

		ip = (struct ip6_hdr *)packet;
		nsp = (struct nd_neighbor_solicit *)(ip + 1);
#define nap ((struct nd_neighbor_advert *)nsp)
#ifdef DEBUG
		fprintf(stderr, "recv for %s#%d\n",
				PRINT6(&pi.ipi6_addr), pi.ipi6_ifindex);
		fprintf(stderr, "recv msg %d/%d lim %d tgt %s ",
				nsp->nd_ns_type, nsp->nd_ns_code, ip->ip6_hlim,
				PRINT6(&nsp->nd_ns_target));
		fprintf(stderr, "src %s ", PRINT6(&ip->ip6_src));
		fprintf(stderr, "look for %s\n", PRINT6(addr));
#endif
		if ((ip->ip6_hlim != ND_DEFAULT_HOPLIM) ||
		    (nsp->nd_ns_code != 0))
			continue;

		cc -= sizeof(*ip);
		switch (nsp->nd_ns_type) {
		    case ND_NEIGHBOR_ADVERT:
			if ((unsigned)cc < sizeof(*nap))
				break;
			if (bcmp(&nap->nd_na_target, addr, sizeof(*addr)))
				break;
			return(1);

		    case ND_NEIGHBOR_SOLICIT:
			if ((unsigned)cc < sizeof(*nsp))
				break;
			if (bcmp(&nsp->nd_ns_target, addr, sizeof(*addr)) ||
			    !IN6_IS_ADDR_UNSPECIFIED(&ip->ip6_src) ||
			    (--*ninflow >= 0))
				break;
			return(1);
		}
	}
}

/* Internet checksum */
static int
in_cksum(addr, len)
	u_short *addr;
	int len;
{
	register int nleft = len;
	register u_short *w = addr;
	register u_short answer;
	u_short odd_byte = 0;
	register int sum = 0;

	/*
	 *  Our algorithm is simple, using a 32 bit accumulator (sum),
	 *  we add sequential 16 bit words to it, and at the end, fold
	 *  back all the carry bits from the top 16 bits into the lower
	 *  16 bits.
	 */
	while (nleft > 1)  {
		sum += *w++;
		nleft -= 2;
	}

	/* mop up an odd byte, if necessary */
	if (nleft == 1) {
		*(u_char *)&odd_byte = *(u_char *)w;
		sum += odd_byte;
	}

	/*
	 * add back carry outs from top 16 bits to low 16 bits
	 */
	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum;				/* truncate to 16 bits */
	return(answer);
}

/*
 * Add all multicast route for an interface
 */
int
dad_add_multi(ifp, tst)
	int ifp, tst;
{
	struct sockaddr *sa;
	int len, r, ret = -1;
#define SA(x)		((struct sockaddr *)(x))
#define ROUNDUP(a) \
        ((a) > 0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))
#define ADVANCE(x, n)	((x) += ROUNDUP(SA(n)->sa_len))
	struct {
		struct rt_msghdr m_rtm;
		char m_space[512];
	} m_rtmsg;
	char *cp = m_rtmsg.m_space;
#define rtm	m_rtmsg.m_rtm
	static int rtm_seq = 2222;

#ifdef DEBUG
	fprintf(stderr, "add all multicast route on ifp#%d\n", ifp);
#endif

	if ((r = socket(PF_ROUTE, SOCK_RAW, 0)) < 0) {
#ifdef DEBUG
		warn("socket (route add multi)");
#endif
		return (ret);
	}
	shutdown(r, 0);

	bzero(&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_flags = RTF_UP|RTF_STATIC;
	rtm.rtm_addrs = RTA_DST|RTA_GATEWAY|RTA_NETMASK;
	rtm.rtm_pid = getpid();
	rtm.rtm_seq = ++rtm_seq;
	rtm.rtm_index = ifp;

	sa = SA(cp);
	sa->sa_family = AF_INET6;
	sa->sa_len = sizeof (struct sockaddr_in6);
	((struct sockaddr_in6 *)sa)->sin6_addr.s6_addr[0] = 0xff;
	ADVANCE(cp, sa);
	sa = SA(cp);
	sa->sa_family = AF_LINK;
	sa->sa_len = sizeof (struct sockaddr_dl);
	((struct sockaddr_dl *)sa)->sdl_index = ifp;
	ADVANCE(cp, sa);
	sa = SA(cp);
	sa->sa_family = AF_UNSPEC;
	sa->sa_len = sizeof (struct sockaddr_in6);
	((struct sockaddr_in6 *)sa)->sin6_addr.s6_addr[0] = 0xff;
	ADVANCE(cp, sa);

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

	if (tst) {
		rtm.rtm_type = RTM_GET;
		if (write(r, &m_rtmsg, len) >= 0) {
#ifdef DEBUG
			fprintf(stderr, "multicast already on ifp#%d\n", ifp);
#endif
			ret = 0;
			goto done;
		}
		if (errno != ESRCH) {
#ifdef DEBUG
			warn("write (route message get multi)");
#endif
			goto done;
		}
	}
	rtm.rtm_type = RTM_ADD;
	if (write(r, &m_rtmsg, len) < len) {
#ifdef DEBUG
		warn("write (route message add multi)");
#endif
		goto done;
	}
	ret = 0;
    done:
	close(r);
	return (ret);
}
