#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <stdio.h>
#include <errno.h>
#include <signal.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/sysctl.h>
#include <sys/reboot.h>

#include <net/if.h>
#include <net/if_dl.h>
#ifndef if_type
#include <net/if_var.h>
#endif
#include <net/route.h>

#include <netinet/in_systm.h>
#define INET6
#include <netinet/in.h>
#include <netinet/in_var.h>
#include <netinet/in6_var.h>
#include <netinet/ip.h>
#include <netinet/ip6.h>
#include <netinet/ip_var.h>
#include <netinet/ip6_var.h>
#include <netinet/ip6_opts.h>

#include <netdb.h>
#include <arpa/inet.h>

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sysexits.h>

#define ROUNDUP(a) \
        ((a) > 0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))
#define ADVANCE(x, n)	((x) += ROUNDUP(((struct sockaddr *)(n))->sa_len))

char *lanname;
char *ctiname;
char buf[64];

struct in_addr src, dst;

struct ifreq ifr;
struct if_laddrreq iflr;

struct {
	struct rt_msghdr m_rtm;
        char m_space[512];
} m_rtmsg;
#define rtm	m_rtmsg.m_rtm

char usage[] = "Usage: cticonfig [-v4lm] [-i lan | -s src] cti target\n";

int u, v, r;			/* Socket file descriptors */

/* Common variables */
int verbose = 0;		/* extra printfs */
int keep_compat = 1;		/* keep compatible addresses */
int install_ll = 1;		/* install link-local addresses */
int install_multi = 1;		/* install the all multicast route */

int main(int, char *[]);
int getaddr4(char *);		/* get src from lanname */
void setcompat(void);		/* setup IPv4-compatible addresses */
void resetif(void);		/* remove IPv4-compatible addresses */
void setlladdr(void);		/* setup link-local addresses */
void setmulti(void);		/* add the all multicast route */
#define satosin6(sa)		((struct sockaddr_in6 *)(sa))
#define INET6_STOP(sa)		inet_ntop(AF_INET6, &satosin6(sa)->sin6_addr, \
					  buf, sizeof buf)

/* 
   - decode argument
   - set compatible addresses
   - reset addresses
   - set link-local addresses
   - set all multicast route
 */
int
main(argc, argv)
	int argc;
	char *argv[];
{
	char ch;
	int hassrc = 0;
	extern char *optarg;
	extern int optind;

	/* Open UDP/IPv6 "u" socket */
	if ((u = socket(PF_INET6, SOCK_DGRAM, 0)) < 0) {
		perror("socket (UDP6)");
		exit(EX_CANTCREAT);
	}

	/* Open UDP/IPv4 "v" socket */
	if ((v = socket(PF_INET, SOCK_DGRAM, 0)) < 0) {
		perror("socket (UDP4)");
		exit(EX_CANTCREAT);
	}

	/* Open Routing "r" socket */
	if ((r = socket(PF_ROUTE, SOCK_RAW, 0)) < 0) {
		perror("socket (ROUTE)");
		exit(EX_CANTCREAT);
	}
	(void) shutdown(r, 0);

	setlinebuf(stdout);

	while ((ch = getopt(argc, argv, "v4lmi:s:")) != EOF) {
		switch (ch) {
		    case 'v':
			verbose++;
			break;
		    case '4':
			keep_compat = 0;
			break;
		    case 'l':
			install_ll = 0;
			break;
		    case 'm':
			install_multi = 0;
			break;
		    case 'i':
			if (hassrc)
				goto usage;
			lanname = optarg;
			if (!getaddr4(lanname)) {
				(void) fprintf(stderr,
					       "bad interface %s\n",
					       lanname);
				exit(EX_DATAERR);
			}
			hassrc++;
			break;
		    case 's':
			if (hassrc)
				goto usage;
			if (!inet_aton(optarg, &src)) {
				(void) fprintf(stderr,
					       "bad source address %s\n",
					       optarg);
				exit(EX_DATAERR);
			}
			hassrc++;
			break;
		    default:
		    usage:
			(void) fprintf(stderr, usage);
			exit(EX_USAGE);
		}
	}

	argc -= optind;
	argv += optind;

	if (hassrc == 0) {
		(void) fprintf(stderr, "needs the source address\n");
		(void) fprintf(stderr, usage);
		exit(EX_USAGE);
	}
	if (argc != 2) {
		(void) fprintf(stderr, "%s paramaters\n",
			       argc < 2 ? "missing" : "extra");
		(void) fprintf(stderr, usage);
		exit(EX_USAGE);
	}
	ctiname = argv[0];
	if (!inet_aton(argv[1], &dst)) {
		(void) fprintf(stderr, "bad target address %s\n", argv[1]);
		exit(EX_DATAERR);
	}

	setcompat();

	if (keep_compat == 0)
		resetif();

	if (install_ll)
		setlladdr();

	if (install_multi)
		setmulti();

	exit(EX_OK);
}

/*
 * Get the IPv4 address of a (LAN) interface
 */
int
getaddr4(name)
	char *name;
{
	struct sockaddr_in *sin = (struct sockaddr_in *)&ifr.ifr_addr;

	bzero((caddr_t)&ifr, sizeof(ifr));
	strncpy(ifr.ifr_name, name, IFNAMSIZ);
	sin->sin_family = AF_INET;
	sin->sin_len = sizeof(*sin);

	if (ioctl(v, SIOCGIFADDR, (caddr_t)&ifr) < 0) {
		perror("ioctl (SIOCGIFADDR)");
		return 0;
	}
	src.s_addr = sin->sin_addr.s_addr;
	if (verbose)
		printf("interface %s address is %s\n",
		       name, inet_ntoa(src));
	return 1;
}

/*
 * Configure the IPv4-compatible addresses
 */
void
setcompat()
{
	register struct sockaddr_in6 *sin;

	bzero((caddr_t)&iflr, sizeof(iflr));
	strncpy(iflr.iflr_name, ctiname, IFNAMSIZ);

	sin = satosin6(&iflr.addr);
	sin->sin6_family = AF_INET6;
	sin->sin6_len = sizeof(*sin);
	bcopy(&src, &sin->sin6_addr.s6_addr[12], 4);

	sin = satosin6(&iflr.dstaddr);
	sin->sin6_family = AF_INET6;
	sin->sin6_len = sizeof(*sin);
	bcopy(&dst, &sin->sin6_addr.s6_addr[12], 4);

	iflr.prefixlen = 128;
	iflr.flags = 0;

	if (ioctl(u, SIOCALIFADDR, (caddr_t)&iflr) < 0) {
		perror("ioctl (SIOCALIFADDR)");
		fprintf(stderr, "can't setup compatible addresses\n");
		exit(EX_OSERR);
	}

	if (verbose) {
		printf("setup %s: %s/128", ctiname, INET6_STOP(&iflr.addr));
		printf(" %s\n", INET6_STOP(&iflr.dstaddr));
	}
}

/*
 * Delete IPv4-compatible addresses
 */
void resetif() {
	register struct sockaddr_in6 *sin;

	bzero((caddr_t)&iflr, sizeof(iflr));
	strncpy(iflr.iflr_name, ctiname, IFNAMSIZ);

	sin = satosin6(&iflr.addr);
	sin->sin6_family = AF_INET6;
	sin->sin6_len = sizeof(*sin);
	bcopy(&src, &sin->sin6_addr.s6_addr[12], 4);

	if (ioctl(u, SIOCDLIFADDR, (caddr_t)&iflr) < 0) {
		perror("ioctl (SIOCDLIFADDR)");
		fprintf(stderr, "can't delete compatible addresses\n");
		exit(EX_OSERR);
	}

	if (verbose)
		printf("delete %s: %s\n", ctiname, INET6_STOP(sin));
}

/*
 * Configure link-local addresses
 */
void setlladdr() {
	register struct sockaddr_in6 *sin;
	int nb;

	nb = atoi(ctiname + 3);
	if ((nb & 0xff) != nb) {
		fprintf(stderr, "%s gives cti<%d>?\n", ctiname, nb);
		nb = 0;
	}

	bzero((caddr_t)&iflr, sizeof(iflr));
	strncpy(iflr.iflr_name, ctiname, IFNAMSIZ);

	sin = satosin6(&iflr.addr);
	sin->sin6_family = AF_INET6;
	sin->sin6_len = sizeof(*sin);
	sin->sin6_addr.s6_addr[0] = 0xfe;
	sin->sin6_addr.s6_addr[1] = 0x80;
	bcopy(&src, &sin->sin6_addr.s6_addr[12], 4);

	sin = satosin6(&iflr.dstaddr);
	sin->sin6_family = AF_INET6;
	sin->sin6_len = sizeof(*sin);
	sin->sin6_addr.s6_addr[0] = 0xfe;
	sin->sin6_addr.s6_addr[1] = 0x80;
	bcopy(&dst, &sin->sin6_addr.s6_addr[12], 4);

	iflr.prefixlen = 128;
	iflr.flags = 0;

	if (ioctl(u, SIOCALIFADDR, (caddr_t)&iflr) < 0) {
		perror("ioctl (SIOCALIFADDR)");
		fprintf(stderr, "can't setup link-local addresses\n");
		exit(EX_OSERR);
	}

	if (verbose) {
		printf("setup %s: %s/128", ctiname, INET6_STOP(&iflr.addr));
		printf(" %s\n", INET6_STOP(&iflr.dstaddr));
	}
}

/*
 * Add the all multicast route
 */
void
setmulti() {
	register char *cp = m_rtmsg.m_space;
	struct sockaddr_in6 *sin;
	struct sockaddr_dl *sdl;
	int len, idx;

	idx = if_nametoindex(ctiname);
	if (idx == 0) {
		fprintf(stderr, "can't find the index\n");
		exit(EX_OSERR);
	}

	bzero(&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = RTM_ADD;
	rtm.rtm_flags = RTF_UP | RTF_STATIC;
	rtm.rtm_addrs = RTA_DST | RTA_GATEWAY | RTA_NETMASK;
	rtm.rtm_pid = getpid();
	rtm.rtm_seq = 1;
	rtm.rtm_index = idx;

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_family = AF_INET6;
	sin->sin6_len = sizeof(*sin);
	sin->sin6_addr.s6_addr[0] = 0xff;
	ADVANCE(cp, sin);

	sdl = (struct sockaddr_dl *)cp;
	sdl->sdl_family = AF_LINK;
	sdl->sdl_len = sizeof(*sdl);
	sdl->sdl_index = idx;
	sdl->sdl_nlen = strlen(ctiname);
	strcpy(sdl->sdl_data, ctiname);
	ADVANCE(cp, sdl);

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_family = AF_UNSPEC;
	sin->sin6_len = sizeof(*sin);
	sin->sin6_addr.s6_addr[0] = 0xff;
	ADVANCE(cp, sin);

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

	if (write(r, (char *)&m_rtmsg, len) < len) {
		perror("write (route message)");
		exit(EX_OSERR);
	}
	if (verbose)
		printf("add direct route to ff00::/8 via %s\n", ctiname);
}
