/*	$NetBSD: ping.c,v 1.20 1995/08/11 22:37:58 cgd Exp $	*/

/*
 * Copyright (c) 1989, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Mike Muuss.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright (c) 1989, 1993\n\
	The Regents of the University of California.  All rights reserved.\n";
#endif /* not lint */

#ifndef lint
#if 0
static char sccsid[] = "@(#)ping.c	8.1 (Berkeley) 6/5/93";
#else
static char rcsid[] = "$NetBSD: ping.c,v 1.20 1995/08/11 22:37:58 cgd Exp $";
#endif
#endif /* not lint */

/*
 *			P I N G . C
 *
 * Using the InterNet Control Message Protocol (ICMP) "ECHO" facility,
 * measure round-trip-delays and packet loss across network paths.
 *
 * Author -
 *	Mike Muuss
 *	U. S. Army Ballistic Research Laboratory
 *	December, 1983
 *
 * Status -
 *	Public Domain.  Distribution Unlimited.
 * Bugs -
 *	More statistics could always be gathered.
 *	This program has to run SUID to ROOT to access the ICMP socket.
 */

#include <sys/param.h>
#include <sys/queue.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <termios.h>

#include <net/if.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip6.h>
#include <netinet/icmp6.h>
#include <netinet/ip6_var.h>
#include <netinet/ip6_opts.h>
#include <netinet/if_ether.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <netdb.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#define	DEFDATALEN	(64 - 8)	/* default data length */
#define	MAXIPLEN	60
#define	MAXICMPLEN	76
#define	MAXPACKET	(65536 - 60 - 8)/* max packet size */
#define	MAXWAIT_DEFAULT	10		/* max seconds to wait for response */
#define	NROUTES		9		/* number of record route slots */

#define	A(bit)		rcvd_tbl[(bit)>>3]	/* identify byte in array */
#define	B(bit)		(1 << ((bit) & 0x07))	/* identify bit in byte */
#define	SET(bit)	(A(bit) |= B(bit))
#define	CLR(bit)	(A(bit) &= (~B(bit)))
#define	TST(bit)	(A(bit) & B(bit))

/* various options */
int options;
#define	F_FLOOD		0x001
#define	F_INTERVAL	0x002
#define	F_NUMERIC	0x004
#define	F_PINGFILLED	0x008
#define	F_QUIET		0x010
#define	F_RROUTE	0x020
#define	F_SO_DEBUG	0x040
#define	F_SO_DONTROUTE	0x080
#define	F_VERBOSE	0x100
#define	F_SADDR		0x200
#define	F_MULTICAST	0x400

/* multicast options */
int moptions;
#define	MULTICAST_NOLOOP	0x001
#define	MULTICAST_HLIM		0x002
#define	MULTICAST_IF		0x004

/*
 * MAX_DUP_CHK is the number of bits in received table, i.e. the maximum
 * number of received sequence numbers we can keep track of.  Change 128
 * to 8192 for complete accuracy...
 */
#define	MAX_DUP_CHK	(8 * 128)
int mx_dup_ck = MAX_DUP_CHK;
char rcvd_tbl[MAX_DUP_CHK / 8];

struct sockaddr_in6 fromwhere;	/* which source address to use */
struct sockaddr_in6 whereto;	/* who to ping */
struct in6_pktinfo pktinfo;	/* interface to use */
struct icmp6_filter filter;	/* ICMPv6 filter */
int datalen = DEFDATALEN;
int s;				/* socket file descriptor */
u_char outpack[MAXPACKET];
char BSPACE = '\b';		/* characters written for flood */
char DOT = '.';
char *myname;
char *hostname;
char hname[64];
int ident;			/* process id to identify our packets */

/* counters */
long npackets;			/* max packets to transmit */
long nreceived;			/* # of packets we got back */
long nrepeats;			/* number of duplicates */
long ntransmitted;		/* sequence # for outbound packets = #sent */
int interval = 1;		/* interval between packets */

/* timing */
int timing;			/* flag to do timing */
int maxwait = MAXWAIT_DEFAULT;	/* max seconds to wait for response */
double tmin = 999999999.0;	/* minimum round trip time */
double tmax = 0.0;		/* maximum round trip time */
double tsum = 0.0;		/* sum of all times, for doing average */

int reset_kerninfo;

void fill __P((char *, char *));
void catcher(), finish(), status();
void pinger();
char *pr_addr __P((struct in6_addr *));
void pr_ndopts __P((char *, int));
void pr_icmph __P((struct icmp6_hdr *, int));
void pr_pack __P((char *, int, struct sockaddr_in6 *));
void pr_retip __P((struct ip6_hdr *));
char *getsrcfor __P((char *));
void usage();

int
main(argc, argv)
	int argc;
	char **argv;
{
	extern int errno, optind;
	extern char *optarg;
	struct timeval timeout;
	struct hostent *hp;
	struct sockaddr_in6 *from;
	struct sockaddr_in6 *to;
	struct termios ts;
	register int i;
	int ch, fdmask, hold, packlen, preload;
	u_char *datap, *packet;
	char *source, *target;
	char lnamebuf[MAXHOSTNAMELEN], hnamebuf[MAXHOSTNAMELEN];
	u_int ttl, loop = 1;

	preload = 0;
	if (tcgetattr (0, &ts) != -1) {
		reset_kerninfo = !(ts.c_lflag & NOKERNINFO);
		ts.c_lflag |= NOKERNINFO;
		tcsetattr (0, TCSANOW, &ts);
	}

	source = NULL;
	datap = &outpack[8 + sizeof(struct timeval)];
	while ((ch = getopt(argc, argv, "I:LS:c:dfi:l:np:qrs:t:vw:")) != EOF)
		switch(ch) {
		case 'c':
			npackets = atoi(optarg);
			if (npackets <= 0)
				errx(1, "bad number of packets to transmit: %s",
				    optarg);
			break;
		case 'd':
			options |= F_SO_DEBUG;
			break;
		case 'f':
			if (getuid())
				errx(1, "%s", strerror(EPERM));
			options |= F_FLOOD;
			setbuf(stdout, (char *)NULL);
			break;
		case 'I':
			memset(&pktinfo, 0, sizeof(struct in6_pktinfo));
			pktinfo.ipi6_ifindex = if_nametoindex(optarg);
			if (pktinfo.ipi6_ifindex == 0)
				 errx(1, "unknown interface: %s", optarg);
			moptions |= MULTICAST_IF;
			break;
		case 'i':		/* wait between sending packets */
			interval = atoi(optarg);
			if (interval <= 0)
				errx(1, "bad timing interval: %s", optarg);
			options |= F_INTERVAL;
			break;
		case 'L':
			moptions |= MULTICAST_NOLOOP;
			loop = 0;
			break;
		case 'l':
			preload = atoi(optarg);
			if (preload < 0)
				errx(1, "bad preload value: %s", optarg);
			break;
		case 'n':
			options |= F_NUMERIC;
			break;
		case 'p':		/* fill buffer with user pattern */
			options |= F_PINGFILLED;
			fill((char *)datap, optarg);
				break;
		case 'q':
			options |= F_QUIET;
			break;
		case 'r':
			options |= F_SO_DONTROUTE;
			break;
		case 'S':
			source = optarg;
			options |= F_SADDR;
			break;
		case 's':		/* size of packet to send */
			datalen = atoi(optarg);
			if (datalen <= 0)
				errx(1, "bad packet size: %s", optarg);
			if (datalen > MAXPACKET)
				errx(1, "packet size too large: %s", optarg);
			break;
		case 't':
			ttl = atoi(optarg);
			if (ttl <= 0)
				errx(1, "bad ttl value: %s", optarg);
			if (ttl > 255)
				errx(1, "ttl value too large: %s", optarg);
			moptions |= MULTICAST_HLIM;
			break;
		case 'v':
			options |= F_VERBOSE;
			break;
		case 'w':
			maxwait = atoi(optarg);
			if (maxwait <= 0)
				errx(1, "bad maxwait value: %s", optarg);
			break;
		default:
			usage();
		}
	argc -= optind;
	argv += optind;

	if (argc != 1)
		usage();
	target = *argv;

	if (source == NULL)
		source = getsrcfor(target);
	memset(&fromwhere, 0, sizeof(struct sockaddr_in6));
	from = &fromwhere;
	from->sin6_len = sizeof(struct sockaddr_in6);
	hp = getipnodebyname(source, AF_INET6, 0, &ch);
	if (!hp)
		errx(1, "unknown source address: %s", source);
	from->sin6_family = hp->h_addrtype;
	memcpy(&from->sin6_addr, hp->h_addr, hp->h_length);
	(void)strncpy(lnamebuf, hp->h_name, sizeof(lnamebuf) - 1);
	myname = lnamebuf;
	freehostent(hp);
	if ((from->sin6_family != AF_INET6) ||
	    (from->sin6_len != sizeof(struct sockaddr_in6)) ||
	    IN6_IS_ADDR_V4MAPPED(&from->sin6_addr))
		errx(1, "no IPv6 address for source: %s", source);

	memset(&whereto, 0, sizeof(struct sockaddr_in6));
	to = &whereto;
	to->sin6_len = sizeof(struct sockaddr_in6);
	hp = getipnodebyname(target, AF_INET6, 0, &ch);
	if (!hp)
		errx(1, "unknown host: %s", target);
	to->sin6_family = hp->h_addrtype;
	memcpy(&to->sin6_addr, hp->h_addr, hp->h_length);
	(void)strncpy(hnamebuf, hp->h_name, sizeof(hnamebuf) - 1);
	hostname = hnamebuf;
	freehostent(hp);
	if ((to->sin6_family != AF_INET6) ||
	    (to->sin6_len != sizeof(struct sockaddr_in6)) ||
	    IN6_IS_ADDR_V4MAPPED(&to->sin6_addr))
		errx(1, "no IPv6 address for host: %s", target);
	if (to->sin6_addr.s6_addr[0] == 0xff)
		options |= F_MULTICAST;

	if (options & F_FLOOD && options & F_INTERVAL)
		errx(1, "-f and -i options are incompatible");

	if (datalen >= sizeof(struct timeval))	/* can we time transfer */
		timing = 1;
	packlen = datalen + MAXIPLEN + MAXICMPLEN;
	if (!(packet = (u_char *)malloc((u_int)packlen)))
		err(1, "malloc");
	if (!(options & F_PINGFILLED))
		for (i = 8; i < datalen; ++i)
			*datap++ = i;

	ident = getpid() & 0xFFFF;

	if ((s = socket(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6)) < 0)
		err(1, "socket");
	hold = 1;
	if (options & F_SO_DEBUG)
		(void)setsockopt(s, SOL_SOCKET, SO_DEBUG, (char *)&hold,
		    sizeof(hold));
	if (options & F_SO_DONTROUTE)
		(void)setsockopt(s, SOL_SOCKET, SO_DONTROUTE, (char *)&hold,
		    sizeof(hold));
	if (moptions & MULTICAST_IF) {
		memcpy(&pktinfo.ipi6_addr,
		       &fromwhere.sin6_addr,
		       sizeof(struct in6_addr));
		if (setsockopt(s, IPPROTO_IPV6, IPV6_PKTINFO, &pktinfo,
			       sizeof(pktinfo)) < 0)
			err(1, "setsockopt IPV6_PKTINFO");
	} else if (bind(s, (struct sockaddr *)&fromwhere,
			sizeof(fromwhere)) < 0) {
		perror("ping: bind");
		exit(1);
	}

	if ((options & F_MULTICAST) &&
	    (moptions & MULTICAST_NOLOOP) &&
	    setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_LOOP, &loop,
		       sizeof(loop)) < 0)
		err(1, "setsockopt IPV6_MULTICAST_LOOP");
	if ((options & F_MULTICAST) &&
	    (moptions & MULTICAST_HLIM) &&
	    setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_HOPS, &ttl,
		       sizeof(ttl)) < 0)
		err(1, "setsockopt IPV6_MULTICAST_HOPS");
	/*
	 * When pinging the broadcast address, you can get a lot of answers.
	 * Doing something so evil is useful if you are trying to stress the
	 * ethernet, or just want to fill the arp cache to get some stuff for
	 * /etc/ethers.
	 */
	hold = 48 * 1024;
	(void)setsockopt(s, SOL_SOCKET, SO_RCVBUF, (char *)&hold,
	    sizeof(hold));

	/* Put an ICMPv6 filter */
	if (options & F_VERBOSE) {
		ICMP6_FILTER_SETPASSALL(&filter);
	} else {
		ICMP6_FILTER_SETBLOCKALL(&filter);
		ICMP6_FILTER_SETPASS(ICMP6_ECHO_REPLY, &filter);
	}
	if (setsockopt(s, IPPROTO_ICMPV6, ICMP6_FILTER, &filter,
		       sizeof(filter)) < 0)
		err(1, "setsockopt ICMP6_FILTER");

	if (to->sin6_family == AF_INET6)
		(void)printf("PING %s (%s): %d data bytes\n", hostname,
		    inet_ntop(AF_INET6, &to->sin6_addr,
			      hname, sizeof(hname)),
		    datalen);
	else
		(void)printf("PING %s: %d data bytes\n", hostname, datalen);

	(void)signal(SIGINT, finish);
	(void)signal(SIGALRM, catcher);
	(void)signal(SIGINFO, status);

	while (preload--)		/* fire off them quickies */
		pinger();

	if ((options & F_FLOOD) == 0)
		catcher();		/* start things going */

	for (;;) {
		struct sockaddr_in6 from;
		register int cc;
		int fromlen;

		if (options & F_FLOOD) {
			pinger();
			timeout.tv_sec = 0;
			timeout.tv_usec = 10000;
			fdmask = 1 << s;
			if (select(s + 1, (fd_set *)&fdmask, (fd_set *)NULL,
			    (fd_set *)NULL, &timeout) < 1)
				continue;
		}
		fromlen = sizeof(from);
		if ((cc = recvfrom(s, (char *)packet, packlen, 0,
		    (struct sockaddr *)&from, &fromlen)) < 0) {
			if (errno == EINTR)
				continue;
			perror("ping: recvfrom");
			continue;
		}
		pr_pack((char *)packet, cc, &from);
		if (npackets && nreceived >= npackets)
			break;
	}
	finish();
	/* NOTREACHED */
	exit(0);	/* Make the compiler happy */
}

/*
 * catcher --
 *	This routine causes another PING to be transmitted, and then
 * schedules another SIGALRM for 1 second from now.
 *
 * bug --
 *	Our sense of time will slowly skew (i.e., packets will not be
 * launched exactly at 1-second intervals).  This does not affect the
 * quality of the delay and loss statistics.
 */
void
catcher()
{
	int waittime;

	pinger();
	(void)signal(SIGALRM, catcher);
	if (!npackets || ntransmitted < npackets)
		alarm((u_int)interval);
	else {
		if (nreceived) {
			waittime = 2 * tmax / 1000;
			if (!waittime)
				waittime = 1;
		} else
			waittime = maxwait;
		(void)signal(SIGALRM, finish);
		(void)alarm((u_int)waittime);
	}
}

/*
 * pinger --
 *	Compose and transmit an ICMP6 ECHO REQUEST packet.  The IP packet
 * will be added on by the kernel.  The ID field is our UNIX process ID,
 * and the sequence number is an ascending integer.  The first 8 bytes
 * of the data portion are used to hold a UNIX "timeval" struct in VAX
 * byte-order, to compute the round-trip time.
 */
void
pinger()
{
	register struct icmp6_hdr *icp;
	register int cc, i;

	icp = (struct icmp6_hdr *)outpack;
	icp->icmp6_type = ICMP6_ECHO_REQUEST;
	icp->icmp6_code = 0;
	icp->icmp6_cksum = 0;
	icp->icmp6_seq = ntransmitted++;
	icp->icmp6_id = ident;			/* ID */

	CLR(icp->icmp6_seq % mx_dup_ck);

	if (timing)
		(void)gettimeofday((struct timeval *)&outpack[8],
		    (struct timezone *)NULL);

	cc = datalen + 8;			/* skips ICMP portion */
	i = sendto(s, (char *)outpack, cc, 0,
		   (struct sockaddr *)&whereto, sizeof(whereto));

	if (i < 0 || i != cc)  {
		if (i < 0)
			perror("ping: sendto");
		(void)printf("ping: wrote %s %d chars, ret=%d\n",
		    hostname, cc, i);
	}
	if (!(options & F_QUIET) && options & F_FLOOD)
		(void)write(STDOUT_FILENO, &DOT, 1);
}

/*
 * pr_pack --
 *	Print out the packet, if it came from us.  This logic is necessary
 * because ALL readers of the ICMPv6 socket get a copy of ALL ICMPv6 packets
 * which arrive ('tis only fair).  This permits multiple copies of this
 * program to be run without having intermingled output (or statistics!).
 */
void
pr_pack(buf, cc, from)
	char *buf;
	int cc;
	struct sockaddr_in6 *from;
{
	register struct icmp6_hdr *icp;
	register int i, j;
	register u_char *cp,*dp;
	struct ip6_hdr *ip;
	struct timeval tv, tp;
	char *pkttime;
	double triptime;
	int dupflag;

	(void)gettimeofday(&tv, (struct timezone *)NULL);

	/* Check the IP header */
	ip = (struct ip6_hdr *)buf;
	if (cc < sizeof(*ip) + sizeof(*icp)) {
		if (options & F_VERBOSE)
			warnx("packet too short (%d bytes) from %s", cc,
			      inet_ntop(AF_INET6, &from->sin6_addr,
					hname, sizeof(hname)));
		return;
	}

	/* Now the ICMPv6 part */
	cc -= sizeof(*ip);
	icp = (struct icmp6_hdr *)(buf + sizeof(*ip));
	/* test the address too */
	if ((icp->icmp6_type == ICMP6_ECHO_REPLY) &&
	    (icp->icmp6_id == ident) &&
	    ((options & F_MULTICAST) ||
	     (bcmp(&from->sin6_addr, &whereto.sin6_addr,
		   sizeof(struct in6_addr)) == 0))) {
		++nreceived;
		if (timing) {
			pkttime = (char *)(icp + 1);
			memcpy(&tp, pkttime, sizeof (tp));
			tvsub(&tv, &tp);
			triptime = ((double)tv.tv_sec) * 1000.0 +
			    ((double)tv.tv_usec) / 1000.0;
			tsum += triptime;
			if (triptime < tmin)
				tmin = triptime;
			if (triptime > tmax)
				tmax = triptime;
		}

		if (TST(icp->icmp6_seq % mx_dup_ck)) {
			++nrepeats;
			--nreceived;
			dupflag = 1;
		} else {
			SET(icp->icmp6_seq % mx_dup_ck);
			dupflag = 0;
		}

		if (options & F_QUIET)
			return;

		if (options & F_FLOOD)
			(void)write(STDOUT_FILENO, &BSPACE, 1);
		else {
			(void)printf("%d bytes from %s: icmp6_seq=%u", cc,
			   inet_ntop(AF_INET6, &from->sin6_addr,
				     hname, sizeof(hname)),
			   icp->icmp6_seq);
			(void)printf(" ttl=%d", ip->ip6_hlim);
			if (timing)
				(void)printf(" time=%.3f ms", triptime);
			if (dupflag)
				(void)printf(" (DUP!)");
			/* check the data */
			cp = (u_char *)(icp + 1) + sizeof (struct timeval);
			dp = &outpack[8 + sizeof (struct timeval)];
			for (i = 8 + sizeof(struct timeval); i < datalen;
			    ++i, ++cp, ++dp) {
				if (*cp != *dp) {
	(void)printf("\nwrong data byte #%d should be 0x%x but was 0x%x",
	    i, *dp, *cp);
					cp = (u_char *)(icp + 1);
					for (i = 8; i < datalen; ++i, ++cp) {
						if ((i % 32) == 8)
							(void)printf("\n\t");
						(void)printf("%x ", *cp);
					}
					break;
				}
			}
		}
	} else {
		/* We've got something other than an ECHOREPLY */
		if (!(options & F_VERBOSE))
			return;
		(void)printf("%d bytes from %s:\n", cc,
		    pr_addr(&from->sin6_addr));
		pr_icmph(icp, cc);
		(void)fflush(stdout);
		return;
	}

	if (!(options & F_FLOOD)) {
		(void)putchar('\n');
		(void)fflush(stdout);
	}
}

/*
 * tvsub --
 *	Subtract 2 timeval structs:  out = out - in.  Out is assumed to
 * be >= in.
 */
tvsub(out, in)
	register struct timeval *out, *in;
{
	if ((out->tv_usec -= in->tv_usec) < 0) {
		--out->tv_sec;
		out->tv_usec += 1000000;
	}
	out->tv_sec -= in->tv_sec;
}

/*
 * status --
 *	Print out statistics when SIGINFO is received.
 */

void
status()
{
	double temp_min = nreceived ? tmin : 0;
	(void)fprintf(stderr, "%ld/%ld packets received (%ld%%) "
		      "%.3f min / %.3f avg / %.3f max\n",
		      nreceived, ntransmitted,
		      (ntransmitted ?
		       100 - (int) (((ntransmitted - nreceived) * 100)
				    / ntransmitted)
		       : 0),
		      temp_min,
		      ((nreceived + nrepeats) ?
		       (tsum / (nreceived + nrepeats))/1000.0
		       : tsum),
		      tmax/ 1000.0);
}

/*
 * finish --
 *	Print out statistics, and give up.
 */
void
finish()
{
	register int i;
	struct termios ts;

	(void)signal(SIGINT, SIG_IGN);
	(void)putchar('\n');
	(void)fflush(stdout);
	(void)printf("--- %s ping statistics ---\n", hostname);
	(void)printf("%ld packets transmitted, ", ntransmitted);
	(void)printf("%ld packets received, ", nreceived);
	if (nrepeats)
		(void)printf("+%ld duplicates, ", nrepeats);
	if (ntransmitted)
		if (nreceived > ntransmitted)
			(void)printf("-- somebody's printing up packets!");
		else
			(void)printf("%d%% packet loss",
			    (int) (((ntransmitted - nreceived) * 100) /
			    ntransmitted));
	(void)putchar('\n');
	if (nreceived && timing) {
		/* Only display average to microseconds */
		i = 1000.0 * tsum / (nreceived + nrepeats);
		(void)printf("round-trip min/avg/max = %.3f/%.3f/%.3f ms\n",
		    tmin, ((double)i) / 1000.0, tmax);
	}
	if (reset_kerninfo && tcgetattr (0, &ts) != -1) {
		ts.c_lflag &= ~NOKERNINFO;
		tcsetattr (0, TCSANOW, &ts);
	}

	if (nreceived)
		exit(0);
	else
		exit(2);
}

/*
 * pr_iph --
 *	Print an IPv6 header.
 */
void
pr_iph(ip)
	struct ip6_hdr *ip;
{
	u_long head = ntohl(ip->ip6_flow);

	(void)printf("Vr DS     Flow   Len  NH Hlim Src      Dst Data\n");
	(void)printf(" %1x  %02x    %05x %04x %02x %02x   ",
		     IPV6_GET_VERSION(head),
		     IPV6_GET_CLASS(head),
		     IPV6_GET_FLOWLABEL(head),
		     ntohs(ip->ip6_plen),
		     ip->ip6_nxt,
		     ip->ip6_hlim);
	(void)printf("from %s ",
		     inet_ntop(AF_INET6, &ip->ip6_src, hname, sizeof(hname)));
	(void)printf("to %s\n",
		     inet_ntop(AF_INET6, &ip->ip6_dst, hname, sizeof(hname)));
}

/*
 * pr_icmph --
 *	Print a descriptive string about an ICMPv6 header.
 */
void
pr_icmph(icp, cc)
	struct icmp6_hdr *icp;
	int cc;
{
	switch(icp->icmp6_type) {
	case ICMP6_DST_UNREACH:
		switch(icp->icmp6_code) {
		case ICMP6_DST_UNREACH_NOROUTE:
			(void)printf("No Route\n");
			break;
		case ICMP6_DST_UNREACH_ADMIN:
			(void)printf("Admin Prohibited\n");
			break;
		case ICMP6_DST_UNREACH_BEYONDSCOPE:
			(void)printf("Beyond Scope\n");
			break;
		case ICMP6_DST_UNREACH_ADDR:
			(void)printf("Address Unreachable\n");
			break;
		case ICMP6_DST_UNREACH_NOPORT:
			(void)printf("Port Unreachable\n");
			break;
		default:
			(void)printf("Unreachable, Bad Code: %d\n",
				     icp->icmp6_code);
			break;
		}
		/* Print returned IP header information */
		pr_retip((struct ip6_hdr *)(icp + 1));
		break;
	case ICMP6_PACKET_TOO_BIG:
		(void)printf("Packet Too Big: MTU %d\n",
			     ntohl(icp->icmp6_mtu));
		pr_retip((struct ip6_hdr *)(icp + 1));
		break;
	case ICMP6_TIME_EXCEEDED:
		switch(icp->icmp6_code) {
		case ICMP6_TIME_EXCEED_TRANSIT:
			(void)printf("Time To Live Exceeded\n");
			break;
		case ICMP6_TIME_EXCEED_REASSEMBLY:
			(void)printf("Frag Reassembly Time Exceeded\n");
			break;
		default:
			(void)printf("Time exceeded, Bad Code: %d\n",
				     icp->icmp6_code);
			break;
		}
		pr_retip((struct ip6_hdr *)(icp + 1));
		break;
	case ICMP6_PARAM_PROB:
		switch(icp->icmp6_code) {
		case ICMP6_PARAMPROB_HEADER:
			(void)printf("Bad Header at 0x%02x\n",
				     ntohl(icp->icmp6_pptr));
			break;
		case ICMP6_PARAMPROB_NEXTHEADER:
			(void)printf("Unknown Next Header at 0x%02x\n",
				     ntohl(icp->icmp6_pptr));
			break;
		case ICMP6_PARAMPROB_OPTION:
			(void)printf("Unknown Option at 0x%02x\n",
				     ntohl(icp->icmp6_pptr));
			break;
		default:
			(void)printf("Parameter Problem, Bad Code: %d\n",
				     icp->icmp6_code);
			break;
		}
		pr_retip((struct ip6_hdr *)(icp + 1));
		break;
	case ICMP6_ECHO_REQUEST:
		(void)printf("Echo Request (%d/%d)\n",
			     ntohs(icp->icmp6_id), ntohs(icp->icmp6_seq));
		break;
	case ICMP6_ECHO_REPLY:
		(void)printf("Echo Reply (%d/%d)\n",
			     ntohs(icp->icmp6_id), ntohs(icp->icmp6_seq));
		break;
	case ICMP6_MEMBERSHIP_QUERY:
		(void)printf("Group Membership Query %s\n",
			     inet_ntop(AF_INET6, icp + 1,
				       hname, sizeof(hname)));
		break;
	case ICMP6_MEMBERSHIP_REPORT:
		(void)printf("Group Membership Report %s\n",
			     inet_ntop(AF_INET6, icp + 1,
				       hname, sizeof(hname)));
		break;
	case ICMP6_MEMBERSHIP_REDUCTION:
		(void)printf("Group Membership Termination %s\n",
			     inet_ntop(AF_INET6, icp + 1,
				       hname, sizeof(hname)));
		break;
	case ND_ROUTER_SOLICIT:
		(void)printf("Router Solicitation\n");
		pr_ndopts((char *)icp + sizeof(struct nd_router_solicit),
			  cc - sizeof(struct nd_router_solicit));
		break;
	case ND_ROUTER_ADVERT:
		(void)printf("Router Advertisement\n");
		pr_ndopts((char *)icp + sizeof(struct nd_router_advert),
			  cc - sizeof(struct nd_router_advert));
		break;
	case ND_NEIGHBOR_SOLICIT:
		(void)printf("Neighbor Solicitation %s",
			     inet_ntop(AF_INET6, icp + 1,
				       hname, sizeof(hname)));
		pr_ndopts((char *)icp + sizeof(struct nd_neighbor_solicit),
			  cc - sizeof(struct nd_neighbor_solicit));
		break;
	case ND_NEIGHBOR_ADVERT:
		(void)printf("Neighbor Advertisement %s\n",
			     inet_ntop(AF_INET6, icp + 1,
				       hname, sizeof(hname)));
		pr_ndopts((char *)icp + sizeof(struct nd_router_advert),
			  cc - sizeof(struct nd_router_advert));
		break;
	case ND_REDIRECT: {
		struct nd_redirect *ndp = (struct nd_redirect *)icp;

		(void)printf("Redirect");
		(void)printf(" for %s",
			     inet_ntop(AF_INET6, &ndp->nd_rd_dst,
				       hname, sizeof(hname)));
		(void)printf(" via %s\n",
			     inet_ntop(AF_INET6, &ndp->nd_rd_target,
				       hname, sizeof(hname)));
		}
		pr_ndopts((char *)icp + sizeof(struct nd_redirect),
			  cc - sizeof(struct nd_redirect));
		break;
	case ICMP6_ROUTER_RENUMBERING:
		(void)printf("Router Renumbering\n");
		break;
	default:
		(void)printf("Bad ICMPv6 type: %d\n", icp->icmp6_type);
	}
}

/*
 * pr_ndopts
 *	Dump Neighbor Discovery options.
 */
void
pr_ndopts(cp, len)
	char *cp;
	int len;
{
	register struct nd_opt_hdr *xp;
	struct nd_opt_prefix_info *pp;
	extern char *ether_ntoa __P((struct ether_addr *));

#define ETHER(x)	((struct ether_addr *)x)

	if (len <= 0)
		return;
	xp = (struct nd_opt_hdr *)cp;
	if (xp->nd_opt_len == 0)
		return;
	cp += xp->nd_opt_len << 3;
	len -= xp->nd_opt_len << 3;
	if (len < 0)
		return;

	switch (xp->nd_opt_type) {
	case ND_OPT_SOURCE_LINKADDR:
		if (xp->nd_opt_len == 1)
			(void)printf("Source LLaddr %s\n",
				     ether_ntoa(ETHER(xp + 1)));
		else
			(void)printf("Source LLaddr\n");
		break;
	case ND_OPT_TARGET_LINKADDR:
		if (xp->nd_opt_len == 1)
			(void)printf("Target LLaddr %s\n",
				     ether_ntoa(ETHER(xp + 1)));
		else
			(void)printf("Target LLaddr\n");
		break;
	case ND_OPT_PREFIX_INFORMATION:
		pp = (struct nd_opt_prefix_info *)xp;
		(void)printf("Prefix Info %s/%d\n",
			     inet_ntop(AF_INET6, &pp->nd_opt_pi_prefix,
				       hname, sizeof(hname)),
			     pp->nd_opt_pi_prefix_len);
		break;
	case ND_OPT_REDIRECTED_HEADER:
		return;
	case ND_OPT_MTU:
	case ND_OPT_SHORTCUT_LIMIT:
	case ND_OPT_MOB_ADVERT_INTERVAL:
	case ND_OPT_MOB_AGENT_INFORMATION:
		break;
		
	default:
		(void)printf("Unknown ND option %d\n", xp->nd_opt_type);
		return;
	}
	pr_ndopts(cp, len);
}

/*
 * pr_addr --
 *	Return an ascii host address as a dotted quad and optionally with
 * a hostname.
 */
char *
pr_addr(addr)
	struct in6_addr *addr;
{
	struct hostent *hp;
	static char buf[80];
	int er;

	if ((options & F_NUMERIC) ||
	    !(hp = getipnodebyaddr(addr, sizeof(struct in6_addr),
				   AF_INET6, &er)))
		snprintf(buf, sizeof buf, "%s",
			 inet_ntop(AF_INET6, addr, hname, sizeof(hname)));
	else {
		snprintf(buf, sizeof buf, "%s (%s)", hp->h_name,
			 inet_ntop(AF_INET6, addr, hname, sizeof(hname)));
		freehostent(hp);
	}
	return(buf);
}

/*
 * pr_retip --
 *	Dump some info on a returned (via ICMPv6) IPv6 packet.
 */
void
pr_retip(ip)
	struct ip6_hdr *ip;
{
	u_char *cp;

	pr_iph(ip);
	cp = (u_char *)ip + sizeof(*ip);

	if (ip->ip6_nxt == 6)
		(void)printf("TCP: from port %u, to port %u (decimal)\n",
			     (*cp * 256 + *(cp + 1)),
			     (*(cp + 2) * 256 + *(cp + 3)));
	else if (ip->ip6_nxt == 17)
		(void)printf("UDP: from port %u, to port %u (decimal)\n",
			     (*cp * 256 + *(cp + 1)),
			     (*(cp + 2) * 256 + *(cp + 3)));
}

void
fill(bp, patp)
	char *bp, *patp;
{
	register int ii, jj, kk;
	int pat[16];
	char *cp;

	for (cp = patp; *cp; cp++)
		if (!isxdigit(*cp))
			errx(1, "patterns must be specified as hex digits");
	ii = sscanf(patp,
	    "%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x",
	    &pat[0], &pat[1], &pat[2], &pat[3], &pat[4], &pat[5], &pat[6],
	    &pat[7], &pat[8], &pat[9], &pat[10], &pat[11], &pat[12],
	    &pat[13], &pat[14], &pat[15]);

	if (ii > 0)
		for (kk = 0;
		    kk <= MAXPACKET - (8 + sizeof(struct timeval) + ii);
		    kk += ii)
			for (jj = 0; jj < ii; ++jj)
				bp[jj + kk] = pat[jj];
	if (!(options & F_QUIET)) {
		(void)printf("PATTERN: 0x");
		for (jj = 0; jj < ii; ++jj)
			(void)printf("%02x", bp[jj] & 0xFF);
		(void)printf("\n");
	}
}

char *
getsrcfor(destname)
	char *destname;
{
	struct sockaddr_in6 sin;
	int er, len = sizeof(sin);
	struct hostent *hp;

	if ((s = socket(AF_INET6, SOCK_DGRAM, 0)) < 0)
		err(1, "UDP socket");

	memset(&sin, 0, sizeof(sin));
	sin.sin6_len = sizeof(struct sockaddr_in6);
	hp = getipnodebyname(destname, AF_INET6, 0, &er);
	if (!hp)
		errx(1, "unknown host: %s", destname);
	sin.sin6_family = hp->h_addrtype;
	memcpy(&sin.sin6_addr, hp->h_addr, hp->h_length);
	sin.sin6_port = htons(NAMESERVER_PORT);
	freehostent(hp);

	if ((sin.sin6_family != AF_INET6) ||
	    (sin.sin6_len != sizeof(struct sockaddr_in6)) ||
	    IN6_IS_ADDR_V4MAPPED(&sin.sin6_addr))
		errx(1, "no IPv6 address for: %s", destname);

	if ((moptions & MULTICAST_IF) &&
	    setsockopt(s, IPPROTO_IPV6, IPV6_PKTINFO, &pktinfo,
		       sizeof(pktinfo)) < 0)
		err(1, "setsockopt IPV6_PKTINFO");

	printf("trying to get source for %s\n", destname);
	if (connect(s, (struct sockaddr *)&sin, len) < 0)
		err(1, "UDP connect");
	memset(&sin, 0, sizeof(sin));
	if (getsockname(s, (struct sockaddr *)&sin, &len) < 0)
		err(1, "getsockname");
	(void)close(s);
	if ((sin.sin6_family != AF_INET6) ||
	    (len != sizeof(struct sockaddr_in6)))
		err(1, "getsockname returns garbage");
	printf("source should be %s\n",
	       inet_ntop(AF_INET6, &sin.sin6_addr, hname, sizeof(hname)));

	return ((char *)inet_ntop(AF_INET6, &sin.sin6_addr,
				  hname, sizeof(hname)));
}

void
usage()
{
	(void)fprintf(stderr,
	    "usage: ping [-dfLnqrv] [-c count] [-i wait] [-l preload]\n\t[-S ifaddr] [-I ifname] [-p pattern] [-s packetsize]\n\t[-t ttl] [-w maxwait] host\n");
	exit(1);
}

