/*
 * Copyright (c) 1989, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Mike Muuss.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright (c) 1989, 1993\n\
	The Regents of the University of California.  All rights reserved.\n";
#endif /* not lint */

#ifndef lint
/*static char sccsid[] = "from: @(#)ping.c	8.1 (Berkeley) 6/5/93";*/
static char *rcsid = "$Id: ping.c,v 1.12 1994/12/18 00:20:51 cgd Exp $";
#endif /* not lint */

/*
 *			P R O B E 6 . C
 *
 * Using the IPv6 No Next Header facility, probe a network path.
 *
 */

#include <sys/param.h>
#include <sys/queue.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/signal.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip6.h>
#include <netinet/ip_var.h>
#include <netinet/ip6_var.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#define	DEFDATALEN	0		/* default data length */
#define	MAXPACKET	(65536 - 40)	/* max packet size */

/* various options */
int options;
#define	F_FLOOD		0x001
#define	F_INTERVAL	0x002
#define	F_FILLED	0x008
#define	F_QUIET		0x010
#define	F_SO_DEBUG	0x040
#define	F_SO_DONTROUTE	0x080
#define	F_VERBOSE	0x100

struct sockaddr_in6 fromwhere;	/* which source address to use */
struct sockaddr_in6 whereto;	/* who to probe */
int datalen = DEFDATALEN;
int s;				/* socket file descriptor */
u_char outpack[MAXPACKET];
char *myname;
char *hostname;
char hname[64];

/* counters */
long npackets;			/* max packets to transmit */
long ntransmitted;		/* sequence # for outbound packets = #sent */
int interval = 1;		/* interval between packets */

void fill __P((char *, char *));
void finish();
void prober();
void usage();

int
main(argc, argv)
	int argc;
	char **argv;
{
	extern int errno, optind;
	extern char *optarg;
	struct hostent *hp;
	struct sockaddr_in6 *from;
	struct sockaddr_in6 *to;
	register int i;
	int ch, hold, ttl;
	u_char *datap;
	char *source, *target;
	char lnamebuf[MAXHOSTNAMELEN], hnamebuf[MAXHOSTNAMELEN];

	source = NULL;
	datap = outpack;
	while ((ch = getopt(argc, argv, "S:c:dfi:p:qrs:t:")) != EOF)
		switch(ch) {
		case 'S':
			source = optarg;
			break;
		case 'c':
			npackets = atoi(optarg);
			if (npackets <= 0)
				errx(1, "bad number of packets to transmit: %s",
				    optarg);
			break;
		case 'd':
			options |= F_SO_DEBUG;
			break;
		case 'f':
			if (getuid())
				errx(1, "%s", strerror(EPERM));
			options |= F_FLOOD;
			break;
		case 'i':		/* wait between sending packets */
			interval = atoi(optarg);
			if (interval <= 0)
				errx(1, "bad timing interval: %s", optarg);
			options |= F_INTERVAL;
			break;
		case 'p':		/* fill buffer with user pattern */
			options |= F_FILLED;
			fill((char *)datap, optarg);
				break;
		case 'q':
			options |= F_QUIET;
			break;
		case 'r':
			options |= F_SO_DONTROUTE;
			break;
		case 's':		/* size of packet to send */
			datalen = atoi(optarg);
			if (datalen <= 0)
				errx(1, "bad packet size: %s", optarg);
			if (datalen > MAXPACKET)
				errx(1, "packet size too large: %s", optarg);
			break;
		case 't':
			ttl = atoi(optarg);
			if (ttl <= 0)
				errx(1, "bad ttl value: %s", optarg);
			if (ttl > 255)
				errx(1, "ttl value too large: %s", optarg);
			break;
		default:
			usage();
		}
	argc -= optind;
	argv += optind;

	if (argc != 1)
		usage();
	target = *argv;

	if (source) {
		memset(&fromwhere, 0, sizeof(struct sockaddr_in6));
		from = &fromwhere;
		from->sin6_len = sizeof(fromwhere);
		hp = getipnodebyname(source, AF_INET6, 0, &ch);
		if (!hp)
			errx(1, "unknown source address: %s", source);
		from->sin6_family = hp->h_addrtype;
		memcpy(&from->sin6_addr, hp->h_addr, hp->h_length);
		(void)strncpy(lnamebuf, hp->h_name, sizeof(lnamebuf) - 1);
		myname = lnamebuf;
		freehostent(hp);
		if ((from->sin6_family != AF_INET6) ||
		    (from->sin6_len != sizeof(struct sockaddr_in6)) ||
		    IN6_IS_ADDR_V4MAPPED(&from->sin6_addr))
			errx(1, "no IPv6 address for source: %s", source);
	}

	memset(&whereto, 0, sizeof(struct sockaddr_in6));
	to = &whereto;
	to->sin6_len = sizeof(whereto);
	hp = getipnodebyname(target, AF_INET6, 0, &ch);
	if (!hp)
		errx(1, "unknown host: %s", target);
	to->sin6_family = hp->h_addrtype;
	memcpy(&to->sin6_addr, hp->h_addr, hp->h_length);
	(void)strncpy(hnamebuf, hp->h_name, sizeof(hnamebuf) - 1);
	hostname = hnamebuf;
	freehostent(hp);
	if ((to->sin6_family != AF_INET6) ||
	    (to->sin6_len != sizeof(struct sockaddr_in6)) ||
	    IN6_IS_ADDR_V4MAPPED(&to->sin6_addr))
		errx(1, "no IPv6 address for host: %s", target);

	if (options & F_FLOOD && options & F_INTERVAL)
		errx(1, "-f and -i options are incompatible");

	if (!(options & F_FILLED))
		for (i = 0; i < datalen; ++i)
			*datap++ = i;

	if ((s = socket(AF_INET6, SOCK_RAW, IP6_NHDR_NONH)) < 0)
		err(1, "socket");
	hold = 1;
	if (options & F_SO_DEBUG)
		(void)setsockopt(s, SOL_SOCKET, SO_DEBUG, (char *)&hold,
		    sizeof(hold));
	if (options & F_SO_DONTROUTE)
		(void)setsockopt(s, SOL_SOCKET, SO_DONTROUTE, (char *)&hold,
		    sizeof(hold));
	if (source &&
	    bind(s, (struct sockaddr *)&fromwhere, sizeof(fromwhere)) < 0) {
		perror("probe6: bind");
		exit(1);
	}

	if (to->sin6_family == AF_INET6)
		(void)printf("PROBE %s (%s): %d data bytes\n", hostname,
		    inet_ntop(AF_INET6, &to->sin6_addr, hname, sizeof(hname)),
		    datalen);
	else
		(void)printf("PROBE %s: %d data bytes\n", hostname, datalen);

	(void)signal(SIGINT, finish);

	for (; !npackets || ntransmitted < npackets; ) {
		prober();
		if (options & F_FLOOD)
			usleep(10000);
		else
			sleep(interval);
	}
	finish();
	exit(0);	/* Make the compiler happy */
}

/*
 * prober --
 *	Compose and transmit a probe packet.
 * The IPv6 packet will be added on by the kernel.
 */
void
prober()
{
	register int cc;
	int i;

	cc = datalen;
	ntransmitted++;
	i = sendto(s, (char *)outpack, cc, 0,
		   (struct sockaddr *)&whereto, sizeof(whereto));

	if (i < 0 || i != cc)  {
		if (i < 0)
			perror("probe6: sendto");
		(void)printf("probe6: wrote %s %d chars, ret=%d\n",
		    hostname, cc, i);
	}
}

/*
 * finish --
 *	Print out statistics, and give up.
 */
void
finish()
{
	(void)signal(SIGINT, SIG_IGN);
	(void)putchar('\n');
	(void)fflush(stdout);
	(void)printf("--- %s probe6 statistics ---\n", hostname);
	(void)printf("%ld packets transmitted, ", ntransmitted);
	(void)putchar('\n');
	exit(0);
}

void
fill(bp, patp)
	char *bp, *patp;
{
	register int ii, jj, kk;
	int pat[16];
	char *cp;

	for (cp = patp; *cp; cp++)
		if (!isxdigit(*cp))
			errx(1, "patterns must be specified as hex digits");
	ii = sscanf(patp,
	    "%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x",
	    &pat[0], &pat[1], &pat[2], &pat[3], &pat[4], &pat[5], &pat[6],
	    &pat[7], &pat[8], &pat[9], &pat[10], &pat[11], &pat[12],
	    &pat[13], &pat[14], &pat[15]);

	if (ii > 0)
		for (kk = 0;
		    kk <= MAXPACKET;
		    kk += ii)
			for (jj = 0; jj < ii; ++jj)
				bp[jj + kk] = pat[jj];
	if (!(options & F_QUIET)) {
		(void)printf("PATTERN: 0x");
		for (jj = 0; jj < ii; ++jj)
			(void)printf("%02x", bp[jj] & 0xFF);
		(void)printf("\n");
	}
}

void
usage()
{
	(void)fprintf(stderr,
	    "usage: probe6 [-dfqr] [-c count] [-i wait]\n\t[-S source] [-p pattern] [-s packetsize] [-t ttl] host\n");
	exit(1);
}
