/*	$NetBSD: arp.c,v 1.11 1995/03/01 11:56:13 chopps Exp $ */

/*
 * Copyright (c) 1984, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Sun Microsystems, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
#ifndef lint
__COPYRIGHT("@(#) Copyright (c) 1984, 1993\n\
	The Regents of the University of California.  All rights reserved.\n");
#endif /* not lint */

#ifndef lint
/*static char sccsid[] = "from: @(#)arp.c	8.2 (Berkeley) 1/2/94";*/
__RCSID("$NetBSD: arp.c,v 1.11 1995/03/01 11:56:13 chopps Exp $");
#endif /* not lint */

/*
 * ndp - display, set, and delete ND protocol table entries
 */

#include <sys/param.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <sys/sysctl.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_types.h>
#include <net/route.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/if_ndp6.h>
#include <arpa/inet.h>

#include <netdb.h>
#include <errno.h>
#include <err.h>
#include <nlist.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <paths.h>

int main(int, char *[]);
int delete(const char *);
void dump(struct in6_addr *);
int loc_ether_aton(const char *, u_char *);
void ether_print(const u_char *);
int file(char *);
void get(const char *);
int getinetaddr(const char *, struct in6_addr *);
void getsocket(void);
int rtmsg(int);
int set(const char *, char **);
void usage(void);

static int pid;
static int nflag;
static int s = -1;
static int ifindex = 0;
static const char *host;
static char cmd = 'g';

int
main(argc, argv)
	int argc;
	char **argv;
{
	int ch;
	extern char *optarg;
	extern int optind;

	pid = getpid();
	while ((ch = getopt(argc, argv, "and:s:S:i:")) != EOF)
		switch((char)ch) {
		case 'd':
		case 's':
		case 'S':
			host = optarg;
		case 'a':
			cmd = ch;
			break;
		case 'i':
			if ((ifindex = atoi(optarg)) <= 0)
				usage();
			break;
		case 'n':
			nflag = 1;
			break;
		case '?':
		default:
			usage();
		}
	argc -= optind;
	argv += optind;
	
	switch (cmd) {
	case 'g':
		if (argc != 1)
			usage();
		get(argv[0]);
		break;

	case 'a':
		if (argc != 0)
			usage();
		dump(NULL);
		break;

	default:
		if ((cmd == 'd') || (cmd == 'S'))
			(void) delete(host);
		if (cmd == 'd')
			break;
		if (argc == 0)
			usage();
		return (set(host, argv) ? 1 : 0);
	}
	return (0);
}

void
getsocket(void)
{
	if (s >= 0)
		return;
	s = socket(PF_ROUTE, SOCK_RAW, 0);
	if (s < 0)
		err(1, "socket");
}

struct	sockaddr_in6 blank_sin = {sizeof(blank_sin), AF_INET6 }, sin_m;
struct	sockaddr_dl blank_sdl = {sizeof(blank_sdl), AF_LINK }, sdl_m;
int	expire_time, found_entry;
struct	{
	struct	rt_msghdr m_rtm;
	char	m_space[512];
}	m_rtmsg;

/*
 * Set an individual ndp entry 
 */
int
set(const char *host, char **argv)
{
	register struct sockaddr_in6 *sin;
	register struct sockaddr_dl *sdl;
	register struct rt_msghdr *rtm;
	u_char *ea;
	char *eaddr;

	sin = &sin_m;
	rtm = &(m_rtmsg.m_rtm);
	eaddr = argv[0];

	getsocket();
	sdl_m = blank_sdl;		/* struct copy */
	sin_m = blank_sin;		/* struct copy */
	if (getinetaddr(host, &sin->sin6_addr) == -1)
		return (1);
	sdl_m.sdl_index = ifindex;
	expire_time = 0;
	if (argv[1] && (strncmp(argv[1], "temp", 4) == 0)) {
		struct timeval time;
		(void)gettimeofday(&time, 0);
		expire_time = time.tv_sec + 2 * 60;
	}
	ea = (u_char *)LLADDR(&sdl_m);
	if (loc_ether_aton(eaddr, ea) == 0)
		sdl_m.sdl_alen = 6;

	if (rtmsg(RTM_GET) < 0) {
		warn("%s", host);
		return (1);
	}
	sin = (struct sockaddr_in6 *)(rtm + 1);
	sdl = (struct sockaddr_dl *)(sin->sin6_len + (char *)sin);
	if (IN6_ARE_ADDR_EQUAL(&sin->sin6_addr, &sin_m.sin6_addr)) {
		if (sdl->sdl_family == AF_LINK &&
		    (rtm->rtm_flags & RTF_LLINFO) &&
		    !(rtm->rtm_flags & RTF_GATEWAY)) switch (sdl->sdl_type) {
		case IFT_ETHER: case IFT_FDDI: case IFT_ISO88023:
		case IFT_ISO88024: case IFT_ISO88025:
			goto overwrite;
		}
		warn("bad network type (%d)\n", sdl->sdl_type);
		return (1);
	}
overwrite:
	if (sdl->sdl_family != AF_LINK) {
		(void)printf("cannot intuit interface index and type for %s\n",
		    host);
		return (1);
	}
	sdl_m.sdl_type = sdl->sdl_type;
	sdl_m.sdl_index = sdl->sdl_index;
	return (rtmsg(RTM_ADD));
}

/*
 * Display an individual ndp entry
 */
void
get(const char *host)
{
	struct sockaddr_in6 *sin;
	char hname[64];

	sin = &sin_m;
	sin_m = blank_sin;		/* struct copy */
	if (getinetaddr(host, &sin->sin6_addr) == -1)
		exit(1);
	dump(&sin->sin6_addr);
	if (found_entry == 0) {
		(void)printf("%s (%s) -- no entry\n", host,
		    inet_ntop(AF_INET6, &sin->sin6_addr, hname, 64));
		exit(1);
	}
}

/*
 * Delete a ndp entry 
 */
int
delete(const char *host)
{
	register struct sockaddr_in6 *sin;
	register struct rt_msghdr *rtm;
	struct sockaddr_dl *sdl;
	char hname[64];

	sin = &sin_m;
	rtm = &m_rtmsg.m_rtm;

	getsocket();
	sin_m = blank_sin;		/* struct copy */
	if (getinetaddr(host, &sin->sin6_addr) == -1)
		return (1);

	if (rtmsg(RTM_GET) < 0) {
		warn("%s", host);
		return (1);
	}
	sin = (struct sockaddr_in6 *)(rtm + 1);
	sdl = (struct sockaddr_dl *)(sin->sin6_len + (char *)sin);
	if (IN6_ARE_ADDR_EQUAL(&sin->sin6_addr, &sin_m.sin6_addr)) {
		if (sdl->sdl_family == AF_LINK &&
		    (rtm->rtm_flags & RTF_LLINFO) &&
		    !(rtm->rtm_flags & RTF_GATEWAY)) switch (sdl->sdl_type) {
		case IFT_ETHER: case IFT_FDDI: case IFT_ISO88023:
		case IFT_ISO88024: case IFT_ISO88025:
			goto delete;
		}
	}
	warnx("delete: can't locate %s", host);
	return (1);
delete:
	if (sdl->sdl_family != AF_LINK) {
		(void)printf("cannot locate %s\n", host);
		return (1);
	}
	if (rtmsg(RTM_DELETE)) 
		return (1);
	(void)printf("%s (%s) deleted\n", host,
		     inet_ntop(AF_INET6, &sin->sin6_addr, hname, 64));
	return (0);
}

/*
 * Dump the entire ndp table
 */
void
dump(struct in6_addr *addr)
{
	int mib[6];
	size_t needed;
	char *host, *lim, *buf, *next;
	struct rt_msghdr *rtm;
	struct sockaddr_in6 *sin;
	struct sockaddr_dl *sdl;
	int er;
	struct hostent *hp;

	mib[0] = CTL_NET;
	mib[1] = PF_ROUTE;
	mib[2] = 0;
	mib[3] = AF_INET6;
	mib[4] = NET_RT_FLAGS;
	mib[5] = RTF_LLINFO;
	if (sysctl(mib, 6, NULL, &needed, NULL, 0) < 0)
		err(1, "route-sysctl-estimate");
	if ((buf = malloc(needed)) == NULL)
		err(1, "malloc");
	if (sysctl(mib, 6, buf, &needed, NULL, 0) < 0)
		err(1, "actual retrieval of routing table");
	lim = buf + needed;
	for (next = buf; next < lim; next += rtm->rtm_msglen) {
		char hbuf[64];

		rtm = (struct rt_msghdr *)next;
		sin = (struct sockaddr_in6 *)(rtm + 1);
		sdl = (struct sockaddr_dl *)(sin + 1);
		if (sdl->sdl_family != AF_LINK)
			continue;
		if (addr) {
			if (!IN6_ARE_ADDR_EQUAL(addr, &sin->sin6_addr))
				continue;
			found_entry = 1;
		} else if (IN6_IS_ADDR_MULTICAST(&sin->sin6_addr))
			continue;
		hp = 0;
		er = 0;
		if (nflag == 0)
			hp = getipnodebyaddr((char *)&sin->sin6_addr,
			    sizeof sin->sin6_addr, AF_INET6, &er);
		if (hp)
			host = hp->h_name;
		else {
			host = "?";
			if (er == TRY_AGAIN)
				nflag = 1;
		}
		(void)printf("%s (%s) at link#%d ", host,
			     inet_ntop(AF_INET6, &sin->sin6_addr,
				       hbuf, sizeof(hbuf)),
			     sdl->sdl_index);
		if (hp)
			freehostent(hp);
		if (sdl->sdl_alen)
			ether_print(LLADDR(sdl));
		else
			(void)printf("(incomplete)");
		if (rtm->rtm_rmx.rmx_expire == 0)
			(void)printf(" permanent");
		if (rtm->rtm_addrs & RTA_NETMASK) {
			sin = (struct sockaddr_in6 *)
				(sdl->sdl_len + (char *)sdl);
			if (!IN6_IS_ADDR_UNSPECIFIED(&sin->sin6_addr))
				(void)printf(" published");
			if (sin->sin6_len != 24)
				(void)printf("(wierd)");
		}
		(void)printf("\n");
	}
}

void
ether_print(const u_char *cp)
{
	(void)printf("%x:%x:%x:%x:%x:%x", cp[0], cp[1], cp[2], cp[3], cp[4],
	    cp[5]);
}

int
loc_ether_aton(const char *a, u_char *n)
{
	int i, o[6];

	i = sscanf(a, "%x:%x:%x:%x:%x:%x", &o[0], &o[1], &o[2], &o[3], &o[4],
	    &o[5]);
	if (i != 6) {
		warnx("invalid Ethernet address '%s'", a);
		return (1);
	}
	for (i=0; i<6; i++)
		n[i] = o[i];
	return (0);
}

void
usage(void)
{
	(void)fprintf(stderr, "usage: ndp [-n] hostname\n");
	(void)fprintf(stderr, "usage: ndp [-n] -a\n");
	(void)fprintf(stderr, "usage: ndp -i if_index -d hostname\n");
	(void)fprintf(stderr, "usage: ndp -i if_index -s hostname ether_addr [temp]\n");
	(void)fprintf(stderr, "usage: ndp -i if_index -S hostname ether_addr [temp]\n");
	exit(1);
}

int
rtmsg(int cmd)
{
	static int seq;
	int rlen;
	register struct rt_msghdr *rtm;
	register char *cp;
	register int l;

	rtm = &m_rtmsg.m_rtm;
	cp = m_rtmsg.m_space;
	errno = 0;

	if (cmd == RTM_DELETE)
		goto doit;
	(void)memset(&m_rtmsg, 0, sizeof(m_rtmsg));
	rtm->rtm_version = RTM_VERSION;
	rtm->rtm_index = ifindex;

	switch (cmd) {
	default:
		errx(1, "internal wrong cmd");
		/*NOTREACHED*/
	case RTM_ADD:
		rtm->rtm_addrs |= RTA_GATEWAY;
		rtm->rtm_rmx.rmx_expire = expire_time;
		rtm->rtm_inits = RTV_EXPIRE;
		rtm->rtm_flags |= (RTF_HOST | RTF_DYNAMIC);
		/* FALLTHROUGH */
	case RTM_GET:
		rtm->rtm_addrs |= RTA_DST;
	}
#define NEXTADDR(w, s) \
	if (rtm->rtm_addrs & (w)) { \
		(void)memcpy(cp, &s, sizeof(s)); cp += sizeof(s);}

	NEXTADDR(RTA_DST, sin_m);
	NEXTADDR(RTA_GATEWAY, sdl_m);

	rtm->rtm_msglen = cp - (char *)&m_rtmsg;
doit:
	l = rtm->rtm_msglen;
	rtm->rtm_seq = ++seq;
	rtm->rtm_type = cmd;
	if ((rlen = write(s, (char *)&m_rtmsg, l)) < 0) {
		if (errno != ESRCH || cmd != RTM_DELETE) {
			warn("writing to routing socket");
			return (-1);
		}
	}
	do {
		l = read(s, (char *)&m_rtmsg, sizeof(m_rtmsg));
	} while (l > 0 && (rtm->rtm_seq != seq || rtm->rtm_pid != pid));
	if (l < 0)
		warn("read from routing socket");
	return (0);
}

int
getinetaddr(const char *host, struct in6_addr *inap)
{
	struct hostent *hp;
	int er;

	hp = getipnodebyname(host, AF_INET6, 0, &er);
	if (hp == NULL) {
		warnx("%s: %s", host, hstrerror(er));
		return (-1);
	}
	(void)memcpy(inap, hp->h_addr, sizeof(*inap));
	freehostent(hp);
	return (0);
}
