#include "ndpd.h"

struct timeval last_rtadv;
struct {
	struct cmsghdr c_cm;
	u_char c_space[64];
} c_cmsg;
#define cm	c_cmsg.c_cm

#define	MAXPACKET	8192	/* max packet size */
char packet[MAXPACKET];
#define RTA_REDIRECT	(RTA_DST | RTA_GATEWAY | RTA_IFA)

/* Send an ICMPv6 Router Solicitation packet on ifp. */
void
send_rtsol(ifp)
	struct interface *ifp;
{
	register struct ip6_hdr *ip;
	register struct nd_router_solicit *rsp;
	register struct nd_opt_hdr *hp;
	struct sockaddr_in6 dst;
	int len, tlen, cc, opt;

	if (ifp->tp) {
		kill_timeout(ifp->tp);
		ifp->tp = NULL;
	}
	if (!IS_UP(ifp))
		return;
	if (ifp->ntosend < 0)
		return;
	else if (ifp->ntosend == 0) {
		ifp->ntosend = -1;
		return;
	}

	bzero((caddr_t)&dst, sizeof(struct sockaddr_in6));
	dst.sin6_len = sizeof(struct sockaddr_in6);
	dst.sin6_family = AF_INET6;
	dst.sin6_addr.s6_addr[0] = 0xff;
	dst.sin6_addr.s6_addr[1] = MADDR6_SCP_LINK;
	dst.sin6_addr.s6_addr[15] = MADDR6_ALLROUTERS;
	dst.sin6_scope_id = ifp->index;

	if (verbose) {
		logtrace("Sending router sol for %s on %s",
			 inet6pr(&ifp->llocal), ifp->name);
	}

	len = sizeof(struct nd_router_solicit);
	if (ifp->hasphy)
		len += ND_OPT_LINKADDR_LEN;
	tlen = len + sizeof(struct ip6_hdr);
	bzero(packet, tlen);

	ip = (struct ip6_hdr *)packet;
	COPY_ADDR6(ifp->llocal, ip->ip6_src);
	COPY_ADDR6(dst.sin6_addr, ip->ip6_dst);
	((struct ip6ovck *)ip)->ih6_pr = IPPROTO_ICMPV6;
	((struct ip6ovck *)ip)->ih6_len = htons(len);

	rsp = (struct nd_router_solicit *)(ip + 1);
	rsp->nd_rs_type = ND_ROUTER_SOLICIT;

	/* other fields are 0 */

	if (ifp->hasphy) {
		hp = (struct nd_opt_hdr *)(rsp + 1);
		hp->nd_opt_type = ND_OPT_SOURCE_LINKADDR;
		hp->nd_opt_len = 1;
		bcopy(ifp->adphy, hp + 1, ifp->lmac);
	}

	/* Compute ICMPv6 checksum here */
	rsp->nd_rs_cksum = in_cksum((u_short *)ip, tlen);

	/* finish IPv6 header */
	ip->ip6_flow = ND_DEFAULT_CLASS | IPV6_VERSION;
	ip->ip6_plen = htons(len);
	ip->ip6_nxt = IPPROTO_ICMPV6;
	ip->ip6_hlim = ND_DEFAULT_HOPLIM;

	opt = ifp->index;
	if (setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_IF,
		       (char *)&opt, sizeof(opt)) < 0) {
		logperror("setsockopt (IPV6_MULTICAST_IF)");
		return;
	}

	cc = sendto(s, packet, tlen, 0,
		    (struct sockaddr *)&dst, sizeof (struct sockaddr_in6));

	if (cc < 0 || cc != tlen)  {
		logerr("wrote %d chars, ret=%d", tlen, cc);
		if (cc < 0) {
			logperror("sendto");
			ifp->ntosend++;		/* keep trying */
		}
	}

	ifp->ntosend--;
	if (ifp->ntosend < 0) {
/* if (ifp->statefeful) set a timemout to start stateful inquiry */
		return;
	}

	(void) gettimeofday(&now, NULL);
	ifp->tp = new_timeout(RTR_SOLICITATION_INTERVAL, 0,
			      (func_to) send_rtsol, ifp);
	if (ifp->tp == NULL)
		abort();
}

/* Send a proxy gratuitous neighbor advertisement */
void
send_grat_adv(pp)
	struct prefix *pp;
{
	register struct ip6_hdr *ip;
	register struct nd_neighbor_advert *nap;
	register struct nd_opt_hdr *hp;
	struct sockaddr_in6 dst;
	int len, tlen, cc, opt;
	long delay;
	struct interface *ifp = pp->ifp;

	if (pp->gtp) {
		kill_timeout(pp->gtp);
		pp->gtp = NULL;
	}
	if (!IS_UP(ifp) || !ifp->hasphy)
		return;
	if (pp->ntosend <= 0)
		return;

	bzero((caddr_t)&dst, sizeof(struct sockaddr_in6));
	dst.sin6_len = sizeof(struct sockaddr_in6);
	dst.sin6_family = AF_INET6;
	dst.sin6_addr.s6_addr[0] = 0xff;
	dst.sin6_addr.s6_addr[1] = MADDR6_SCP_LINK;
	dst.sin6_addr.s6_addr[15] = MADDR6_ALLNODES;
	dst.sin6_scope_id = ifp->index;

	if (verbose) {
		logtrace("Sending gratuitous nd adv for %s on %s",
			 inet6pr(&pp->addr), ifp->name);
	}

	len = sizeof(*nap) + ND_OPT_LINKADDR_LEN;
	tlen = len + sizeof(struct ip6_hdr);
	bzero(packet, tlen);

	ip = (struct ip6_hdr *)packet;
	COPY_ADDR6(ifp->llocal, ip->ip6_src);
	COPY_ADDR6(dst.sin6_addr, ip->ip6_dst);
	((struct ip6ovck *)ip)->ih6_pr = IPPROTO_ICMPV6;
	((struct ip6ovck *)ip)->ih6_len = htons(len);

	nap = (struct nd_neighbor_advert *)(ip + 1);
	nap->nd_na_type = ND_NEIGHBOR_ADVERT;
	nap->nd_na_flags_reserved = ND_NA_FLAG_OVERRIDE;
	COPY_ADDR6(pp->addr, nap->nd_na_target);

	hp = (struct nd_opt_hdr *)(nap + 1);
	hp->nd_opt_type = ND_OPT_TARGET_LINKADDR;
	hp->nd_opt_len = 1;
	bcopy(ifp->adphy, hp + 1, ifp->lmac);

	/* Compute ICMPv6 checksum here */
	nap->nd_na_cksum = in_cksum((u_short *)ip, tlen);

	/* finish IPv6 header */
	ip->ip6_flow = ND_DEFAULT_CLASS | IPV6_VERSION;
	ip->ip6_plen = htons(len);
	ip->ip6_nxt = IPPROTO_ICMPV6;
	ip->ip6_hlim = ND_DEFAULT_HOPLIM;

	opt = ifp->index;
	if (setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_IF,
		       (char *)&opt, sizeof(opt)) < 0) {
		logperror("setsockopt (IPV6_MULTICAST_IF)");
		return;
	}

	cc = sendto(s, packet, tlen, 0,
		    (struct sockaddr *)&dst, sizeof (struct sockaddr_in6));

	if (cc < 0 || cc != tlen) {
		logerr("wrote %d chars, ret=%d", tlen, cc);
		if (cc < 0)
			logperror("sendto");
	}

	pp->ntosend--;
	if (pp->ntosend == 0)
		return;
	delay = random() % MAX_ANYCAST_DELAY_TIME;

	(void) gettimeofday(&now, NULL);

	pp->gtp = new_timeout(MS_TO_SEC(delay), MS_TO_USEC(delay),
			      (func_to) send_grat_adv, pp);
}

/* Receive an ICMPv6 (Router Advertisement or Redirect) packet */
/* TODO: mobiles need some ICMPv6 errors like unreachables */
void
recv_icmp()
{
	register struct ip6_hdr *ip;
	register struct icmp6_hdr *icp;
	struct interface *ifp;
	struct msghdr msg;
	struct iovec iov;
	struct in6_pktinfo pi;
	int cc;

	bzero((caddr_t)&msg, sizeof(struct msghdr));
	bzero((caddr_t)&iov, sizeof(struct iovec));
	msg.msg_iov = &iov;
	msg.msg_iov->iov_base = packet;
	msg.msg_iov->iov_len = MAXPACKET;
	msg.msg_iovlen = 1;
	msg.msg_control = (caddr_t)&c_cmsg;
	msg.msg_controllen = sizeof(c_cmsg);

	if ((cc = recvmsg(s, &msg, 0)) < 0) {
		if (verbose)
			logperror("read (ICMPv6)");
		return;
	}
	if (cc == 0) {
		if (verbose)
			logerr("read ICMPv6 return 0");
		return;
	}
	if (msg.msg_flags & (MSG_TRUNC|MSG_CTRUNC)) {
		if (verbose)
			logerr("read ICMPv6 truncated");
		return;
	}
	bcopy(c_cmsg.c_space, &pi, sizeof(pi));
	if ((msg.msg_controllen == 0) ||
	    (cm.cmsg_len != (sizeof(struct cmsghdr) + sizeof(pi))) ||
	    (cm.cmsg_level != IPPROTO_IPV6) ||
	    (cm.cmsg_type != IPV6_PKTINFO) ||
	    (pi.ipi6_ifindex == 0)) {
		if (verbose)
			logerr("read ICMPv6 doesn't give interface");
		return;
	}

	FORALLIFP(ifp)
		if (ifp->index == pi.ipi6_ifindex)
			break;
	if (ifp == NULL || !IS_UP(ifp))
		return;

	ip = (struct ip6_hdr *)packet;
	if (cc < sizeof(*ip) + sizeof(*icp)) {
		if (verbose)
			logerr("ICMPv6: too short (%d)", cc);
		return;
	}
	if (!IN6_IS_ADDR_LINKLOCAL(&ip->ip6_src) ||
	    (ip->ip6_hlim != ND_DEFAULT_HOPLIM))
		return;

	cc -= sizeof(struct ip6_hdr);
	icp = (struct icmp6_hdr *)(ip + 1);

	switch (icp->icmp6_type) {
	    case ND_REDIRECT:
		recv_redirect(ifp, ip, cc);
		break;
	    case ND_ROUTER_ADVERT:
		recv_rtadv(ifp, ip, cc);
		break;
	    /* Other types are filtered */
	}
	return;
}

void
recv_rtadv(ifp, ip, len)
	struct interface *ifp;
	register struct ip6_hdr *ip;
	int len;
{
	register struct nd_router_advert *rap;
	register struct nd_opt_hdr *hp;
	struct nd_opt_prefix_info *pip;
	struct nd_opt_mtu *mp;
	struct prefix *pp, *pp2;
	struct router *rp;
	struct in6_addr mask, rta;
	int newthings = 0, gotrta = 0, changed = 0, pass = 1;
	u_int v;
	u_int32_t mtu;
	struct ifreq ifr;

	rap = (struct nd_router_advert *)(ip + 1);
	if (rap->nd_ra_code != 0) {
		if (verbose)
			logerr("ICMPv6 Rt Adv: code=%d", rap->nd_ra_code);
		return;
	}
	if ((unsigned)len < sizeof(*rap)) {
		if (verbose)
			logerr("ICMPv6 Rt Adv: too short %d", len);
			return;
	}

	if (verbose) {
		logtrace("receive ICMPv6 Rt Adv on %s from %s to %s (len=%d)",
			 ifp->name,
			 inet6pr(&ip->ip6_src),
			 inet6pr(&ip->ip6_dst),
			 len + sizeof(struct ip6_hdr));
	}
	(void) gettimeofday(&now, NULL);
	last_rtadv = now;

	if (mobile) {
		if (ifp->ntosend >= MAX_RTR_SOLICITATIONS)
			ifp->ntosend = 1;
	} else {
		if (ifp->tp) {
			kill_timeout(ifp->tp);
			ifp->tp = NULL;
		}
		ifp->ntosend = -1;
	}

	FORALLRP(rp, ifp)
		if (SAME_ADDR6(rp->addr.sin6_addr, ip->ip6_src))
			break;
	if (rp == NULL) {
		if (verbose)
			logtrace("allocate router");

		if (Calloc(rp, router) == NULL) {
			if (verbose)
				logerr("out of memory");
			return;
		}
		rp->ifp = ifp;

		rp->addr.sin6_len = sizeof(struct sockaddr_in6);
		rp->addr.sin6_family = AF_INET6;
		COPY_ADDR6(ip->ip6_src, rp->addr.sin6_addr);
		rp->addr.sin6_scope_id = ifp->index;

		rp->phy.sdl_len = sizeof(struct sockaddr_dl);
		rp->phy.sdl_family = AF_LINK;
		rp->phy.sdl_index = ifp->index;
		if (!ifp->hasphy)
			rp->phy.sdl_alen = NOPHY;

		TAILQ_INSERT_TAIL(&ifp->routers, rp, rt_list);
		newthings = 1;
	}
	cur_rt = rp;
	rp->last = last_rtadv;
	rp->lifetime = ntohs(rap->nd_ra_router_lifetime);
	rp->nflags = rap->nd_ra_flags_reserved;
	rp->nflags &= ND_RA_FLAG_ALL;
	if (verbose)
		logtrace("router %s %x %d",
			 inet6pr(&rp->addr.sin6_addr),
			 (int)rap->nd_ra_flags_reserved, (int)rp->lifetime);

	if ((v = rap->nd_ra_curhoplimit) && CurHopLimit != v)
		setCurHopLimit(v);
	if ((v = ntohl(rap->nd_ra_reachable)) && BaseReachableTime != v)
		setReachableTime(v);
	if ((v = ntohl(rap->nd_ra_retransmit)) && RetransTimer != v)
		setRetransTimer(v);

	len -= sizeof(*rap);
	hp = (struct nd_opt_hdr *)(rap + 1);
	while (len) switch (hp->nd_opt_type) {
	    case ND_OPT_SOURCE_LINKADDR:
		if (!ifp->hasphy) {
			if (verbose)
				logtrace("unexpected source lladdr");
			goto next;
		}
		if (hp->nd_opt_len != ((ifp->lmac+7)/8)) {
			if (verbose)
				logerr("bad LL address length (%d)",
				       hp->nd_opt_len << 3);
			goto bad;
		}
		if (len < (hp->nd_opt_len << 3)) {
			if (verbose)
				logerr("Rt Adv underflow");
			goto bad;
		}
		rp->phy.sdl_alen = ifp->lmac;
		bcopy(hp + 1, LLADDR(&rp->phy), ifp->lmac);
		goto next;

	    case ND_OPT_PREFIX_INFORMATION:
		pip = (struct nd_opt_prefix_info *)hp;
		/* check length */
		if (pip->nd_opt_pi_len != 4) {
			if (verbose)
				logerr("bad PREF length (%d)",
				       pip->nd_opt_pi_len << 3);
			goto next;
		}
		if (len < (pip->nd_opt_pi_len << 3)) {
			if (verbose)
				logerr("Rt Adv underflow");
			goto next;
		}

		/* new router address flag */
		if (pip->nd_opt_pi_flags_reserved & ND_OPT_PI_FLAG_RTADDR) {
			COPY_ADDR6(pip->nd_opt_pi_prefix, rta);
			gotrta |= 1;
		}
		mk_mask(&mask, pip->nd_opt_pi_prefix_len);
		apply_mask(&pip->nd_opt_pi_prefix, &mask);

		/* ignore if no flag is set */
		if ((pip->nd_opt_pi_flags_reserved & ND_OPT_PI_FLAG_CFG) == 0)
			goto next;
		/* ignore strange and link-local */
		if ((pip->nd_opt_pi_prefix.s6_addr[0] == 0) ||
		    (pip->nd_opt_pi_prefix.s6_addr[0] == 0xff) ||
		    IN6_IS_ADDR_LINKLOCAL(&pip->nd_opt_pi_prefix)) {
			if (verbose)
				logerr("bad PREF prefix %s",
				       inet6pr(&pip->nd_opt_pi_prefix));
			goto next;
		}
		/* RFCs say plen + ltoken == 128 */
		if ((pip->nd_opt_pi_prefix_len > 128) ||
		    (pip->nd_opt_pi_prefix_len < 16) ||
		    ((pip->nd_opt_pi_prefix_len != 128) &&
		     (pip->nd_opt_pi_prefix_len > (128 - 8 * ifp->ltoken)))) {
			if (verbose)
				logerr("strange PREF plen (%d)",
				       pip->nd_opt_pi_prefix_len);
			goto next;
		}
		pip->nd_opt_pi_valid_time = ntohl(pip->nd_opt_pi_valid_time);
		pip->nd_opt_pi_preferred_time =
			ntohl(pip->nd_opt_pi_preferred_time);
		if (pip->nd_opt_pi_valid_time <
				pip->nd_opt_pi_preferred_time) {
			if (verbose)
				logerr("bad PREF ilife = %d, dlife = %d",
				       pip->nd_opt_pi_valid_time,
				       pip->nd_opt_pi_preferred_time);
			goto next;
		}
		/* ignore pref si not stateless or statefull ?? */
		if ((ifp->state != STATELESS) && (ifp->state != STATEFULL))
			goto next;

		if (verbose)
			logtrace("prefix %s/%d %x (%u/%u)",
				 inet6pr(&pip->nd_opt_pi_prefix),
				 pip->nd_opt_pi_prefix_len,
				 pip->nd_opt_pi_flags_reserved,
				 pip->nd_opt_pi_valid_time,
				 pip->nd_opt_pi_preferred_time);

		FORALLPP(pp, ifp)
			if (SAME_ADDR6(pip->nd_opt_pi_prefix, pp->pref) &&
			    (pip->nd_opt_pi_prefix_len == pp->plen))
				break;
		if (pp == NULL) {
			if (verbose)
				logtrace("allocate prefix");

			if (Calloc(pp, prefix) == NULL) {
				if (verbose)
					logerr("out of memory");
				goto bad;
			}
			pp->ifp = ifp;
			COPY_ADDR6(pip->nd_opt_pi_prefix, pp->pref);
			pp->plen = pip->nd_opt_pi_prefix_len;
			mk_pref(pp);
			mk_addr(pp);
			/*
			 * search again:
			 * it may be a known prefix (from startup)
			 * but with different flags (A changed to AL)
			 * or a fake home address
			 */
			FORALLPP(pp2, ifp) {
				if (pp2->plen != 128 ||
				    !SAME_ADDR6(pp2->pref, pp->addr))
					continue;
				if ((pp2->pflags & ND_OPT_PI_FLAG_ONLINK) == 0)
					break;
			}
			if (pp2) {
				pp2->plen = pp->plen;
				mk_pref(pp2);
				free(pp);
				pp = pp2;
				if ((pp->mflags & MOB_ISFA) == 0)
					newthings = 0;
				if (verbose)
					logtrace("reuse prefix");
			} else {
				/* ignore unknown and invalid */
				if (pip->nd_opt_pi_valid_time == 0) {
					free(pp);
					goto next;
				}
				if (FIRST(ifp) == NULL)
					COPY_ADDR6(pp->addr, ifp->addr);
				TAILQ_INSERT_TAIL(&ifp->prefixes,
						  pp, prf_list);
				if (newthings)
					newthings++;
			}
		} else
		/* there are some cases where old things are in fact new */
		if (pp->nflags & ND_OPT_PI_FLAG_RECV) {
			if (verbose)
				logtrace("skipping already seen prefix %s/%d",
					 inet6pr(&pp->pref), pp->plen);
			goto next;
		} else if ((pp->mflags & MOB_ISFA) == 0)
			newthings = 0;
		pp->last = last_rtadv;
		pp->nflags = pip->nd_opt_pi_flags_reserved;
		if (pp->nflags & ND_OPT_PI_FLAG_RTADDR) {
			if (pp->mflags & MOB_ISHA) {
				COPY_ADDR6(rta, rp->rta);
				gotrta |= 2;
			}
			if (pp->ifp->flags & IFF_POINTOPOINT)
				COPY_ADDR6(rta, pp->peer);
		}
		pp->nflags &= ND_OPT_PI_FLAG_CFG;
		pp->nflags |= ND_OPT_PI_FLAG_RECV;
		pp->deplife = pip->nd_opt_pi_preferred_time;
#ifdef PARANOID_WITH_RA
		if ((pip->nd_opt_pi_valid_time > MINVALLIFE) ||
		    (pip->nd_opt_pi_valid_time > pp->invlife)) {
			pp->invlife = pip->nd_opt_pi_valid_time;
			goto next;
		}
		if (pp->invlife <= MINVALLIFE)
			goto next;
		pp->invlife = MINVALLIFE;
		goto next;
#else
		pp->invlife = pip->nd_opt_pi_valid_time;
		goto next;
#endif

	    case ND_OPT_MTU:
		mp = (struct nd_opt_mtu *)hp;
		if (mp->nd_opt_mtu_len != 1 || mp->nd_opt_mtu_reserved) {
			if (verbose)
				logerr("bad MTU ext (len %d res %d)",
				       mp->nd_opt_mtu_len << 3,
				       mp->nd_opt_mtu_reserved);
			goto next;
		}
		if (len < (mp->nd_opt_mtu_len << 3)) {
			if (verbose)
				logerr("Rt Adv underflow");
			goto next;
		}
		mtu = ntohl(mp->nd_opt_mtu_mtu);
		if (mtu < IP6_MMTU) {
			if (verbose)
				logerr("bad MTU (%d)", mtu);
			goto next;
		}
		if (mtu == ifp->mtu)
			goto next;
		if (verbose)
			logtrace("changing mtu of %s from %lu to %u",
				 ifp->name,  ifp->mtu, mtu);
		bzero((caddr_t)&ifr, sizeof(ifr));
		strncpy(ifr.ifr_name, ifp->name, IFNAMSIZ);
		ifr.ifr_mtu = mtu;
		if (ioctl(u, SIOCSIFMTU, (caddr_t)&ifr) < 0) {
			logerr("ioctl (SIOCSIFMTU): "
			       "MTU %d on interface %s: %s",
				mtu, ifp->name, strerror(errno));
			goto next;
		}
		ifp->mtu = mtu;
		goto next;

	    default:
	    next:
		if (hp->nd_opt_len == 0) {
			if (verbose)
				logerr("null option length (%d)",
				       hp->nd_opt_type);
			goto bad;
		}
		len -= hp->nd_opt_len << 3;
		if (len < 0) {
			if (verbose)
				logerr("Rt Adv underflow");
			goto bad;
		}
		if (len == 0)
			break;
		hp = (struct nd_opt_hdr *)
			((caddr_t)hp + (hp->nd_opt_len << 3));
	}

	/* try to set the router address */
	switch (gotrta) {
	    case 0:
		/* no router address */
		if ((rp->nflags & ND_RA_FLAG_HAGENT) &&
		    IS_ANYADDR6(rp->rta) &&
		    verbose)
			logtrace("router %s with H flag but without addresses",
				 inet6pr(&rp->addr.sin6_addr));
		break;

	    case 1:
		/* a random router address (can we select the best one?) */
		if (IS_ANYADDR6(rp->rta))
			COPY_ADDR6(rta, rp->rta);
		break;

	    case 3:
		/* good: a router home address */
		break;

	    default:
		/* should never happen! */
		if (verbose)
			logerr("recv_rtadv: gotrta %d?", gotrta);
	}

	/*
	 * Update all changed prefixes, begin by good ones
	 */

    again:
	FORALLPP(pp, ifp)
		if (pp->pflags != pp->nflags) {
			if ((pass > 1) ||
			    ((pp->nflags & ND_OPT_PI_FLAG_AUTO) &&
			     (pp->deplife != 0) &&
			     (pp->invlife != 0))) {
				update_prefix(pp);
				changed++;
				/* update_prefix can change the chain! */
				goto again;
			}
		}
	if (pass == 1) {
		pass++;
		goto again;
	}
	if (changed) {
		sort_prefix(ifp);
		if (mobile && newthings && !booting)
			mobile_newrtadv();
	}
    bad:
	update_router(rp);

	if (verbose)
		logtrace("ICMPv6 Rt Adv: end");
}

void
recv_redirect(ifp, ip, len)
	struct interface *ifp;
	struct ip6_hdr *ip;
	int len;
{
	struct sockaddr_in6 ifa;
	register struct nd_redirect *rdp;
	register char *cp = m_rtmsg.m_space;
	struct router *rp;
	struct nd_opt_hdr *xp;
	caddr_t mac = NULL;
	struct sockaddr_in6 *sin;
	struct sockaddr_dl *sdl;
	register int i;
	int gw_redirect = 0;
	u_long mtu;
	short flags;

	rdp = (struct nd_redirect *)(ip + 1);

	if (rdp->nd_rd_code != 0) {
		if (verbose)
			logerr("ICMPv6 Redirect: code=%d",
			       rdp->nd_rd_code);
		return;
	}
	if ((unsigned)len < sizeof(*rdp)) {
		if (verbose)
			logerr("ICMPv6 Redirect: too short %d", len);
		return;
	}

	if (IN6_IS_ADDR_MULTICAST(&rdp->nd_rd_dst)) {
		if (verbose)
			logerr("ICMPv6 Redirect: illegal dst %s",
			       inet6pr(&rdp->nd_rd_dst));
		return;
	}

	if (SAME_ADDR6(rdp->nd_rd_target, rdp->nd_rd_dst)) {
		if (!ifp->hasphy) {
			if (verbose)
				printf("unespected lladdr redirect\n");
			return;
		}
		FORALLRP(rp, ifp)
		    if (SAME_ADDR6(rdp->nd_rd_target, rp->addr.sin6_addr)) {
			if (verbose)
				logtrace("link-layer redirect for router %s",
					 inet6pr(&rdp->nd_rd_dst));
			rp->phy.sdl_alen = 0;
			get_rtphy(rp);
			return;
		    }
		len -= sizeof(*rdp);
		xp = (struct nd_opt_hdr *)(rdp + 1);
		while (len && (mac == NULL)) switch (xp->nd_opt_type) {
		    case ND_OPT_TARGET_LINKADDR:
			if (xp->nd_opt_len < 1) {
				if (verbose)
					logerr("bad LL address length (%d)",
					       xp->nd_opt_len << 3);
				return;
			}
			if (len < (xp->nd_opt_len << 3)) {
				if (verbose)
					logerr("Redirect underflow");
				return;
			}
			mac = (caddr_t)(xp + 1);
			break;

		    default:
			if (xp->nd_opt_len == 0) {
				if (verbose)
					logerr("null option length (%d)",
					       xp->nd_opt_type);
				return;
			}
			len -= xp->nd_opt_len << 3;
			if (len < 0) {
				if (verbose)
					logerr("Redirect underflow");
				return;
			}
			xp = (struct nd_opt_hdr *)
				    ((caddr_t)xp + (xp->nd_opt_len << 3));
		}

	} else {
		gw_redirect = 1;
		if (!IN6_IS_ADDR_LINKLOCAL(&rdp->nd_rd_target)) {
			if (verbose)
				logerr("ICMPv6 Redirect: illegal gw %s",
				       inet6pr(&rdp->nd_rd_target));
			return;
		}

	}

	if (verbose) {
		if (gw_redirect)
			logtrace("redirect for dest %s via new gateway %s",
				 inet6pr(&rdp->nd_rd_dst),
				 inet6pr(&rdp->nd_rd_target));
		else if (mac)
			logtrace("redirect for dest %s direct new address %s",
				 inet6pr(&rdp->nd_rd_dst),
				 mac2ascii(mac, ifp->lmac));
		else
			logtrace("redirect for dest %s on link (no address)",
				 inet6pr(&rdp->nd_rd_dst));
	}

	bzero((caddr_t)&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = RTM_GET;
	rtm.rtm_flags = RTF_UP | RTF_GATEWAY | RTF_HOST | RTF_STATIC;
	rtm.rtm_addrs = RTA_DST | RTA_IFP;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;
	rtm.rtm_index = ifp->index;

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_family = AF_INET6;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	COPY_ADDR6(rdp->nd_rd_dst, sin->sin6_addr);
	ADVANCE(cp, sin);

	sdl = (struct sockaddr_dl *)cp;
	sdl->sdl_family = AF_LINK;
	sdl->sdl_len = sizeof(struct sockaddr_dl);
	ADVANCE(cp, sdl);

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;
	cp = (caddr_t)&m_rtmsg;

	DUMPR("recv_redirect GET");
	if (write(r, cp, len) < len) {
		if (verbose)
			logperror("write (recv_redirect GET)");
		return;
	}

	for (;;) {
		if (debug)
			bzero(cp, MAX_MSG_ROUTE);
		len = recv(r, cp, MAX_MSG_ROUTE, MSG_PEEK);
		if (len < 0) {
			if (verbose)
				logperror("recv (recv_redirect GET)");
			return;
		}
		if ((rtm.rtm_seq != rtm_seq) || (rtm.rtm_pid != pid))
			recv_route();
		else
			break;
	}
	if (debug)
		bzero(cp, MAX_MSG_ROUTE);
	(void)gettimeofday(&now, NULL);
	len = read(r, cp, MAX_MSG_ROUTE);
	if (len < 0) {
		if (verbose)
			logperror("read (recv_redirect GET)");
		return;
	} else if (len  == 0) {
		if (verbose)
			logerr("read (recv_redirect GET) return 0");
		return;
	}
	if ((rtm.rtm_seq != rtm_seq) || (rtm.rtm_pid != pid)) {
		if (verbose)
			logerr("lost (recv_redirect GET)");
		return;
	}

	if (rtm.rtm_version != RTM_VERSION) {
		if (verbose)
			logerr("bad message version %d", rtm.rtm_version);
		return;
	}
	if (rtm.rtm_msglen > len) {
		if (verbose)
			logerr("bad message length (%d > %d)",
			       rtm.rtm_msglen, len);
		return;
	}
	if (rtm.rtm_errno) {
		if (verbose)
			logerr("get route failed (%s)",
			       strerror(rtm.rtm_errno));
		return;
	}
	flags = rtm.rtm_flags;
	if ((flags & RTF_GATEWAY) == 0) {
		if (gw_redirect && verbose)
			logerr("redirect for a direct route");
		return;
	}
	if (rtm.rtm_index != ifp->index) {
		if (verbose)
			logerr("bad interface index (%d != %d)",
			       rtm.rtm_index, ifp->index);
		return;
	}
	if ((rtm.rtm_addrs & RTA_REDIRECT) != RTA_REDIRECT) {
		if (verbose)
			logerr("not enough addresses for redirect");
		return;
	}
	if ((mtu = rtm.rtm_rmx.rmx_mtu) < IP6_MMTU)
		mtu = ifp->mtu;
	cp = m_rtmsg.m_space;
	for (i = 0; i < RTAX_MAX; i++) {
		if ((rtm.rtm_addrs & (1 << i)) == 0)
			continue;
		sin = (struct sockaddr_in6 *)cp;
		switch (i) {
		    case RTAX_DST:
			if ((sin->sin6_family != AF_INET6) ||
			    ((rtm.rtm_flags & RTF_HOST) &&
			     !SAME_ADDR6(rdp->nd_rd_dst, sin->sin6_addr)))
				return;
			goto nxt;

		    case RTAX_GATEWAY:
			if ((sin->sin6_family != AF_INET6) ||
			    !SAME_ADDR6(ip->ip6_src, sin->sin6_addr))
				return;
			goto nxt;

		    case RTAX_IFA:
			if (sin->sin6_family != AF_INET6)
				return;
			bcopy(cp, (caddr_t)&ifa, sin->sin6_len);
			goto nxt;

		    default:
		    nxt:
			ADVANCE(cp, sin);
		}
	}

	bzero((caddr_t)&m_rtmsg, sizeof(m_rtmsg));
	cp = m_rtmsg.m_space;
	rtm.rtm_version = RTM_VERSION;
	if (flags & RTF_HOST) {
		rtm.rtm_type = RTM_CHANGE;
		rtm.rtm_flags = flags | RTF_MODIFIED;
	} else {
		rtm.rtm_type = RTM_ADD;
		flags = rtm.rtm_flags = RTF_UP | RTF_HOST | RTF_DYNAMIC;
		rtm.rtm_rmx.rmx_mtu = mtu;
		rtm.rtm_inits = RTV_MTU;
	}
	rtm.rtm_addrs = RTA_DST | RTA_GATEWAY | RTA_IFP | RTA_IFA;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;
	rtm.rtm_index = ifp->index;
	if (gw_redirect || mac) {
		rtm.rtm_rmx.rmx_expire = now.tv_sec + 10*60;
		rtm.rtm_inits |= RTV_EXPIRE;
	}

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_family = AF_INET6;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	COPY_ADDR6(rdp->nd_rd_dst, sin->sin6_addr);
	ADVANCE(cp, sin);

	if (gw_redirect) {
		rtm.rtm_flags |= RTF_GATEWAY;
		sin = (struct sockaddr_in6 *)cp;
		sin->sin6_family = AF_INET6;
		sin->sin6_len = sizeof(struct sockaddr_in6);
		COPY_ADDR6(rdp->nd_rd_target, sin->sin6_addr);
		ADVANCE(cp, sin);
	} else {
		sdl = (struct sockaddr_dl *)cp;
		sdl->sdl_family = AF_LINK;
		sdl->sdl_len = sizeof(struct sockaddr_dl);
		sdl->sdl_index = ifp->index;
		if (mac) {
			sdl->sdl_alen = ifp->lmac;
			bcopy(mac, LLADDR(sdl), ifp->lmac);
		} else
			sdl->sdl_alen = 0;
		ADVANCE(cp, sdl);
	}

	sdl = (struct sockaddr_dl *)cp;
	sdl->sdl_family = AF_LINK;
	sdl->sdl_len = sizeof(struct sockaddr_dl);
	sdl->sdl_index = ifp->index;
	ADVANCE(cp, sdl);

	sin = (struct sockaddr_in6 *)cp;
	bcopy((caddr_t)&ifa, cp, ifa.sin6_len);
	ADVANCE(cp, sin);

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

	DUMPR("recv_redirect ADD/CHANGE");
	if (write(r, (char *)&m_rtmsg, len) < len) {
		if (verbose)
			logperror("write (recv_redirect ADD/CHANGE)");
		return;
	}
}
