#include "ndpd.h"

int rtm_seq = 0;

u_int CurHopLimit;		/* initialize in kernel */
u_int KeepTimer;
u_int BaseReachableTime;
u_int RetransTimer;

static struct interface *new_if(struct if_msghdr *, int);
static struct if_msghdr *get_ifinfo(int);
static int refresh_if(struct interface *, int);
u_int getkerncst(int, u_int);

void
recv_route()
{
	struct if_msghdr *ifm;
	struct ifa_msghdr *ifam;
	register struct sockaddr_in6 *sa, *sin = NULL;
	register struct sockaddr_dl *sdl;
	register struct interface *ifp;
#ifdef nomore
	struct prefix *pp;
#endif
	struct in6_addr addr;
	struct router *rp;
	char *cp;
	int cc, i, is_up, was_up, first;
	u_char plen = 0;

	cp = (caddr_t)&m_rtmsg;
	if (debug)
		bzero(cp, MAX_MSG_ROUTE);
	cc = read(r, cp, MAX_MSG_ROUTE);
	if (cc < 0) {
		if (verbose)
			logperror("read (route message)");
		return;
	} else if (cc == 0) {
		if (verbose)
			logerr("read (route message) return 0");
		return;
	}

	if (rtm.rtm_version != RTM_VERSION) {
		if (verbose)
			logerr("bad message version %d", rtm.rtm_version);
		return;
	}

	/* ignore our messages */

	if (rtm.rtm_pid == pid)
		return;

	if (verbose)
		logtrace("receive route (%d,%d)",
			 rtm.rtm_msglen, rtm.rtm_type);
	DUMPR("received");

	switch (rtm.rtm_type) {
	    case RTM_NEWADDR:
		ifam = (struct ifa_msghdr *)cp;
		if (verbose)
		    	logtrace("receive route newaddr on #%d",
				 ifam->ifam_index);
		FORALLIFP(ifp)
		    if (ifp->index == ifam->ifam_index)
			break;
		if (ifp == NULL)
			ifp = new_if((struct if_msghdr *)cp, 1);
		if (ifp == NULL)
			return;
#ifdef IFA_FIRST
		first = ifam->ifam_flags & IFA_FIRST;
#else
		first = 0;
#endif
		cp += sizeof(struct ifa_msghdr);
		sin = NULL;
		for (i = 0; i < RTAX_MAX; i++) {
			if ((ifam->ifam_addrs & (1 << i)) == 0)
				continue;
			sa = (struct sockaddr_in6 *)cp;
			ADVANCE(cp, sa);
			switch (i) {
			    case RTAX_IFA:
				if (sa->sin6_family != AF_INET6)
					return;
				sin = sa;
				break;

			    case RTAX_NETMASK:
				if (sa->sin6_family != AF_UNSPEC)
					return;
				plen = get_prefixlen(sa);
				break;
			}
		}
		if ((sin == NULL) ||
		    (plen < 3) ||
		    (sin->sin6_addr.s6_addr[0] == 0) ||
		    (sin->sin6_addr.s6_addr[0] == 0xff))
			return;
		if (verbose)
			logtrace("get new address %s/%d on %s",
				 inet6pr(&sin->sin6_addr), plen, ifp->name);
		if (IS_LINKLADDR6(sin->sin6_addr) ) {
			if (!refresh_if(ifp, 1))
				break;
			was_up = IS_UP(ifp);
			if (ifp->state != STATENOLL) {
				if (!first)	/* an other LL in front */
					return;
				if (bcmp(&sin->sin6_addr, &ifp->llocal,
					 sizeof ifp->llocal) != 0) {
					logerr("changing LL addr of %s",
					       ifp->name);
				}
				localroute(ifp, &ifp->llocal, 0);
				was_up = 0;	/* force RTsol if IFF_UP */
			} else if (prefixstatic(ifp->name))
				ifp->state = STATESTATIC;
			else
				ifp->state = STATELESS;
			COPY_ADDR6(sin->sin6_addr, ifp->llocal);
			localroute(ifp, &ifp->llocal, 1);
			if (FIRST(ifp) == NULL)
				COPY_ADDR6(ifp->llocal, ifp->addr);
			goto change_if;
		}
		COPY_ADDR6(sin->sin6_addr, addr);
		localroute(ifp, &addr, 1);
#if noyet
		if (first &&
		    bcmp(&sin->sin6_addr, &ifp->addr, sizeof ifp->addr) != 0) {
			struct prefix *pp;

			FORALLPP(pp, ifp)
				if (SAME_ADDR6(pp->addr, sin->sin6_addr))
					break;
			/* accept manual change */
			if (pp && pp != FIRST(ifp))
				set_first_prefix(pp);
		}
#endif
		/* otherwise do nothing */
		/* we admit all new addresses are done by the program */
		return;

#ifdef nomore
		if (Calloc(pp, prefix) == NULL) {
			if (verbose)
				logerr("out of memory");
			return;
		}

		pp->alen = pp->plen = plen;
		COPY_ADDR6(sin->sin6_addr, pp->addr);
		COPY_ADDR6(pp->addr, pp->pref);
		mk_pref(pp);
		pp->ifp = ifp;
		pp->invlife = INFLIFE;
		pp->deplife = DEFPREFLIFE;
		pp->nflags = ND_OPT_PI_FLAG_AUTO;
		if (pp->plen != 128)
			pp->nflags |= ND_OPT_PI_FLAG_ONLINK;
		pp->pflags = pp->nflags;
		TAILQ_INSERT_TAIL(&ifp->prefixes, pp, prf_list);
		update_prefix(pp);
		sort_prefix(ifp);
		return;
#endif

	    case RTM_DELADDR:
		ifam = (struct ifa_msghdr *)cp;
		if (verbose)
			logtrace("receive route deladdr on #%d",
				 ifam->ifam_index);
		FORALLIFP(ifp)
		    if (ifp->index == ifam->ifam_index)
			break;
		if (ifp == NULL)
			return;
		cp += sizeof(struct ifa_msghdr);
		for (i = 0; i < RTAX_MAX; i++) {
			if ((ifam->ifam_addrs & (1 << i)) == 0)
				continue;
			sa = (struct sockaddr_in6 *)cp;
			ADVANCE(cp, sa);
			switch (i) {
			    case RTAX_IFA:
				if (sa->sin6_family != AF_INET6)
					return;
				sin = sa;
				break;

			    case RTAX_NETMASK:
				if (sa->sin6_family != AF_UNSPEC)
					return;
				plen = get_prefixlen(sa);
				if (plen < 3)
					return;
				break;
			}
		}
		if (IS_LINKLADDR6(sin->sin6_addr) ) {
			(void) refresh_if(ifp, 0);
			was_up = IS_UP(ifp);
			if (ifp->state != STATENOLL) {
				if (bcmp(&sin->sin6_addr, &ifp->llocal,
					 sizeof ifp->llocal) != 0)
					return;
				logerr("deleting LL addr of %s", ifp->name);
				ifp->state = STATENOLL;
			}
			localroute(ifp, &ifp->llocal, 0);
			goto change_if;
		}
		/* otherwise do nothing */
		if ((sin->sin6_addr.s6_addr[0] == 0) ||
		    (sin->sin6_addr.s6_addr[0] == 0xff))
			return;
		COPY_ADDR6(sin->sin6_addr, addr);
		localroute(ifp, &addr, 0);
		/* we admit all delete addresses are done by the program */
		return;

#ifdef nomore
		FORALLPP(pp, ifp)
			if (SAME_ADDR6(pp->addr, sin->sin6_addr))
				break;
		if (pp == NULL)
			return;
		prefix_inval(pp);
		return;
#endif

	    case RTM_IFINFO:
		ifm = (struct if_msghdr *)cp;
		if (verbose)
			logtrace("receive route ifinfo for #%d",
				 ifm->ifm_index);
		FORALLIFP(ifp)
		    if (ifp->index == ifm->ifm_index)
			break;
#define IFF_FIXED (IFF_BROADCAST|IFF_MULTICAST|IFF_LOOPBACK|IFF_POINTOPOINT)
		if (ifp) {
			if ((ifp->flags ^ ifm->ifm_flags) & IFF_FIXED) {
				logerr("unsupported flag change for %s",
				       ifp->name);
				abort();
			}
			was_up = IS_UP(ifp);
		} else {
			was_up = 0;
			ifp = new_if(ifm, 1);
			if (ifp == NULL)
				return;
		}

		ifp->flags = ifm->ifm_flags;
    change_if:
		is_up = IS_UP(ifp);
		if (was_up && !is_up) {
			if (verbose)
				logtrace("interface %s down", ifp->name);
			if_down(ifp);
			return;
		}

		if (!(is_up && !was_up))
			return;

		if (verbose)
			logtrace("interface %s up", ifp->name);
		(void) gettimeofday(&now, NULL);
		if_up(ifp);

		if (!new_timeout(5, 0, (func_to) reset_default, 0))
			abort();
		return;

	    case RTM_RTLOST:
#define RTA_LOST	(RTA_DST | RTA_GATEWAY)
		if ((rtm.rtm_addrs & RTA_LOST) != RTA_LOST)
			return;
		cp += sizeof(struct rt_msghdr);
		sin = (struct sockaddr_in6 *)cp;
		if (sin->sin6_family != AF_INET6)
			return;
		ADVANCE(cp, sin);
		sdl = (struct sockaddr_dl *)cp;
		if ((sdl->sdl_family != AF_LINK) ||
		    (sdl->sdl_index == 0))
			return;

		FORALLIFP(ifp)
		    if (ifp->index == sdl->sdl_index)
			break;
		if (ifp == NULL)
			return;
		if (verbose)
			logtrace("%s router %s on %s",
				 sdl->sdl_alen != 0 ? "lost" : "unreachable",
				 inet6pr(&sin->sin6_addr),
				 ifp->name);
		FORALLRP(rp, ifp)
			if (SAME_ADDR6(rp->addr.sin6_addr, sin->sin6_addr))
				break;
		if (rp == NULL)
			return;
		if ((ifp->hasphy) && (sdl->sdl_alen == 0))
			router_unreachable(rp, 1);
		else
			router_invalid(rp);
		return;

	    case RTM_LOSING:
		if ((rtm.rtm_addrs & RTA_LOST) != RTA_LOST)
			return;
		cp += sizeof(struct rt_msghdr);
		sin = (struct sockaddr_in6 *)cp;
		if (sin->sin6_family != AF_INET6)
			return;
		ADVANCE(cp, sin);

		if (verbose)
			logtrace("lost peer %s", inet6pr(&sin->sin6_addr));
		lost_mobile(sin);
		return;
	}
}


int mib[] = { CTL_NET, PF_ROUTE, 0, AF_INET6, NET_RT_IFLIST, 0 };

void
init()
{
	register struct if_msghdr *ifm;
	register struct interface *ifp;
	register struct prefix *pp;
	struct sockaddr_in6 *sin;
	int i;
	char *buf, *cp, *ap;
	unsigned bufsize;

	CurHopLimit = getCurHopLimit();
	KeepTimer = getKeepTimer();
	BaseReachableTime = getReachableTime();
	RetransTimer = getRetransTimer();

	TAILQ_INIT(&timeouts);
	TAILQ_INIT(&interfaces);
	TAILQ_INIT(&dynifs);
	TAILQ_INIT(&movifs);
	TAILQ_INIT(&homeaddrs);

	ifp = NULL;

	if (sysctl(mib, 6, NULL, &bufsize, NULL, 0) < 0) {
		logperror("sysctl0");
		exit(4);
	}
	buf = (char *)malloc(bufsize);
	if (buf == NULL) {
		logerr("out of memory");
		exit(4);
	}
	if (sysctl(mib, 6, buf, &bufsize, NULL, 0) < 0) {
		logperror("sysctl");
		exit(4);
	}

	for (cp = buf; cp < buf + bufsize; cp += ifm->ifm_msglen) {
		ifm = (struct if_msghdr *)cp;

		if (ifm->ifm_type == RTM_IFINFO) {
			ifp = new_if(ifm, 1);
			continue;
		}
		if (ifp == NULL)
			continue;
		if (ifm->ifm_type != RTM_NEWADDR) {
			if (verbose)
				logerr("bad message type (%d)",
				       ifm->ifm_type);
			continue;
		}
		if ((ifm->ifm_addrs & RTA_IFA) == 0) {
			if (verbose)
				logerr("no ifa for %s", ifp->name);
			continue;
		}
		if (verbose)
			logtrace("get NEWADDR");

		if (Calloc(pp, prefix) == NULL) {
			if (verbose)
				logerr("out of memory");
			continue;
		}

		ap = cp + sizeof(struct ifa_msghdr);
		for (i = 0; i < RTAX_MAX; i++) {
			if ((ifm->ifm_addrs & (1 << i)) == 0)
				continue;
			sin = (struct sockaddr_in6 *)ap;
			ADVANCE(ap, sin);
			switch (i) {
			    case RTAX_NETMASK:
				if (sin->sin6_family != AF_UNSPEC)
					goto skip;
				pp->alen = pp->plen = get_prefixlen(sin);
				break;

			    case RTAX_IFA:
				if (sin->sin6_family != AF_INET6)
					goto skip;
				COPY_ADDR6(sin->sin6_addr, pp->addr);
				break;

			    case RTAX_BRD:
				if ((ifm->ifm_flags & IFF_POINTOPOINT) == 0)
					break;
				if (sin->sin6_family != AF_INET6)
					goto skip;
				COPY_ADDR6(sin->sin6_addr, pp->peer);
				break;
			}
		}
		if ((pp->plen < 3) || IS_ANYADDR6(pp->addr))
			goto skip;
		if (verbose)
			logtrace("interface %s got addr %s plen %d addrs %x",
				 ifp->name,
				 inet6pr(&pp->addr),
				 pp->plen, ifm->ifm_addrs);
		if (IS_LINKLADDR6(pp->addr)) {
			if (IS_ANYADDR6(ifp->llocal))
				COPY_ADDR6(pp->addr, ifp->llocal);
			goto skip;
		}
		if ((pp->addr.s6_addr[0] == 0) ||
		    (pp->addr.s6_addr[0] == 0xff) ||
		    (pp->plen < 16) ||
		    (pp->plen != 128 &&
		     pp->plen > (128 - 8 * ifp->ltoken))) {
			if (verbose)
				logerr("ignored interface address %s",
				       inet6pr(&pp->addr));
			goto skip;
		}
		pp->ifp = ifp;
		COPY_ADDR6(pp->addr, pp->pref);
		mk_pref(pp);
		pp->invlife = INFLIFE;
		pp->deplife = DEFPREFLIFE;
		pp->nflags = ND_OPT_PI_FLAG_AUTO;
		if (pp->plen != 128)
			pp->nflags |= ND_OPT_PI_FLAG_ONLINK;
		pp->pflags = pp->nflags;
		TAILQ_INSERT_TAIL(&ifp->prefixes, pp, prf_list);
		continue;
    skip:
		free(pp);
	}
	free(buf);

	if (loop_if == NULL)
		logtrace("no loopback");

	FORALLIFP(ifp) {
		if (IS_ANYADDR6(ifp->llocal)) {
			if (verbose)
				logtrace("ignoring interface %s without LL",
					 ifp->name);
			continue;
		} else if (prefixstatic(ifp->name))
			ifp->state = STATESTATIC;
		else
			ifp->state = STATELESS;
		if (FIRST(ifp) == NULL)
			COPY_ADDR6(ifp->llocal, ifp->addr);
		else
			COPY_ADDR6(FIRST(ifp)->addr, ifp->addr);
		/* update_prefix will only set the timeout */
		FORALLPP(pp, ifp)
			update_prefix(pp);
		sort_prefix(ifp);
		prompt(ifp);
	}
}

static struct if_msghdr *
get_ifinfo(index)
	int index;
{
	register struct if_msghdr *ifm = NULL;
	char *buf;
	unsigned bufsize;

	mib[5] = index;

	if (sysctl(mib, 6, NULL, &bufsize, NULL, 0) < 0) {
		if (verbose)
			logperror("get_ifinfo: sysctl0");
		return ifm;
	}
	buf = (char *)malloc(bufsize);
	if (buf == NULL) {
		if (verbose)
			logerr("out of memory");
		return ifm;
	}
	if (sysctl(mib, 6, buf, &bufsize, NULL, 0) < 0) {
		if (verbose)
			logperror("get_ifinfo: sysctl");
		free(buf);
		return ifm;
	}
	ifm = (struct if_msghdr *)buf;

	if ((ifm->ifm_type != RTM_IFINFO) || (ifm->ifm_index != index)) {
		ifm = NULL;
		free(buf);
	}
	return ifm;
}

static struct interface *
new_if(ifm, first)
	register struct if_msghdr *ifm;
	int first;
{
	register struct interface *ifp = NULL;
	struct sockaddr_dl *sdl;
	int hasphy = 0, ltoken = 0, lmac = 0;

	if ((ifm->ifm_addrs != (1 << RTAX_IFP)) ||
	    (ifm->ifm_type != RTM_IFINFO)) {
		struct if_msghdr *nifm;

		if (first) {
			nifm = get_ifinfo((int)ifm->ifm_index);
			if (nifm == NULL)
				return ifp;
			ifp = new_if(nifm, 0);
			free(nifm);
			return ifp;
		} else if (verbose)
			logerr("can't get ifinfo for if#%d", ifm->ifm_index);
		return ifp;
	}

	sdl = (struct sockaddr_dl *)((char *)ifm + sizeof(struct if_msghdr));
	if ((sdl->sdl_family != AF_LINK) || (sdl->sdl_nlen == 0)) {
		if (verbose)
			logerr("bad sdl address");
		return ifp;
	}

	if ((loop_if == NULL) &&
	    (ifm->ifm_flags & IFF_LOOPBACK) &&
	    (sdl->sdl_nlen >= 3) &&
	    (bcmp(sdl->sdl_data, "lo0", sdl->sdl_nlen) == 0))
		goto create;

	if ((ifm->ifm_flags & IFF_LOOPBACK) ||
	    (ifm->ifm_data.ifi_mtu < IP6_MMTU))
		return ifp;

	if (ifm->ifm_flags & IFF_POINTOPOINT) {
		if (sdl->sdl_nlen > 3 &&
		   (bcmp(sdl->sdl_data, "cti", 3) == 0))
			ltoken = SITTOKEN;
		else if (ifm->ifm_data.ifi_type == IFT_PPP)
			ltoken = PPPTOKEN;
		/* else ... not supported */
	} else if ((ifm->ifm_flags & IFF_MULTICAST) == IFF_MULTICAST) {
		/* everything has 48 bit mac addr ... */
		ltoken = MACTOKEN;
		lmac = MACMAC;
		if (ifm->ifm_data.ifi_addrlen != lmac)
			return ifp;
		/* ignore ATM interfaces (should work but disturb) ??? */
		if (ifm->ifm_data.ifi_type == IFT_ATM)
			return ifp;
		hasphy = 1;
	} /* else ... not supported */
	if (ltoken == 0)
		return ifp;

	if (ifm->ifm_addrs != RTA_IFP) {
		if (verbose)
			logerr("bad ifinfo (no ifp#%d)",
			       ifm->ifm_index);
		return ifp;
	}
	if ((sdl->sdl_family != AF_LINK) || (sdl->sdl_nlen == 0)) {
		if (verbose)
			logerr("bad sdl address");
		return ifp;
	}
	if (hasphy && sdl->sdl_alen != lmac) {
		if (verbose)
			logerr("unsupported type for interface %.*s",
				sdl->sdl_nlen, sdl->sdl_data);
		return ifp;
	}

    create:
	if (Calloc(ifp, interface) == NULL) {
		logerr("out of memory");
		abort();
	}
	if ((ifm->ifm_flags & IFF_LOOPBACK) == 0)
		TAILQ_INSERT_TAIL(&interfaces, ifp, if_list);
	TAILQ_INIT(&ifp->routers);
	TAILQ_INIT(&ifp->prefixes);
	ifp->mtu = ifm->ifm_data.ifi_mtu;
	ifp->flags = ifm->ifm_flags;
	ifp->index = ifm->ifm_index;
	ifp->state = STATENOLL;
	ifp->hasphy = hasphy;
	ifp->ltoken = ltoken;
	ifp->lmac = lmac;
	if (hasphy)
		bcopy(LLADDR(sdl), (caddr_t)ifp->adphy, lmac);
	bcopy(sdl->sdl_data, ifp->name, sdl->sdl_nlen);
	ifp->stateful = ND_RA_FLAG_CFG;
	ifp->ntosend = MAX_RTR_SOLICITATIONS;
	if (verbose)
		logtrace("got interface %s#%d", ifp->name, ifp->index);
	if (ifm->ifm_flags & IFF_LOOPBACK) {
		loop_if = ifp;
		ifp = NULL;
	}
	return ifp;
}

/*
 * Refresh copy of interface parameters on link-local address event
 */

static int
refresh_if(ifp, onnew)
	struct interface *ifp;
	int onnew;
{
	struct if_msghdr *ifm;
	struct sockaddr_dl *sdl;
	struct ifreq ifr;

	/* refresh flags first */

	bzero((caddr_t)&ifr, sizeof(ifr));
	strncpy(ifr.ifr_name, ifp->name, IFNAMSIZ);

	if (ioctl(u, SIOCGIFFLAGS, (caddr_t)&ifr) < 0) {
		logperror("ioctl (SIOCGIFFLAGS)");
		return 0;
	}
	ifp->flags = ifr.ifr_flags;

	if (!onnew || !is_dyn(ifp) ||
	    (ifp->prefixes.tqh_first != NULL) ||
	    (ifp->routers.tqh_first != NULL))
		return 1;

	/* argh! perhaps it is a real new interface */

	if (verbose)
		logtrace("perhaps new interface %s#%d", ifp->name, ifp->index);

	ifm = get_ifinfo((int)ifp->index);
	if (ifm == NULL) {
		logerr("can get infos for interface %s", ifp->name);
		return 0;
	}
	if ((ifm->ifm_addrs != RTA_IFP) || (ifm->ifm_type != RTM_IFINFO)) {
		logerr("bad infos for interface %s", ifp->name);
		goto bad;
	}
	if ((ifm->ifm_flags ^ ifp->flags) & IFF_FIXED) {
		logerr("illegal flags change for %s", ifp->name);
		goto bad;
	}
	sdl = (struct sockaddr_dl *)((char *)ifm + sizeof(struct if_msghdr));
	if ((sdl->sdl_family != AF_LINK) ||
	    (sdl->sdl_nlen != strlen(ifp->name)) ||
	    bcmp(sdl->sdl_data, ifp->name, sdl->sdl_nlen)) {
		logerr("wrong sdl for %s", ifp->name);
		goto bad;
	}
	if (ifp->hasphy) {
		if (ifm->ifm_data.ifi_addrlen != ifp->lmac) {
			logerr("bad lmac for %s", ifp->name);
			goto bad;
		}
		if (verbose &&
		    bcmp(LLADDR(sdl), (caddr_t)ifp->adphy, ifp->lmac))
			logtrace("MAC address change for %s", ifp->name);
		bcopy(LLADDR(sdl), (caddr_t)ifp->adphy, ifp->lmac);
	}
	free(ifm);
	return 1;
    bad:
	free(ifm);
	return 0;
}

void
get_main_if()
{
	register char *cp = m_rtmsg.m_space;
	struct sockaddr_in6 *sin;
	struct sockaddr_dl *sdl;
	int kidx, len;
	struct interface *ifp;

	/* try the kernel sysct */

	if ((kidx = main_interface(NULL)) != 0) {
		FORALLIFP(ifp)
			if (ifp->index == kidx)
				break;

		if (ifp && IS_UP(ifp)) {
			main_if = ifp;
			if (verbose)
				logtrace("(kernel) main interface is %s",
					 ifp->name);
			return;
		}
	}

	/* try the IPv6 default route */

	bzero((caddr_t)&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = RTM_GET;
	rtm.rtm_flags = RTF_UP | RTF_GATEWAY | RTF_STATIC;
	rtm.rtm_addrs = RTA_DST | RTA_NETMASK | RTA_IFP;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	ADVANCE(cp, sin);

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_UNSPEC;
	ADVANCE(cp, sin);

	sdl = (struct sockaddr_dl *)cp;
	sdl->sdl_len = sizeof(struct sockaddr_dl);
	sdl->sdl_family = AF_LINK;
	ADVANCE(cp, sdl);

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

	cp = (caddr_t)&m_rtmsg;

	DUMPR("get_main_if");
	if (write(r, cp, len) < len) {
		if (verbose)
			logperror("write (get_main_if)");
		return;
	}

	for (;;) {
		if (debug)
			bzero(cp, MAX_MSG_ROUTE);
		len = recv(r, cp, MAX_MSG_ROUTE, MSG_PEEK);
		if (len < 0) {
			if (verbose)
				logperror("recv (get_main_if)");
			return;
		}
		if ((rtm.rtm_seq != rtm_seq) || (rtm.rtm_pid != pid))
			recv_route();
		else
			break;
	}
	if (debug)
		bzero(cp, MAX_MSG_ROUTE);
	len = read(r, cp, MAX_MSG_ROUTE);
	if (len < 0) {
		if (verbose)
			logperror("read (get_main_if)");
		return;
	} else if (len  == 0) {
		if (verbose)
			logerr("read (get_main_if) return 0");
		return;
	}
	if ((rtm.rtm_seq != rtm_seq) || (rtm.rtm_pid != pid)) {
		if (verbose)
			logerr("lost (get_main_if)");
		return;
	}

	if (rtm.rtm_version != RTM_VERSION) {
		if (verbose)
			logerr("bad message version %d", rtm.rtm_version);
		return;
	}
	if (rtm.rtm_msglen > len) {
		if (verbose)
			logerr("bad message length (%d > %d)",
			       rtm.rtm_msglen, len);
		return;
	}
	if (rtm.rtm_errno) {
		if (verbose)
			logerr("get route (get_main_if) failed (%s)",
			       strerror(rtm.rtm_errno));
		return;
	}
	if ((rtm.rtm_addrs & RTA_IFP) == 0) {
		if (verbose)
			logerr("no IFP in route");
		return;
	}
	if (rtm.rtm_index == 0) {
		if (verbose)
			logerr("no IFP index in route");
		return;
	}

	FORALLIFP(ifp)
	    if (ifp->index == rtm.rtm_index)
		break;

	if (ifp && IS_UP(ifp))
		main_if = ifp;
	if (verbose) {
		if (ifp)
			logtrace("main interface is %s", ifp->name);
		else
			logtrace("no main interface?");
	}
}

void
setCurHopLimit(hop)
u_int hop;
{
#ifdef notyet
	static int mib[] = { CTL_NET, PF_INET, IPPROTO_IP, IPCTL_DEFTTL};
	int val;
#endif

	if (hop < MINHOP || hop > MAXHOP)
		return;
	if (verbose)
#ifndef notyet
		logerr("(don't) setting CurHop to %d", hop);
#else
		logerr("setting CurHop to %d", hop);
	/* don't clobber a shared parameter ? */
	if (sysctl(mib, (sizeof mib)/(sizeof (int)),
		NULL, NULL, &hop, sizeof(hop)) < 0) {
		if (verbose)
			logerr("can't set CurHopLimit");
	}
#endif
	CurHopLimit = hop;
}

/* XXXX the function SHOULD be called on every few hours ... */
void
setReachableTime(basereach)
u_int basereach;
{
	static int mib[] = { CTL_NET, PF_INET6, IPPROTO_IP, IP6CTL_REACHABLE};
	u_int val;

	val = RTTOKERN(MIN_RANDOM_FACTOR(basereach) +
		       random() % DELTA_RANDOM_FACTOR(basereach));
	if (val < MINKREACH)
		val = MINKREACH;
	else if (val > MAXKREACH)
		val = MAXKREACH;
	if (verbose)
		logerr("setting Reachable time to %u sec (msg=%u)",
			val, basereach);
	if (sysctl(mib, (sizeof mib)/(sizeof (int)),
		NULL, NULL, &val, sizeof(val)) < 0) {
		if (verbose)
			logerr("can't set ReachableTime");
	}
	BaseReachableTime = basereach;
}

void
setRetransTimer(retrans)
u_int retrans;
{
	static int mib[] = { CTL_NET, PF_INET6, IPPROTO_IP, IP6CTL_RETRANS};
	u_int val;

	val = RTTOKERN(retrans);
	if (val < MINKRETRANS)
		val = MINKRETRANS;
	else if (val > MAXKRETRANS)
		val = MAXKRETRANS;
	if (sysctl(mib, (sizeof mib)/(sizeof (int)),
		NULL, NULL, &val, sizeof(val)) < 0) {
		if (verbose)
			logerr("can't set RetransTimer");
	}
	if (verbose)
		logerr("setting RetransTimer to %u sec (msg=%u)",
		       val, retrans);
	RetransTimer = retrans;
}

u_int
getkerncst(typ, def)
	int typ;
	u_int def;
{
	static int mib[] = { CTL_NET, PF_INET6, IPPROTO_IP, IP6CTL_RETRANS};
	u_int val, len = sizeof val;

#define POS_TYP	(((sizeof mib) / (sizeof (int))) - 1)

	mib[POS_TYP] = typ;
	if (sysctl(mib, (sizeof mib)/(sizeof(int)), &val, &len, NULL, NULL) < 0
	   || len != sizeof val)
		val = 0;
	if (val == 0)
		val = def;
	return val;
}

u_int
getKeepTimer()
{
	return getkerncst(IP6CTL_KEEP, KEEP_TIMER);
}

u_int
getReachableTime()
{
	return getkerncst(IP6CTL_REACHABLE, REACHABLE_TIME);
}

u_int
getRetransTimer()
{
	return getkerncst(IP6CTL_RETRANS, RETRANS_TIMER);
}

u_int
getCurHopLimit()
{
	return getkerncst(IP6CTL_DEFTTL, 0);
}

void
localroute(ifp, addr, set)
	struct interface *ifp;
	struct in6_addr *addr;
	int set;
{
	register char *cp = m_rtmsg.m_space;
	struct sockaddr_in6 *sin;
	struct sockaddr_dl *sdl;
	int len;

	if ((ifp->flags & IFF_POINTOPOINT) == 0)
		return;
	bzero((caddr_t)&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = set ? RTM_ADD : RTM_DELETE;
	rtm.rtm_flags = RTF_UP | RTF_HOST | RTF_LOCAL;
	rtm.rtm_addrs = RTA_DST | RTA_GATEWAY;
	rtm.rtm_index = ifp->index;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(*addr, sin->sin6_addr);
	ADVANCE(cp, sin);

	sdl = (struct sockaddr_dl *)cp;
	sdl->sdl_family = AF_LINK;
	sdl->sdl_len = sizeof(struct sockaddr_dl);
	sdl->sdl_index = ifp->index;
	ADVANCE(cp, sdl);

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

    retry:
	/* if exists do a change */
	DUMPR("localroute");
	if (write(r, (char *)&m_rtmsg, len) < len) {
		if (rtm.rtm_type == RTM_ADD && errno == EEXIST) {
			rtm.rtm_type = RTM_CHANGE;
			rtm.rtm_seq = ++rtm_seq;
			goto retry;
		} else if (verbose)
			logperror("write (localroute)");
		return;
	}
}

/*
 * Test if an interface is dynamic/removable
 * (set ifp pointer as a side effect)
 */
int
is_dyn(ifp)
	struct interface *ifp;
{
	register struct dynif *dip = NULL;

	for (dip = dynifs.tqh_first; dip; dip = dip->di_list.tqe_next) {
		if (dip->ifp && (dip->ifp == ifp))
			return 1;
		if (strncmp(dip->name, ifp->name, IFNAMSIZ) != 0)
			continue;
		if (dip->ifp) {
			if (verbose)
				logerr("is_dyn mismatch (%s != %s)",
				       dip->name, dip->ifp->name);
			dip->ifp = NULL;
			continue;
		}
		dip->ifp = ifp;
		return 1;
	}
	return 0;
}

/*
 * Test if an interface is dynamic/movable
 * (set ifp pointer as a side effect)
 */
int
is_mov(ifp)
	struct interface *ifp;
{
	register struct dynif *dip = NULL;

	for (dip = movifs.tqh_first; dip; dip = dip->di_list.tqe_next) {
		if (dip->ifp && (dip->ifp == ifp))
			return 1;
		if (strncmp(dip->name, ifp->name, IFNAMSIZ) != 0)
			continue;
		if (dip->ifp) {
			if (verbose)
				logerr("is_mov mismatch (%s != %s)",
				       dip->name, dip->ifp->name);
			dip->ifp = NULL;
			continue;
		}
		dip->ifp = ifp;
		return 1;
	}
	return 0;
}
