#include "ndpd.h"
struct timeval now;
int s, r, u;			/* Socket file descriptors */
int pid;

int verbose = 0;
int debug = 0;
int timestamps = 0;
int mobile = 0;

char usage[] = "Usage: ndpd-host [-d] [-v] [-t] [-m] [-n] [-i ifname,...]\n";

static int nopref = 0;
static int interfmax = 0;
static int interflen = 0;
static char *interflist = 0;

int main __P((int, char *[]));
static char *dump_if_state __P((char));
static char *dump_ha_state __P((char));

int
main(argc, argv)
	int argc;
	char *argv[];
{
	char ch;
	u_int val;
	int nfds, cc;
	fd_set rfds;
	struct timeval tv;
	struct timeout *tp;
	struct interface *ifp;
	extern int optind;
	extern char *optarg;
	struct icmp6_filter filter;

	pid = getpid();
	(void) setpriority(PRIO_PROCESS, pid, PRIO_MIN + 1);
	(void) gettimeofday(&now, NULL);

	while ((ch = getopt(argc, argv, "dvtmni:")) != EOF)
		switch (ch) {
		    case 'd':
			debug = 1;
			break;
		    case 'v':
			verbose++;
			break;
		    case 't':
			timestamps = 1;
			break;
		    case 'm':
			mobile = 1;
			break;
		    case 'n':
			nopref = 1;
			break;
		    case 'i':
		    {
			char *p = optarg;
#define INTFCHAR(x) (isalnum(x) || ((x)=='-') || ((x)=='_'))

			interfmax += strlen(p) +1 ;
			interflist = realloc(interflist, interfmax);
			if (interflist == NULL) {
				(void) fprintf(stderr, "malloc\n");
				exit(1);
			}
			interflist[interflen] = 0;
			while (*p) {
				while (INTFCHAR(*p))
					interflist[interflen++] = *p++;
				interflist[interflen++] = 0;
				if (*p == 0)
					break;
				while (*p && !INTFCHAR(*p))
					p++;
			}
			break;
		    }
		    default:
			(void) fprintf(stderr, usage);
			exit(1);
		}

	argc -= optind;
	argv += optind;

	if (argc != 0) {
		(void) fprintf(stderr, "Extra parameters\n");
		(void) fprintf(stderr, usage);
		exit(1);
	}

	srandom((int)gethostid());
	setlinebuf(stdout);

	if ((s = socket(PF_INET6, SOCK_RAW, IPPROTO_ICMPV6)) < 0) {
		logperror("socket (ICMP)");
		exit(2);
	}

	val = 1;
	if (setsockopt(s, IPPROTO_IPV6, IP_HDRINCL,
		       &val, sizeof(val)) < 0) {
		logperror("setsockopt (IP_HDRINCL)");
		exit(3);
	}

	val = 1;
	if (setsockopt(s, IPPROTO_IPV6, IPV6_RECVPKTINFO,
		       &val, sizeof(val)) < 0) {
		logperror("setsockopt (IPV6_RECVPKTINFO)");
		exit(3);
	}

	val = ND_DEFAULT_HOPLIM;
	if (setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_HOPS,
		       &val, sizeof(val)) < 0) {
		logperror("setsockopt (IPV6_MULTICAST_HOPS)");
		exit(3);
	}

	val = 0;
	if (setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_LOOP,
		       &val, sizeof(val)) < 0) {
		logperror("setsockopt (IPV6_MULTICAST_LOOP)");
		exit(3);
	}

	/*
	 * Only interested in a few icmp:
	 * ND_REDIRECT ND_ROUTER_ADVERT
	 */
	ICMP6_FILTER_SETBLOCKALL(&filter);
	ICMP6_FILTER_SETPASS(ND_REDIRECT, &filter);
	ICMP6_FILTER_SETPASS(ND_ROUTER_ADVERT, &filter);
	if (setsockopt(s, IPPROTO_IPV6, ICMP6_FILTER,
		       &filter, sizeof(filter)) < 0) {
		logperror("setsockopt (ICMP6_FILTER)");
		exit(3);
	}

	/* protect the ICMPv6 socket against mobility stuff */

	disable_binding(s);
	disable_homeaddr(s);

	if ((r = socket(PF_ROUTE, SOCK_RAW, 0)) < 0) {
		logperror("socket (ROUTE)");
		exit(2);
	}

	if ((u = socket(PF_INET6, SOCK_DGRAM, 0)) < 0) {
		logperror("socket (UDP)");
		exit(2);
	}
	(void) shutdown(u, 0);

	init();
	mobility_init();

	if (interfaces.tqh_first == NULL) {
		if (verbose)
			logerr("No IPV6 interface !!");
		exit(1);
	}

	(void) signal(SIGUSR1, set_verbose);
	(void) signal(SIGUSR2, set_verbose);
	(void) signal(SIGINT, dump);
	(void) signal(SIGTERM, finish);
	(void) signal(SIGQUIT, finish);

	get_main_if();
	if ((main_if != NULL) && (main_if != interfaces.tqh_first)) {
		TAILQ_REMOVE(&interfaces, main_if, if_list);
		TAILQ_INSERT_HEAD(&interfaces, main_if, if_list);
	}

	(void) gettimeofday(&now, NULL);

	FORALLIFP(ifp)
		if (IS_UP(ifp))
			if_up(ifp);

	(void) gettimeofday(&now, NULL);
	if (!new_timeout(WAIT_INIT, 0, (func_to) reset_default, 0))
		abort();

	for (;;) {
		run_timeouts();

		FD_ZERO(&rfds);
		FD_SET(s, &rfds);
		nfds = s;
		FD_SET(r, &rfds);
		nfds = r > nfds ? r : nfds;
		FD_SET(m, &rfds);
		nfds = m > nfds ? m : nfds;
		if (b >= 0) {
			FD_SET(b, &rfds);
			nfds = b > nfds ? b : nfds;
		}

		if ((tp = timeouts.tqh_first)) {
			tv.tv_usec = tp->tv.tv_usec + 1000000 - now.tv_usec;
			tv.tv_sec = tp->tv.tv_sec - now.tv_sec;
			NORMTIME(tv);
			if (tv.tv_sec > 0)
				tv.tv_sec -= 1;
			else
				tv.tv_usec = 10000;
		} else {
			tv.tv_sec = 3;
			tv.tv_usec = 0;
		}

		cc = select(nfds + 1, &rfds, NULL, NULL, &tv);
		if (cc < 0) {
			if (errno != EINTR && verbose)
				logperror("select");
			continue;
		}
		if (cc == 0)
			continue;
		(void) gettimeofday(&now, NULL);
		if (FD_ISSET(r, &rfds))
			recv_route();
		if (FD_ISSET(s, &rfds))
			recv_icmp();
		if (FD_ISSET(m, &rfds))
			recv_destopt();
		if ((b >= 0) && FD_ISSET(b, &rfds))
			recv_bpf();
		if (verbose)
			logerr("end of main loop reached");
	}
	/*NOTREACHED*/
}

/* IS_UP(ifp) is true */

void
if_up(ifp)
	struct interface *ifp;
{
	struct prefix *pp;

	localroute(ifp, &ifp->llocal, 1);
	FORALLPP(pp, ifp)
		localroute(ifp, &pp->addr, 1);
	if (!IS_UP(ifp))
		return;

	if (mobile)
		mobile_newif(ifp);

	if_restart(ifp);
}

/* Send Router Solicitations */

void
if_restart(ifp)
	struct interface *ifp;
{
	struct prefix *pp;
	long delay = 0;

	/* not immediately when booting */

	if (booting) {
		delay = MAX_RTR_SOLICITATION_DELAY * M_SECOND;
		delay = random() % delay;
	}

	/* assert ourselves for each addresses */

	FORALLPP(pp, ifp) {
		if (pp->gtp) {
			kill_timeout(pp->gtp);
			pp->gtp = NULL;
		}
		pp->ntosend = MAX_ADVERT_REXMIT;
		if (delay != 0) {
			pp->gtp = new_timeout(MS_TO_SEC(delay),
					      MS_TO_USEC(delay),
					      (func_to) send_grat_adv,
					      pp);
			if (pp->gtp == NULL)
				abort();
		} else
			send_grat_adv(pp);
	}

	/* send some router solicitations */

	if (ifp->tp) {
		kill_timeout(ifp->tp);
		ifp->tp = NULL;
	}
	ifp->ntosend = MAX_RTR_SOLICITATIONS;

	if (delay != 0) {
		ifp->tp = new_timeout(MS_TO_SEC(delay), MS_TO_USEC(delay),
				      (func_to) send_rtsol, ifp);
		if (ifp->tp == NULL)
			abort();
	} else
		send_rtsol(ifp);
}

/* IS_UP(ifp) is unknown */

void
if_down(ifp)
	struct interface *ifp;
{
	struct prefix *pp;

	if (ifp->tp) {
		kill_timeout(ifp->tp);
		ifp->tp = NULL;
	}
	ifp->ntosend = -1;

	localroute(ifp, &ifp->llocal, 0);
	FORALLPP(pp, ifp)
		localroute(ifp, &pp->addr, 0);

	if (is_dyn(ifp) && !IS_UP(ifp)) {
		if (verbose)
			logtrace("Flush lost interface %s", ifp->name);
		/* flush prefixes */
		while ((pp = FIRST(ifp)) != NULL) {
			while (pp->prf_list.tqe_next != NULL)
				pp = pp->prf_list.tqe_next;
			if (pp->mflags & MOB_ISHA) {
				prefix_deprec(pp);
				pp->mflags &= ~(MOB_ISHA | MOB_ISFA);
				if (pp->hap)
					TAILQ_REMOVE(&pp->hap->homeprefs,
						     pp, mob_link);
				pp->hap = NULL;
			}
			prefix_inval(pp);
		}
		/* finish */
		flushif(ifp);
	} else if (!is_dyn(ifp) && !IS_UP(ifp) && verbose)
		logerr("Lost an IPv6 interface (%s)", ifp->name);

	if (ifp == def_if)
		reset_default();
}

void
set_verbose(sig)
	int sig;
{
	if (sig == SIGUSR1)
		verbose++;
	else
		verbose = 0;
}

static const char wday_name[7][3] = {
    "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
};
static const char mon_name[12][3] = {
    "Jan", "Feb", "Mar", "Apr", "May", "Jun",
    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};

void dump(s)
	int s;
{
	register struct timeout *tp;
	register struct interface *ifp;
	register struct dynif *dip;
	register struct homeaddr *hap;
	struct tm *tmnow;

	(void) gettimeofday(&now, NULL);
	tmnow = localtime(&now.tv_sec);

	logdebug("DUMP at %.3s %.3s%3d %02d:%02d:%02d.%06d %d",
		 wday_name[tmnow->tm_wday],
		 mon_name[tmnow->tm_mon],
		 tmnow->tm_mday, tmnow->tm_hour,
		 tmnow->tm_min, tmnow->tm_sec,
		 now.tv_usec,
		 1900 + tmnow->tm_year);

	logdebug("\nTIMEOUT LIST:");
	for (tp = timeouts.tqh_first; tp; tp = tp->tm_list.tqe_next)
		dump_timeout(tp);

	logdebug("\nINTERFACE LIST:");
	FORALLIFP(ifp)
		dump_interface(ifp);
	ifp = loop_if;
	if (ifp) {
		logdebug("loopback/discard interface:");
		dump_interface(ifp);
	}
	logdebug("\n");

	logdebug("last RtAdv at %s", tvpr(&last_rtadv));
	if (def_if)
		logdebug("default interface is %s", def_if->name);
	else
		logdebug("no current default interface");
	if (main_if)
		logdebug("main interface is %s", main_if->name);
	else
		logdebug("no main interface");

	logdebug("default source (link#%d) = %s",
		 def_ifa.index,
		 inet6pr(&def_ifa.addr));

	if (def_rt)
		logdebug("default router is %s",
			 inet6pr(&def_rt->addr.sin6_addr));
	else
		logdebug("no default router");

	logdebug("\nDYNAMIC/REMOVABLE INTERFACE LIST:");
	if (dynifs.tqh_first)
		printf("\t");
	for (dip = dynifs.tqh_first; dip; ) {
		if (dip->ifp)
			printf("%s (%lx)", dip->name, (u_long)(dip->ifp));
		else
			printf("%s", dip->name);
		dip = dip->di_list.tqe_next;
		if (dip)
			printf(", ");
	}
	if (dynifs.tqh_first)
		printf("\n");

	logdebug("\nMOVABLE INTERFACE LIST:");
	if (movifs.tqh_first)
		printf("\t");
	for (dip = movifs.tqh_first; dip; ) {
		if (dip->ifp)
			printf("%s (%lx)", dip->name, (u_long)(dip->ifp));
		else
			printf("%s", dip->name);
		dip = dip->di_list.tqe_next;
		if (dip)
			printf(", ");
	}
	if (movifs.tqh_first)
		printf("\n");

	logdebug("\nHOME ADDRESS LIST:");
	FORALLHA(hap)
		dump_homeaddr(hap);

	logdebug("\nMOBILE LIST:");
	dump_mobiles();
}

void
finish(s)
	int s;
{
	struct interface *ifp;

	mobility_cleanup();
	mobile = 0;

	FORALLIFP(ifp)
		if (IS_UP(ifp))
			if_down(ifp);

	cleanup_default();

	if (verbose)
		logtrace("finished!");

	exit(0);
}

int
prefixstatic(name)
char *name;
{
	char *p = interflist;

	if (p == 0)
		return (nopref);
	for (; p < &interflist[interflen]; p += strlen(p) + 1) {
		if (strcmp(name, p) == 0)
			return (nopref);
	}
	return (!nopref);
}

void
dump_prefix(pp)
	register struct prefix *pp;
{
	logdebug("prefix (%lx):", (u_long)pp);
	logdebug("\tinterface=%s", pp->ifp->name);
	logdebug("\tprefix=%s/%d", inet6pr(&pp->pref), pp->plen);
	logdebug("\t[mask=%s]", inet6pr(&pp->pmask));
	logdebug("\taddress=%s/%d", inet6pr(&pp->addr), pp->alen);
	if (pp->ifp->flags & IFF_POINTOPOINT)
		logdebug("\tpeer address=%s", inet6pr(&pp->peer));
	if (pp->hap)
		logdebug("\thomeaddr=%lx (link=%lx)",
			 (u_long)(pp->hap), (u_long)(pp->mob_link.tqe_next));
	logdebug("\ttimeout=%lx\trenew=%u", (u_long)(pp->tp), pp->rendelay);
	logdebug("\tlast seen=%s", tvpr(&pp->last));
	logdebug("\tvalidity=%s\tpreference=%s",
		lifepr(pp->invlife), lifepr(pp->deplife));
	logdebug("\tflags=%x (next=%x)", pp->pflags, pp->nflags);
	logdebug("\tmobility=%x [%s]\tadvtosend=%d when %lx",
		 pp->mflags, dump_mflags(pp->mflags),
		 pp->ntosend, (u_long)(pp->gtp));
}

void
dump_router(rp)
	register struct router *rp;
{
	logdebug("router (%lx):", (u_long)rp);
	logdebug("\tinterface=%s", rp->ifp->name);
	logdebug("\taddress=%s", inet6pr(&rp->addr.sin6_addr));
	logdebug("\tethernet=(link#%d) %s",
		 rp->phy.sdl_index,
		 mac2ascii(LLADDR(&rp->phy), rp->phy.sdl_alen));
	logdebug("\trouter address=%s", inet6pr(&rp->rta));
	logdebug("\ttimeout=%lx", (u_long)(rp->tp));
	logdebug("\tlast seen=%s", tvpr(&rp->last));
	logdebug("\tlifetime=%u", rp->lifetime);
	logdebug("\tflags=%x (next=%x)", rp->aflags, rp->nflags);
	logdebug("\treachable time=%u.%03d\tretrans timers=%u.%03d",
		 rp->reach / 1000, rp->reach % 1000,
		 rp->retrans / 1000, rp->retrans % 1000);
}

void
dump_interface(ifp)
	register struct interface *ifp;
{
	register struct prefix *pp;
	register struct router *rp;

	logdebug("interface (%lx):", (u_long)ifp);
	logdebug("\tname=%s(if#%d)", ifp->name, ifp->index);
	logdebug("\tlink-local address=%s", inet6pr(&ifp->llocal));
	logdebug("\tbest address=%s", inet6pr(&ifp->addr));
	logdebug("\ttoken=%s",
		 mac2ascii(&ifp->llocal.s6_addr[16-ifp->ltoken], ifp->ltoken));
	logdebug("\tlgtoken %d lgmac %d adphy %s", ifp->ltoken, ifp->lmac,
		 ifp->hasphy ? mac2ascii(ifp->adphy, ifp->lmac) : "NONE");
	logdebug("\tstate=%x [%s]\tstateful=%x",
		ifp->state, dump_if_state(ifp->state), ifp->stateful);
	logdebug("\tRtSol to send=%d when %lx",
		 ifp->ntosend, (u_long)(ifp->tp));
	logdebug("\tMTU=%lu\tflags=%x", ifp->mtu, ifp->flags);

	logdebug("\nPREFIX LIST:");
	FORALLPP(pp, ifp)
		dump_prefix(pp);

	logdebug("\nROUTER LIST:");
	FORALLRP(rp, ifp)
		dump_router(rp);

	logdebug("\n");
}

void
dump_homeaddr(hap)
	register struct homeaddr *hap;
{
	struct prefix *pp;

	logdebug("home address (%lx):", (u_long)hap);
	logdebug("\taddress=%s", inet6pr(&hap->addr));
	logdebug("\tstate=%d [%s], flags=%d",
		 hap->state, dump_ha_state(hap->state), hap->flags);
	if (HAPP(hap))
		FORALLHP(pp, hap)
			logdebug("\thome prefix=%s/%d on %s",
				 inet6pr(&pp->addr), pp->plen, pp->ifp->name);
	else
		logdebug("\tno home prefix");
	if (hap->capp)
		logdebug("\tcare-of prefix=%s/%d on %s",
			 inet6pr(&hap->capp->addr),
			 hap->capp->plen, hap->capp->ifp->name);
	else
		logdebug("\tno care-of prefix");
}

char *
dump_mflags(mflags)
	u_char mflags;
{
	static char pbuf[16];

	bzero(pbuf, sizeof(pbuf));
	snprintf(pbuf, sizeof(pbuf) - 1, "%s%s%s%s",
		 mflags & MOB_ISHA ? "H" : "",
		 mflags & MOB_ISCA ? "C" : "",
		 mflags & MOB_ISFA ? "F" : "",
		 mflags & MOB_HASFL ? "R" : "");
	return pbuf;
}

static char *
dump_if_state(state)
	char state;
{
	switch (state) {
	    case STATENOLL:
		return "no LL";

	    case STATELESS:
		return "stateless";

	    case STATEFULL:
		return "statefull";

	    case STATESTATIC:
		return "static";

	    default:
		return "???";
	}
}

static char *
dump_ha_state(state)
	char state;
{
	switch (state) {
	    case HAST_UNKNOWN:
		return "unknown";

	    case HAST_IN:
		return "in/at home";

	    case HAST_OUT:
		return "out/in visit";

	    default:
		return "???";
	}
}


	
