#include "util_ndpd.h"

#define WAIT_INIT	30

struct prefix {
	TAILQ_ENTRY(prefix) prf_list;	/* prefix list */
	TAILQ_ENTRY(prefix) mob_link;	/* prefix link for mobility */
	struct in6_addr pref;		/* prefix itself */
	struct in6_addr pmask;		/* prefix mask */
	struct in6_addr addr;		/* address */
	struct in6_addr peer;		/* peer address for p2p */
	struct interface *ifp;		/* up pointer (interface/list) */
	struct homeaddr *hap;		/* up pointer (homeaddr/link) */
	struct timeout *tp;		/* timeout */
	struct router *offrp;		/* off-link router */
	struct timeval last;		/* last seen */
	u_int invlife;			/* invalidation lifetime */
	u_int deplife;			/* deprecation lifetime */
	u_int rendelay;			/* renew delay */
	u_char plen;			/* length (as a prefix) */
	u_char alen;			/* length (as an address) */
	u_char pflags;			/* flags */
	u_char nflags;			/* new flags */
	u_char mflags;			/* mobility flags */
#define MOB_ISHA		1	/* is a home address prefix */
#define MOB_ISCA		2	/* is a care-of address prefix */
#define MOB_ISFA		4	/* is a fake address */
#define MOB_HASFL		8	/* has a forced length */
	struct timeout *gtp;		/* when to send gratuitous Nd Adv */
	int ntosend;			/* Gratuitous Nd Adv to send */
};

struct router {
	TAILQ_ENTRY(router) rt_list;	/* router list */
	struct interface *ifp;		/* up pointer */
	struct sockaddr_in6 addr;	/* designated address */
	struct sockaddr_dl phy;		/* data-link address */
	struct in6_addr rta;		/* router address */
	struct timeout *tp;		/* timeout */
	struct timeval last;		/* last seen */
	u_short lifetime;		/* lifetime */
	u_char aflags;			/* autoconfig flags */
	u_char nflags;			/* new flags */
	u_int reach;			/* ND reachable time */
	u_int retrans;			/* ND reachable retrans timer */
};
#define FORALLRP(rp, ifp)	\
	for (rp = BEST(ifp); rp; rp = rp->rt_list.tqe_next)

struct interface {
	TAILQ_ENTRY(interface) if_list;	/* list of interfaces */
	TAILQ_HEAD(, router) routers;	/* routers */
	TAILQ_HEAD(, prefix) prefixes;	/* prefixes */
	struct in6_addr llocal;		/* link-local address */
	struct in6_addr addr;		/* best address */
	u_char adphy[MAXTOKEN];		/* MAC address, left just. */
	char state;			/* state, LL present, accept prefix? */
#define STATENOLL	0
#define STATELESS	1
#define STATEFULL	2
#define STATESTATIC	3
	u_char stateful;		/* stateful flags */
	u_char ltoken;			/* token length */
	u_char lmac;			/* MAC length */
	char hasphy;			/* is there a phy addr ? */
	u_long mtu;			/* IPv6 MTU */
	short flags;			/* copy of kernel flags */
	u_int index;			/* kernel index */
	char name[IFNAMSIZ];		/* name */
	struct timeout *tp;		/* when to send Rt Sol */
	int ntosend;			/* Rt Sol to send */
};
struct interface *def_if, *main_if, *loop_if;
#define BEST(ifp)	(ifp->routers.tqh_first)

struct {
	struct in6_addr	addr;		/* default source address */
	u_int index;			/* default interface index */
} def_ifa;
struct router *def_rt, *cur_rt;

struct dynif {
	TAILQ_ENTRY(dynif) di_list;	/* list of dynamic interfaces */
	struct interface *ifp;		/* interface when known */
	char name[IFNAMSIZ];		/* name */
};

struct homeaddr {
	TAILQ_ENTRY(homeaddr) ha_list;	/* list of home address */
	struct in6_addr addr;		/* a home address */
	char state;			/* at home/in visit */
	char flags;			/* flags (unused) */
	TAILQ_HEAD(, prefix) homeprefs;	/* list of home prefixes */	
	struct prefix *capp;		/* care-of address prefix */
#define HAST_UNKNOWN	0
#define HAST_IN		1
#define HAST_OUT	2
};
TAILQ_HEAD(, homeaddr) homeaddrs;
#define PRIMHA		homeaddrs.tqh_first
#define FORALLHA(hap)	\
	for(hap = PRIMHA; hap; hap = hap->ha_list.tqe_next)
#define HAPP(hap)	(hap->homeprefs.tqh_first)
#define FORALLHP(happ, hap)	\
	for(happ = HAPP(hap); happ; happ = happ->mob_link.tqe_next)

TAILQ_HEAD(, interface) interfaces;
TAILQ_HEAD(, dynif) dynifs, movifs;
TAILQ_HEAD(, timeout) timeouts;

/* router purge/depreciate operations */

#define PURGE_NOP	0	/* no operation */
#define DEPREC_OLDER	1	/* depreciate older */
#define PURGE_OLDER	2	/* purge older */
#define DEPREC_OTHER	3	/* depreciate not current */
#define PURGE_OTHER	4	/* purge not current */
#define PURGE_ALL	5	/* purge all */

/* route flush operations */ 

#define FLUSH_NOP	0	/* no operation */
#define FLUSH_IF	1	/* all routes using an interface */
#define FLUSH_IFA	2	/* all routes using an interface address */
#define FLUSH_GW	3	/* all routes using a gateway */
#define FLUSH_ND	4	/* all unresolved neighbor on an interface */
#define FLUSH_NDPF	5	/* all neighbors in a prefix */
#define FLUSH_DYN	6	/* all dynamic routes using a gateway */
#define FLUSH_ALL	7	/* all IPv6 routes */

extern int s, r, u, m, b;		/* Socket file descriptors */

extern int verbose;
extern int debug;
extern int timestamps;
extern int mobile;

extern int booting;
extern struct timeval last_rtadv;

/* kernel values, time in seconds */
extern u_int CurHopLimit;

#define RTTOKERN(x)	((x)/1000)
#define MAXKREACH	3600			/* in seconds */
#define MINKREACH	2			/* in seconds */
#define REACHABLE_TIME	30000	/* ms, not used, read kernel */
#define DELTA_RANDOM_FACTOR(x)	(1+(x))		/* 1.5 - 0.5, include bound */
#define MIN_RANDOM_FACTOR(x)	(((x)+1)/2)
extern u_int BaseReachableTime;

#define KEEP_TIMER	120			/* in seconds */
extern u_int KeepTimer;

#define MAXKRETRANS	3600			/* in seconds */
#define MINKRETRANS	1			/* in seconds */
#define RETRANS_TIMER	1000	/* ms, not used, read kernel */
extern u_int RetransTimer;

void send_rtsol(struct interface *);
void send_grat_adv(struct prefix *);
void recv_icmp(void);
void recv_rtadv(struct interface *, struct ip6_hdr *, int);
void recv_redirect(struct interface *, struct ip6_hdr *, int);
void recv_route(void);
void init(void), set_verbose(int), dump(int), finish(int);
void get_main_if(void);
int is_dyn(struct interface *);
int is_mov(struct interface *);
void update_prefix(struct prefix *);
void prefix_deprec(struct prefix *);
void prefix_inval(struct prefix *);
void sort_prefix(struct interface *);
void add_prefix(struct prefix *);
void del_prefix(struct prefix *, int);
void prompt(struct interface *);
void add_prefroute(struct prefix *);
void del_prefroute(struct prefix *);
void update_router(struct router *);
void router_invalid(struct router *);
void router_unreachable(struct router *, int);
void purge_routers(struct interface *, struct timeval *, int);
void sort_router(struct interface *, int);
void get_rtphy(struct router *);
void del_router(struct router *, int);
void reset_default(void);
void cleanup_default(void);
void del_default(void);
void change_default(void);
void flushif(struct interface *);
void flush(int, struct in6_addr *, struct in6_addr *, u_int);
void dump_prefix(struct prefix *);
void dump_router(struct router *);
void dump_interface(struct interface *);
void dump_homeaddr(struct homeaddr *);
char *dump_mflags(u_char);
void setCurHopLimit(u_int);
void setReachableTime(u_int);
void setRetransTimer(u_int);
u_int getKeepTimer(void);
u_int getReachableTime(void);
u_int getRetransTimer(void);
u_int getCurHopLimit(void);
void localroute(struct interface *, struct in6_addr *, int);
void if_up(struct interface *);
void if_restart(struct interface *);
void if_down(struct interface *);
int prefixstatic(char *);

void disable_binding(int);
void disable_homeaddr(int);
void lost_mobile(struct sockaddr_in6 *);
void recv_destopt(void);
void recv_bpf(void);
void mobility_init(void);
void mobile_booted(void);
void mobility_cleanup(void);
void mobile_zappref(struct prefix *);
void mobile_chgpref(struct interface *);
void mobile_chgif(void);
void mobile_chgha(struct homeaddr *, struct prefix *);
int mobile_chgrt(struct prefix *, int, int);
void mobile_newif(struct interface *);
void mobile_newrtadv(void);
void dump_mobiles(void);
