#include "ndpd.h"

int booting = 1;
static struct if_laddrreq ifr;
static void reset_default_restart(void);
static void flushone(int, struct in6_addr *, struct in6_addr *, u_int);
static char *dump_purge_op(int);
static char *dump_flush_op(int);

/*
 * A prefix has changed (new one, different flags, different timers)
 */

void
update_prefix(pp)
	register struct prefix *pp;
{
	int vanishing = 0;

	(void)gettimeofday(&now, NULL);

	if (pp->tp) {
		kill_timeout(pp->tp);
		pp->tp = NULL;
	}

	pp->nflags &= ~ND_OPT_PI_FLAG_RECV;

	if (((pp->nflags & ND_OPT_PI_FLAG_AUTO) == 0) ||
	    (pp->deplife == 0) || (pp->invlife == 0))
		vanishing = 1;

	/* Home address prefixes (vanishing) */

	if (mobile && vanishing && (pp->mflags & MOB_ISHA)) {
		struct homeaddr *hap;
		struct prefix *happ;

		FORALLHA(hap)
			FORALLHP(happ, hap)
				if (pp == happ) {
					mobile_chgha(hap, pp);
					goto done_v;
				}
	    done_v:
	}

	if (pp->invlife == 0 || !pp->nflags) {	 /* destroy immediately */
		prefix_inval(pp);
		return;
	}

	if (pp->pflags == pp->nflags)
		goto nochange;

	if (pp->pflags & ND_OPT_PI_FLAG_AUTO)
		del_prefix(pp, 1);
	if (pp->pflags == ND_OPT_PI_FLAG_ONLINK)
		del_prefroute(pp);
	else if ((pp->pflags & ND_OPT_PI_FLAG_AUTO) &&
		 (pp->alen == 128) && (pp->plen != 128))
		del_prefroute(pp);
	if (pp->nflags & ND_OPT_PI_FLAG_AUTO) {
		if (pp->nflags & ND_OPT_PI_FLAG_ONLINK)
			pp->alen = pp->plen;
		else
			pp->alen = 128;
		add_prefix(pp);
		if (pp == FIRST(pp->ifp))
			prompt(pp->ifp);
		if ((pp->alen == 128) && (pp->plen != 128))
			add_prefroute(pp);
	} else if (pp->nflags == ND_OPT_PI_FLAG_ONLINK)
		add_prefroute(pp);

	pp->pflags = pp->nflags;

    nochange:

	/* Home address prefixes (not vanishing) */

	if (mobile && !vanishing && (pp->mflags & MOB_ISHA)) {
		struct homeaddr *hap;
		struct prefix *happ;

		FORALLHA(hap)
			FORALLHP(happ, hap)
				if (pp == happ) {
					mobile_chgha(hap, pp);
					goto done_nv;
				}
	    done_nv:
	}

	if (pp->invlife == 0) {
		prefix_inval(pp);
		return;
	}

	if (pp->deplife == 0) {
		prefix_deprec(pp);
		return;
	}

	if ((pp->deplife != INFLIFE) && (pp->deplife < pp->invlife)) {
		pp->tp = new_timeout(pp->deplife, 0,
				     (func_to) prefix_deprec, pp);
		return;
	}

	if (pp->invlife != INFLIFE)
		pp->tp = new_timeout(pp->invlife, 0,
				     (func_to) prefix_inval, pp);
	return;
}

/*
 * A prefix is deprecated
 */

void
prefix_deprec(pp)
	register struct prefix *pp;
{
	struct prefix *first = FIRST(pp->ifp);

	(void)gettimeofday(&now, NULL);

	if (pp->tp) {
		kill_timeout(pp->tp);
		pp->tp = NULL;
	}

	/* Home address prefixes */

	if (mobile && (pp->mflags & MOB_ISHA)) {
		struct homeaddr *hap;
		struct prefix *happ;

		FORALLHA(hap)
			FORALLHP(happ, hap)
				if (pp == happ) {
					pp->deplife = 0;
					mobile_chgha(hap, pp);
					goto done_d;
				}
	    done_d:
	} else {
		TAILQ_REMOVE(&pp->ifp->prefixes, pp, prf_list);
		TAILQ_INSERT_TAIL(&pp->ifp->prefixes, pp, prf_list);
	}

	if (pp->invlife != INFLIFE) {
		if (pp->invlife <= pp->deplife) {
			prefix_inval(pp);
			return;
		}
		pp->tp = new_timeout(pp->invlife - pp->deplife, 0,
				     (func_to) prefix_inval, pp);
	}

	if (verbose)
		logtrace("deprecated prefix %s on %s",
			 inet6pr(&pp->addr), pp->ifp->name);

	if (pp == first) {
		sort_prefix(pp->ifp);
		if (mobile)
			mobile_zappref(pp);
	}
}

/*
 * A prefix is invalidated
 */

void
prefix_inval(pp)
	register struct prefix *pp;
{
	struct prefix *first;

	first = FIRST(pp->ifp);

	if (pp->tp) {
		kill_timeout(pp->tp);
		pp->tp = NULL;
	}
	if (pp->gtp) {
		kill_timeout(pp->gtp);
		pp->gtp = NULL;
	}
	pp->ntosend = 0;

	if (mobile && (pp->mflags & MOB_ISHA)) {
		logerr("invalidated a home address prefix??");
		return;
	}

	if (pp->pflags == ND_OPT_PI_FLAG_ONLINK)
		del_prefroute(pp);
	else if (pp->pflags & ND_OPT_PI_FLAG_AUTO) {
		if ((pp->alen == 128) && (pp->plen != 128))
			del_prefroute(pp);
		del_prefix(pp, 1);
	}
	TAILQ_REMOVE(&pp->ifp->prefixes, pp, prf_list);

	if (verbose)
		logtrace("invalided prefix %s on %s",
			 inet6pr(&pp->addr), pp->ifp->name);

	if (pp == first) {
		sort_prefix(pp->ifp);
		if (mobile)
			mobile_zappref(pp);
	} else if (mobile && (pp->mflags & MOB_ISCA))
		mobile_zappref(pp);

	free(pp);
}

/*
 * Sort the prefix list:
 * 	order is (not) site-local, deprecated life and invalid life
 */

void
sort_prefix(ifp)
	register struct interface *ifp;
{
	register struct prefix *pp, *pp2;
	int changed = 0;

	if (FIRST(ifp) && !SAME_ADDR6(FIRST(ifp)->addr, ifp->addr))
		changed++;

#define SWAP_PREFIX	\
	if (pp == FIRST(ifp)) \
		changed++; \
	TAILQ_REMOVE(&ifp->prefixes, pp2, prf_list); \
	TAILQ_INSERT_BEFORE(pp, pp2, prf_list); \
	goto again;

    again:

	for (pp = FIRST(ifp); pp; pp = pp2) {
		pp2 = pp->prf_list.tqe_next;
		if (pp2 == NULL)
			break;
		if (pp->mflags & MOB_ISHA)
			continue;
		if (IS_SITELADDR6(pp->addr)) {
			if (!IS_SITELADDR6(pp2->addr)) {
				SWAP_PREFIX
			}
			if ((pp->deplife < pp2->deplife) ||
			    ((pp->deplife == pp2->deplife) &&
			     (pp->invlife < pp2->invlife))) {
				SWAP_PREFIX
			}
		} else if (IS_SITELADDR6(pp2->addr)) {
			continue;
		} else if ((pp->deplife < pp2->deplife) ||
			   ((pp->deplife == pp2->deplife) &&
			    (pp->invlife < pp2->invlife))) {
			SWAP_PREFIX
		}
	}

	if (FIRST(ifp) == NULL) {
		COPY_ADDR6(ifp->llocal, ifp->addr);
		changed++;
	} else if (changed)
		COPY_ADDR6(FIRST(ifp)->addr, ifp->addr);

	if (changed == 0)
		return;

	if (mobile && (ifp == def_if))
		for (pp = FIRST(ifp); pp; pp = pp2) {
			pp2 = pp->prf_list.tqe_next;
			if (pp2 == NULL)
				break;
			if (((pp2->mflags & (MOB_ISHA | MOB_ISCA)) != 0) &&
			    ((pp->mflags & (MOB_ISHA | MOB_ISCA)) == 0)) {
				mobile_chgpref(ifp);
				break;
			}
		}

	prompt(ifp);
	if ((booting == 0) &&
	    (((pp = FIRST(ifp)) == NULL) ||
	     IS_SITELADDR6(pp->addr) ||
	     (pp->deplife == 0)))
		reset_default();
}

/*
 * Add a prefix to an interface (ie add an interface address)
 */

void
add_prefix(pp)
	register struct prefix *pp;
{
	register struct sockaddr_in6 *sin;
	int cc, p2p = pp->ifp->flags & IFF_POINTOPOINT;

	if (mobile && mobile_chgrt(pp, RTM_ADD, ND_OPT_PI_FLAG_AUTO)) {
		if (verbose)
			logtrace("cancel add prefix %s/%d on %s",
				 inet6pr(&pp->addr), pp->alen, pp->ifp->name);
		return;
	}

	bzero((caddr_t)&ifr, sizeof(ifr));
	strncpy(ifr.iflr_name, pp->ifp->name, IFNAMSIZ);

	sin = (struct sockaddr_in6 *)&ifr.addr;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(pp->addr, sin->sin6_addr);

	if (p2p) {
		if (pp->mflags & MOB_ISFA)
			mk_local_peer(pp);
		else
			mk_peer(pp);
		sin = (struct sockaddr_in6 *)&ifr.dstaddr;
		sin->sin6_len = sizeof(struct sockaddr_in6);
		sin->sin6_family = AF_INET6;
		COPY_ADDR6(pp->peer, sin->sin6_addr);
	}

	ifr.prefixlen = pp->alen;

	DUMPR("SIOCALIFADDR-add_prefix");
	cc = ioctl(u, SIOCALIFADDR, (caddr_t)&ifr);
	if ((cc < 0) && (errno != EEXIST)) {
		logerr("ioctl (SIOCALIFADDR) prefix %s/%d%s%s "
		       "on interface %s: %s",
		       inet6pr(&pp->addr),
		       pp->alen,
		       p2p ? ":" : "",
		       p2p ? inet6pr(&pp->peer) : "",
		       pp->ifp->name, strerror(errno));
	} else if (verbose)
		logtrace("add %sprefix %s/%d%s%s on interface %s",
			 cc < 0 ? "extra " : "",
			 inet6pr(&pp->addr),
			 pp->alen,
			 p2p ? ":" : "",
			 p2p ? inet6pr(&pp->peer) : "",
			 pp->ifp->name);
}

/*
 * Delete a prefix from an interface (ie remove an interface address)
 */

void
del_prefix(pp, flshtyp)
	register struct prefix *pp;
	int flshtyp;
{
	register struct sockaddr_in6 *sin;

	if (mobile && mobile_chgrt(pp, RTM_DELETE, ND_OPT_PI_FLAG_AUTO)) {
		if (verbose)
			logtrace("cancel del prefix %s/%d on %s",
				 inet6pr(&pp->addr), pp->alen, pp->ifp->name);
		return;
	}


	bzero((caddr_t)&ifr, sizeof(ifr));
	strncpy(ifr.iflr_name, pp->ifp->name, IFNAMSIZ);

	sin = (struct sockaddr_in6 *)&ifr.addr;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(pp->addr, sin->sin6_addr);

	ifr.prefixlen = pp->alen;

	DUMPR("SIOCDLIFADDR-del_prefix");
	if (ioctl(u, SIOCDLIFADDR, (caddr_t)&ifr) < 0) {
		if ((errno == EADDRNOTAVAIL) && (pp->nflags == 0) &&
		    (pp->pflags == 0) && (pp->mflags == 0))
			return;
		logerr("ioctl (SIOCDLIFADDR) prefix %s/%d on interface %s: %s",
		       inet6pr(&pp->addr), pp->alen,
		       pp->ifp->name, strerror(errno));
	} else if (verbose)
		logtrace("delete prefix %s/%d on interface %s",
			 inet6pr(&pp->addr), pp->alen, pp->ifp->name);

	/* from on-link to off-link should not loose the default... */
	if ((pp->ifp->index == def_ifa.index) &&
	    SAME_ADDR6(pp->addr, def_ifa.addr))
		CLR_ADDR6(def_ifa.addr);

	flush(flshtyp ? FLUSH_IFA : FLUSH_NDPF,
	      &pp->addr, &pp->pmask, pp->ifp->index);
}

/*
 * Move the address associated with the first prefix at the first place
 */

void
prompt(ifp)
	register struct interface *ifp;
{
	register struct sockaddr_in6 *sin;

	bzero((caddr_t)&ifr, sizeof(ifr));
	strncpy(ifr.iflr_name, ifp->name, IFNAMSIZ);

	sin = (struct sockaddr_in6 *)&ifr.addr;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(ifp->addr, sin->sin6_addr);
	ifr.flags = IFLR_FIRST;

	DUMPR("SIOCALIFADDR-prompt");
	if (ioctl(u, SIOCALIFADDR, (caddr_t)&ifr) < 0) {
		logerr("ioctl (SIOCALIFADDR) prefix %s on interface %s: %s",
		       inet6pr(&ifp->addr), ifp->name, strerror(errno));
	} else if (verbose)
		logtrace("prompt prefix %s on interface %s",
			 inet6pr(&ifp->addr), ifp->name);

	if (ifp == def_if) {
		if (BEST(ifp) != NULL)
			change_default();
		else
			cleanup_default();
		if (mobile)
			mobile_chgpref(ifp);
	}
}

/*
 * Add an off-link prefix to an interface (ie add a route)
 */

void
add_prefroute(pp)
	register struct prefix *pp;
{
	register char *cp = m_rtmsg.m_space;
	struct interface *ifp = pp->ifp;
	struct sockaddr_in6 *sin;
	struct sockaddr_dl *sdl;
	int len;

	if (pp->plen == 128) {
		if (verbose)
			logtrace("cancel add prefroute for %s with length 128",
				 inet6pr(&pp->addr));
		return;
	}

	if (mobile && mobile_chgrt(pp, RTM_ADD, ND_OPT_PI_FLAG_ONLINK)) {
		if (verbose)
			logtrace("cancel add prefroute for %s/%d on %s",
				 inet6pr(&pp->addr), pp->plen, pp->ifp->name);
		return;
	}

	bzero((caddr_t)&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = RTM_ADD;
	rtm.rtm_flags = RTF_UP;
	if (ifp->hasphy && (ifp->ltoken == MACTOKEN))
		rtm.rtm_flags |= RTF_CLONING;
	rtm.rtm_addrs = RTA_DST | RTA_GATEWAY | RTA_NETMASK | RTA_IFP;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;
	rtm.rtm_index = ifp->index;

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(pp->pref, sin->sin6_addr);
	ADVANCE(cp, sin);

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(ifp->addr, sin->sin6_addr);
	ADVANCE(cp, sin);

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_UNSPEC;
	COPY_ADDR6(pp->pmask, sin->sin6_addr);
	ADVANCE(cp, sin);

	sdl = (struct sockaddr_dl *)cp;
	sdl->sdl_family = AF_LINK;
	sdl->sdl_len = sizeof(struct sockaddr_dl);
	sdl->sdl_index = ifp->index;
	ADVANCE(cp, sdl);

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

	DUMPR("add_prefroute");
	if (write(r, (char *)&m_rtmsg, len) < len) {
		if (verbose)
			logperror("write (add_prefroute)");
	} else if (verbose)
		logtrace("add prefix route %s/%d on %s",
			 inet6pr(&pp->addr), pp->plen, ifp->name);
}

/*
 * Delete an off-link prefix from an interface (ie remove a route)
 */

void
del_prefroute(pp)
	register struct prefix *pp;
{
	register char *cp = m_rtmsg.m_space;
	struct sockaddr_in6 *sin;
	int len;

	if (pp->plen == 128) {
		if (verbose)
			logtrace("cancel del prefroute for %s with length 128",
				 inet6pr(&pp->addr));
		return;
	}

	if (mobile && mobile_chgrt(pp, RTM_DELETE, ND_OPT_PI_FLAG_ONLINK)) {
		if (verbose)
			logtrace("cancel del prefroute for %s/%d on %s",
				 inet6pr(&pp->addr), pp->plen, pp->ifp->name);
		return;
	}

	bzero((caddr_t)&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = RTM_DELETE;
	rtm.rtm_flags = RTF_UP;
	rtm.rtm_addrs = RTA_DST | RTA_NETMASK;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;
	rtm.rtm_index = pp->ifp->index;

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(pp->pref, sin->sin6_addr);
	ADVANCE(cp, sin);

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_UNSPEC;
	COPY_ADDR6(pp->pmask, sin->sin6_addr);
	ADVANCE(cp, sin);

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

	DUMPR("del_prefroute");
	if (write(r, (char *)&m_rtmsg, len) < len) {
		if (verbose)
			logperror("write (del_prefroute)");
	} else if (verbose)
		logtrace("delete prefix route %s/%d",
			 inet6pr(&pp->addr), pp->plen);
}

/*
 * A route has changed
 * (new one, different flags, invalid, get/lost its address)...
 */

void
update_router(rp)
	register struct router *rp;
{
	(void)gettimeofday(&now, NULL);

	if (rp->tp) {
		kill_timeout(rp->tp);
		rp->tp = NULL;
	}

	if (rp->aflags != (rp->nflags & ND_RA_FLAG_CFG)) {
		/* managed flag change */
#ifdef notyet
		if (rp->ifp->state != STATENOLL)
			stateful(rp->aflags, rp->nflags);
#endif
		rp->ifp->stateful = (rp->nflags & ND_RA_FLAG_CFG);
	}
	rp->aflags = rp->nflags & ND_RA_FLAG_CFG;

	if (rp->lifetime == 0) {
		router_invalid(rp);
		return;
	}

	if (rp->phy.sdl_alen && (BEST(rp->ifp)->phy.sdl_alen == 0)) {
		TAILQ_REMOVE(&rp->ifp->routers, rp, rt_list);
		TAILQ_INSERT_HEAD(&rp->ifp->routers, rp, rt_list);
		if (rp->ifp == def_if)
			change_default();
	}

	rp->tp = new_timeout(rp->lifetime, 0,
			     (func_to) router_invalid, rp);

	sort_router(rp->ifp, 1);

	if (rp == cur_rt)
		cur_rt = NULL;
}

/*
 * A router has become invalid (no more a router, timer expired)
 */

void
router_invalid(rp)
	register struct router *rp;
{
	if (rp->tp) {
		kill_timeout(rp->tp);
		rp->tp = NULL;
	}

	if (verbose)
		logtrace("invalided router %s",
			 inet6pr(&rp->addr.sin6_addr));
	del_router(rp, 1);
	TAILQ_REMOVE(&rp->ifp->routers, rp, rt_list);
	if (rp == cur_rt)
		cur_rt = NULL;
	if (IS_UP(rp->ifp))
		sort_router(rp->ifp, 1);
	free(rp);
}

/*
 * A router has become unreachable (ie its address is unknown)
 */

void
router_unreachable(rp, do_sort)
	register struct router *rp;
	int do_sort;
{
	(void)gettimeofday(&now, NULL);

	if (rp->lifetime < BaseReachableTime) {
		router_invalid(rp);
		return;
	}

	if (rp->tp) {
		kill_timeout(rp->tp);
		rp->tp = NULL;
	}

	rp->phy.sdl_alen = 0;
	rp->lifetime = BaseReachableTime;
	if (verbose)
		logtrace("unreachable router %s (keep=%d)",
			 inet6pr(&rp->addr.sin6_addr), BaseReachableTime);

	rp->tp = new_timeout(rp->lifetime, 0,
			     (func_to) router_invalid, rp);

	if (rp->rt_list.tqe_next) {
		/* move this unreachable router at the end of the list */

		TAILQ_REMOVE(&rp->ifp->routers, rp, rt_list);
		TAILQ_INSERT_TAIL(&rp->ifp->routers, rp, rt_list);

		if (do_sort)
			sort_router(rp->ifp, 1);
	}
}

/*
 * Purge old routers on an interface
 */

void
purge_routers(ifp, before, op)
	struct interface *ifp;
	struct timeval *before;
	int op;
{
	struct router *rp, *rp1 = NULL;

	if (verbose)
		logtrace("purge routers op=%d [%s] before=%s on %s",
			 op, dump_purge_op(op),
			 tvpr(before), ifp->name);

	/* sanity */
	if ((op == PURGE_OTHER) || (op == DEPREC_OTHER)) {
		if (cur_rt == NULL) {
			logerr("purge routers %d [%s]: no current router",
			       op, dump_purge_op(op));
			abort();
		}
		if (cur_rt->ifp != ifp) {
			logerr("purge routers %d [%s]: interface mismatch",
			       op, dump_purge_op(op));
			logerr("%s != %s", cur_rt->ifp->name, ifp->name);
			abort();
		}
	}

	switch (op) {
	    case PURGE_NOP:
		/* no operation */ 
		return;

	    case DEPREC_OLDER:
		/* depreciate all routers older than before */
		/* first phase: invalid unreachables */
	    again_do:
		FORALLRP(rp, ifp)
			if ((rp->phy.sdl_alen == 0) &&
			    !AFTER(rp->last, *before))
				rp1 = rp;
		if (rp1 != NULL) {
			router_invalid(rp1);
			rp1 = NULL;
			goto again_do;
		}
		/* second phase: make older unreachable */
	    again_du:
		FORALLRP(rp, ifp)
			if ((rp->phy.sdl_alen != 0) &&
			    !AFTER(rp->last, *before))
				rp1 = rp;
		if (rp1 != NULL) {
			router_unreachable(rp1, 0);
			del_router(rp1, 0);
			rp1 = NULL;
			goto again_du;
		}
		return;

	    case PURGE_OLDER:
		/* purge all routers older than before */
	    again_po:
		FORALLRP(rp, ifp)
			if (!AFTER(rp->last, *before))
				rp1 = rp;
		if (rp1 != NULL) {
			router_invalid(rp1);
			rp1 = NULL;
			goto again_po;
		}
		return;

	    case DEPREC_OTHER:
		/* depreciate all but not the current routers */
		/* first phase: invalid unreachables */
	    again_d:
		FORALLRP(rp, ifp)
			if ((rp != cur_rt) && (rp->phy.sdl_alen == 0))
				rp1 = rp;
		if (rp1 != NULL) {
			router_invalid(rp1);
			rp1 = NULL;
			goto again_d;
		}
		/* second phase: make not current unreachable */
	    again_u:
		FORALLRP(rp, ifp)
			if ((rp != cur_rt) && (rp->phy.sdl_alen != 0))
				rp1 = rp;
		if (rp1 != NULL) {
			router_unreachable(rp1, 0);
			del_router(rp1, 0);
			rp1 = NULL;
			goto again_u;
		}
		/* last phase: make not current old */
		FORALLRP(rp, ifp)
			if (rp != cur_rt) {
				rp->last.tv_sec = cur_rt->last.tv_sec - 1;
				rp->last.tv_usec =
					cur_rt->last.tv_usec + 999999;
				NORMTIME(rp->last);
			}
		return;

	    case PURGE_OTHER:
		/* purge all but not the current routers on the interface */
	    again:
		FORALLRP(rp, ifp)
			if (rp != cur_rt)
				rp1 = rp;
		if (rp1 != NULL) {
			router_invalid(rp1);
			rp1 = NULL;
			goto again;
		}
		return;

	    case PURGE_ALL:
		/* purge all routers on the interface */
		while ((rp = BEST(ifp)) != NULL) {
			while (rp->rt_list.tqe_next != NULL)
				rp = rp->rt_list.tqe_next;
			router_invalid(rp);
		}
		return;
	}
}

/*
 * Sort the router list:
 * 	order is only based on the availability of the link-layer address
 */

void
sort_router(ifp, reset)
	register struct interface *ifp;
	int reset;
{
	register struct router *rp, *rp2;
	int changed = 0;

#define	SWAP_ROUTER	\
  	if (rp == BEST(ifp)) \
  		changed++; \
	TAILQ_REMOVE(&ifp->routers, rp2, rt_list); \
	TAILQ_INSERT_BEFORE(rp, rp2, rt_list); \
	goto again;

    again:

	for (rp = BEST(ifp); rp; rp = rp2) {
		rp2 = rp->rt_list.tqe_next;
		if (rp2 == NULL)
			break;
		if (rp->phy.sdl_alen == 0) {
			if (rp2->phy.sdl_alen) {
				SWAP_ROUTER
			}
		} else if (rp2->phy.sdl_alen == 0) {
			continue;
		}
	}

	if (booting) {
		reset_default();
		return;
	}

	if ((changed == 0) && (reset == 0))
		return;

	if (ifp == def_if)
		reset_default();
}

/*
 * Try to get the link-layer address of a router
 */

void
get_rtphy(rp)
	register struct router *rp;
{
	register char *cp = m_rtmsg.m_space;
	struct sockaddr_in6 *sin;
	struct sockaddr_dl *sdl;
	register int i;
	int len;

	if (!rp->ifp->hasphy) {
		rp->phy.sdl_alen = NOPHY;
		return;
	}

	bzero((caddr_t)&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = RTM_GET;
	rtm.rtm_flags = RTF_UP | RTF_GATEWAY | RTF_HOST | RTF_STATIC;
	rtm.rtm_addrs = RTA_DST | RTA_IFP;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;
	rtm.rtm_index = rp->ifp->index;

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_family = AF_INET6;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	COPY_ADDR6(rp->addr.sin6_addr, sin->sin6_addr);
	ADVANCE(cp, sin);

	sdl = (struct sockaddr_dl *)cp;
	sdl->sdl_family = AF_LINK;
	sdl->sdl_len = sizeof(struct sockaddr_dl);
	sdl->sdl_index = rp->ifp->index;
	ADVANCE(cp, sdl);

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;
	cp = (caddr_t)&m_rtmsg;

	DUMPR("get_rtphy(write)");
	if (write(r, cp, len) < len) {
		if (verbose)
			logperror("write (get_rtphy)");
		return;
	}

	for (;;) {
		if (debug)
			bzero(cp, MAX_MSG_ROUTE);
		len = recv(r, cp, MAX_MSG_ROUTE, MSG_PEEK);
		if (len < 0) {
			if (verbose)
				logperror("recv (get_rtphy)");
			return;
		}
		if ((rtm.rtm_seq != rtm_seq) || (rtm.rtm_pid != pid))
			recv_route();
		else
			break;
	}
	if (debug)
		bzero(cp, MAX_MSG_ROUTE);
	len = read(r, cp, MAX_MSG_ROUTE);
	if (len < 0) {
		if (verbose)
			logperror("read (get_rtphy)");
		return;
	} else if (len  == 0) {
		if (verbose)
			logerr("read (get_rtphy) return 0");
		return;
	}
	if ((rtm.rtm_seq != rtm_seq) || (rtm.rtm_pid != pid)) {
		if (verbose)
			logerr("lost (get_rtphy)");
		return;
	}

	if (rtm.rtm_version != RTM_VERSION) {
		if (verbose)
			logerr("bad message version %d", rtm.rtm_version);
		return;
	}
	if (rtm.rtm_msglen > len) {
		if (verbose)
			logerr("bad message length (%d > %d)",
			       rtm.rtm_msglen, len);
		return;
	}
	if (rtm.rtm_errno) {
		if (verbose)
			logerr("get route failed (%s)",
			       strerror(rtm.rtm_errno));
		return;
	}
	DUMPR("get_rtphy(read)");
	if (rtm.rtm_flags & RTF_CLONING) {
		if (verbose)
			logtrace("get_rtphy & RTF_CLONING");
#if 0
		router_invalid(rp);
#endif
		return;
	}

#define RTF_LINKLOCAL	(RTF_UP | RTF_HOST | RTF_LLINFO)
	if ((rtm.rtm_flags & RTF_LINKLOCAL) != RTF_LINKLOCAL) {
		if (verbose)
			logerr("bad link-local route");
		return;
	}
	if (rtm.rtm_index != rp->ifp->index) {
		if (verbose)
			logerr("bad interface index (%d != %d)",
			       rtm.rtm_index, rp->ifp->index);
		return;
	}
	if ((rtm.rtm_addrs & RTA_GATEWAY) == 0) {
		if (verbose)
			logerr("not enough addresses for redirect");
		return;
	}
	cp = m_rtmsg.m_space;
	for (i = 0; i < RTAX_MAX; i++) {
		if ((rtm.rtm_addrs & (1 << i)) == 0)
			continue;
		switch (i) {
		    case RTAX_DST:
			sin = (struct sockaddr_in6 *)cp;
			if ((sin->sin6_family != AF_INET6) ||
			    !SAME_ADDR6(rp->addr.sin6_addr, sin->sin6_addr))
				return;
			ADVANCE(cp, sin);
			break;

		    case RTAX_GATEWAY:
			sdl = (struct sockaddr_dl *)cp;
			if ((sdl->sdl_family != AF_LINK) ||
			    (sdl->sdl_alen != rp->ifp->lmac))
				return;
			rp->phy.sdl_alen = sdl->sdl_alen;
			bcopy(LLADDR(sdl), LLADDR(&rp->phy), sdl->sdl_alen);
			ADVANCE(cp, sdl);
			break;

		    default:
			sin = (struct sockaddr_in6 *)cp;
			ADVANCE(cp, sin);
		}
	}

	if (verbose)
		logtrace("%sget physical address %s of %s",
			 rp->phy.sdl_alen ? "" : "try to ",
			 mac2ascii(LLADDR(&rp->phy), rp->phy.sdl_alen),
			 inet6pr(&rp->addr.sin6_addr));

	if (rp->phy.sdl_alen) {
		sort_router(rp->ifp, 0);
		return;
	}
}

/*
 * Delete a router
 */

void
del_router(rp, do_flush)
	register struct router *rp;
	int do_flush;
{
	register char *cp = m_rtmsg.m_space;
	struct sockaddr_in6 *sin;
	int len;

	bzero((caddr_t)&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = RTM_DELETE;
	rtm.rtm_flags = RTF_UP | RTF_GATEWAY | RTF_HOST | RTF_STATIC;
	rtm.rtm_addrs = RTA_DST;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;
	rtm.rtm_index = rp->ifp->index;

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_family = AF_INET6;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	COPY_ADDR6(rp->addr.sin6_addr, sin->sin6_addr);
	ADVANCE(cp, sin);

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;
	cp = (caddr_t)&m_rtmsg;

	DUMPR("del_router");
	if (write(r, cp, len) < len) {
		if (errno == ESRCH) {
			if (verbose)
				logtrace("delete router %s already done",
					 inet6pr(&rp->addr.sin6_addr));
		} else {
			if (verbose)
				logperror("write (del_router)");
			return;
		}
	} else if (verbose)
		logtrace("delete router %s", inet6pr(&rp->addr.sin6_addr));

	if (do_flush)
		flush(FLUSH_DYN, &rp->addr.sin6_addr, NULL, rp->ifp->index);
}

/*
 * The default router can be changed, compute (ie sort) things:
 * 	- first try to get the best interface:
 *		1 - up/nd enabled
 * 		2 - has at least a router
 * 		3 - first router has a link-layer address
 * 		4 - has a prefix
 * 		5 - has a not deprecated prefix
 * 		6 - has a global prefix
 * 		7 - is the best interface
 *		8 - is the main interface
 */

int in_reset_default = 0;	/* avoid running twice */
struct timeout *tp_reset_default;

static void
reset_default_restart()
{
	tp_reset_default = NULL;
	reset_default();
}

int mobile_has_booted = 0;

void
reset_default()
{
	register struct interface *ifp, *mob_if, *best_if;
	register struct router *rp, *rp2;
	register struct prefix *pp;
	int changing, goodies = 0;

	if (in_reset_default) {		/* already running */
		in_reset_default = -1;
		return;
	}
	if (tp_reset_default) {		/* suppress the scheduled call */
		kill_timeout(tp_reset_default);
		tp_reset_default = NULL;
	}
	mob_if = best_if = main_if;

	/* compute the prefered interface for mobility */

	if (mobile) {
		struct homeaddr *hap = PRIMHA;

		/* no home address? */
		if (hap == NULL)
			goto restarted;

		/* the main interface has a real home-prefix? */
		FORALLHP(pp, hap)
			if ((pp->ifp == mob_if) &&
			    ((pp->mflags & MOB_ISFA) == 0))
				goto restarted;

		/* try first real home-prefix */
		pp = HAPP(hap);
		if (pp && ((pp->mflags & MOB_ISFA) == 0)) {
			mob_if = best_if = pp->ifp;
			if (verbose)
				logtrace("mobility change initial best "
					 "interface for home to %s",
					 pp->ifp->name);
			goto restarted;
		}

		/* try the care-of prefix */
		pp = hap->capp;
		if (pp != NULL) {
			mob_if = best_if = pp->ifp;
			if (verbose)
				logtrace("mobility change initial best "
					 "interface for care-of to %s",
					 pp->ifp->name);
			goto restarted;
		}
		mob_if = NULL;
	} else
		mob_if = NULL;

    restarted:

	(void)gettimeofday(&now, NULL);
	in_reset_default = 1;
	FORALLIFP(ifp) {
		if (IS_UP(ifp)) {
			if (goodies == 0) {
				best_if = ifp;
				goodies++;
			}
		} else
			continue;
		/* beware : can call recursive calls to reset_default */
		for (rp = BEST(ifp); rp; rp = rp2) {
			rp2 = rp->rt_list.tqe_next;
			if (rp->phy.sdl_alen == 0)
				get_rtphy(rp);
		}
		if (in_reset_default < 0)
			goto restarted;
		if ((rp = BEST(ifp))) {
			if (goodies == 1) {
				best_if = ifp;
				goodies++;
			}
		} else
			continue;
		if ((rp = BEST(ifp))->phy.sdl_alen) {
			if (goodies == 2) {
				best_if = ifp;
				goodies++;
			}
		} else
			continue;

		pp = FIRST(ifp);
		while (pp && (pp->mflags & MOB_ISFA))
			pp = pp->prf_list.tqe_next;
		if (pp) {
			if (goodies == 3) {
				best_if = ifp;
				goodies++;
			}
		} else
			continue;
		if (pp->deplife > 0) {
			if (goodies == 4) {
				best_if = ifp;
				goodies++;
			}
		} else
			continue;
		if (!IS_SITELADDR6(pp->addr)) {
			if (goodies == 5) {
				best_if = ifp;
				goodies++;
			}
		} else
			continue;
		if (ifp == mob_if) {
			if (goodies == 6) {
				best_if = ifp;
				goodies++;
			}
		} /* no else continue here! */
		if (ifp == def_if) {
			if (goodies == 6) {
				best_if = ifp;
				goodies++;
			}
		} else
			continue;
		if (ifp == main_if) {
			best_if = ifp;
			goodies++;
			break;
		}
	}
	in_reset_default = 0;

	/* turn into a daemon at the first call */
	if (booting) {
		if (verbose)
			logtrace("first call to reset_default");
		booting = 0;
		if (!debug) {
			daemon(0, 0);
			pid = getpid();
			initlog();
		}
	}

	/* we had no main interface ? */
	if (goodies && main_if == NULL)
		main_if = best_if;

	changing = def_if != best_if;

	if (goodies <= 1) {
		/* no useable router: everything is on-link */
		if (def_if) {
			cleanup_default();
			/* is there an useable interface? */
			if (goodies > 0)
				def_if = best_if;
			else {
				def_if = NULL;
				changing++;
			}
			if (mobile && changing)
				mobile_chgif();
		}
	} else {
		def_if = best_if;
		if (mobile_has_booted == 0)
			mobile_booted();
		mobile_has_booted = 1;
		change_default();
		if (mobile && changing)
			mobile_chgif();
	}

	/* everything is good ? */
	if (goodies >= 6)
		return;

	/* retry in order to get better */
	(void)gettimeofday(&now, NULL);

	tp_reset_default = new_timeout((goodies <= 2 ? 1 : 10), 0,
				       (func_to) reset_default_restart, 0);
}

/*
 * Reput the direct/everything on-link default route
 */

void
cleanup_default()
{
	register char *cp = m_rtmsg.m_space;
	register struct interface *ifp;
	register struct sockaddr_in6 *sin;
	int len;

	ifp = (main_if ? main_if : def_if);
	if (ifp == NULL)
		return;
	del_default();

	/* mobile_chgif will do it */
	if (mobile && PRIMHA && (PRIMHA->state != HAST_UNKNOWN))
		return;

	bzero((caddr_t)&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = RTM_ADD;
	rtm.rtm_flags = RTF_UP | RTF_STATIC;
	if (ifp->hasphy && (ifp->ltoken == MACTOKEN))
		rtm.rtm_flags |= RTF_CLONING;
	rtm.rtm_addrs = RTA_DST | RTA_GATEWAY | RTA_NETMASK;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;
	rtm.rtm_index = ifp->index;

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	ADVANCE(cp, sin);

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(ifp->llocal, sin->sin6_addr);
	ADVANCE(cp, sin);

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_UNSPEC;
	ADVANCE(cp, sin);

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

	DUMPR("cleanup_default");
	if (write(r, (char *)&m_rtmsg, len) < len)
		logperror("write (cleanup_default)");
	else if (verbose)
		logtrace("cleanup default route");
}

/*
 * Delete the default route
 */

void
del_default()
{
	register char *cp = m_rtmsg.m_space;
	struct sockaddr_in6 *sin;
	int len;

	bzero((caddr_t)&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = RTM_DELETE;
	rtm.rtm_flags = RTF_UP | RTF_GATEWAY | RTF_STATIC;
	rtm.rtm_addrs = RTA_DST | RTA_NETMASK;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	ADVANCE(cp, sin);

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_UNSPEC;
	ADVANCE(cp, sin);

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

	DUMPR("del_default");
	if (write(r, (char *)&m_rtmsg, len) == len)
		if (verbose)
			logtrace("delete default route");

	def_rt = NULL;
}

/*
 * Change the default route (both the gateway and the interface address)
 */

void
change_default()
{
	register struct interface *ifp;
	register struct router *rp;
	register char *cp;
	struct sockaddr_in6 *sin;
	struct sockaddr_dl *sdl;
	int len, type = RTM_CHANGE;
	static int safety = 0;

	if ((ifp = def_if) == NULL) {
		logerr("no default interface?");
		abort();
	}
	if ((rp = BEST(ifp)) == NULL) {
		logerr("no default router?");
		abort();
	}
	if (def_rt == NULL) {
		del_default();
		if (def_ifa.index)
			flush(FLUSH_ND, NULL, NULL, def_ifa.index);
		else if (main_if)
			flush(FLUSH_ND, NULL, NULL, main_if->index);
		type = RTM_ADD;
	} else if ((rp == def_rt) &&
		   (ifp->index == def_ifa.index) &&
		   SAME_ADDR6(ifp->addr, def_ifa.addr) &&
		   (safety++ < 10)) {
		if (verbose)
			logtrace("unnecessary change_default for (%lx,%lx)?",
				 (u_long)ifp, (u_long)rp);
		return;
	} else if ((ifp->index != def_ifa.index) && def_rt)
		flush(FLUSH_DYN, &def_rt->addr.sin6_addr, NULL, def_ifa.index);

	if (verbose) {
		if (safety <= 10) 
			logtrace("will change default");
		else
			logtrace("will force default by safety");
	}

    retry:

	bzero((caddr_t)&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = type;
	rtm.rtm_flags = RTF_UP | RTF_GATEWAY | RTF_STATIC;
	rtm.rtm_addrs = RTA_DST | RTA_GATEWAY | RTA_NETMASK;
	rtm.rtm_addrs |= RTA_IFP | RTA_IFA;

	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;
	rtm.rtm_rmx.rmx_mtu = ifp->mtu;
	rtm.rtm_inits = RTV_MTU;
	rtm.rtm_index = ifp->index;

	cp = m_rtmsg.m_space;

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	ADVANCE(cp, sin);

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(rp->addr.sin6_addr, sin->sin6_addr);
	ADVANCE(cp, sin);

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_UNSPEC;
	ADVANCE(cp, sin);

	sdl = (struct sockaddr_dl *)cp;
	sdl->sdl_family = AF_LINK;
	sdl->sdl_len = sizeof(struct sockaddr_dl);
	sdl->sdl_index = ifp->index;
	ADVANCE(cp, sdl);

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(ifp->addr, sin->sin6_addr);
	ADVANCE(cp, sin);

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

	DUMPR("change_default");
	if (write(r, (char *)&m_rtmsg, len) < len) {
		if ((type == RTM_CHANGE) && (errno == ESRCH)) {
			type = RTM_ADD;
			goto retry;
		}
		if (verbose)
			logperror("write (change_default)");
		return;
	} else if (verbose)
		logtrace("change default via %s with %s on link#%d",
			 inet6pr(&rp->addr.sin6_addr),
			 inet6pr(&ifp->addr),
			 ifp->index);
	def_rt = rp;
	def_ifa.index = ifp->index;
	COPY_ADDR6(ifp->addr, def_ifa.addr);
	safety = 0;
}

/*
 * Flush any IPv6 stuff for an interface
 */

void
flushif(ifp)
	register struct interface *ifp;
{
	struct sockaddr_in6 *sin;

	/* flush routers */

	purge_routers(ifp, NULL, PURGE_ALL);

	/* prefixes already done */

	/* flush the link-local address */

	bzero((caddr_t)&ifr, sizeof(ifr));
	strncpy(ifr.iflr_name, ifp->name, IFNAMSIZ);

	sin = (struct sockaddr_in6 *)&ifr.addr;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(ifp->llocal, sin->sin6_addr);

	DUMPR("SIOCDLIFADDR-flushif");
	if (ioctl(u, SIOCDLIFADDR, (caddr_t)&ifr) < 0) {
		logperror("ioctl (SIOCDLIFADDR) on LL prefix");
	} else if (verbose)
		logtrace("delete LL prefix on interface %s", ifp->name);

	/* flush all routes */

	flush(FLUSH_IF, NULL, NULL, ifp->index);

	/* reset state to no link-local */

	ifp->state = STATENOLL;
	ifp->stateful = ND_RA_FLAG_CFG;
}

/*
 * Flush a matching route
 */

static void
flushone(op, addr, mask, index)
	int op;
	struct in6_addr *addr, *mask;
	u_int index;
{
	register char *cp = m_rtmsg.m_space;
	struct sockaddr *sa;
	struct sockaddr_in6 *sin, *dst = NULL;
	struct sockaddr_dl *sdl;
	int i, len, plen = 128;

	DUMPR("flushone get");

	switch (op) {
	    case FLUSH_ALL:
		goto delete;
	    case FLUSH_NOP:
		return;
	    default:
		if (rtm.rtm_index != index)
			return;
	}
	switch (op) {
	    case FLUSH_IF:
		goto delete;
	    case FLUSH_GW:
		if ((rtm.rtm_flags & RTF_GATEWAY) == 0)
			return;
		break;
	    case FLUSH_ND:
	    case FLUSH_NDPF:
		if ((rtm.rtm_flags & RTF_LLINFO) == 0)
			return;
		/* fallin */
	    case FLUSH_DYN:
		if ((rtm.rtm_flags & (RTF_DYNAMIC | RTF_HOST)) == 0)
			return;
	}

	for (i = 0; i < RTAX_MAX; i++) {
		if ((rtm.rtm_addrs & (1 << i)) == 0)
			continue;
		switch (i) {
		    case RTAX_DST:
			dst = (struct sockaddr_in6 *)cp;
			if (dst->sin6_family != AF_INET6)
				return;
#define T(x)	(((dst->sin6_addr.s6_addr[x] ^ addr->s6_addr[x]) \
			& mask->s6_addr[x]) != 0)

			if ((op == FLUSH_NDPF) &&
			    (T(0) || T(1) || T(2) || T(3) ||
			     T(4) || T(5) || T(6) || T(7)))
				return;
#undef T			    
			ADVANCE(cp, dst);
			break;

		    case RTAX_GATEWAY:
			sa = (struct sockaddr *)cp;
			switch (op) {
			    case FLUSH_GW:
			    case FLUSH_DYN:
				sin = (struct sockaddr_in6 *)cp;
				if ((sin->sin6_family == AF_INET6) &&
				    SAME_ADDR6(*addr, sin->sin6_addr))
					goto delete;
				return;
			    case FLUSH_ND:
				sdl = (struct sockaddr_dl *)cp;
				if ((sdl->sdl_family == AF_LINK) &&
				    (sdl->sdl_alen == 0))
					goto delete;
				return;
			    case FLUSH_NDPF:
				sdl = (struct sockaddr_dl *)cp;
				if (sdl->sdl_family != AF_LINK)
					return;
			}
			ADVANCE(cp, sa);
			break;

		    case RTAX_NETMASK:
			sin = (struct sockaddr_in6 *)cp;
			if (sin->sin6_family != AF_UNSPEC)
				return;
			plen = get_prefixlen(sin);
			ADVANCE(cp, sin);
			break;

		    case RTAX_IFA:
			sin = (struct sockaddr_in6 *)cp;
			if (sin->sin6_family != AF_INET6)
				return;
			if (((op == FLUSH_IFA) || (op == FLUSH_NDPF)) &&
			    SAME_ADDR6(*addr, sin->sin6_addr))
				goto delete;
			return;

		    default:
			sa = (struct sockaddr *)cp;
			ADVANCE(cp, sa);
			break;
		}
	}

	if (verbose)
		logerr("flushone in unreachable?");
	return;

   delete:

	rtm.rtm_type = RTM_DELETE;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;
	len = rtm.rtm_msglen;

	DUMPR("flushone delete");
	if (write(r, (char *)&m_rtmsg, len) < len) {
		if (verbose)
			logperror("write (flushone)");
		return;
	} else if (verbose) {
		if (dst)
			logtrace("flush route to %s/%d",
				 inet6pr(&dst->sin6_addr), plen);
		else
			logtrace("flush route");
	}
}

/*
 * Get the unicast IPv6 route table
 */

#undef rtm

void
flush(op, addr, mask, index)
	int op;
	struct in6_addr *addr, *mask;
	u_int index;
{
	register struct rt_msghdr *rtm;
	char *buf, *cp;
	unsigned bufsize;
	int mib[6];

	if (verbose)
		logtrace("flush op=%d [%s] index=%d addr=%s",
			 op, dump_flush_op(op), index,
			 addr ? inet6pr(addr) : "none");

	mib[0] = CTL_NET;
	mib[1] = PF_ROUTE;
	mib[2] = 0;
	mib[3] = AF_INET6;
	mib[4] = NET_RT_FLAGS;
	mib[5] = RTF_UP;

	if (sysctl(mib, 6, NULL, &bufsize, NULL, 0) < 0) {
		if (verbose)
			logperror("flush: sysctl0");
		return;
	}
	bufsize = (3 * bufsize) / 2 + sizeof(m_rtmsg);
	buf = (char *)malloc(bufsize);
	if (buf == NULL) {
		if (verbose)
			logerr("out of memory");
		return;
	}
	if (sysctl(mib, 6, buf, &bufsize, NULL, 0) < 0) {
		if (verbose)
			logperror("flush: sysctl");
		free(buf);
		return;
	}

	for (cp = buf; cp < buf + bufsize; cp += rtm->rtm_msglen) {
		rtm = (struct rt_msghdr *)cp;

		if ((rtm->rtm_msglen < sizeof(*rtm)) ||
		    (rtm->rtm_version != RTM_VERSION) ||
		    (rtm->rtm_type != RTM_GET)) {
			if (verbose)
				logerr("flush: lost synchro");
			free(buf);
			return;
		}
		if (cp + rtm->rtm_msglen > buf + bufsize) {
			if (verbose)
				logerr("flush: trailing garbage");
			free(buf);
			return;
		}
		if (rtm->rtm_msglen > sizeof(m_rtmsg)) {
			if (verbose)
				logerr("flush: very large (%d) message",
				       (int)rtm->rtm_msglen);
			continue;
		}

		bcopy(cp, &m_rtmsg, rtm->rtm_msglen);
		flushone(op, addr, mask, index);
	}
	free(buf);
}

/* translate purge operation for human beings */

static char *
dump_purge_op(op)
	int op;
{
	switch (op) {
	    case PURGE_NOP:
		return "NOP";

	    case DEPREC_OLDER:
		return "Dold";

	    case PURGE_OLDER:
		return "Pold";

	    case DEPREC_OTHER:
		return "D<>";

	    case PURGE_OTHER:
		return "P<>";

	    case PURGE_ALL:
		return "ALL";

	    default:
		return "???";
	}
}

/* translate flush operation for human beings */

static char *
dump_flush_op(op)
	int op;
{
	switch (op) {
	    case FLUSH_NOP:
		return "NOP";

	    case FLUSH_IF:
		return "IF";

	    case FLUSH_IFA:
		return "IFA";

	    case FLUSH_GW:
		return "GW";

	    case FLUSH_ND:
		return "ND";

	    case FLUSH_NDPF:
		return "NDPF";

	    case FLUSH_DYN:
		return "DYN";

	    case FLUSH_ALL:
		return "ALL";

	    default:
		return "???";
	}
}
