#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <signal.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/sysctl.h>
#include <sys/uio.h>
#include <sys/resource.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_types.h>
#ifndef IF_QFULL
#include <net/if_var.h>
#endif
#include <net/route.h>

#include <netinet/in_systm.h>
#define INET6
#include <netinet/in.h>
#include <netinet/in6_var.h>
#include <netinet/ip6.h>
#include <netinet/icmp6.h>
#include <netinet/ip_var.h>
#include <netinet/ip6_var.h>
#include <netinet/ip6_opts.h>

#include <netdb.h>
#include <arpa/inet.h>

#ifdef TRACE
void DUMP(const char *, void *, int);
#define DUMPR(x) DUMP(x, &rtm, rtm.rtm_msglen)
#else
#define DUMPR(x)
#endif

#define NPR6BUF	8	/* max printing computed strings in parallel */

#ifndef TAILQ_INSERT_BEFORE
#define	TAILQ_INSERT_BEFORE(listelm, elm, field) {			\
	(elm)->field.tqe_prev = (listelm)->field.tqe_prev;		\
	(elm)->field.tqe_next = (listelm);				\
	*(listelm)->field.tqe_prev = (elm);				\
	(listelm)->field.tqe_prev = &(elm)->field.tqe_next;		\
}
#endif

#define Calloc(p, t)	(p = (struct t *)calloc(1,sizeof(struct t)))

/* for some sanity checks */
#define MAXHOP	255
#define MINHOP	1

#define MAXTOKEN	12	/* That is a max */
#define MACTOKEN	8	/* eth, fddi, ...*/
#define MACMAC		6
#define PPPTOKEN	8	/* ppp */
#define SITTOKEN	8	/* sit point to point */
#define SITSIT		4

/* millisecond to timeval */
#define M_SECOND			1000	/* ms per sec */
#define MU_SECOND			1000	/* microsec per M_SECOND */
#define MS_TO_SEC(x)			((x)/M_SECOND)
#define MS_TO_USEC(x)			(((x)%M_SECOND)*MU_SECOND)

/* from RFC 2461
	MAX_INITIAL_RTR_ADVERT_INTERVAL  16 seconds
	MAX_INITIAL_RTR_ADVERTISEMENTS    3 transmissions
	MAX_FINAL_RTR_ADVERTISEMENTS      3 transmissions
	MIN_DELAY_BETWEEN_RAS             3 seconds
	MAX_RA_DELAY_TIME                 .5 seconds
	AdvSendAdvertisements	FALSE
	MaxRtrAdvInterval Default: 600 seconds
	MinRtrAdvInterval Default: 0.33 * MaxRtrAdvInterval
	AdvManagedFlag Default: FALSE
	AdvOtherConfigFlag Default: FALSE
	AdvLinkMTU Default: 0
	AdvReachableTime Default: 0
	AdvRetransTimer Default: 0
	AdvMaxHopLimit Default:  value in most recent "Assigned Numbers" RFC
	AdvDefaultLifetime Default: 3 * MaxRtrAdvInterval
	AdvPrefixList Each prefix has an associated:
	       AdvOnLinkFlag Default:        TRUE
	       AdvPreferredLifetime Default: 7 days
	       AdvValidLifetime Default:    30 days
	       ValidLifetime Threshold:      2 hours
   from mobility draft 8
	INITIAL_BINDACK_TIMEOUT		  1 second
	MAX_BINDACK_TIMEOUT		256 seconds
	MAX_UPDATE_RATE			  1 per second
	SLOW_UPDATE_RATE		  1 per 10 seconds
	MAX_FAST_UPDATES		  5 transmissions
	MAX_ADVERT_REXMIT		  3 transmissions
*/

#define MAX_RTR_SOLICITATIONS 		3
#define RTR_SOLICITATION_INTERVAL 	4.0
#define MAX_RTR_SOLICITATION_DELAY	1.0

#define MAX_ANYCAST_DELAY_TIME		(M_SECOND)
#define MAX_INITIAL_RTR_ADVERT_INTERVAL	(16*M_SECOND)
#define MAX_INITIAL_RTR_ADVERTISEMENTS	3
#define MAX_FINAL_RTR_ADVERTISEMENTS	3
#define MIN_DELAY_BETWEEN_RAS		(3*M_SECOND)
#define MAX_RA_DELAY_TIME		(M_SECOND/2)
#define DEF_RTR_ADV_INTERVAL		(600*M_SECOND)
#define MAX_RTR_ADV_INTERVAL		(1800*M_SECOND)
#define MIN_RTR_ADV_INTERVAL		(4*M_SECOND)
#define MAX_RT_LIFE			0xffff

#define INITIAL_BINDACK_TIMEOUT		1
#define MAX_BINDACK_TIMEOUT		256
#define MAX_ADVERT_REXMIT		3

/* Router Adv flags */
#define ND_RA_FLAG_ALL	\
	(ND_RA_FLAG_MANAGED |	\
	 ND_RA_FLAG_OTHER |	\
	 ND_RA_FLAG_HAGENT)
#define ND_RA_FLAG_CFG		(ND_RA_FLAG_MANAGED | ND_RA_FLAG_OTHER)

/* Prefix announces */
#define INFLIFE		0xffffffff
#define DEFPREFLIFE	(7 * 24 * 60 * 60)
#define DEFVALLIFE	(30 * 24 * 60 * 60)
#define MINVALLIFE	(2 * 60 * 60)
#define ND_OPT_PI_FLAG_ALL	\
	(ND_OPT_PI_FLAG_ONLINK |	\
	 ND_OPT_PI_FLAG_AUTO |		\
	 ND_OPT_PI_FLAG_RTADDR |	\
	 ND_OPT_PI_FLAG_SITE)
#define ND_OPT_PI_FLAG_CFG	(ND_OPT_PI_FLAG_ONLINK | ND_OPT_PI_FLAG_AUTO)
#define ND_OPT_PI_FLAG_RECV	1	/* NOT in ND_OPT_PI_FLAG_ALL */

/* Prefix traversal */
#define LOOPALLPP(init, pp)	for (init; pp; pp = pp->prf_list.tqe_next)
#define FORALLPP(pp, ifp)	LOOPALLPP(pp = FIRST(ifp), pp)
#define FIRST(ifp)	(ifp->prefixes.tqh_first)

/* Interface traversal */
#define IS_UP(ifp)	((ifp->state != STATENOLL) && (ifp->flags & IFF_UP))
#define FORALLIFP(ifp)	\
	for (ifp = interfaces.tqh_first; ifp; ifp = ifp->if_list.tqe_next)
/* in the sdl_alen field of phy address, means no address */
#define NOPHY	255

typedef void (*func_to) (void *);
struct timeout {
	TAILQ_ENTRY(timeout) tm_list;	/* timeout list */
	struct timeval tv;		/* when */
	func_to func;			/* function */
	void *arg;			/* argument */
	int status;			/* 0: inactive, >0: in timeouts list */
					/* <0: being expired (do not free!) */
	int refcnt;			/* refcount: if 0, expire=>free */
};

/* route socket */
#define MAX_MSG_ROUTE	2048
struct {
       	struct rt_msghdr m_rtm;
        char m_space[MAX_MSG_ROUTE - sizeof(struct rt_msghdr)];
} m_rtmsg;
#define rtm	m_rtmsg.m_rtm
extern int rtm_seq;
extern int pid;

#define ROUNDUP(a) \
        ((a) > 0 ? (1 + (((a) - 1) | (sizeof(int32_t) - 1))) : sizeof(int32_t))
#define ADVANCE(x, n) (x += ROUNDUP(((struct sockaddr *)n)->sa_len))

extern struct timeval now;

#define NORMTIME(tv) \
	while ((tv).tv_usec >= 1000000) { \
		(tv).tv_sec += 1; \
		(tv).tv_usec -= 1000000; \
	}

/* tv1 == tv2 ? */
#define SIMULT(tv1, tv2) \
	(((tv1).tv_sec == (tv2).tv_sec) && ((tv1).tv_usec == (tv2).tv_usec))

/* tv1 >= tv2 ? */
#define AFTER(tv1, tv2) \
	(((tv1).tv_sec > (tv2).tv_sec) || \
	 (((tv1).tv_sec == (tv2).tv_sec) && ((tv1).tv_usec >= (tv2).tv_usec)))

struct prefix;
struct interface;
int in_cksum(u_short *, int);
struct timeout *new_timeout(int, int, func_to, void *);
void set_timeout(struct timeout *, int, int, func_to, void *);
void add_timeout(struct timeout *);
void dump_timeout(struct timeout *);
void kill_timeout(struct timeout *);
void run_timeouts(void);
void apply_mask(struct in6_addr *, struct in6_addr *);
void mk_mask(struct in6_addr *, int);
void mk_pref(struct prefix *);
u_char get_prefixlen(struct sockaddr_in6 *);
char *mac2ascii(caddr_t, int);
const char *lifepr(u_int32_t);
#ifdef NDPD_ROUTER
union datelife;
const char *datelifepr(int, union datelife *);
#else
void mk_addr(struct prefix *);
void mk_peer(struct prefix *);
void mk_local_peer(struct prefix *);
#endif
void initlog(void);
void logerr(char *, ...), logtrace(char *, ...), logdebug(char *, ...);
void logperror(char *);
const char *inet6pr(struct in6_addr *);
const char *tvpr(struct timeval *);
