/* Manage mobile cache entry in kernel */

#include "ndpd.h"
#include "mobility.h"

/* Add a binding in kernel */

void
add_binding(mop)
	struct mobile *mop;
{
	struct ip6_mobopt mopt;

	bzero(&mopt, sizeof(mopt));
	bcopy(&mop->careof, &mopt.mo_addr, sizeof(mopt.mo_addr));
	bcopy(&mop->home, &mopt.mo_filt, sizeof(mopt.mo_filt));
	mopt.mo_flen = FULL_MATCHING;

	if (setsockopt(m, IPPROTO_DSTOPTS, IP6_DOPT_RTADD,
		       (char *)&mopt, sizeof(mopt)) < 0) {
		if (verbose)
			logperror("setsockopt (IP6_DOPT_RTADD)");
	} else if (verbose)
		logtrace("add_binding for %s via %s",
			 inet6pr(&mop->home), inet6pr(&mop->careof));
}

/* Delete a binding in kernel */

void
del_binding(mop)
	struct mobile *mop;
{
	struct ip6_mobopt mopt;

	bzero(&mopt, sizeof(mopt));
	bcopy(&mop->home, &mopt.mo_filt, sizeof(mopt.mo_filt));
	mopt.mo_flen = FULL_MATCHING;

	if (setsockopt(m, IPPROTO_DSTOPTS, IP6_DOPT_RTDELETE,
		       (char *)&mopt, sizeof(mopt)) < 0) {
		if (verbose)
			logperror("setsockopt (IP6_DOPT_RTDELETE)");
	} else if (verbose)
		logtrace("del_binding for %s", inet6pr(&mop->home));
}

/* Change a binding in kernel */

void
change_binding(mop)
	struct mobile *mop;
{
	del_binding(mop);
	add_binding(mop);
}

/* Disable a binding in kernel */

void
disable_binding(fd)
	int fd;
{
	struct ip6_mobopt mopt;

	bzero(&mopt, sizeof(mopt));
	mopt.mo_flen = PCB_MATCHING;

	if (setsockopt(fd, IPPROTO_DSTOPTS, IP6_DOPT_RTDISABLE,
		       (char *)&mopt, sizeof(mopt)) < 0) {
		if (verbose)
			logperror("setsockopt (IP6_DOPT_RTDISABLE)");
	}
}

/* Flush the host route */

void
del_route(addr)
	struct in6_addr *addr;
{
	register char *cp = m_rtmsg.m_space;
	struct sockaddr_in6 *sin;
	int len;

	if (IN6_IS_ADDR_UNSPECIFIED(addr) ||
	    IN6_IS_ADDR_MULTICAST(addr) ||
	    IN6_IS_ADDR_LINKLOCAL(addr) ||
	    IN6_IS_ADDR_SITELOCAL(addr)) {
		if (verbose)
			logerr("del_route (mobility) on bad address %s",
			       inet6pr(addr));
		return;
	}

	bzero((caddr_t)&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = RTM_DELETE;
	rtm.rtm_flags = RTF_UP | RTF_HOST;
	rtm.rtm_addrs = RTA_DST;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(*addr, sin->sin6_addr);
	ADVANCE(cp, sin);

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

	DUMPR("del_route (mobility)");
	if (write(r, (char *)&m_rtmsg, len) < len) {
		if (verbose && (errno != ESRCH))
			logperror("write (del_route in mobility)");
		return;
	} else if (verbose)
		logtrace("delete mobility route %s", inet6pr(addr));
}
