/* Security Association management */

#include "ndpd.h"
#include "mobility.h"
#include <netinet/ipsec.h>

/* IPsec policy */

char *ipsec_policy;
int ipsec_len;

int is_secure = -1;

static void build_policy(void);

/* Is IPsec in use for mobility */

int
check_ipsec()
{
	int mib[] = { CTL_NET, PF_INET6, IPPROTO_AH, IPSECCTL_MOBILITY };
	int val, valsize, cc;

	valsize = sizeof(int);
	cc = sysctl(mib, 4, &val, &valsize, NULL, 0);
	if (cc < 0) {
		if ((errno == ENOENT) && verbose)
			logtrace("kernel without IPsec");
		logperror("sysctl (net.inet6.ipsec.mobility)");
		return (0);
	}
	if (val < 0) {
		logerr("sysctl net.inet6.ipsec.mobility -> %d", val);
		return (0);
	}
	if (val > 1) {
		if (verbose)
			logtrace("sysctl net.inet6.ipsec.mobility -> %d", val);
		val = 1;
	}

	if (val != 0)
		build_policy();
	return (val);
}	

static void
build_policy()
{
	static struct sadb_x_policy p;

	bzero((caddr_t)&p, sizeof(p));

	p.sadb_x_policy_len = sizeof(p) >> 3;
	p.sadb_x_policy_exttype = SADB_X_EXT_POLICY;
	p.sadb_x_policy_type = IPSEC_POLICY_ENTRUST;
	p.sadb_x_policy_dir = IPSEC_DIR_OUTBOUND;
	p.sadb_x_policy_reserved = 0;

	ipsec_len = sizeof(p);
	ipsec_policy = (char *)&p;
}

int
get_security(mop)
	struct mobile *mop;
{
	if (is_secure == -1)
		is_secure = check_ipsec();
	return (1);
}

void
del_security(mop)
	struct mobile *mop;
{
	if (mop->sap)
		free(mop->sap);
	mop->sap = NULL;
}
