/* Sending */

#include "ndpd.h"
#include "mobility.h"

#define MAXOPTIONS	1024	/* max options size */
static char opts[MAXOPTIONS];

/* Send an acknowledgement (TODO: AH) */

void
send_bnd_ack(mop, status, refresh)
	struct mobile *mop;
	u_char status;
	u_int refresh;
{
	struct sockaddr_in6 sin;
	register u_char *cp;
	char babuf[16];
	int d, cc;

	/* get a fresh socket */

	if ((d = socket(PF_INET6, SOCK_RAW, IPPROTO_DSTOPTS)) < 0) {
		logperror("socket (send binding ack)");
		return;
	}
	(void) shutdown(d, 0);

	/* bind it */

	bzero((caddr_t)&sin, sizeof(sin));
	sin.sin6_family = AF_INET6;
	sin.sin6_len = sizeof(sin);
	COPY_ADDR6(mop->me, sin.sin6_addr);

	if (bind(d, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
		logperror("bind (send binding ack)");
		(void) close(d);
		return;
	}

	/* connect it */

	COPY_ADDR6(mop->home, sin.sin6_addr);

	if (connect(d, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
		logperror("connect (send binding ack)");
		(void) close(d);
		return;
	}

	/* security policy */

	if ((ipsec_len > 0) &&
	    (setsockopt(d, IPPROTO_IPV6, IPV6_IPSEC_POLICY,
			ipsec_policy, ipsec_len) < 0)) {
		logperror("setsockopt (binding ack policy)");
		(void) close(d);
		return;
	}

	/* fill a destination option header with the binding ack */

	cp = (u_char *)babuf;

	SET8(IP6_NHDR_NONH, cp);
	SET8(1, cp);
	SET8(OPT6_PAD_1, cp);
	SET8(OPT6_BINDING_ACK, cp);
	SET8(11, cp);
	SET8(status, cp);
	SET16(mop->seq, cp);
	SET32(mop->lifetime, cp);
	SET32(refresh, cp);

	if (status >= OPT6_BNDA_UNSPEC) {

		/* fill routing header */

		bzero(opts, 40);

		cp = (u_char *)opts;

		SET8(IP6_NHDR_RT, cp);
		SET8(2, cp);
		SET8(IP6_LSRRT, cp);
		SET8(1, cp);
		SET32(0, cp);
		SETADDR(mop->home, cp);
		SETADDR(mop->careof, cp);

		/* set options */

		if (setsockopt(d, IPPROTO_IP, IP_OPTIONS, opts, 40) < 0) {
			if (verbose)
				logperror("setsockopt (send binding ack)");
			(void) close(d);
			return;
		}
	}

	/* send */

	cc = write(d, babuf, sizeof(babuf));

	if (verbose) {
		if (cc != sizeof(babuf))
			logperror("write (send binding ack)");
		else
			logtrace("send binding ack [%d] to %s",
				 status, inet6pr(&mop->home));
	}
	(void) close(d);
}

/* Send a binding request */

void
send_bnd_req(mop)
	struct mobile *mop;
{
	struct sockaddr_in6 sin;
	register u_char *cp;
	char brbuf[8];
	int d, cc;

	/* get a fresh socket */

	if ((d = socket(PF_INET6, SOCK_RAW, IPPROTO_DSTOPTS)) < 0) {
		logperror("socket (send binding request)");
		return;
	}
	(void) shutdown(d, 0);

	/* bind it */

	bzero((caddr_t)&sin, sizeof(sin));
	sin.sin6_family = AF_INET6;
	sin.sin6_len = sizeof(sin);
	COPY_ADDR6(mop->me, sin.sin6_addr);

	if (bind(d, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
		logperror("bind (send binding request)");
		(void) close(d);
		return;
	}

	/* connect it */

	COPY_ADDR6(mop->home, sin.sin6_addr);

	if (connect(d, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
		logperror("connect (send binding request)");
		(void) close(d);
		return;
	}

	/* security policy: not needed */

	/* fill a destination option header with the binding request */

	cp = (u_char *)brbuf;

	SET8(IP6_NHDR_NONH, cp);
	SET8(0, cp);
	SET8(OPT6_BINDING_REQ, cp);
	SET8(0, cp);
	SET8(OPT6_PAD_N, cp);
	SET8(2, cp);
	SET16(0, cp);

	/* send */

	cc = write(d, brbuf, sizeof(brbuf));

	if (verbose) {
		if (cc != sizeof(brbuf))
			logperror("write (send binding request)");
		else
			logtrace("send binding request to %s",
				 inet6pr(&mop->home));
	}
	(void) close(d);
}

#ifdef HOME_AGENT

/* Send a home agent discovery reply */

void
send_ha_disc(mop, pp)
	struct mobile *mop;
	struct prefix *pp;
{
	struct hagent *hap;
	struct sockaddr_in6 sin;
	register u_char *cp;
	char *babuf;
	int d, n, me, size, cc;

	/* compute the buffer size */

	n = me = 0;
	if (pp->hacnt > 0) {
		n = pp->hacnt;
		if (n > 12)
			n = 12;
		FORALLHA(hap, pp)
			if (pp->ifp->opt.hapref < hap->hapref)
				me++;
			else
				break;
		if (me) {
			n++;
			me = n - me;
			if (me <= 0)
				me = 1;
		}
	}
	size = 24 + (n * 16);

	/* allocate the buffer */

	if (verbose)
		logtrace("send ha discovery reply (size=%d, %d items)",
			 size, n);
	if (me && verbose)
		logtrace("including me at rank %d", me);

	if ((babuf = malloc(size)) == NULL) {
		logerr("out of memory");
		return;
	}

	/* get a fresh socket */

	if ((d = socket(PF_INET6, SOCK_RAW, IPPROTO_DSTOPTS)) < 0) {
		logperror("socket (send ha discovery reply)");
		free(babuf);
		return;
	}
	(void) shutdown(d, 0);

	/* bind it */

	bzero((caddr_t)&sin, sizeof(sin));
	sin.sin6_family = AF_INET6;
	sin.sin6_len = sizeof(sin);
	COPY_ADDR6(mop->me, sin.sin6_addr);

	if (bind(d, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
		logperror("bind (send ha discovery reply)");
		free(babuf);
		(void) close(d);
		return;
	}

	/* connect it */

	COPY_ADDR6(mop->home, sin.sin6_addr);

	if (connect(d, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
		logperror("connect (send ha discovery reply)");
		free(babuf);
		(void) close(d);
		return;
	}

	/* TODO: security policy */

	/* disable binding */

	disable_binding(d);

	/* fill a destination option header with the binding ack */

	cp = (u_char *)babuf;
	hap = pp->hagents.tqh_first;

	SET8(IP6_NHDR_NONH, cp);
	SET8(2 * (n + 1), cp);
	SET8(OPT6_PAD_1, cp);
	SET8(OPT6_BINDING_ACK, cp);
	SET8(19 + 16 * n, cp);
	SET8(OPT6_BNDA_DYNDIS, cp);
	SET16(mop->seq, cp);
	SET32(0, cp);
	SET32(0, cp);
	SET8(MOB6_SUBOPT_PADN, cp);
	SET8(4, cp);
	SET32(0, cp);
	/* phased-out sub-option! */
	SET8(3, cp);
	SET8(n * 16, cp);
	while (n > 0) {
		if (me == n) {
			SETADDR(pp->addr, cp);
			n--;
		} else {
			SETADDR(hap->addr, cp);
			hap = hap->ha_list.tqe_next;
			n--;
		}
	}

	/* fill routing header */

	bzero(opts, 40);

	cp = (u_char *)opts;

	SET8(IP6_NHDR_RT, cp);
	SET8(2, cp);
	SET8(IP6_LSRRT, cp);
	SET8(1, cp);
	SET32(0, cp);
	SETADDR(mop->home, cp);
	SETADDR(mop->careof, cp);

	/* set options */

	if (setsockopt(d, IPPROTO_IP, IP_OPTIONS, opts, 40) < 0) {
		if (verbose)
			logperror("setsockopt (send ha discovery reply)");
		free(babuf);
		(void) close(d);
		return;
	}

	/* send */

	cc = write(d, babuf, size);

	if (verbose) {
		if (cc != size)
			logperror("write (send ha discovery reply)");
		else
			logtrace("send ha discovery reply to %s for %s",
				 inet6pr(&mop->home), inet6pr(&mop->me));
	}
	free(babuf);
	(void) close(d);
}
#endif
