/* Common part */

#include "ndpd.h"
#include "mobility.h"

static void dump_mobile(struct mobile *);
#ifdef HOME_AGENT
static void dump_tunnel(struct tunnel *);
#endif
#ifdef MOBILE
static void dump_corr(struct corr *);
#endif

void
dump_mobiles()
{
	register struct mobile *mop;
#ifdef HOME_AGENT
	struct tunnel *tup;
#endif
#ifdef MOBILE
	struct corr *cop;
#endif

	FORALLMOB(mop)
		dump_mobile(mop);
	logdebug("<<<kernel mobile");
	kvm_mobiles();
	logdebug("kernel mobile>>>");
#ifdef HOME_AGENT
	FORALLTUN(tup)
		dump_tunnel(tup);
#endif
#ifdef MOBILE
	FORALLCOR(cop)
		dump_corr(cop);
#endif
}

static void
dump_mobile(mop)
	struct mobile *mop;
{
	logdebug("mobile (%lx):", (u_long)mop);
	logdebug("\thome address=%s", inet6pr(&mop->home));
	logdebug("\tcare-of address=%s", inet6pr(&mop->careof));
	logdebug("\tmy address=%s", inet6pr(&mop->me));
	logdebug("\ttimeout=%lx", (u_long)(mop->tp));
	logdebug("\tlast refresh=%lu (%ld)",
		 mop->refresh, now.tv_sec - mop->refresh);
	logdebug("\tlifetime=%s", lifepr(mop->lifetime));
	logdebug("\tseq=%d", mop->seq);
	logdebug("\tflags=%x [%s]", mop->flags, bnd_upd_flg(mop->flags));
	logdebug("\tplen=%x", mop->plen);
	logdebug("\tsap=%lx", (u_long)(mop->sap));
#ifdef HOME_AGENT
	if (mop->pp) {
		logdebug("\tprefix (%lx)", (u_long)(mop->pp));
		logdebug("\t\t%s/%d on %s",
			 inet6pr(&mop->pp->pref), mop->pp->plen,
			 mop->pp->ifp->name);
	}
	if (mop->ap) {
		logdebug("\tproxy (%lx)", (u_long)(mop->ap));
		logdebug("\t\t%s, type=%d",
			 inet6pr(&mop->ap->addr), mop->ap->type);
	}
	if (mop->tup)
		logdebug("\ttunnel=%s (%lx)",
			 mop->tup->name, (u_long)(mop->tup));
#endif
}

#ifdef HOME_AGENT

static void
dump_tunnel(tup)
	struct tunnel *tup;
{
	logdebug("tunnel (name=%s, state=%d)", tup->name, tup->state);
}
#endif

#ifdef MOBILE

static void
dump_corr(cop)
	struct corr *cop;
{
	logdebug("correspondent (%lx):", (u_long)cop);
	if (cop->type & CORR_PREFIX)
		logdebug("\taddr=%s/%d", inet6pr(&cop->addr), cop->plen);
	else
		logdebug("\taddr=%s", inet6pr(&cop->addr));
	logdebug("\tcare-of=%s", inet6pr(&cop->careof));
	logdebug("\thome=%s", inet6pr(&cop->home));
	logdebug("\ttype=%x [%s]", cop->type, corr_type(cop->type));
	logdebug("\tflags=%x [%s]", cop->flags, bnd_upd_flg(cop->flags));
	logdebug("\tseq=%d, life=%d", cop->seq, cop->lifetime);
	logdebug("\tlast seen=%lu (%ld)", cop->last, now.tv_sec - cop->last);
	logdebug("\ttimeout=%lx", (u_long)(cop->tp));
	logdebug("\tretrans counter=%d", cop->retcnt);
	logdebug("\tretrans timeout=%lx", (u_long)(cop->rettp));
	logdebug("\tsap=%lx", (u_long)(cop->sap));
}
#endif

/* print bits */

char *
bnd_upd_flg(flags)
	u_char flags;
{
	static char pbuf[16];

	bzero(pbuf, sizeof(pbuf));
	snprintf(pbuf, sizeof(pbuf) - 1, "%s%s%s%s",
		 flags & OPT6_BNDU_ACK ? "A" : "",
		 flags & OPT6_BNDU_HR ? "H" : "",
		 flags & OPT6_BNDU_RT ? "R" : "",
		 flags & OPT6_BNDU_DAD ? "D" : "");
	return pbuf;
}

#ifdef MOBILE

char *
corr_type(type)
	u_int type;
{
	static char pbuf[16];

	bzero(pbuf, sizeof(pbuf));
	snprintf(pbuf, sizeof(pbuf) - 1, "%s%s%s%s%s",
		 type & CORR_PREFIX ? "P" : "",
		 type & CORR_SETHA ? "A" : "",
		 type & CORR_CLONING ? "C" : "",
		 type & CORR_USEBND ? "U" : "",
		 type & CORR_PROTECT ? "-" : "");
	return pbuf;
}
#endif
