/* Retransmission management for a mobile node */

#include "ndpd.h"
#include "mobility.h"

#ifdef MOBILE

static void schedule_ret(struct corr *);

/* start retransmit stuff */

void
start_retrans(cop)
	struct corr *cop;
{
	struct timeout *tp;

	(void)gettimeofday(&now, NULL);

	/* allocate a timeout */

	if (Calloc(tp, timeout) == NULL) {
		if (verbose)
			logerr("out of memory");
		return;
	}

	/* cleanup state */

	if (cop->rettp)
		cancel_retrans(cop);
	cop->retcnt = 0;

	/* schedule first retransmit */

	cop->rettp = tp;
	set_timeout(tp, INITIAL_BINDACK_TIMEOUT, 0,
		    (func_to) schedule_ret, cop);
}

/* Cancel retranmit stuff */

void
cancel_retrans(cop)
	struct corr *cop;
{
	if (cop->rettp) {
		kill_timeout(cop->rettp);
		cop->rettp = NULL;
	}
}

/* Schedule a retransmit */

static int backoff[11] = { 1, 1, 2, 4, 8, 16, 32, 64, 128, 256, 256 };

#define BACKOFF(x)	\
	((x) >= 10 ? MAX_BINDACK_TIMEOUT : backoff[x])

static void
schedule_ret(cop)
	struct corr *cop;
{
	(void)gettimeofday(&now, NULL);

	/* cancel timeout */

	if (cop->rettp == NULL) {
		if (verbose)
			logerr("schedule without timeout?");
		return;
	}

	cop->rettp->refcnt++;	/* keep timeout structure */
	kill_timeout(cop->rettp);

	/* resend the binding update or home registration */

	cop->seq++;
	cop->retcnt++;

	if (cop->flags & OPT6_BNDU_HR)
		send_home_reg(cop);
	else
		send_bnd_upd(cop);

	cop->rettp->refcnt--;

	/* give up for binding requests */

	if ((cop->retcnt >= 10) && ((cop->flags & OPT6_BNDU_HR) == 0)) {
		if (verbose)
			logerr("giving up for bind update to %s",
			       inet6pr(&cop->addr));
		cancel_retrans(cop);
		return;
	}

	/* schedule next retrans */

	set_timeout(cop->rettp,
		    BACKOFF(cop->retcnt - (cop->flags & OPT6_BNDU_HR ? 1 : 0)),
		    0,
		    (func_to) schedule_ret,
		    cop);
}
#endif
