/* Sending for a mobile node */

#ifdef MOBILE

#include "ndpd.h"
#include "mobility.h"

static void add_home_reg(struct corr *);
static void del_home_reg(struct corr *, struct interface *);
static struct interface *home_if(struct corr *);

static void force_nd(struct corr *);

/* Send a binding update (TODO: AH) */

void
send_bnd_upd(cop)
	struct corr *cop;
{
	struct sockaddr_in6 sin;
	register u_char *cp;
	char bubuf[32];
	int d, cc;

	if (verbose) {
		logtrace("will send binding update");
		logtrace("addr=%s", inet6pr(&cop->addr));
		logtrace("home=%s", inet6pr(&cop->home));
		logtrace("care-of=%s", inet6pr(&cop->careof));
		logtrace("flags=%x [%s] life=%u seq=%d retcnt=%d",
			 cop->flags, bnd_upd_flg(cop->flags),
			 cop->lifetime, cop->seq, cop->retcnt);
	}

	/* hack: flush old route */

	if (cop->retcnt == 0)
		del_route(&cop->addr);

	/* get a fresh socket */

	if ((d = socket(PF_INET6, SOCK_RAW, IPPROTO_DSTOPTS)) < 0) {
		logperror("socket (send binding update)");
		return;
	}
	(void) shutdown(d, 0);

	/* bind it (note: applies ourself home address stuff) */

	bzero((caddr_t)&sin, sizeof(sin));
	sin.sin6_family = AF_INET6;
	sin.sin6_len = sizeof(sin);
	COPY_ADDR6(cop->careof, sin.sin6_addr);

	if (bind(d, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
		logperror("bind (send binding update)");
		(void) close(d);
		return;
	}

	/* connect it */

	COPY_ADDR6(cop->addr, sin.sin6_addr);

	if (connect(d, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
		logperror("connect (send binding update)");
		(void) close(d);
		return;
	}

	/* security policy */

	if ((ipsec_len > 0) &&
	    (setsockopt(d, IPPROTO_IPV6, IPV6_IPSEC_POLICY,
			ipsec_policy, ipsec_len) < 0)) {
		logperror("setsockopt (binding update policy)");
		(void) close(d);
		return;
	}

	/* disable any home-address automatic option */

	disable_homeaddr(d);

	/*
	 * Fill a destination option header with the binding update
	 * and a home address options.
	 */

	cp = (u_char *)bubuf;

	SET8(IP6_NHDR_NONH, cp);
	SET8(3, cp);

	SET8(OPT6_BINDING_UPD, cp);
	SET8(8, cp);
	SET8(cop->flags, cp);
	SET8(0, cp);
	SET16(cop->seq, cp);
	SET32(cop->lifetime, cp);

	SET8(OPT6_PAD_1, cp);
	SET8(OPT6_PAD_1, cp);

	SET8(OPT6_HOME_ADDR, cp);
	SET8(16, cp);
	SETADDR(cop->home, cp);

	/* send */

	cc = write(d, bubuf, sizeof(bubuf));

	if (verbose) {
		if (cc != sizeof(bubuf))
			logperror("write (send binding update)");
		else
			logtrace("sent binding update to %s",
				 inet6pr(&cop->addr));
	}
	(void) close(d);
}

/* Send a home registration (TODO: AH) */

void
send_home_reg(agp)
	struct corr *agp;
{
	struct interface *ifp;

	if (IN6_ARE_ADDR_EQUAL(&agp->careof, &agp->home) &&
	    ((ifp = home_if(agp)) != NULL))
		del_home_reg(agp, ifp);
	else
		add_home_reg(agp);

}

static void
add_home_reg(agp)
	struct corr *agp;
{
	struct sockaddr_in6 sin;
	register u_char *cp;
	char bubuf[32];
	int d, cc;

	if (verbose) {
		logtrace("will send home registration (add)");
		logtrace("addr=%s", inet6pr(&agp->addr));
		logtrace("home=%s", inet6pr(&agp->home));
		logtrace("care-of=%s", inet6pr(&agp->careof));
		logtrace("flags=%x [%s] life=%u seq=%d retcnt=%d",
			 agp->flags, bnd_upd_flg(agp->flags),
			 agp->lifetime, agp->seq, agp->retcnt);
	}

	/* hack: flush old route */

	if (agp->retcnt == 0)
		del_route(&agp->addr);

	/* get a fresh socket */

	if ((d = socket(PF_INET6, SOCK_RAW, IPPROTO_DSTOPTS)) < 0) {
		logperror("socket (add home registration)");
		return;
	}
	(void) shutdown(d, 0);

	/* bind it (note: applies ourself home address stuff) */

	bzero((caddr_t)&sin, sizeof(sin));
	sin.sin6_family = AF_INET6;
	sin.sin6_len = sizeof(sin);
	COPY_ADDR6(agp->careof, sin.sin6_addr);

	if (bind(d, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
		logperror("bind (add home registration)");
		(void) close(d);
		return;
	}

	/* connect it */

	COPY_ADDR6(agp->addr, sin.sin6_addr);

	if (connect(d, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
		logperror("connect (add home registration)");
		(void) close(d);
		return;
	}

	/* security policy */

	if ((ipsec_len > 0) &&
	    (setsockopt(d, IPPROTO_IPV6, IPV6_IPSEC_POLICY,
			ipsec_policy, ipsec_len) < 0)) {
		logperror("setsockopt (home registration policy)");
		(void) close(d);
		return;
	}

	/* disable any home-address automatic option */

	disable_homeaddr(d);

	/*
	 * Fill a destination option header with the home registration
	 * and the home address options.
	 */

	cp = (u_char *)bubuf;

	SET8(IP6_NHDR_NONH, cp);
	SET8(3, cp);

	SET8(OPT6_BINDING_UPD, cp);
	SET8(8, cp);
	SET8(agp->flags, cp);
	SET8(0, cp);	/* was agp->plen */
	SET16(agp->seq, cp);
	SET32(agp->lifetime, cp);

	SET8(OPT6_PAD_1, cp);
	SET8(OPT6_PAD_1, cp);

	SET8(OPT6_HOME_ADDR, cp);
	SET8(16, cp);
	SETADDR(agp->home, cp);

	/* send */

	cc = write(d, bubuf, sizeof(bubuf));

	if (verbose) {
		if (cc != sizeof(bubuf))
			logperror("write (add home registration)");
		else
			logtrace("sent home registration (add) to %s",
				 inet6pr(&agp->addr));
	}
	(void) close(d);

	/* setup tunnel */

	if (agp->retcnt == 0)
		madd_tunnel(agp);
}

static void
del_home_reg(agp, ifp)
	struct corr *agp;
	struct interface *ifp;
{
	struct sockaddr_in6 sin;
	register u_char *cp;
	char bubuf[56];
	int d, cc;

	if (verbose) {
		logtrace("will send home registration (delete)");
		logtrace("addr=%s", inet6pr(&agp->addr));
		logtrace("home=%s", inet6pr(&agp->home));
		logtrace("care-of=%s", inet6pr(&agp->careof));
		logtrace("source=%s", inet6pr(&ifp->llocal));
		logtrace("flags=%x [%s] life=%u seq=%d retcnt=%d",
			 agp->flags, bnd_upd_flg(agp->flags),
			 agp->lifetime, agp->seq, agp->retcnt);
	}

	/* hack: try to force a neighbor entry on come back */

	force_nd(agp);

	/* get a fresh socket */

	if ((d = socket(PF_INET6, SOCK_RAW, IPPROTO_DSTOPTS)) < 0) {
		logperror("socket (del home registration)");
		return;
	}
	(void) shutdown(d, 0);

	/* bind it (note: applies ourself home address stuff) */

	bzero((caddr_t)&sin, sizeof(sin));
	sin.sin6_family = AF_INET6;
	sin.sin6_len = sizeof(sin);
	COPY_ADDR6(ifp->llocal, sin.sin6_addr);
	sin.sin6_scope_id = ifp->index;

	if (bind(d, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
		logperror("bind (del home registration)");
		(void) close(d);
		return;
	}

	/* connect it */

	COPY_ADDR6(agp->addr, sin.sin6_addr);

	if (connect(d, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
		logperror("connect (del home registration)");
		(void) close(d);
		return;
	}

	/* security policy */

	if ((ipsec_len > 0) &&
	    (setsockopt(d, IPPROTO_IPV6, IPV6_IPSEC_POLICY,
			ipsec_policy, ipsec_len) < 0)) {
		logperror("setsockopt (home deregistration policy)");
		(void) close(d);
		return;
	}

	/* disable any home-address automatic option */

	disable_homeaddr(d);

	/*
	 * Fill a destination option header with the home registration
	 * (including the care-of address) and the home address options.
	 */

	cp = (u_char *)bubuf;

	SET8(IP6_NHDR_NONH, cp);
	SET8(6, cp);

	SET8(OPT6_BINDING_UPD, cp);
	SET8(28, cp);
	SET8(agp->flags, cp);
	SET8(0, cp);	/* was agp->plen */
	SET16(agp->seq, cp);
	SET32(agp->lifetime, cp);

	SET8(MOB6_SUBOPT_PAD1, cp);
	SET8(MOB6_SUBOPT_PAD1, cp);

	SET8(MOB6_SUBOPT_COA, cp);
	SET8(16, cp);
	SETADDR(agp->careof, cp);

	SET8(OPT6_PAD_N, cp);
	SET8(4, cp);
	SET32(0, cp);

	SET8(OPT6_HOME_ADDR, cp);
	SET8(16, cp);
	SETADDR(agp->home, cp);

	/* send */

	cc = write(d, bubuf, sizeof(bubuf));

	if (verbose) {
		if (cc != sizeof(bubuf))
			logperror("write (del home registration)");
		else
			logtrace("sent home registration (delete) to %s",
				 inet6pr(&agp->addr));
	}
	(void) close(d);
}

/* find the home interface where is the home agent */

static struct interface *
home_if(agp)
	struct corr *agp;
{
	struct homeaddr *hap;
	struct prefix *pp;
	struct interface *ifp;

	/* find the home address */

	FORALLHA(hap)
		if (IN6_ARE_ADDR_EQUAL(&agp->home, &hap->addr))
			break;
	if (hap == NULL) {
		logerr("home_if: no home address");
		return NULL;
	}
	if (hap->state != HAST_IN) {
		logerr("home_if: %s not at home", inet6pr(&hap->addr));
		return NULL;
	}

	/* find the home prefix */

	pp = HAPP(hap);
	if (pp == NULL) {
		logerr("home_if: no home prefix");
		return NULL;
	}
	if ((pp->mflags & MOB_ISHA) == 0)
		logerr("home_if: %s/%d is not a home prefix",
		       inet6pr(&pp->addr), pp->plen);
	if (pp->mflags & MOB_ISFA)
		logerr("home_if: %s/%d is a fake home prefix",
		       inet6pr(&pp->addr), pp->plen);
	if ((pp->alen == 128) && verbose)
		logtrace("home_if: %s/%d is not a real prefix (alen=%d)",
			 inet6pr(&pp->addr), pp->plen, pp->alen);

	/* find the home interface */

	ifp = pp->ifp;

	if (ifp == NULL) {
		logtrace("home_if: %s/%d has no interface",
			 inet6pr(&pp->addr), pp->plen);
		return NULL;
	}
	if (!IS_UP(ifp) && verbose)
		logtrace("home_if: home interface %s is not up",
			 ifp->name);

	/* finished */

	if (verbose)
		logtrace("home interface is %s (prefix %s/%d)",
			 ifp->name, inet6pr(&pp->addr), pp->plen);

	return ifp;
}

/*
 * Force a neighbor cache entry for the home agent when at home
 */

static void
force_nd(agp)
	struct corr *agp;
{
	register char *cp;
	register struct router *rp;
	struct homeaddr *hap;
	struct sockaddr_in6 *sin;
	struct sockaddr_dl *sdl;
	int len, type = RTM_ADD;

	FORALLHA(hap)
		if (IN6_ARE_ADDR_EQUAL(&agp->home, &hap->addr))
			break;
	if (hap == NULL) {
		if (verbose)
			logerr("force_nd: can't find the home address %s",
			       inet6pr(&agp->home));
		return;
	}
	if (HAPP(hap) == NULL) {
		if (verbose)
			logtrace("force_nd: no home prefix");
		return;
	}
	if (HAPP(hap)->mflags & MOB_ISFA) {
		if (verbose)
			logerr("force_nd: home prefix is fake");
		return;
	}
	if (HAPP(hap)->plen == 128) {
		if (verbose)
			logtrace("force_nd: home prefix length is 128");
		return;
	}
	if (HAPP(hap)->ifp == NULL) {
		if (verbose)
			logerr("force_nd: can't find interface for prefix %s",
			       inet6pr(&HAPP(hap)->addr));
		return;
	}
	if (!HAPP(hap)->ifp->hasphy) {
		if (verbose)
			logtrace("force_nd: no phy on interface %s",
				 HAPP(hap)->ifp->name);
		return;
	}
	FORALLRP(rp, HAPP(hap)->ifp)
		if (IN6_ARE_ADDR_EQUAL(&agp->addr, &rp->rta))
			break;
	if (rp == NULL) {
		if (verbose)
			logtrace("force_nd: can't find home agent "
				 "%s as a router",
				 inet6pr(&agp->addr));
		return;
	}
	if (rp->phy.sdl_alen == 0) {
		if (verbose)
			logtrace("force_nd: home agent %s has "
				 "no physical address",
				 inet6pr(&agp->addr));
		return;
	}

	/* do the route! */

	(void) gettimeofday(&now, NULL);

    retry:

	bzero((caddr_t)&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = type;
	rtm.rtm_flags = RTF_UP | RTF_HOST;
	rtm.rtm_addrs = RTA_DST | RTA_GATEWAY | RTA_IFP | RTA_IFA;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;
	rtm.rtm_index = rp->ifp->index;
	rtm.rtm_rmx.rmx_expire = now.tv_sec + 5;
	rtm.rtm_inits = RTV_EXPIRE;

	cp = m_rtmsg.m_space;

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(rp->rta, sin->sin6_addr);
	ADVANCE(cp, sin);

	sdl = (struct sockaddr_dl *)cp;
	sdl->sdl_family = AF_LINK;
	sdl->sdl_len = sizeof(struct sockaddr_dl);
	sdl->sdl_index = rp->ifp->index;
	sdl->sdl_alen = rp->phy.sdl_alen;
	bcopy(LLADDR(&rp->phy), LLADDR(sdl), sdl->sdl_alen);
	ADVANCE(cp, sdl);

	sdl = (struct sockaddr_dl *)cp;
	sdl->sdl_family = AF_LINK;
	sdl->sdl_len = sizeof(struct sockaddr_dl);
	sdl->sdl_index = rp->ifp->index;
	ADVANCE(cp, sdl);

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(agp->home, sin->sin6_addr);
	ADVANCE(cp, sin);

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

	DUMPR("force_nd");
	if (write(r, (char *)&m_rtmsg, len) < len) {
		if ((type == RTM_ADD) && (errno == EEXIST)) {
			type = RTM_CHANGE;
			goto retry;
		}
		if (verbose)
			logperror("write (force_nd)");
		return;
	} else if (verbose)
		logtrace("force_nd for %s on %s#%d",
			 inet6pr(&rp->rta), rp->ifp->name, rp->ifp->index);
}
#endif
