/* Booted mobile node */

#ifdef MOBILE

#include "ndpd.h"
#include "mobility.h"

/* Determine if we are at home or in visit */

void
where_at_boot(hap)
	struct homeaddr *hap;
{
	struct interface *ifp;
	struct prefix *pp = NULL;

	FORALLIFP(ifp)
	    if (IS_UP(ifp))
		for(pp = ifp->prefixes.tqh_first; pp;
		    pp = pp->prf_list.tqe_next)
		    if (IN6_ARE_ADDR_EQUAL(&pp->addr, &hap->addr))
			goto found;

	/* mobile in visit */

	if ((mscope & SCP_PREFIX) == 0)
		hap->state = HAST_OUT;
	return;

   found:

	/* mobile at home */

	hap->state = HAST_IN;
	pp->mflags = MOB_ISHA;
	if (pp->hap) {
		/* should never happen! */
		logerr("where_at_boot: prefix %s is already a home prefix",
		       inet6pr(&pp->addr));
		TAILQ_REMOVE(&pp->hap->homeprefs, pp, mob_link);
		pp->hap = NULL;
	}
	TAILQ_INSERT_HEAD(&hap->homeprefs, pp, mob_link);
	pp->hap = hap;

	/* prompt the home prefix */

	TAILQ_REMOVE(&ifp->prefixes, pp, prf_list);
	TAILQ_INSERT_HEAD(&ifp->prefixes, pp, prf_list);
}

/* Add a home prefix on an interface */

void
add_home_prefix(hap, ifp)
	struct homeaddr *hap;
	struct interface *ifp;
{
	struct prefix *pp;

	/* TODO: support more than one home address */

	if (hap != PRIMHA)
		return;

	if (!IS_UP(ifp))
		return;

	/* already done? */
	if ((hap->state == HAST_IN) && (HAPP(hap) == FIRST(ifp))) {
		pp = HAPP(hap);
		goto done;
	}

	if (Calloc(pp, prefix) == NULL) {
		if (verbose)
			logerr("out of memory");
		return;
	}

	if (verbose)
		logtrace("adding home prefix %s on interface %s",
			 inet6pr(&hap->addr), ifp->name);

	if ((ifp == def_if) || (HAPP(hap) == NULL))
		TAILQ_INSERT_HEAD(&hap->homeprefs, pp, mob_link);
	else
		TAILQ_INSERT_TAIL(&hap->homeprefs, pp, mob_link);
	pp->ifp = ifp;
	pp->hap = hap;
	COPY_ADDR6(hap->addr, pp->pref);
	COPY_ADDR6(hap->addr, pp->addr);
	pp->alen = pp->plen = 128;
	mk_mask(&pp->pmask, 128);
	pp->invlife = INFLIFE;
	pp->deplife = DEFPREFLIFE;
	pp->pflags = pp->nflags = ND_OPT_PI_FLAG_AUTO;
	pp->mflags = MOB_ISHA | MOB_ISFA;
	TAILQ_INSERT_HEAD(&ifp->prefixes, pp, prf_list);

	if (discarding)
		homeroute(NULL, RTM_DELETE);
	add_prefix(pp);

    done:
	if (ifp == def_if)
		mobile = 0;
	COPY_ADDR6(hap->addr, ifp->addr);
	prompt(ifp);
	mobile = 1;
	pp->ntosend = MAX_ADVERT_REXMIT;
	send_grat_adv(pp);
}

/* Final process when the default route is here */

void
read_final()
{
	struct homeaddr *hap = PRIMHA;
	struct corr *cop;
	int i;

	/* the good time to protect candidate care-of prefixes */

	protect_all();

	if (hap == NULL)
		return;

	/* unknown state is in fact in visit */

	FORALLHA(hap)
		if (hap->state == HAST_UNKNOWN)
			hap->state = HAST_OUT;

	/* manage only one home address */

	hap = PRIMHA;

	/* find the home agent */

	FORALLCOR(cop) {
		if ((cop->flags & OPT6_BNDU_HR) == 0)
			goto next;
		for (i = 0; i < 8; i++)
			if (hap->addr.s6_addr[i] != cop->addr.s6_addr[i])
				goto next;
		break;
	    next:
	}
	if (cop == NULL) {
		cop = correspondents.tqh_first;
		if (cop && (cop->flags & OPT6_BNDU_HR))
			logerr("no matching home agent");
	} else if (cop != correspondents.tqh_first) {
		/* put the home agent in front */
		TAILQ_REMOVE(&correspondents, cop, cr_list);
		TAILQ_INSERT_HEAD(&correspondents, cop, cr_list);
	}

	/* First: we are virtually at home */

	FORALLCOR(cop) {
		COPY_ADDR6(hap->addr, cop->careof);
		COPY_ADDR6(hap->addr, cop->home);

		/* a good time to set up security */
		mget_security(cop);
	}

	/* Second: are we really at home ? */

	if (hap->state == HAST_OUT)
		major_move(HAPP(hap), 0);
}
#endif
