/* Manage home-address cache entry in kernel (mobile part) */

#ifdef MOBILE

#include "ndpd.h"
#include "mobility.h"

/* Add/disable a home-address for a prefix in kernel */

void
add_homeaddr(cop)
	struct corr *cop;
{
	struct ip6_mobopt mopt;

	bzero(&mopt, sizeof(mopt));
	bcopy(&cop->careof, &mopt.mo_addr, sizeof(mopt.mo_addr));
	bcopy(&cop->addr, &mopt.mo_filt, sizeof(mopt.mo_filt));
	mopt.mo_flen = cop->plen;

	if ((cop->type & CORR_SETHA) == 0)
		goto disabling;

	/* add a home-address option output policy */

	if (verbose)
		logtrace("add homeaddr for %s/%d via %s",
			 inet6pr(&cop->addr),
			 cop->plen,
			 inet6pr(&cop->careof));

	if (setsockopt(m, IPPROTO_DSTOPTS, IP6_DOPT_HAADD,
		       (char *)&mopt, sizeof(mopt)) < 0) {
		if (verbose)
			logperror("setsockopt (IP6_DOPT_HAADD)");
	}
	return;

	/* disable a home-address option output policy */

    disabling:
	if (verbose)
		logtrace("disable homeaddr for %s/%d via %s",
			 inet6pr(&cop->addr),
			 cop->plen,
			 inet6pr(&cop->careof));

	if (setsockopt(m, IPPROTO_DSTOPTS, IP6_DOPT_HADISABLE,
		       (char *)&mopt, sizeof(mopt)) < 0) {
		if (verbose)
			logperror("setsockopt (IP6_DOPT_HADISABLE)");
	}
}

/* Delete a home-address in kernel */

void
del_homeaddr(cop)
	struct corr *cop;
{
	struct ip6_mobopt mopt;

	bzero(&mopt, sizeof(mopt));
	bcopy(&cop->addr, &mopt.mo_filt, sizeof(mopt.mo_filt));
	mopt.mo_flen = cop->plen;

	/* delete a home-address option output policy */

	if (verbose)
		logtrace("del homeaddr for %s/%d",
			 inet6pr(&cop->addr), cop->plen);

	if (setsockopt(m, IPPROTO_DSTOPTS, IP6_DOPT_HADELETE,
		       (char *)&mopt, sizeof(mopt)) < 0) {
		if (verbose)
			logperror("setsockopt (IP6_DOPT_HADELETE)");
	}
}

#ifdef notneeded
/* Change a home-address in kernel */

void
change_homeaddr(cop)
	struct corr *cop;
{
	del_homeaddr(cop);
	add_homeaddr(cop);
}
#endif

#endif
