/* Deal with interface changes */

#include "ndpd.h"
#include "mobility.h"

static void lastroute(void);

/* default interface has changed */

void
mobile_chgif()
{
#ifdef MOBILE
	register struct corr *cop;
	register struct homeaddr *hap = PRIMHA;
	struct prefix *newpp, *oldpp = NULL;

	if (!mobile || multi ||
	    (hap == NULL) || (hap->state == HAST_UNKNOWN))
		return;

	if (verbose)
		logtrace("mobile_chgif to %s",
			 def_if ? def_if->name : "none");

	homeroute(def_if, RTM_CHANGE);

	if (def_if == NULL) {
		logtrace("no default interface, use discard");
		FORALLHA(hap) {
			FORALLHP(newpp, hap) {
				/*
				 * something to do?
				 */
			}
			hap->capp = NULL;
		}
		FORALLCOR(cop)
			cancel_retrans(cop);
		lastroute();
		return;
	}

	/* check if there is something to do */

	newpp = FIRST(def_if);
	if (newpp && (newpp->mflags & MOB_ISHA) &&
	    ((newpp->mflags & MOB_ISFA) == 0)) {
		if (verbose)
			logtrace("mobile_chgif: already at home");
		return;
	}

	cop = correspondents.tqh_first;

	/* find the new care-of address */

	FORALLPP(newpp, def_if)
		if ((newpp->mflags & MOB_ISHA) == 0)
			break;
	if (newpp == NULL) {
		if (verbose)
			logtrace("mobile_chgif: no care-of address");
		return;
	}

	/* find a home address */

	/* first case: care-of address moves between interfaces */

	FORALLHA(hap) {
		oldpp = hap->capp;
		if (oldpp && (oldpp->ifp != def_if))
			break;
	}
	if (hap)
		goto found;

	/* second case: no care-of address */

	FORALLHA(hap) {
		oldpp = hap->capp;
		if (oldpp == NULL)
			break;
	}
	if (hap == NULL) {
		if (verbose)
			logtrace("mobile_chgif: no home address");
		return;
	}

    found:

	/* Trivial cases where nothing will be done */

	if (cop == NULL) {
		if (verbose)
			logtrace("mobile_chgif: no correspondent");
		return;
	}

	/* change care-of prefix */

	if (oldpp)
		oldpp->mflags &= ~MOB_ISCA;
	newpp->mflags |= MOB_ISCA;
	FORALLHA(hap)
		if (hap->capp == oldpp)
			hap->capp = newpp;

	/* do it! */

	FORALLCOR(cop)
		mobile_move(cop, &newpp->addr);
#else
	lastroute();
#endif
}

#ifdef MOBILE

/* (re)set the route to the home address */

int discarding = 0;

void
homeroute(ifp, cmd)
	struct interface *ifp;
	int cmd;
{
	register char *cp;
	struct homeaddr *hap = PRIMHA;
	struct prefix *pp;
	struct sockaddr_in6 *sin;
	struct sockaddr_dl *sdl;
	int len, useloop = ifp == NULL;

	if (hap == NULL)
		return;
	if (useloop)
		ifp = loop_if;
	if (ifp == NULL) {
		if (verbose)
			logtrace("no loopback for home route");
		return;
	}

	if (verbose)
		logtrace("%s home route for %s to %s",
			 cmd == RTM_CHANGE ? "change" :
			    (cmd == RTM_ADD ? "add" : "delete"),
			 inet6pr(&hap->addr), ifp->name);

	/* delete current route */

	if (cmd & RTM_DELETE) {

		bzero((caddr_t)&m_rtmsg, sizeof(m_rtmsg));
		rtm.rtm_version = RTM_VERSION;
		rtm.rtm_type = RTM_DELETE;
		rtm.rtm_flags = RTF_UP | RTF_HOST | RTF_LOCAL;
		rtm.rtm_addrs = RTA_DST;
		rtm.rtm_pid = pid;
		rtm.rtm_seq = ++rtm_seq;
		cp = m_rtmsg.m_space;

		sin = (struct sockaddr_in6 *)cp;
		sin->sin6_len = sizeof(struct sockaddr_in6);
		sin->sin6_family = AF_INET6;
		COPY_ADDR6(hap->addr, sin->sin6_addr);
		ADVANCE(cp, sin);

		rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

		if ((write(r, (char *)&m_rtmsg, len) < len) && verbose)
			logtrace("delete homeroute: %s", strerror(errno));
	}

	/* (re)set interface address to on loopback */

	if (useloop && (cmd & RTM_ADD)) {
		if (Calloc(pp, prefix) == NULL) {
			if (verbose)
				logerr("out of memory");
			/* route add shall fail... */
			return;
		}
		if (verbose)
			logtrace("adding home prefix %s on interface %s",
				 inet6pr(&hap->addr), ifp->name);
		pp->ifp = ifp;
		COPY_ADDR6(hap->addr, pp->pref);
		COPY_ADDR6(hap->addr, pp->addr);
		pp->alen = pp->plen = 128;
		mk_mask(&pp->pmask, 128);
		pp->invlife = INFLIFE;
		pp->deplife = INFLIFE;
		pp->pflags = pp->nflags = ND_OPT_PI_FLAG_AUTO;
		pp->mflags = MOB_ISHA | MOB_ISFA;
		TAILQ_INSERT_HEAD(&ifp->prefixes, pp, prf_list);
		add_prefix(pp);
		COPY_ADDR6(hap->addr, ifp->addr);
		prompt(ifp);

		discarding = 1;

		/* route was added by kernel */
		return;
	}
	if (discarding && (cmd & RTM_DELETE)) {
		pp = FIRST(loop_if);
		if (pp == NULL) {
			logerr("discarding without a prefix");
			abort();
		}
		if (pp->tp) {
			if (verbose)
				logerr("timeout on discarding prefix");
			kill_timeout(pp->tp);
		}
		del_prefix(pp, 1);
		TAILQ_REMOVE(&loop_if->prefixes, pp, prf_list);
		if (verbose)
			logtrace("removed discarding prefix %s on %s",
				 inet6pr(&pp->addr), loop_if->name);
		free(pp);

		discarding = 0;
	}

	/* add new route */

	if ((cmd & RTM_ADD) == 0)
		return;

	bzero((caddr_t)&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_flags = RTF_UP | RTF_HOST | RTF_LOCAL;
	rtm.rtm_type = RTM_ADD;
	rtm.rtm_addrs = RTA_DST | RTA_GATEWAY | RTA_IFP | RTA_IFA;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;
	cp = m_rtmsg.m_space;

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(hap->addr, sin->sin6_addr);
	ADVANCE(cp, sin);

	sdl = (struct sockaddr_dl *)cp;
	sdl->sdl_family = AF_LINK;
	sdl->sdl_len = sizeof(struct sockaddr_dl);
	sdl->sdl_index = ifp->index;
	ADVANCE(cp, sdl);

	sdl = (struct sockaddr_dl *)cp;
	sdl->sdl_family = AF_LINK;
	sdl->sdl_len = sizeof(struct sockaddr_dl);
	sdl->sdl_index = ifp->index;
	ADVANCE(cp, sdl);

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(hap->addr, sin->sin6_addr);
	ADVANCE(cp, sin);

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

	DUMPR("homeroute");
	if ((write(r, (char *)&m_rtmsg, len) < len) && verbose)
		logtrace("add homeroute: %s", strerror(errno));

}

static void
lastroute()
{
	register char *cp = m_rtmsg.m_space;
	register struct sockaddr_in6 *sin;
	register struct sockaddr_dl *sdl;
	struct homeaddr *hap = PRIMHA;
	int len;

	/* last chance default route */

	if (hap == NULL) {
		logerr("no home address for last chance default route");
		return;
	}
	if (loop_if == NULL) {
		if (verbose)
			logtrace("no loopback for last chance");
		return;
	}

	bzero((caddr_t)&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = RTM_ADD;
	rtm.rtm_flags = RTF_UP | RTF_STATIC | RTF_BLACKHOLE;
	rtm.rtm_addrs = RTA_DST| RTA_GATEWAY| RTA_NETMASK| RTA_IFP| RTA_IFA;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;
	rtm.rtm_index = loop_if->index;

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	ADVANCE(cp, sin);

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(hap->addr, sin->sin6_addr);
	ADVANCE(cp, sin);

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_UNSPEC;
	ADVANCE(cp, sin);

	sdl = (struct sockaddr_dl *)cp;
	sdl->sdl_family = AF_LINK;
	sdl->sdl_len = sizeof(struct sockaddr_dl);
	sdl->sdl_index = loop_if->index;
	ADVANCE(cp, sdl);

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(hap->addr, sin->sin6_addr);
	ADVANCE(cp, sin);

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

	DUMPR("lastroute");
	if (write(r, (char *)&m_rtmsg, len) < len) {
		if (verbose)
			logperror("write (lastroute)");
		return;
	} else if (verbose)
		logtrace("last chance default route");
}

#else

static void
lastroute()
{
	register char *cp = m_rtmsg.m_space;
	register struct interface *ifp = main_if;
	register struct sockaddr_in6 *sin;
	int len;

	if (ifp == NULL)
		return;

	/* from cleanup_default() */

	bzero((caddr_t)&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = RTM_ADD;
	rtm.rtm_flags = RTF_UP | RTF_STATIC;
	if (ifp->hasphy && (ifp->ltoken == MACTOKEN))
		rtm.rtm_flags |= RTF_CLONING;
	rtm.rtm_addrs = RTA_DST | RTA_GATEWAY | RTA_NETMASK;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;
	rtm.rtm_index = ifp->index;

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	ADVANCE(cp, sin);

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(ifp->llocal, sin->sin6_addr);
	ADVANCE(cp, sin);

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_UNSPEC;
	ADVANCE(cp, sin);

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

	DUMPR("lastroute");
	if (write(r, (char *)&m_rtmsg, len) < len) {
		if (verbose)
			logperror("write (lastroute)");
		return;
	} else if (verbose)
		logtrace("last chance default route");
}

#endif
