/* Support for major movements (from/to home) */

#ifdef MOBILE

#include "ndpd.h"
#include "mobility.h"

static void do_purge(struct interface *);

/* alerted by home prefix zap/change */

void
major_move(pp, purge)
	struct prefix *pp;
	int purge;
{
	register struct corr *cop;
	register struct homeaddr *hap;
	struct prefix *happ, *cp;

	cop = correspondents.tqh_first;

	/* find the first associated home address (note: double break) */

	FORALLHA(hap)
		FORALLHP(happ, hap)
			if (pp == happ)
				goto found;
    found: 

	/* Trivial cases where nothing will be done */

	if ((hap == NULL) || (hap->state == HAST_UNKNOWN)) {
		if (verbose)
			logtrace("major_move: no home address");
		return;
	}

	if ((pp == NULL) || ((pp->mflags & MOB_ISHA) == 0)) {
		if (verbose)
			logtrace("major_move: unused interface");
		return;
	}

	/* purge routers (mobile Ethernet) */

	if (purge)
		do_purge(pp->ifp);

	if (cop == NULL) {
		if (verbose)
			logtrace("major_move: no correspondent");
		return;
	}

	/* First case: from home to remote */

	if (hap->state == HAST_OUT) {
		if (def_if == NULL) {
			logerr("major_move: no interfaces");
			return;
		}
		cp = FIRST(def_if);
		while (cp && (cp->mflags & MOB_ISHA))
			cp = cp->prf_list.tqe_next;
		if (cp == NULL) {
			if (verbose)
				logtrace("major_move: found no care-of");
			return;
		}
		cp->mflags |= MOB_ISCA;
		FORALLCOR(cop)
			mobile_move(cop, &cp->addr);
		FORALLHA(hap)
			FORALLHP(happ, hap)
				if (pp == happ) {
					hap->capp = cp;
					hap->state = HAST_OUT;
				}
		return;
	}

	/* Second case: from remote to home */

	if (hap->state == HAST_IN) {
		FORALLCOR(cop)
			mobile_move(cop, &pp->addr);
		FORALLHA(hap)
			FORALLHP(happ, hap)
				if (pp == happ) {
					if (hap->capp != NULL)
						hap->capp->mflags &= ~MOB_ISCA;
					hap->capp = NULL;
					hap->state = HAST_IN;
				}
		return;
	}
}

/* Purge routers because it is a mobile Ethernet! */

static void
do_purge(ifp)
	struct interface *ifp;
{
	/* sanities */

	if (ifp == NULL) {
		logerr("do purge: no interface");
		abort();
	}
	if (cur_rt == NULL) {
		if (verbose)
			logtrace("do purge on %s: no current router",
				 ifp->name);
		return;
	}
	if (cur_rt->ifp != ifp) {
		if (verbose)
			logerr("do purge: interface mismatch %s != %s",
			       cur_rt->ifp->name, ifp->name);
		return;
	}

	/* do it with a not too aggressive way */

	purge_routers(ifp, &last_rtadv, DEPREC_OTHER);
}
#endif
