/* Protection of candidate care-of prefixes against spurious mobile stuff */

#ifdef MOBILE

#include "ndpd.h"
#include "mobility.h"

/* protect everything */

void
protect_all()
{
	struct interface *ifp;
	struct prefix *pp;

	if (!mobile)
		return;

	FORALLIFP(ifp)
		FORALLPP(pp, ifp)
			if ((pp->mflags & MOB_ISHA) == 0)
				add_protect(pp);
}

/* protect a prefix against spurious mobile stuff */

void
add_protect(pp)
	register struct prefix *pp;
{
	struct corr *cop, *cop0 = correspondents.tqh_first;

	if (cop0 == NULL)
		return;

	FORALLCOR(cop)
		if ((cop->type & CORR_PROTECT) &&
		    IN6_ARE_ADDR_EQUAL(&pp->addr, &cop->addr))
			break;
	if (cop != NULL) {
		if (verbose)
			logtrace("add protect: has already %s",
				 inet6pr(&pp->addr));
		return;
	}

	if (Calloc(cop, corr) == NULL) {
		if (pp->alen != 128) {
			logerr("out of memory");
			abort();
		}
		if (verbose)
			logerr("out of memory");
		return;
	}

	cop->type = CORR_PROTECT;
	if (pp->alen != 128) {
		COPY_ADDR6(pp->pref, cop->addr);
		COPY_ADDR6(pp->pmask, cop->mask);
		cop->type |= CORR_PREFIX;
	} else
		COPY_ADDR6(pp->addr, cop->addr);
	COPY_ADDR6(cop0->home, cop->home);
	COPY_ADDR6(cop0->home, cop->careof);
	cop->plen = pp->alen;

	if (verbose)
		logtrace("add protect %s/%d", inet6pr(&pp->addr), pp->alen);
	add_corr(cop, 0);
}

/* unprotect a prefix against spurious mobile stuff */

void
del_protect(pp)
	register struct prefix *pp;
{
	struct corr *cop;

	FORALLCOR(cop)
		if ((cop->type & CORR_PROTECT) &&
		    IN6_ARE_ADDR_EQUAL(&pp->addr, &cop->addr))
			break;
	if (cop == NULL) {
		if (verbose)
			logtrace("delete protect: can't found %s",
				 inet6pr(&pp->addr));
		return;
	}

	if (verbose)
		logtrace("delete protect %s", inet6pr(&pp->addr));
	del_corr(cop);
}
#endif
