#include "ndpd.h"

struct {
	struct cmsghdr c_cm;
	u_char c_space[64];
} c_cmsg;
#define cm	c_cmsg.c_cm

#define MAXPACKET	8192	/* max packet size */
char *bigbuffer = NULL;
int bigbufferlen = MAXPACKET;

static void send_rtadv(struct interface *);
static void recv_rtsol(struct interface *, struct ip6_hdr *, int);
static void recv_rtadv(struct interface *, struct ip6_hdr *, int);
static struct nd_opt_agti *get_agti(struct nd_router_advert *, int);
static void recv_ndsol(struct interface *, struct ip6_hdr *, int);
static void sol_rtadv(struct interface *);
static void add_pending(struct anycast *, struct in6_addr *, long);
static void send_apadv(struct presp *);
static int already_err(struct interface *, struct in6_addr *, int, void *);
static void err_timeout(struct timeout *);
static void expire_err(void);

/* Send an ICMPv6 Router Advertisment packet on ifp. */
/*  always use multicast - restart timer */
static void
send_rtadv(ifp)
	struct interface *ifp;
{
	register struct ip6_hdr *ip;
	register struct nd_router_advert *rap;
	register struct nd_opt_hdr *hp;
	register struct prefix *pp;
	struct sockaddr_in6 dst;
	int len, cc;
	long delay;
	int dorouter, opt, needrta;
	u_int32_t vlife, plife;

	if (ifp->tp) {
		kill_timeout(ifp->tp);
		ifp->tp = NULL;
	}
	if (!IS_UP(ifp) || ifp->opt.no_rtadv)
		return;
	dorouter = is_router();
	needrta = ifp->opt.flags & ND_RA_FLAG_HAGENT;
	bzero((caddr_t)&dst, sizeof(struct sockaddr_in6));
	dst.sin6_len = sizeof(struct sockaddr_in6);
	dst.sin6_family = AF_INET6;
	dst.sin6_addr.s6_addr[0] = 0xff;
	dst.sin6_addr.s6_addr[1] = MADDR6_SCP_LINK;
	dst.sin6_addr.s6_addr[15] = MADDR6_ALLNODES;
	dst.sin6_scope_id = ifp->index;

	if (verbose) {
		logtrace("Sending RtAdv for %s/%s",
			 ifp->name, inet6pr(&ifp->llocal));
	}

	pp = UFIRST(ifp);
    more:
	ip = (struct ip6_hdr *)bigbuffer;
	len = sizeof(*rap);
	bzero(bigbuffer, sizeof(*ip) + len);

	((struct ip6ovck *)ip)->ih6_pr = IPPROTO_ICMPV6;
	COPY_ADDR6(ifp->llocal, ip->ip6_src);
	COPY_ADDR6(dst.sin6_addr, ip->ip6_dst);

	rap = (struct nd_router_advert *)(ip + 1);
	rap->nd_ra_type = ND_ROUTER_ADVERT;

	if (dorouter)
		rap->nd_ra_router_lifetime = htons(ifp->opt.life);
	rap->nd_ra_curhoplimit = MaxHop;
	rap->nd_ra_retransmit = htonl(ifp->opt.retrans);
	rap->nd_ra_reachable = htonl(ifp->opt.reach);
	rap->nd_ra_flags_reserved = ifp->opt.flags;

	hp = (struct nd_opt_hdr *)(rap + 1);
	if (dorouter && ifp->hasphy) {
		bzero(hp, ND_OPT_LINKADDR_LEN);
		hp->nd_opt_type = ND_OPT_SOURCE_LINKADDR;
		hp->nd_opt_len = 1;
		bcopy(ifp->adphy, hp + 1, ifp->lmac);
		len += hp->nd_opt_len << 3;
		hp = (struct nd_opt_hdr *)
			((caddr_t)hp + (hp->nd_opt_len << 3));
	}

	if (ifp->opt.extra & ADVMTU) {
		struct nd_opt_mtu *mp = (struct nd_opt_mtu *)hp;
		u_int32_t mtu;

		mtu = (ifp->opt.extra & ADVMTUIF) ? ifp->mtu : ifp->opt.mtuadv;
		bzero(mp, sizeof (*mp));
		mp->nd_opt_mtu_type = ND_OPT_MTU;
		mp->nd_opt_mtu_len = 1;
		mp->nd_opt_mtu_mtu = htonl(mtu);
		len += hp->nd_opt_len << 3;
		hp = (struct nd_opt_hdr *)
			((caddr_t)hp + (hp->nd_opt_len << 3));
	}
	if (ifp->opt.extra & ADVADVINT) {
		struct nd_opt_advit *itp = (struct nd_opt_advit *)hp;

		bzero(itp, sizeof (*itp));
		itp->nd_opt_ait_type = ND_OPT_MOB_ADVERT_INTERVAL;
		itp->nd_opt_ait_len = 1;
		itp->nd_opt_ait_it = htonl(ifp->opt.maxrtadv);
		len += hp->nd_opt_len << 3;
		hp = (struct nd_opt_hdr *)
			((caddr_t)hp + (hp->nd_opt_len << 3));
	}
	if ((ifp->opt.extra & ADVHAINF) &&
	    (ifp->opt.flags & ND_RA_FLAG_HAGENT) &&
	    ((ifp->opt.hapref != 0) || (ifp->opt.halife != ifp->opt.life))) {
		struct nd_opt_agti *aip = (struct nd_opt_agti *)hp;

		bzero(aip, sizeof (*aip));
		aip->nd_opt_hai_type = ND_OPT_MOB_AGENT_INFORMATION;
		aip->nd_opt_hai_len = 1;
		aip->nd_opt_hai_pref = htons(ifp->opt.hapref);
		aip->nd_opt_hai_life = htons(ifp->opt.halife);
		len += hp->nd_opt_len << 3;
		hp = (struct nd_opt_hdr *)
			((caddr_t)hp + (hp->nd_opt_len << 3));
	}

	LOOPALLPP(, pp) {
		struct nd_opt_prefix_info *pip;

		if (len + sizeof(*pip) + sizeof(*ip) > IP6_MMTU)
			break;
		pip = (struct nd_opt_prefix_info *)hp;
		bzero(pip, sizeof(*pip));
		pip->nd_opt_pi_type = ND_OPT_PREFIX_INFORMATION;
		pip->nd_opt_pi_len = 4;
		if ((pp->plen < 16) ||
		    ((pp->plen != 128) && (pp->plen > 128 - 8 * ifp->ltoken)))
			pip->nd_opt_pi_prefix_len = 128 - 8 * ifp->ltoken;
		else
			pip->nd_opt_pi_prefix_len = pp->plen;
		pip->nd_opt_pi_flags_reserved =
			pp->flags & ND_OPT_PI_FLAG_ALL;
		if (pp->valpref.val_date)
			vlife = (now.tv_sec >= pp->valpref.val.date)?  0 :
					pp->valpref.val.date - now.tv_sec;
		else
			vlife = pp->valpref.val.life;
		if (pp->valpref.pref_date)
			plife = (now.tv_sec >= pp->valpref.pref.date)?  0 :
					pp->valpref.pref.date - now.tv_sec;
		else
			plife = pp->valpref.pref.life;
		pip->nd_opt_pi_valid_time = htonl(vlife);
		pip->nd_opt_pi_preferred_time = htonl(plife);
		if (pp->flags & ND_OPT_PI_FLAG_RTADDR) {
			COPY_ADDR6(pp->addr, pip->nd_opt_pi_prefix);
			needrta = 0;
		} else
			COPY_ADDR6(pp->pref, pip->nd_opt_pi_prefix);
		if (pp->flags & ND_OPT_PI_FLAG_SITE)
			pip->nd_opt_pi_reserved_site_plen = htonl(pp->slen);
		len += hp->nd_opt_len << 3;
		hp = (struct nd_opt_hdr *)
			((caddr_t)hp + (hp->nd_opt_len << 3));
	}

	/* Home agents need at least one router address */
	if (needrta) {
		static int tenth = 0;

		if (verbose || !tenth)
			logtrace("home agent without a router address on %s",
				 ifp->name);
		tenth = (tenth + 1) % 10;
	}

	/* Compute ICMPv6 checksum here */
	((struct ip6ovck *)ip)->ih6_len = htons(len);
	rap->nd_ra_cksum = in_cksum((u_short *)ip, len + sizeof(*ip));

	/* finish IPv6 header */
	ip->ip6_flow = ND_DEFAULT_CLASS | IPV6_VERSION;
	ip->ip6_plen = htons(len);
	ip->ip6_nxt = IPPROTO_ICMPV6;
	ip->ip6_hlim = ND_DEFAULT_HOPLIM;

	opt = ifp->index;
	if (setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_IF,
		       (char *)&opt, sizeof(opt)) < 0) {
		if (verbose)
			logperror("setsockopt (IPV6_MULTICAST_IF)");
		goto done;
	}

	len += sizeof(struct ip6_hdr);
	cc = sendto(s, bigbuffer, len, 0,
		    (struct sockaddr *)&dst, sizeof (struct sockaddr_in6));

	if (cc < 0 || cc != len)  {
		logerr("send_rtadv: wrote %d chars, ret=%d", len, cc);
		if (cc < 0) {
			logperror("sendto");
			goto done;
		}
	}
	if (pp)
		goto more;
    done:
	(void) gettimeofday(&now, NULL);
	ifp->last = now;
	delay = ifp->opt.minrtadv +
		random() % (ifp->opt.maxrtadv - ifp->opt.minrtadv);
	if (ifp->ntosend < MAX_INITIAL_RTR_ADVERTISEMENTS - 1) {
		ifp->ntosend++;
		if (delay > MAX_INITIAL_RTR_ADVERT_INTERVAL)
			delay = MAX_INITIAL_RTR_ADVERT_INTERVAL;
		}
	ifp->tp = new_timeout(MS_TO_SEC(delay), MS_TO_USEC(delay),
			      (func_to) send_rtadv, ifp);
	if (ifp->tp == NULL)
		abort();
}

/* (Re)start sending RTAdv periodically */
void
start_rtadv(ifp)
	struct interface *ifp;
{
	ifp->ntosend = 0;
	send_rtadv(ifp);
}

/* respond to RTSol : advance date of next RTAdv */
static void
sol_rtadv(ifp)
	struct interface *ifp;
{
	struct timeval tv;
	long delay;
	struct timeout *tp;

	(void) gettimeofday(&now, NULL);
	tv.tv_sec = ifp->last.tv_sec + MS_TO_SEC(MIN_DELAY_BETWEEN_RAS);
	tv.tv_usec = ifp->last.tv_usec + MS_TO_USEC(MIN_DELAY_BETWEEN_RAS);
	NORMTIME(tv);
	if (AFTER(tv, now))
		tv = now;
	delay = random() % MAX_RA_DELAY_TIME;
	tv.tv_sec = tv.tv_sec + MS_TO_SEC(delay);
	tv.tv_usec = tv.tv_usec + MS_TO_USEC(delay);
	NORMTIME(tv);

	tp = ifp->tp;
	if (!tp || tp->status < 0 || AFTER(tv, tp->tv)) /* only if running */
		return;
	tp->refcnt++;	/* do not free tp */
	kill_timeout(tp);
	tp->refcnt--;
	tp->tv = tv;
	add_timeout(tp);	/* keep same descriptor */
}

void
stop_rtadv(ifp)
	struct interface *ifp;
{
	/* TO BE WRITTEN.  What must we do ???? */
}

/* Send an anycast/proxy neighbor advertisement */
static void
send_apadv(rp)
	struct presp *rp;
{
	register struct ip6_hdr *ip;
	register struct nd_neighbor_advert *nap;
	register struct nd_opt_hdr *lp;
	struct msghdr msg;
	struct iovec iov;
	struct sockaddr_in6 dst;
	int len, cc;
	struct anycast *ap = rp->which;
	struct interface *ifp = ap->ifp;
	struct in6_pktinfo *pi;

	if (!IS_UP(ifp) || !ifp->hasphy || NO_ANYCAST(ifp))
		goto done;

	bzero((caddr_t)&dst, sizeof(struct sockaddr_in6));
	dst.sin6_len = sizeof(struct sockaddr_in6);
	dst.sin6_family = AF_INET6;
	COPY_ADDR6(rp->peer, dst.sin6_addr);
	dst.sin6_scope_id = ifp->index;

	ip = (struct ip6_hdr *)bigbuffer;
	len = sizeof(*nap) + ND_OPT_LINKADDR_LEN;
	bzero(bigbuffer, sizeof(struct ip6_hdr) + len);

	((struct ip6ovck *)ip)->ih6_pr = IPPROTO_ICMPV6;
	((struct ip6ovck *)ip)->ih6_len = htons(len);
	COPY_ADDR6(ifp->llocal, ip->ip6_src);
	COPY_ADDR6(dst.sin6_addr, ip->ip6_dst);

	nap = (struct nd_neighbor_advert *)(ip + 1);
	nap->nd_na_type = ND_NEIGHBOR_ADVERT;
	if (ap->type != IP6ANY_HOST_PROXY)
		nap->nd_na_flags_reserved = ND_NA_FLAG_ROUTER;
	if (!IN6_IS_ADDR_MULTICAST(&rp->peer))
		nap->nd_na_flags_reserved |= ND_NA_FLAG_SOLICITED;
	COPY_ADDR6(ap->addr, nap->nd_na_target);

	lp = (struct nd_opt_hdr *)(nap + 1);
	lp->nd_opt_type = ND_OPT_TARGET_LINKADDR;
	lp->nd_opt_len = 1;
	bcopy(ifp->adphy, lp + 1, ifp->lmac);

	/* Compute ICMPv6 checksum here */
	nap->nd_na_cksum = in_cksum((u_short *)ip, len + sizeof(*ip));

	/* finish IPv6 header */
	ip->ip6_flow = ND_DEFAULT_CLASS | IPV6_VERSION;
	ip->ip6_plen = htons(len);
	ip->ip6_nxt = IPPROTO_ICMPV6;
	ip->ip6_hlim = ND_DEFAULT_HOPLIM;

	len += sizeof(struct ip6_hdr);
	bzero((caddr_t)&msg, sizeof(struct msghdr));
	bzero((caddr_t)&iov, sizeof(struct iovec));
	bzero((caddr_t)&c_cmsg, sizeof(c_cmsg));
	msg.msg_name = (caddr_t)&dst;
	msg.msg_namelen = sizeof(struct sockaddr_in6);
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	iov.iov_base = bigbuffer;
	iov.iov_len = len;
	msg.msg_control = (caddr_t)&c_cmsg;
	msg.msg_controllen =
		sizeof(struct cmsghdr) + sizeof(struct in6_pktinfo);
	cm.cmsg_len = msg.msg_controllen;
	cm.cmsg_level = IPPROTO_IPV6;
	cm.cmsg_type = IPV6_PKTINFO;
	pi = (struct in6_pktinfo *)&c_cmsg.c_space;
	bcopy(&ifp->llocal, &pi->ipi6_addr, sizeof(struct in6_addr));
	pi->ipi6_ifindex = ifp->index;

	cc = sendmsg(s, &msg, 0);

	if (cc < 0 || cc != len) {
		logerr("send_apadv: wrote %d chars, ret=%d", len, cc);
		if (cc < 0) {
			logperror("send_apadv: sendmsg");
			goto done;
		}
	} else if (verbose)
		logtrace("send apadv for %s from %s to %s with flag=%x",
			 inet6pr(&ap->addr),
			 inet6pr(&ifp->llocal),
			 inet6pr(&rp->peer),
			 ntohl(nap->nd_na_flags_reserved));
    done:
	if (rp->tp)
		cancel_anycast(rp);
	else
		free(rp);
}

/* Send an anycast/proxy gratuitous neighbor advertisement */
void
send_grat_apadv(ap)
	struct anycast *ap;
{
	register struct ip6_hdr *ip;
	register struct nd_neighbor_advert *nap;
	register struct nd_opt_hdr *lp;
	struct msghdr msg;
	struct iovec iov;
	struct sockaddr_in6 dst;
	int len, cc;
	long delay;
	struct interface *ifp = ap->ifp;
	struct in6_pktinfo *pi;

	if (ap->tp) {
		kill_timeout(ap->tp);
		ap->tp = NULL;
	}
	if (!IS_UP(ifp) || !ifp->hasphy || NO_ANYCAST(ifp))
		return;
	if (ap->ntosend <= 0)
		return;

	bzero((caddr_t)&dst, sizeof(struct sockaddr_in6));
	dst.sin6_len = sizeof(struct sockaddr_in6);
	dst.sin6_family = AF_INET6;
	dst.sin6_addr.s6_addr[0] = 0xff;
	dst.sin6_addr.s6_addr[1] = MADDR6_SCP_LINK;
	dst.sin6_addr.s6_addr[15] = MADDR6_ALLNODES;
	dst.sin6_scope_id = ifp->index;

	ip = (struct ip6_hdr *)bigbuffer;
	len = sizeof(*nap) + ND_OPT_LINKADDR_LEN;
	bzero(bigbuffer, sizeof(struct ip6_hdr) + len);

	((struct ip6ovck *)ip)->ih6_pr = IPPROTO_ICMPV6;
	((struct ip6ovck *)ip)->ih6_len = htons(len);
	COPY_ADDR6(ifp->llocal, ip->ip6_src);
	COPY_ADDR6(dst.sin6_addr, ip->ip6_dst);

	nap = (struct nd_neighbor_advert *)(ip + 1);
	nap->nd_na_type = ND_NEIGHBOR_ADVERT;
	if (ap->type != IP6ANY_HOST_PROXY)
		nap->nd_na_flags_reserved = ND_NA_FLAG_ROUTER;
	if (ap->type != IP6ANY_ANYCAST)
		nap->nd_na_flags_reserved |= ND_NA_FLAG_OVERRIDE;
	COPY_ADDR6(ap->addr, nap->nd_na_target);

	lp = (struct nd_opt_hdr *)(nap + 1);
	lp->nd_opt_type = ND_OPT_TARGET_LINKADDR;
	lp->nd_opt_len = 1;
	bcopy(ifp->adphy, lp + 1, ifp->lmac);

	/* Compute ICMPv6 checksum here */
	nap->nd_na_cksum = in_cksum((u_short *)ip, len + sizeof(*ip));

	/* finish IPv6 header */
	ip->ip6_flow = ND_DEFAULT_CLASS | IPV6_VERSION;
	ip->ip6_plen = htons(len);
	ip->ip6_nxt = IPPROTO_ICMPV6;
	ip->ip6_hlim = ND_DEFAULT_HOPLIM;

	len += sizeof(struct ip6_hdr);
	bzero((caddr_t)&msg, sizeof(struct msghdr));
	bzero((caddr_t)&iov, sizeof(struct iovec));
	bzero((caddr_t)&c_cmsg, sizeof(c_cmsg));
	msg.msg_name = (caddr_t)&dst;
	msg.msg_namelen = sizeof(struct sockaddr_in6);
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	iov.iov_base = bigbuffer;
	iov.iov_len = len;
	msg.msg_control = (caddr_t)&c_cmsg;
	msg.msg_controllen =
		sizeof(struct cmsghdr) + sizeof(struct in6_pktinfo);
	cm.cmsg_len = msg.msg_controllen;
	cm.cmsg_level = IPPROTO_IPV6;
	cm.cmsg_type = IPV6_PKTINFO;
	pi = (struct in6_pktinfo *)&c_cmsg.c_space;
	bcopy(&ifp->llocal, &pi->ipi6_addr, sizeof(struct in6_addr));
	pi->ipi6_ifindex = ifp->index;

	cc = sendmsg(s, &msg, 0);
	if (cc < 0 || cc != len) {
		logerr("send_grat_apadv: wrote %d chars, ret=%d", len, cc);
		if (cc < 0)
			logperror("sendmsg");
	} else if (verbose)
		logtrace("send grat apadv for %s on %s from %s with flag=%x",
			 inet6pr(&ap->addr),
			 ifp->name,
			 inet6pr(&ifp->llocal),
			 ntohl(nap->nd_na_flags_reserved));

	ap->ntosend--;
	if (ap->ntosend == 0)
		return;
	delay = random() % MAX_ANYCAST_DELAY_TIME;

	(void) gettimeofday(&now, NULL);

	ap->tp = new_timeout(MS_TO_SEC(delay), MS_TO_USEC(delay),
			     (func_to) send_grat_apadv, ap);
}

/* Receive an ICMPv6 packet */
void
recv_icmp()
{
	register struct ip6_hdr *ip;
	register struct icmp6_hdr *icp;
	struct interface *ifp;
	struct msghdr msg;
	struct iovec iov;
	struct in6_pktinfo pi;
	int cc;

	bzero((caddr_t)&msg, sizeof(struct msghdr));
	bzero((caddr_t)&iov, sizeof(struct iovec));
	msg.msg_iov = &iov;
	msg.msg_iov->iov_base = bigbuffer;
	msg.msg_iov->iov_len = bigbufferlen;
	msg.msg_iovlen = 1;
	msg.msg_control = (caddr_t)&c_cmsg;
	msg.msg_controllen = sizeof(c_cmsg);

	if ((cc = recvmsg(s, &msg, 0)) < 0) {
		if (verbose)
			logperror("read (ICMPv6)");
		return;
	}
	if (cc == 0) {
		if (verbose)
			logerr("read ICMPv6 return 0");
		return;
	}
	if (msg.msg_flags & (MSG_TRUNC|MSG_CTRUNC)) {
		if (verbose)
			logerr("read ICMPv6 truncated");
		return;
	}
	bcopy(c_cmsg.c_space, &pi, sizeof(pi));
	if ((msg.msg_controllen == 0) ||
	    (cm.cmsg_len != (sizeof(struct cmsghdr) + sizeof(pi))) ||
	    (cm.cmsg_level != IPPROTO_IPV6) ||
	    (cm.cmsg_type != IPV6_PKTINFO) ||
	    (pi.ipi6_ifindex == 0)) {
		if (verbose)
			logerr("read ICMPv6 doesn't give interface");
		return;
	}

	FORALLIFP(ifp)
		if (ifp->index == pi.ipi6_ifindex)
			break;
	if (ifp == NULL || !IS_UP(ifp) || ifp->opt.no_rtadv)
		return;

	ip = (struct ip6_hdr *)bigbuffer;
	if (cc < sizeof(*ip) + sizeof(*icp)) {
		if (verbose)
			logerr("ICMPv6: too short (%d)", cc);
		return;
	}
	if (ip->ip6_hlim != ND_DEFAULT_HOPLIM) {
		if (verbose)
			logtrace("no HOPS (%d)", ND_DEFAULT_HOPLIM);
		return;
	}
	cc -= sizeof(struct ip6_hdr);
	icp = (struct icmp6_hdr *)(ip + 1);

	switch (icp->icmp6_type) {
	    case ND_ROUTER_ADVERT:
		if (!IN6_IS_ADDR_LINKLOCAL(&ip->ip6_src))
			return;
		recv_rtadv(ifp, ip, cc);
		break;
	    case ND_ROUTER_SOLICIT:
		if (!IN6_IS_ADDR_LINKLOCAL(&ip->ip6_src))
			return;
		recv_rtsol(ifp, ip, cc);
		break;
	    case ND_NEIGHBOR_SOLICIT:
		recv_ndsol(ifp, ip, cc);
		break;
	    /* Other types are filtered */
	}
}

/* Receive an RTSol packet */
static void
recv_rtsol(ifp, ip, cc)
	struct interface *ifp;
	register struct ip6_hdr *ip;
	int cc;
{
	register struct nd_router_solicit *rsp;
	register struct nd_opt_hdr *hp;

	rsp = (struct nd_router_solicit *)(ip + 1);
	if (rsp->nd_rs_code != 0) {
		if (verbose)
			logerr("ICMPv6 RtSol: code=%d", rsp->nd_rs_type);
		return;
	}
	if ((unsigned)cc < sizeof(*rsp)) {
		if (verbose)
			logerr("ICMPv6 RtSol: too short %d", cc);
			return;
	}

	if (verbose) {
		logtrace("receive ICMPv6 RtSol on %s/%s to %s (len=%d)",
			 ifp->name, inet6pr(&ip->ip6_src),
			 inet6pr(&ip->ip6_dst),
			 cc + sizeof(struct ip6_hdr));
	}

	cc -= sizeof(*rsp);
	hp = (struct nd_opt_hdr *)(rsp + 1);
	while (cc) switch (hp->nd_opt_type) {
	    /* Nothing to do, multicast response
	    case ND_OPT_SOURCE_LINKADDR:
	     */
	    default:
		if (hp->nd_opt_len == 0) {
			if (verbose)
				logerr("null option length (%d)",
				       hp->nd_opt_type);
			return;
		}
		cc -= hp->nd_opt_len << 3;
		if (cc < 0) {
			if (verbose)
				logerr("RtSol underflow");
			return;
		}
		if (cc == 0)
			break;
		hp = (struct nd_opt_hdr *)
			((caddr_t)hp + (hp->nd_opt_len << 3));
	}
	sol_rtadv(ifp);
}

/* Receive an RTAdv packet - check option conflicts */
static void
recv_rtadv(ifp, ip, cc)
	struct interface *ifp;
	register struct ip6_hdr *ip;
	int cc;
{
	register struct nd_router_advert *rap;
	register struct nd_opt_hdr *hp;
	struct nd_opt_agti *aip = NULL;
	int len = cc;
	char msg[200], *p = msg;

	rap = (struct nd_router_advert *)(ip + 1);
	if (rap->nd_ra_code != 0)
		return;
	if ((unsigned)len < sizeof(*rap))
		return;

	if (verbose) {
		logtrace("receive ICMPv6 RtAdv on %s/%s",
			 ifp->name, inet6pr(&ip->ip6_src));
	}

#define SETANDDIF(x, y)	((x) && (y) && ((x) != (y)))
#define FLAGS(rap)	(rap->nd_ra_flags_reserved & ND_RA_FLAG_CFG)
	rap->nd_ra_reachable = ntohl(rap->nd_ra_reachable);
	rap->nd_ra_retransmit = ntohl(rap->nd_ra_retransmit);
	if (SETANDDIF(MaxHop, rap->nd_ra_curhoplimit))
		p += sprintf(p, " hop %d!=%d", MaxHop, rap->nd_ra_curhoplimit);
	if ((ifp->opt.flags & ND_RA_FLAG_CFG) != FLAGS(rap))
		p += sprintf(p, " flags 0x%x!=0x%x",
			     FLAGS(rap), ifp->opt.flags & ND_RA_FLAG_CFG);
	if (SETANDDIF(ifp->opt.retrans, rap->nd_ra_retransmit))
		p += sprintf(p, " Retrans %d!=%d",
		             rap->nd_ra_retransmit, ifp->opt.retrans);
	if (SETANDDIF(ifp->opt.reach, rap->nd_ra_reachable))
		p += sprintf(p, " ReachTime %d!=%d",
		             rap->nd_ra_reachable, ifp->opt.reach);
	if (p != msg) {
	    if (!already_err(ifp, &ip->ip6_src, 0, rap))
		logerr("RtAdv from %s/%s in!=conf:%s",
		       ifp->name, inet6pr(&ip->ip6_src), msg);
	    p = msg;
	}
#undef SETANDDIF
#undef FLAGS

	len -= sizeof(*rap);
	hp = (struct nd_opt_hdr *)(rap + 1);
	while (len) switch (hp->nd_opt_type) {
	    case ND_OPT_MTU: {
		u_int32_t mtu;
		struct nd_opt_mtu *mp = (struct nd_opt_mtu *)hp;

		if ((ifp->opt.extra & ADVMTU) == 0)
			goto next;
		mtu = (ifp->opt.extra & ADVMTUIF) ? ifp->mtu : ifp->opt.mtuadv;
		mp->nd_opt_mtu_mtu = ntohl(mp->nd_opt_mtu_mtu);
		if (mp->nd_opt_mtu_mtu != mtu)
		    if (!already_err(ifp, &ip->ip6_src, ND_OPT_MTU, hp))
			logerr("RtAdv from %s/%s in!=conf: mtu %d!=%d",
				       ifp->name, inet6pr(&ip->ip6_src),
				       mp->nd_opt_mtu_mtu, mtu);
		goto next;
	    }
	    case ND_OPT_PREFIX_INFORMATION:
	    {
		struct nd_opt_prefix_info *pip;
		struct prefix *pp;
		u_int32_t life, lifein;
		struct in6_addr rta, mask;
		int dosite;

		pip = (struct nd_opt_prefix_info *)hp;
		if (pip->nd_opt_pi_flags_reserved & ND_OPT_PI_FLAG_RTADDR)
			COPY_ADDR6(pip->nd_opt_pi_prefix, rta);
		mk_mask(&mask, pip->nd_opt_pi_prefix_len);
		apply_mask(&pip->nd_opt_pi_prefix, &mask);

		LOOPALLPP(pp = UFIRST(ifp), pp)
		    if ((pp->plen == pip->nd_opt_pi_prefix_len) &&
			SAME_ADDR6(pp->pref, pip->nd_opt_pi_prefix))
			break;
		if (pp == NULL)
			goto next;
		dosite = pip->nd_opt_pi_flags_reserved & ND_OPT_PI_FLAG_SITE;
		dosite &= pp->flags & ND_OPT_PI_FLAG_SITE;
		if ((pip->nd_opt_pi_flags_reserved & ND_OPT_PI_FLAG_RTADDR) &&
		    (pp->hacnt != -1)) {
			if (aip == NULL)
				aip = get_agti(rap, cc);
			update_hagent(pp, aip, &rta);
		}
		pip->nd_opt_pi_flags_reserved &= ND_OPT_PI_FLAG_CFG;
		pip->nd_opt_pi_valid_time = ntohl(pip->nd_opt_pi_valid_time);
		pip->nd_opt_pi_preferred_time =
			ntohl(pip->nd_opt_pi_preferred_time);
		pip->nd_opt_pi_reserved_site_plen =
			ntohl(pip->nd_opt_pi_reserved_site_plen);
		pip->nd_opt_pi_reserved_site_plen &= 0xff;

		if (pip->nd_opt_pi_flags_reserved !=
				(pp->flags & ND_OPT_PI_FLAG_CFG))
			p += sprintf(p, " flags 0x%x!=0x%x",
				     pp->flags & ND_OPT_PI_FLAG_CFG,
				     pip->nd_opt_pi_flags_reserved);

		lifein = pip->nd_opt_pi_valid_time;
		if (pp->valpref.val_date) {
			life = (now.tv_sec >= pp->valpref.val.date)? 0 :
				pp->valpref.val.date - now.tv_sec;
			if (abs(lifein - life) <= DELTALIFE)
				life = lifein;
		} else
			life = pp->valpref.val.life;
		if (life != lifein)
			p += sprintf(p, " val_life %d!=%d", life, lifein);

		lifein = pip->nd_opt_pi_preferred_time;
		if (pp->valpref.pref_date) {
			life = (now.tv_sec >= pp->valpref.pref.date)? 0 :
				pp->valpref.pref.date - now.tv_sec;
			if (abs(lifein - life) <= DELTALIFE)
				life = lifein;
		} else
			life = pp->valpref.pref.life;
		if (life != lifein)
			p += sprintf(p, " pref_life %d!=%d", life, lifein);

		if (dosite && (pip->nd_opt_pi_reserved_site_plen != pp->slen))
			p += sprintf(p, " site_Plength %d!=%d",
				     pip->nd_opt_pi_reserved_site_plen,
				     pp->slen);

		if (p != msg) {
		    if (!already_err(ifp,
				     &ip->ip6_src,
				     ND_OPT_PREFIX_INFORMATION, hp))
			logerr("RtAdv from %s/%s in!=conf pref %s/%d:%s",
			       ifp->name, inet6pr(&ip->ip6_src),
			       inet6pr(&pp->addr), pp->plen,
			       msg);
		    p = msg;
		}
		goto next;
	    }
	    case ND_OPT_MOB_AGENT_INFORMATION:
		aip = (struct nd_opt_agti *)hp;
		aip->nd_opt_hai_pref = ntohs(aip->nd_opt_hai_pref);
		aip->nd_opt_hai_life = ntohs(aip->nd_opt_hai_life);
		if (verbose)
			logtrace("found agti (pref=%d,life=%u)",
				 aip->nd_opt_hai_pref,
				 aip->nd_opt_hai_life);
		goto next;
	    default:
	    next:
		if (hp->nd_opt_len == 0) {
			if (verbose)
				logerr("null option length (%d)",
				       hp->nd_opt_type);
			return;
		}
		len -= hp->nd_opt_len << 3;
		if (len < 0)
			return;
		if (len == 0)
			break;
		hp = (struct nd_opt_hdr *)
			((caddr_t)hp + (hp->nd_opt_len << 3));
	}
}

/* Get home agent information option */

static struct nd_opt_agti *
get_agti(rap, len)
	struct nd_router_advert *rap;
	int len;
{
	register struct nd_opt_hdr *hp;
	register struct nd_opt_agti *aip;
	static struct nd_opt_agti ai;

	/* default values */

	bzero(&ai, sizeof(ai));
	ai.nd_opt_hai_pref = 0;
	if (rap->nd_ra_flags_reserved & ND_RA_FLAG_HAGENT)
		ai.nd_opt_hai_life = ntohs(rap->nd_ra_router_lifetime);
	else {
		/* H flag is not set: kill it! */
		ai.nd_opt_hai_life = 0;
		return (&ai);
	}

	/* scan for the option */

	len -= sizeof(*rap);
	hp = (struct nd_opt_hdr *)(rap + 1);
	while (len) switch (hp->nd_opt_type) {
	    case ND_OPT_MOB_AGENT_INFORMATION:
		aip = (struct nd_opt_agti *)hp;
		aip->nd_opt_hai_pref = ntohs(aip->nd_opt_hai_pref);
		aip->nd_opt_hai_life = ntohs(aip->nd_opt_hai_life);
		if (verbose)
			logtrace("get_agti: found in packet");
		return (aip);
	    default:
		if (hp->nd_opt_len == 0)
			return (&ai);
		len -= hp->nd_opt_len << 3;
		if (len <= 0)
			return (&ai);
		hp = (struct nd_opt_hdr *)
			((caddr_t)hp + (hp->nd_opt_len << 3));
	}
	return (&ai);
}

/* Receive an NDSol packet */
static void
recv_ndsol(ifp, ip, cc)
	struct interface *ifp;
	register struct ip6_hdr *ip;
	int cc;
{
	register struct nd_neighbor_solicit *nsp;
	register struct nd_opt_hdr *hp;
	struct prefix *pp;
	struct anycast *ap = NULL;
	long delay;

	nsp = (struct nd_neighbor_solicit *)(ip + 1);
	if (nsp->nd_ns_code != 0) {
		if (verbose)
			logerr("ICMPv6 NDSol: code=%d", nsp->nd_ns_code);
		return;
	}
	if ((unsigned)cc < sizeof(*nsp)) {
		if (verbose)
			logerr("ICMPv6 NDSol: too short %d", cc);
			return;
	}

	if (verbose) {
		logtrace("receive ICMPv6 NDSol on %s/%s for %s to %s (len=%d)",
			 ifp->name, inet6pr(&ip->ip6_src),
			 inet6pr(&nsp->nd_ns_target),
			 inet6pr(&ip->ip6_dst),
			 cc + sizeof(struct ip6_hdr));
	}

	cc -= sizeof(*nsp);
	hp = (struct nd_opt_hdr *)(nsp + 1);
	while (cc) switch (hp->nd_opt_type) {
	    /* Nothing to do, multicast response
	    case ND_OPT_SOURCE_LINKADDR:
	     */
	    default:
		if (hp->nd_opt_len == 0) {
			if (verbose)
				logerr("null option length (%d)",
				       hp->nd_opt_type);
			return;
		}
		cc -= hp->nd_opt_len << 3;
		if (cc < 0) {
			if (verbose)
				logerr("NDSol underflow");
			return;
		}
		if (cc == 0)
			break;
		hp = (struct nd_opt_hdr *)
			((caddr_t)hp + (hp->nd_opt_len << 3));
	}

	FORALLPP(pp, ifp)
		FORALLPANY(ap, pp)
			if (SAME_ADDR6(ap->addr, nsp->nd_ns_target))
				break;

	if (ap == NULL)
		FORALLOANY(ap, ifp)
			if (SAME_ADDR6(ap->addr, nsp->nd_ns_target))
				break;

	if (ap == NULL)
		return;

	if (IS_ANYADDR6(ip->ip6_src)) {
		bzero((caddr_t)&ip->ip6_src, sizeof(struct in6_addr));
		ip->ip6_src.s6_addr[0] = 0xff;
		ip->ip6_src.s6_addr[1] = MADDR6_SCP_LINK;
		ip->ip6_src.s6_addr[15] = MADDR6_ALLNODES;
	}
	if ((ap->type == IP6ANY_ANYCAST) &&
	    IN6_IS_ADDR_MULTICAST(&ip->ip6_dst))
		delay = random() % MAX_ANYCAST_DELAY_TIME;
	else
		delay = 0;
	add_pending(ap, &ip->ip6_src, delay);
}

/*
 * Build a pending response for an anycast/proxy response
 * to a neighbor sollicitation.
 */
static void
add_pending(ap, peer, delay)
	struct anycast *ap;
	struct in6_addr *peer;
	long delay;
{
	register struct presp *rp;
	struct timeout *tp;

	if (Calloc(rp, presp) == NULL) {
		logerr("out of memory");
		abort();
	}
	rp->which = ap;
	COPY_ADDR6(*peer, rp->peer);
	if (delay == 0) {
		send_apadv(rp);
		return;
	}

	TAILQ_INSERT_TAIL(&ap->presps, rp, prs_list);

	(void) gettimeofday(&now, NULL);
	tp = new_timeout(MS_TO_SEC(delay), MS_TO_USEC(delay),
			 (func_to) send_apadv, rp);
	if (tp == NULL)
		abort();
	rp->tp = tp;
}

/*
 * Cancel a pending response for an anycast/proxy response
 * to a neighbor sollicitation.
 */
void
cancel_anycast(rp)
	register struct presp *rp;
{
	if (rp->tp == NULL) {
		if (verbose)
			logerr("pending response without a timeout?");
	} else {
		kill_timeout(rp->tp);
		rp->tp = NULL;
	}

	TAILQ_REMOVE(&rp->which->presps, rp, prs_list);
	free(rp);
}

/*
 * avoid multiple error mismatch messages
 */
#define ERR_KEEP	6*3600
#define ERR_PURGE	20*60
struct err_msg {
	TAILQ_ENTRY(err_msg) er_list;	/* list of messages */
	u_int index;			/* kernel ifp index */
	struct in6_addr addr;		/* source address */
	int type;			/* err msg */
	time_t date;			/* when created */
	union {
#define SKIPICMP ((int)&(((struct nd_router_advert *)0)->nd_ra_curhoplimit))
		char rtadv[sizeof(struct nd_router_advert)];
		struct nd_opt_prefix_info pref;
		struct nd_opt_mtu mtu;
	} data;
};
static TAILQ_HEAD(, err_msg) err_msg;
static struct timeout *err_inited;
#define FORALLERR(er)	\
	for (er = err_msg.tqh_first; er; er = er->er_list.tqe_next)

void
init_err()
{
	TAILQ_INIT(&err_msg);
	(void)gettimeofday(&now, NULL);
	err_inited = new_timeout(ERR_PURGE, 0, (func_to) err_timeout, 0);
	if (err_inited) {
		err_inited->refcnt = 1;	/* keep on expire */
		err_inited->arg = err_inited;
	}
}

static void
expire_err()
{
	struct err_msg *er;

	while ((er = err_msg.tqh_first)) {
		if (now.tv_sec - er->date <= ERR_KEEP)
			break;
		TAILQ_REMOVE(&err_msg, er, er_list);
	}
}

/* timout process on err tables */
static void
err_timeout(tp)
	struct timeout *tp;
{
	tp->tv.tv_sec += ERR_PURGE;
	expire_err();
	add_timeout(tp);
}

void
dump_err()
{
	register struct err_msg *er;

	FORALLERR(er) {
		logdebug("error descr (%lx):", (u_long)er);
		logdebug("\tfrom if#%d addr=%s",
			 er->index,
			 inet6pr(&er->addr));
		logdebug("\ttype=%d created=%lu", er->type, er->date);
		if (er->type == ND_OPT_PREFIX_INFORMATION)
			logdebug("\tprefix=%s/%d",
				 inet6pr(&er->data.pref.nd_opt_pi_prefix),
				 er->data.pref.nd_opt_pi_prefix_len);
	}
}

static int
already_err(ifp, src, type, info)
	struct interface *ifp;
	struct in6_addr *src;
	int type;
	void *info;
{
	struct err_msg *er;
	struct nd_opt_prefix_info *pip;
	struct nd_opt_mtu *mp;
	int len;

	if (!err_inited)
		return 0;
	if (type == 0) {
		((struct nd_router_advert *)info)->nd_ra_router_lifetime = 0;
		len = sizeof er->data.rtadv;
	} else if (type == ND_OPT_MTU) {
		len = sizeof er->data.mtu;
	} else if (type == ND_OPT_PREFIX_INFORMATION) {
		len = sizeof er->data.pref;
	} else {
		logerr("internal error #1");
		abort();
	}
	expire_err();
	FORALLERR(er) {
		if (er->type != type ||
		    !(er->index == ifp->index && SAME_ADDR6(er->addr, *src)))
			continue;
		 if (type != ND_OPT_PREFIX_INFORMATION)
			break;
		 pip = (struct nd_opt_prefix_info *)info;
		 if (er->data.pref.nd_opt_pi_prefix_len ==
			pip->nd_opt_pi_prefix_len &&
		     SAME_ADDR6(er->data.pref.nd_opt_pi_prefix,
				pip->nd_opt_pi_prefix))
			break;
	}
	if (er) {
		switch(type) {		/* same error */
		    case 0:	/* ICMPv6 RT ADV */
			if (memcmp(((char *)&er->data.rtadv) + SKIPICMP,
				   ((char *)info) + SKIPICMP,
				   sizeof(struct nd_router_advert) - SKIPICMP)
			      == 0)
				return 1;
			break;
		    case ND_OPT_MTU:
			mp = (struct nd_opt_mtu *)info;
			if (mp->nd_opt_mtu_mtu == er->data.mtu.nd_opt_mtu_mtu)
				return 1;
			break;
		    case ND_OPT_PREFIX_INFORMATION: {
			int d, i;

			pip = (struct nd_opt_prefix_info *)info;
			if (er->data.pref.nd_opt_pi_flags_reserved !=
					pip->nd_opt_pi_flags_reserved)
				break;
			i = now.tv_sec - er->date;
			d = pip->nd_opt_pi_valid_time -
				er->data.pref.nd_opt_pi_valid_time;
			if (d && abs(d - i) > DELTALIFE)
				break;
			d = pip->nd_opt_pi_preferred_time -
				er->data.pref.nd_opt_pi_preferred_time;
			if (d && abs(d - i) > DELTALIFE)
				break;
			return 1;
		    }
		}
		TAILQ_REMOVE(&err_msg, er, er_list);	/* reinit */
	}
	if (Calloc(er, err_msg) == NULL) {
		if (verbose)
			logerr("out of memory");
		return 0;
	}
	er->index = ifp->index;
	COPY_ADDR6(*src, er->addr);
	er->type = type;
	er->date = now.tv_sec;
	memcpy(&er->data, info, len);
	TAILQ_INSERT_TAIL(&err_msg, er, er_list);
	return 0;
}
