#include "ndpd.h"

int rtm_seq = 0;
u_short lo_index = 0;

static struct interface *new_if(struct if_msghdr *, int);
static struct if_msghdr *get_ifinfo(int);
static void refresh_flags(struct interface *);
void setppflags(struct prefix *);

/* interface changes */
void
recv_route()
{
	struct if_msghdr *ifm;
	struct ifa_msghdr *ifam;
	register struct sockaddr_in6 *sa, *sin = NULL;
	register struct interface *ifp;
	register struct prefix *pp;
	int cc, i;
	char *cp;
	int is_up, was_up, first, new;
	u_char plen = 0;

	cp = (caddr_t)&m_rtmsg;
	if (debug)
		bzero(cp, MAX_MSG_ROUTE);
	cc = read(r, cp, MAX_MSG_ROUTE);
	if (cc < 0) {
		if (verbose)
			logperror("read (route message)");
		return;
	} else if (cc == 0) {
		if (verbose)
			logerr("read (route message) return 0");
		return;
	}

	if (rtm.rtm_version != RTM_VERSION) {
		if (verbose)
			logerr("bad message version %d", rtm.rtm_version);
		return;
	}

	if (verbose)
		logtrace("receive route (%d,%d)",
			 rtm.rtm_msglen, rtm.rtm_type);

	switch (rtm.rtm_type) {
	    case RTM_NEWADDR:
		ifam = (struct ifa_msghdr *)cp;
		if (verbose)
		    	logtrace("receive route newaddr on #%d",
				 ifam->ifam_index);
#ifdef IFA_FIRST
		first = ifam->ifam_flags & IFA_FIRST;
#else
		first = 0;
#endif
		FORALLIFP(ifp)
		    if (ifp->index == ifam->ifam_index)
			break;
		if (ifp == NULL)
			ifp = new_if((struct if_msghdr *)cp, 1);
		if (ifp == NULL)
			return;
		cp += sizeof *ifam;
		sin = NULL;
		for (i = 0; i < RTAX_MAX; i++) {
			if ((ifam->ifam_addrs & (1 << i)) == 0)
				continue;
			sa = (struct sockaddr_in6 *)cp;
			ADVANCE(cp, sa);
			switch (i) {
			    case RTAX_IFA:
				if (sa->sin6_family != AF_INET6)
					return;
				sin = sa;
				break;

			    case RTAX_NETMASK:
				if (sa->sin6_family != AF_UNSPEC)
					return;
				plen = get_prefixlen(sa);
				break;
			}
		}
		if ((sin == NULL) ||
		    (plen < 3) ||
		    (sin->sin6_addr.s6_addr[0] == 0) ||
		    (sin->sin6_addr.s6_addr[0] == 0xff))
			return;
		if (IS_LINKLADDR6(sin->sin6_addr) ) {
			refresh_flags(ifp);
			was_up = IS_UP(ifp);
			if (ifp->state != STATENOLL) {
				if (!first)	/* an other LL in front */
					return;
				if (bcmp(&sin->sin6_addr,
					 &ifp->llocal,
					 sizeof ifp->llocal) != 0) {
					logerr("changing LL addr of %s",
					       ifp->name);
				}
				localroute(ifp, &ifp->llocal, 0);
				was_up = 0;	/* force RTsol if IFF_UP */
			} else
				ifp->state = STATELESS;
			COPY_ADDR6(sin->sin6_addr, ifp->llocal);
			localroute(ifp, &ifp->llocal, 1);
			goto change_if;
		}
		FORALLPP(pp, ifp)
		    if ((pp->plen == plen) &&
			SAME_ADDR6(pp->addr, sin->sin6_addr))
			break;
		if (pp) {	
			if (verbose)
				logtrace("reget address %s/%d on %s first %d",
					 inet6pr(&sin->sin6_addr), plen,
					 ifp->name, first);
			if (!first)
				return;
			TAILQ_REMOVE(&ifp->prefixes, pp, prf_list);
			new = 0;
			goto reorderpp;
		}

		if (verbose)
			logtrace("get new address %s/%d on %s",
				 inet6pr(&sin->sin6_addr), plen, ifp->name);
		if (Calloc(pp, prefix) == NULL) {
			if (verbose)
				logerr("out of memory");
			return;
		}

		if ((plen < 16) ||
		    ((plen != 128) && (plen > 128 - 8 * ifp->ltoken))) {
			if (verbose)
				logerr("bad prefix length %d (should be %d)",
				       plen, 128 - 8 * ifp->ltoken);
		}
		pp->plen = plen;
		COPY_ADDR6(sin->sin6_addr, pp->addr);
		COPY_ADDR6(pp->addr, pp->pref);
		mk_pref(pp);
		pp->ifp = ifp;
		setppflags(pp);
		TAILQ_INIT(&pp->anycasts);
		TAILQ_INIT(&pp->hagents);
		pp->hacnt = -1;
		if (FIRST(ifp) == 0)
			first = 1;
		new = 1;
	reorderpp:
		if (first)
			TAILQ_INSERT_HEAD(&ifp->prefixes, pp, prf_list);
		else
			TAILQ_INSERT_TAIL(&ifp->prefixes, pp, prf_list);
		if (new)
			localroute(ifp, &pp->addr, 1);
		update_prefix(pp);
		update_advprefix(pp);
		if (IS_UP(ifp)) {
			if (optrip) {
				if (first)
					update_rip(ifp);
				if (new)
					add_route_prefix(pp, ifp, 0);
			}
			if (new)
				add_anycast_prefix(ifp, pp, NULL);
		}
		return;

	    case RTM_DELADDR:
		ifam = (struct ifa_msghdr *)cp;
		if (verbose)
			logtrace("receive route deladdr on #%d",
				 ifam->ifam_index);
		FORALLIFP(ifp)
		    if (ifp->index == ifam->ifam_index)
			break;
		if (ifp == NULL)
			return;
		cp += sizeof *ifam;
		for (i = 0; i < RTAX_MAX; i++) {
			if ((ifam->ifam_addrs & (1 << i)) == 0)
				continue;
			sa = (struct sockaddr_in6 *)cp;
			ADVANCE(cp, sa);
			switch (i) {
			    case RTAX_IFA:
				if (sa->sin6_family != AF_INET6)
					return;
				sin = sa;
				break;

			    case RTAX_NETMASK:
				if (sa->sin6_family != AF_UNSPEC)
					return;
				plen = get_prefixlen(sa);
				break;
			}
		}
		if (plen < 3)
			return;
		if (IS_LINKLADDR6(sin->sin6_addr) ) {
			refresh_flags(ifp);
			was_up = IS_UP(ifp);
			if (ifp->state != STATENOLL) {
				logerr("deleting LL addr of %s", ifp->name);
				ifp->state = STATENOLL;
			}
			localroute(ifp, &ifp->llocal, 0);
			goto change_if;
		}
		FORALLPP(pp, ifp)
		    if ((pp->plen == plen) &&
			SAME_ADDR6(pp->addr, sin->sin6_addr))
			break;
		if (pp == NULL)
			return;
		localroute(ifp, &pp->addr, 0);
		del_route_prefix(pp, ifp);
		del_anycast_prefix(ifp, pp, NULL);
		first = (pp == FIRST(ifp));
		prefix_inval(pp);	/* and free pp */
		if (optrip && first)
			update_rip(ifp);
		return;

	    case RTM_IFINFO:
		ifm = (struct if_msghdr *)cp;
		if (verbose)
			logtrace("receive route ifinfo for #%d",
				 ifm->ifm_index);
		FORALLIFP(ifp)
		    if (ifp->index == ifm->ifm_index)
			break;
#define IFF_FIXED (IFF_BROADCAST|IFF_MULTICAST|IFF_LOOPBACK|IFF_POINTOPOINT)
		if (ifp) {
			if ((ifp->flags ^ ifm->ifm_flags) & IFF_FIXED) {
				logerr("unsupported flag change for %s",
					ifp->name);
				abort();
			}
			was_up = IS_UP(ifp);
		} else {
			was_up = 0;
			ifp = new_if(ifm, 1);
			if (ifp == NULL)
				return;
			read_gateway(1);	/* setoptions */
		}

		ifp->flags = ifm->ifm_flags;
    change_if:
		is_up = IS_UP(ifp);
		if (was_up && !is_up) {
			if (verbose)
				logtrace("interface %s down", ifp->name);
			if_down(ifp);
			return;
		}

		if (!(is_up && !was_up))
			return;

		if (verbose)
			logtrace("interface %s up", ifp->name);

		if_up(ifp);
		return;

	    case RTM_LOSING:
#define RTA_LOST	(RTA_DST | RTA_GATEWAY)
		if ((rtm.rtm_addrs & RTA_LOST) != RTA_LOST)
			return;
		cp += sizeof(struct rt_msghdr);
		sin = (struct sockaddr_in6 *)cp;
		if (sin->sin6_family != AF_INET6)
			return;
		ADVANCE(cp, sin);

		if (verbose)
			logtrace("lost peer %s", inet6pr(&sin->sin6_addr));
		lost_mobile(sin);
		return;
	}
}


static int mib[] = { CTL_NET, PF_ROUTE, 0, AF_INET6, NET_RT_IFLIST, 0 };

/* init tables, get all interfaces */
void
init()
{
	register struct if_msghdr *ifm;
	register struct interface *ifp;
	register struct prefix *pp;
	struct sockaddr_in6 *sa, *sin;
	int i, plen, count = 0;
	char *buf, *cp, *ap;
	unsigned bufsize;

	TAILQ_INIT(&interfaces);
	TAILQ_INIT(&timeouts);
	TAILQ_INIT(&routes);
	TAILQ_INIT(&Freeroutes);
	ifp = NULL;

	if (sysctl(mib, sizeof mib/sizeof(int), NULL, &bufsize, NULL, 0) < 0) {
		logperror("sysctl0");
		exit(4);
	}
	buf = (char *)malloc(bufsize);
	if (buf == NULL) {
		logerr("out of memory");
		exit(4);
	}
	if (sysctl(mib, sizeof mib/sizeof(int), buf, &bufsize, NULL, 0) < 0) {
		logperror("sysctl");
		exit(4);
	}

	for (cp = buf; cp < buf + bufsize; cp += ifm->ifm_msglen) {
		ifm = (struct if_msghdr *)cp;

		if (ifm->ifm_type == RTM_IFINFO) {
			ifp = new_if(ifm, 1);
			continue;
		}
		if (ifp == NULL)
			continue;
		if (ifm->ifm_type != RTM_NEWADDR) {
			if (verbose)
				logerr("bad message type (%d)",
				       ifm->ifm_type);
			continue;
		}
		if ((ifm->ifm_addrs & RTA_IFA) == 0) {
			if (verbose)
				logerr("no ifa for %s", ifp->name);
			continue;
		}

		plen = 0;
		sin = NULL;
		ap = cp + sizeof(struct ifa_msghdr);
		for (i = 0; i < RTAX_MAX; i++) {
			if ((ifm->ifm_addrs & (1 << i)) == 0)
				continue;
			sa = (struct sockaddr_in6 *)ap;
			ADVANCE(ap, sa);
			switch (i) {
			    case RTAX_NETMASK:
				if (sa->sin6_family != AF_UNSPEC)
					goto skip;
				plen = get_prefixlen(sa);
				break;

			    case RTAX_IFA:
				if (sa->sin6_family != AF_INET6)
					goto skip;
				sin = sa;
				break;
			}
		}
		count++;
		if ((sin == NULL) ||
		    (plen < 3) ||
		    (sin->sin6_addr.s6_addr[0] == 0) ||
		    (sin->sin6_addr.s6_addr[0] == 0xff)) {
	    skip:
			continue;
		}
		if (IS_LINKLADDR6(sin->sin6_addr)) {
			if (IS_ANYADDR6(ifp->llocal))
				COPY_ADDR6(sin->sin6_addr, ifp->llocal);
			continue;
		}
		if (Calloc(pp, prefix) == NULL) {
			if (verbose)
				logerr("out of memory");
			continue;
		}

		if ((plen < 16) ||
		    ((plen != 128) && (plen > 128 - 8 * ifp->ltoken))) {
			if (verbose)
				logerr("bad prefix length %d (should be %d)",
				       plen, 128 - 8 * ifp->ltoken);
		}
		pp->ifp = ifp;
		COPY_ADDR6(sin->sin6_addr, pp->addr);
		COPY_ADDR6(pp->addr, pp->pref);
		pp->plen = plen;
		mk_pref(pp);
		setppflags(pp);
		TAILQ_INIT(&pp->anycasts);
		TAILQ_INIT(&pp->hagents);
		pp->hacnt = -1;
		TAILQ_INSERT_TAIL(&ifp->prefixes, pp, prf_list);
		if (verbose)
			logtrace("interface %s got addr %s plen %d addrs %x",
				 ifp->name,
				 inet6pr(&pp->addr),
				 pp->plen, ifm->ifm_addrs);
	}
	free(buf);

	if (count == 0) {
		logerr("found no IPv6 address!?");
		exit(5);
	}

	FORALLIFP(ifp) {
		if (IS_ANYADDR6(ifp->llocal)) {
			if (verbose)
				logtrace("ignoring interface %s without LL",
					 ifp->name);
		} else
			ifp->state = STATELESS;
		FORALLPP(pp, ifp)
			update_prefix(pp);
	}

	bigbuffer = malloc(bigbufferlen);
	if (!bigbuffer) {
		logerr("out of memory");
		exit(3);
	}
}

static struct if_msghdr *
get_ifinfo(index)
	int index;
{
	register struct if_msghdr *ifm = NULL;
	char *buf;
	unsigned bufsize;

	mib[5] = index;

	if (sysctl(mib, sizeof mib/sizeof(int), NULL, &bufsize, NULL, 0) < 0) {
		if (verbose)
			logperror("get_ifinfo: sysctl0");
		return ifm;
	}
	buf = (char *)malloc(bufsize);
	if (buf == NULL) {
		if (verbose)
			logerr("out of memory");
		return ifm;
	}
	if (sysctl(mib, sizeof mib/sizeof(int), buf, &bufsize, NULL, 0) < 0) {
		if (verbose)
			logperror("get_ifinfo: sysctl");
		free(buf);
		return ifm;
	}
	ifm = (struct if_msghdr *)buf;

	if ((ifm->ifm_type != RTM_IFINFO) || (ifm->ifm_index != index)) {
		ifm = NULL;
		free(buf);
	}
	return ifm;
}

static struct interface *
new_if(ifm, first)
	register struct if_msghdr *ifm;
	int first;
{
	register struct interface *ifp = NULL;
	struct sockaddr_dl *sdl;
	int hasphy = 0, ltoken = 0, lmac = 0;

	if ((ifm->ifm_addrs != (1 << RTAX_IFP)) ||
	    (ifm->ifm_type != RTM_IFINFO)) {
		struct if_msghdr *nifm;

		if (first) {
			nifm = get_ifinfo((int)ifm->ifm_index);
			if (nifm == NULL)
				return ifp;
			ifp = new_if(nifm, 0);
			free(nifm);
			return ifp;
		} else if (verbose)
			logerr("can't get ifinfo for if#%d", ifm->ifm_index);
		return ifp;
	}

	sdl = (struct sockaddr_dl *)((char *)ifm + sizeof(struct if_msghdr));
	if ((sdl->sdl_family != AF_LINK) || (sdl->sdl_nlen == 0)) {
		if (verbose)
			logerr("bad sdl address");
		return ifp;
	}

	if ((lo_index == 0) &&
	    (ifm->ifm_flags & IFF_LOOPBACK) &&
	    (sdl->sdl_nlen >= 3) &&
	    (bcmp(sdl->sdl_data, "lo0", sdl->sdl_nlen) == 0))
		lo_index = ifm->ifm_index;

	if ((ifm->ifm_flags & IFF_LOOPBACK) ||
	    (ifm->ifm_data.ifi_mtu < IP6_MMTU))
		return ifp;

	if (ifm->ifm_flags & IFF_POINTOPOINT) {
		if (sdl->sdl_nlen > 3 &&
		   (bcmp(sdl->sdl_data, "cti", 3) == 0))
			ltoken = SITTOKEN;
		else if (ifm->ifm_data.ifi_type == IFT_PPP)
			ltoken = PPPTOKEN;
		/* else ... not supported */
	} else if ((ifm->ifm_flags & IFF_MULTICAST) == IFF_MULTICAST) {
		/* everything has 48 bit mac addr ... */
		ltoken = MACTOKEN;
		lmac = MACMAC;
		if (ifm->ifm_data.ifi_addrlen != lmac)
			return ifp;
		/* ignore ATM interfaces (should work but disturb) ??? */
		if (ifm->ifm_data.ifi_type == IFT_ATM)
			return ifp;
		hasphy = 1;
	} /* else ... not supported */
	if (ltoken == 0)
		return ifp;

	if (ifm->ifm_addrs != RTA_IFP) {
		if (verbose)
			logerr("bad ifinfo (no ifp#%d)",
			       ifm->ifm_index);
		return ifp;
	}
	if ((sdl->sdl_family != AF_LINK) || (sdl->sdl_nlen == 0)) {
		if (verbose)
			logerr("bad sdl address");
		return ifp;
	}
	if (hasphy && sdl->sdl_alen != lmac) {
		if (verbose)
			logerr("unsupported type for interface %.*s",
				sdl->sdl_nlen, sdl->sdl_data);
		return ifp;
	}

	if (Calloc(ifp, interface) == NULL) {
		logerr("out of memory");
		abort();
	}
	TAILQ_INSERT_TAIL(&interfaces, ifp, if_list);
	TAILQ_INIT(&ifp->prefixes);
	TAILQ_INIT(&ifp->fprefixes);
	TAILQ_INIT(&ifp->oanycasts);
	ifp->mtu = ifm->ifm_data.ifi_mtu;
	ifp->metric = ifm->ifm_data.ifi_metric;
	if (bigbufferlen < ifp->mtu) {
		bigbufferlen = ifp->mtu;
		if (bigbuffer) {
			free(bigbuffer);
			bigbuffer = malloc(bigbufferlen);
			if (bigbuffer == NULL) {
				logerr("out of memory");
				abort();
			}
		}
	}
	ifp->flags = ifm->ifm_flags;
	ifp->index = ifm->ifm_index;
	ifp->state = STATENOLL;
	ifp->hasphy = hasphy;
	ifp->ltoken = ltoken;
	ifp->lmac = lmac;
	ifp->opt = defopt;
	if (hasphy)
		bcopy(LLADDR(sdl), (caddr_t)ifp->adphy, lmac);
	bcopy(sdl->sdl_data, ifp->name, sdl->sdl_nlen);
	ifp->ntosend = 0;
	if (verbose)
		logtrace("got interface %s#%d", ifp->name, ifp->index);
	return ifp;
}

/*
 * Refresh copy of interface flags on link-local address event
 */

static void
refresh_flags(ifp)
	struct interface *ifp;
{
	struct ifreq ifr;

	bzero((caddr_t)&ifr, sizeof(ifr));
	strncpy(ifr.ifr_name, ifp->name, IFNAMSIZ);

	if (ioctl(u, SIOCGIFFLAGS, (caddr_t)&ifr) < 0) {
		logperror("ioctl (SIOCGIFFLAGS)");
		return;
	}
	ifp->flags = ifr.ifr_flags;
}

int
is_router()
{
	static int mib[] = { CTL_NET, PF_INET6, IPPROTO_IP, IP6CTL_FORWARDING};
	int val;
	size_t size = sizeof(val);

	if (sysctl(mib, sizeof mib/sizeof(int), &val, &size, 0, 0) < 0 ||
			size != sizeof(val)) {
		if (verbose)
			logerr("can't test ip6_forwarding");
		return (1);
	}
	return(val);
}

void
setppflags(pp)
	struct prefix *pp;
{
	pp->flags = ND_OPT_PI_FLAG_AUTO;
	if (pp->plen != 128)
		pp->flags |= ND_OPT_PI_FLAG_ONLINK; /* ??? Alain conteste */
	if (pp->ifp && (pp->ifp->flags & IFF_POINTOPOINT))
		pp->flags |= ND_OPT_PI_FLAG_RTADDR; /* help peer */
}

/* set a route for a p2p local addr through localhost */

void
localroute(ifp, addr, set)
	struct interface *ifp;
	struct in6_addr *addr;
	int set;
{
	register char *cp = m_rtmsg.m_space;
	struct sockaddr_in6 *sin;
	struct sockaddr_dl *sdl;
	int len;

	if (((ifp->flags & IFF_POINTOPOINT) == 0) ||
	    (set && ((ifp->flags & IFF_UP) == 0)))
		return;
	if (debug)
		logdebug("%s local route for %s (link#%d)",
			 set ? "add" : "delete",
			 inet6pr(addr),
			 ifp->index);
	bzero((caddr_t)&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = set ? RTM_ADD : RTM_DELETE;
	rtm.rtm_flags = RTF_UP | RTF_HOST | RTF_LOCAL;
	rtm.rtm_addrs = RTA_DST | RTA_GATEWAY;
	rtm.rtm_index = ifp->index;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(*addr, sin->sin6_addr);
	ADVANCE(cp, sin);

	sdl = (struct sockaddr_dl *)cp;
	sdl->sdl_family = AF_LINK;
	sdl->sdl_len = sizeof(struct sockaddr_dl);
	sdl->sdl_index = ifp->index;
	ADVANCE(cp, sdl);

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

    retry:
	/* if exists do a change */
#ifdef TRACE
	DUMP("localroute", (char *)&m_rtmsg, len);
#endif
	if (write(r, (char *)&m_rtmsg, len) < len) {
		if ((rtm.rtm_type == RTM_ADD) && (errno == EEXIST)) {
			rtm.rtm_type = RTM_CHANGE;
			rtm.rtm_seq = ++rtm_seq;
			goto retry;
		} else if (verbose)
			logperror("write (localroute)");
		return;
	}
}
