#include "ndpd.h"

struct timeval now;
int s, r, u;			/* Socket file descriptors */
int one = 1;

int pid;
u_int MaxHop;
int optg = 0;
int optrip = 0;
int optripinst = 1;
u_int deplife = INFLIFE;
u_int invlife = INFLIFE;

int verbose = 0;
int debug = 0;
int timestamps = 0;
int mobile = 0;
char usage[] =  "Usage: ndpd-router [-dtvprMORSgnqs] "
		"[-H hapref[/halife]]\n\t[-D max[/min[/life]]] "
		"[-T reach/retr/hop] [-u ripport] [-f file]\n";
char *gateway6 = GATEWAY6;
struct option defopt, defoptini;

int main(int, char *[]);
static void sighup(int);
static void sigalrm(int);
static int sig_gateway;

static struct ipv6_mreq mr_val;

int
main(argc, argv)
	int argc;
	char *argv[];
{
	char ch;
	int nfds, cc;
	fd_set rfds;
	struct timeval tv;
	struct timeout *tp;
	struct interface *ifp;
	extern int optind;
	extern char *optarg;
	u_int on;
	int optr = 0;
	struct icmp6_filter filter;

	pid = getpid();
	(void) setpriority(PRIO_PROCESS, pid, PRIO_MIN + 2);
	(void) gettimeofday(&now, NULL);
	defopt.optsplit = 1;	/* poisoning reverse */
	defopt.valpref.val.life = DEFVALLIFE;
	defopt.valpref.pref.life = DEFPREFLIFE;
	/* ...val_date = ...pref_date = 0; */

	while ((ch = getopt(argc, argv, "ndvtrpP:MOH:D:T:gsqRSu:f:")) != EOF)
		switch (ch) {
		case 'S':
			defopt.optsplit = 0;
			break;
		case 'R':
			defopt.optsplit = -1;
			break;
		case 's':
			optrip = 1;
			break;
		case 'n':
			optripinst = 0;
			break;
		case 'q':
			optrip = -1;
			break;
		case 'g':
			optg = 1;
			break;
		case 'u':
			if (!(ripport = atoi(optarg)))
				goto err;
			break;
		case 'd':
			debug = 1;
			break;
		case 'v':
			verbose++;
			break;
		case 't':
			timestamps = 1;
			break;
		case 'r':
			optr = 1;
			break;
		case 'p':
			defopt.nopref = 1;
			break;
		case 'M':
			defopt.flags |= ND_RA_FLAG_MANAGED;
			break;
		case 'O':
			defopt.flags |= ND_RA_FLAG_OTHER;
			break;
		case 'H':
			defopt.flags |= ND_RA_FLAG_HAGENT;
			{
				char *p;
				u_int i = 0;

				defopt.hapref = strtol(optarg, &p, 0);
				if (*p)
					i = strtoul(p + 1, 0, 0);
				if (i > MAX_RT_LIFE)
					defopt.halife = MAX_RT_LIFE;
				else if (i > 0)
					defopt.halife = i;
				if (defopt.hapref || defopt.halife)
					defopt.extra |= ADVHAINF;
			}
			break;
		case 'D':
			if (*optarg == 'n')
				defopt.no_rtadv = 1;
			else {
				char *p;
				u_int i;

				defopt.maxrtadv = strtoul(optarg, &p, 0);
				if (*p)
					defopt.minrtadv =
						strtoul(p + 1, &p, 0);
				else
					defopt.minrtadv = 0;
				if (*p) {
					i = strtoul(p + 1, 0, 0);
					if (i > MAX_RT_LIFE)
						i = MAX_RT_LIFE;
					defopt.life = i;
					if (!optr)
						optr = -1;
				}
			}
			break;
		case 'T':
			{
				char *p;
				u_int i;

				if ((i = strtoul(optarg, &p, 0))) {
					if (i < 2)
						i = 2;
					if (i > 3600)
						i = 3600;
					defopt.reach = i*1000; /* millisec */
				}
				if (*p && (i = strtoul(p + 1, &p, 0))) {
					if (i < 2)
						i = 2;
					if (i > 3600)
						i = 3600;
					defopt.retrans = i*1000; /* millisec */
				}
				if (*p && (i = strtoul(p + 1, 0, 0))) {
					if (i > MAXHOP)
						i = MAXHOP;
					MaxHop = i;
				}
			}
			break;
		case 'P':
			{
				char *p;
				int ret;

				p = strchr(optarg, '/');
				if (p == 0 || p != optarg) {
					if (p)
						*p = 0;
					ret = convvalpref(optarg,
						&defopt.valpref.val_date,
						&defopt.valpref.val);
					if (p)
						*p = '/';
					if (ret < 0)
						goto err;
					defopt.valpref.pref= defopt.valpref.val;
					defopt.valpref.pref_date =
					    defopt.valpref.val_date;
				}
				if (p)
					if (convvalpref(p+1,
						&defopt.valpref.pref_date,
						&defopt.valpref.pref) < 0)
						goto err;
			}
			break;
		case 'f':
			gateway6 = optarg;
			break;
		default:
		err:
			(void) fprintf(stderr, usage);
			exit(1);
		}

	argc -= optind;
	argv += optind;

	if (argc != 0) {
		(void) fprintf(stderr, "Extra parameters\n");
		(void) fprintf(stderr, usage);
		exit(1);
	}
	defopt.maxrtadv *= M_SECOND;	/* milliseconds */
	defopt.minrtadv *= M_SECOND;	/* milliseconds */
	if (defopt.maxrtadv < MIN_RTR_ADV_INTERVAL ||
	    defopt.maxrtadv > MAX_RTR_ADV_INTERVAL)
		defopt.maxrtadv = DEF_RTR_ADV_INTERVAL;
	if (defopt.minrtadv < (1 * M_SECOND) ||
	    defopt.minrtadv > ((defopt.maxrtadv * 3) / 4))
		defopt.minrtadv = defopt.maxrtadv / 3;
	if (!optr)
		defopt.life = 10 * (defopt.maxrtadv / M_SECOND);
	if (optr > 0)
		defopt.life = 0;
	defoptini = defopt;

	srandom((int)gethostid());
	setlinebuf(stdout);

	if ((s = socket(PF_INET6, SOCK_RAW, IPPROTO_ICMPV6)) < 0) {
		logperror("socket (ICMP)");
		exit(2);
	}

	on = 1;
	if (setsockopt(s, IPPROTO_IPV6, IP_HDRINCL,
		       &on, sizeof(on)) < 0) {
		logperror("setsockopt (IP_HDRINCL)");
		exit(3);
	}

	on = 1;
	if (setsockopt(s, IPPROTO_IPV6, IPV6_RECVPKTINFO,
		       &on, sizeof(on)) < 0) {
		logperror("setsockopt (IPV6_RECVPKTINFO)");
		exit(3);
	}

	on = ND_DEFAULT_HOPLIM;
	if (setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_HOPS,
		       &on, sizeof(on)) < 0) {
		logperror("setsockopt (IPV6_MULTICAST_HOPS)");
		exit(3);
	}

	on = 0;
	if (setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_LOOP,
		       &on, sizeof(on)) < 0) {
		logperror("setsockopt (IPV6_MULTICAST_LOOP)");
		exit(3);
	}

	/*
	 * Only interested in a few icmp:
	 * ND_NEIGHBOR_SOLICIT, ND_ROUTER_ADVERT, ND_ROUTER_SOLICIT,
	 */
	ICMP6_FILTER_SETBLOCKALL(&filter);
	ICMP6_FILTER_SETPASS(ND_NEIGHBOR_SOLICIT, &filter);
	ICMP6_FILTER_SETPASS(ND_ROUTER_ADVERT, &filter);
	ICMP6_FILTER_SETPASS(ND_ROUTER_SOLICIT, &filter);
	if (setsockopt(s, IPPROTO_IPV6, ICMP6_FILTER,
		       &filter, sizeof(filter)) < 0) {
		logperror("setsockopt (ICMP6_FILTER)");
		exit(3);
	}

	/* protect the ICMPv6 socket against mobility stuff */

	disable_binding(s);
	disable_homeaddr(s);

	if ((r = socket(PF_ROUTE, SOCK_RAW, 0)) < 0) {
		logperror("socket (ROUTE)");
		exit(2);
	}

	if ((u = socket(PF_INET6, SOCK_DGRAM, 0)) < 0) {
		logperror("socket (UDP)");
		exit(2);
	}
	(void) shutdown(u, 0);

	init();
	mobility_init();
	init_err();

	if (!interfaces.tqh_first) {
		if (verbose) logerr("No IPV6 interface !!");
		exit(1);
	}

	(void) signal(SIGUSR1, set_verbose);
	(void) signal(SIGUSR2, set_verbose);
	(void) signal(SIGINT, dump);
	(void) signal(SIGTERM, finish);
	(void) signal(SIGQUIT, finish);
	(void) signal(SIGHUP, sighup);
	(void) signal(SIGALRM, sigalrm);

	if (!debug) {
		daemon(0, 0);
		initlog();
		pid = getpid();
	}

	bzero(&mr_val.ipv6mr_multiaddr, sizeof (mr_val.ipv6mr_multiaddr));
	mr_val.ipv6mr_multiaddr.s6_addr[0] = 0xff;
	mr_val.ipv6mr_multiaddr.s6_addr[1] = MADDR6_SCP_LINK;

	(void) gettimeofday(&now, NULL);

	if (optrip)
                init_rip();
	read_gateway(-1);
	clear_rip_changes();

	FORALLIFP(ifp)
		if (IS_UP(ifp))
			if_up(ifp);

	for (;;) {
		run_timeouts();

		FD_ZERO(&rfds);
		FD_SET(s, &rfds);
		nfds = s;
		FD_SET(r, &rfds);
		nfds = r > nfds ? r : nfds;
		FD_SET(m, &rfds);
		nfds = m > nfds ? m : nfds;
		if (optrip) {
			FD_SET(rip_in, &rfds);
			nfds = rip_in > nfds ? rip_in : nfds;
		}

		if ((tp = timeouts.tqh_first)) {
			tv.tv_usec = tp->tv.tv_usec + 1000000 - now.tv_usec;
			tv.tv_sec = tp->tv.tv_sec - now.tv_sec;
			NORMTIME(tv);
			if (tv.tv_sec > 0)
				tv.tv_sec -= 1;
			else
				tv.tv_usec = 10000;
		} else {
			tv.tv_sec = 3;
			tv.tv_usec = 0;
		}

		cc = select(nfds + 1, &rfds, NULL, NULL, &tv);

		new_gateway();
		if (sig_gateway) {
			sig_gateway = 0;
			continue;
		}
		if (cc < 0) {
			if (errno != EINTR && verbose)
				logperror("select");
			continue;
		}
		if (cc == 0)
			continue;
		(void) gettimeofday(&now, NULL);
		if (FD_ISSET(r, &rfds))
			recv_route();
		if (FD_ISSET(s, &rfds))
			recv_icmp();
		if (FD_ISSET(m, &rfds))
			recv_destopt();
		if (FD_ISSET(rip_in, &rfds))
			recv_rippacket();
		if (verbose)
			logerr("end of main loop reached");
	}
	/*NOTREACHED*/
}

void
set_verbose(sig)
	int sig;
{
	if (sig == SIGUSR1)
		verbose++;
	else
		verbose = 0;
}

static const char wday_name[7][3] = {
    "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
};
static const char mon_name[12][3] = {
    "Jan", "Feb", "Mar", "Apr", "May", "Jun",
    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};

void dump(s)
	int s;
{
	register struct timeout *tp;
	register struct interface *ifp;
	register struct Route *rp;
	struct tm *tmnow;
	int saverr = errno;

	(void) gettimeofday(&now, NULL);
	tmnow = localtime(&now.tv_sec);

	logdebug("DUMP at %.3s %.3s%3d %02d:%02d:%02d.%06d %d",
		 wday_name[tmnow->tm_wday],
		 mon_name[tmnow->tm_mon],
		 tmnow->tm_mday, tmnow->tm_hour,
		 tmnow->tm_min, tmnow->tm_sec,
		 now.tv_usec,
		 1900 + tmnow->tm_year);

	logdebug("timeout list:");
	for (tp = timeouts.tqh_first; tp; tp = tp->tm_list.tqe_next)
		dump_timeout(tp);

	logdebug("interface list:");
	FORALLIFP(ifp)
		dump_interface(ifp);

	logdebug("route list:");
	FORALLRP(rp)
		dump_route(rp);

	logdebug("mobile list:");
	dump_mobiles();

	logdebug("known error list:");
	dump_err();

	errno = saverr;
}

void
finish(s)
	int s;
{
	struct interface *ifp;

	mobility_cleanup();

	FORALLIFP(ifp)
		if (IS_UP(ifp))
			if_down(ifp);

	if (verbose)
		logtrace("finished!");

	exit(0);
}

static void
sighup(s)
	int s;
{
	sig_gateway = 1;
}

static void
sigalrm(s)
	int s;
{
	struct interface *ifp;

	FORALLIFP(ifp)
		if (IS_UP(ifp))
			start_rtadv(ifp);
}

void
if_rtadv(ifp, up)
	struct interface *ifp;
	int up;
{
	mr_val.ipv6mr_interface = ifp->index;
	mr_val.ipv6mr_multiaddr.s6_addr[15] = MADDR6_ALLROUTERS;
	if (up) {
		if (setsockopt(s, IPPROTO_IPV6, IPV6_JOIN_GROUP,
			       &mr_val, sizeof(mr_val)) < 0) {
			logerr("setsockopt (IPV6_JOIN_GROUP(%d)) on %s: %s",
				mr_val.ipv6mr_multiaddr.s6_addr[15],
				ifp->name,
				strerror(errno));
			if_down(ifp);
			ifp->flags &= ~IFF_UP;
		} else
			start_rtadv(ifp);
	} else {
		stop_rtadv(ifp);
		(void) setsockopt(s, IPPROTO_IPV6, IPV6_LEAVE_GROUP,
			       &mr_val, sizeof(mr_val));
	}
}

void
if_apadv(ifp, up)
	struct interface *ifp;
	int up;
{
	struct prefix *pp;

	if (up) {
		FORALLPP(pp, ifp)
			add_anycast_prefix(ifp, pp, NULL);
		add_anycast_prefix(ifp, NULL, NULL);
	} else {
		FORALLPP(pp, ifp)
			del_anycast_prefix(ifp, pp, NULL);
		del_anycast_prefix(ifp, NULL, NULL);
	}
}

void
anycast_multi(ap, up)
	struct anycast *ap;
	int up;
{
	struct interface *ifp = ap->ifp;

	mr_val.ipv6mr_interface = ifp->index;
	mr_val.ipv6mr_multiaddr.s6_addr[11] = 0x1;
	bcopy((caddr_t)&ap->addr.s6_addr[12],
	      (caddr_t)&mr_val.ipv6mr_multiaddr.s6_addr[12],
	      4);
	if (ifp->ltoken == 8)
		mr_val.ipv6mr_multiaddr.s6_addr[12] = 0xff;

	if (up) {
		if ((setsockopt(s, IPPROTO_IPV6, IPV6_JOIN_GROUP,
				&mr_val, sizeof(mr_val)) < 0) &&
		    (errno != EADDRINUSE))
			logperror("setsockopt (IPV6_JOIN_GROUP)");
	} else
		(void) setsockopt(s, IPPROTO_IPV6, IPV6_LEAVE_GROUP,
				  &mr_val, sizeof(mr_val));
	bzero((caddr_t)&mr_val.ipv6mr_multiaddr.s6_addr[11], 5);
}

void
if_ripin(ifp, up)
	struct interface *ifp;
	int up;
{
	mr_val.ipv6mr_interface = ifp->index;
	mr_val.ipv6mr_multiaddr.s6_addr[15] = MADDR6_RIP;

	if (up) {
		if (setsockopt(rip_in, IPPROTO_IPV6, IPV6_JOIN_GROUP,
				&mr_val, sizeof(mr_val)) < 0) {
			logerr("setsockopt (IPV6_JOIN_GROUP(%d)) on %s: %s",
				mr_val.ipv6mr_multiaddr.s6_addr[15],
				ifp->name,
				strerror(errno));
			if_down(ifp);
			ifp->flags &= ~IFF_UP;
		}
	} else {
		(void) setsockopt(rip_in, IPPROTO_IPV6, IPV6_LEAVE_GROUP,
				  &mr_val, sizeof(mr_val));
		del_route_ifp(ifp);
	}
}

void
if_ripout(ifp, up)
	struct interface *ifp;
	int up;
{
	if (up)
		start_rip(ifp);
	else
		stop_rip(ifp);
}

void
if_up(ifp)		/* IS_UP(ifp) is true */
	struct interface *ifp;
{
	struct prefix *pp;

	localroute(ifp, &ifp->llocal, 1);
	FORALLPP(pp, ifp)
		localroute(ifp, &pp->addr, 1);
	if (!NO_ANYCAST(ifp))
		if_apadv(ifp, 1);
	if (!ifp->opt.no_rtadv)
		if_rtadv (ifp, 1);
	if (!IS_UP(ifp) || !optrip)
		return;
	FORALLPP(pp, ifp)
		add_route_prefix(pp, ifp, 0);
	if (!ifp->opt.no_ripin)
		if_ripin(ifp, 1);
	if (!IS_UP(ifp))
		return;
	if (!ifp->opt.no_ripout)
		if_ripout(ifp, 1);
}


void
if_down(ifp)	/* IS_UP(ifp) is unknown */
	struct interface *ifp;
{
	struct prefix *pp;

	localroute(ifp, &ifp->llocal, 0);
	FORALLPP(pp, ifp)
		localroute(ifp, &pp->addr, 0);
	if (!NO_ANYCAST(ifp))
		if_apadv(ifp, 0);
	if (!ifp->opt.no_rtadv)
		if_rtadv(ifp, 0);
	if (!optrip)
		return;
	FORALLPP(pp, ifp)
		del_route_prefix(pp, ifp);
	if (!ifp->opt.no_ripin)
		if_ripin(ifp, 0);
	if (!ifp->opt.no_ripout)
		if_ripout(ifp, 0);
}

void
dump_prefix(pp)
	register struct prefix *pp;
{
	register struct anycast *ap;
	register struct hagent *hap;

	logdebug("prefix (%lx):", (u_long)pp);
	logdebug("\tinterface=%s", pp->ifp->name);
	logdebug("\tprefix=%s with len=%d and mask=%s",
		 inet6pr(&pp->pref), pp->plen, inet6pr(&pp->pmask));
	if (pp->flags & ND_OPT_PI_FLAG_SITE)
		logdebug("\tsite Plength=%d", pp->slen);
	logdebug("\tif address=%s", inet6pr(&pp->addr));
	logdebug("\tvalidity=%s\tpreference=%s\tflags=%x",
		 datelifepr(pp->valpref.val_date, &pp->valpref.val),
		 datelifepr(pp->valpref.pref_date, &pp->valpref.pref),
		 pp->flags);
	logdebug("anycast/proxy list:");
	FORALLPANY(ap, pp)
		dump_anycast(ap);
	if (pp->hacnt >= 0) {
		logdebug("home agent list (count=%d):", pp->hacnt);
		FORALLHA(hap, pp)
			dump_hagent(hap);
	} else
		logdebug("no home agent list");
}

void
dump_anycast(ap)
	register struct anycast *ap;
{
	register struct presp *rp;

	logdebug("anycast/proxy (%lx):", (u_long)ap);
	if (ap->prefix)
		logdebug("\tparent prefix=%lx", (u_long)ap->prefix);
	logdebug("\tinterface=%lx", (u_long)ap->ifp);
	logdebug("\taddress=%s", inet6pr(&ap->addr));
	logdebug("\ttype=%d\tadvtosend=%d at %lx",
		 ap->type, ap->ntosend, (u_long)(ap->tp));
	logdebug("pending response list:");
	for (rp = ap->presps.tqh_first; rp; rp = rp->prs_list.tqe_next)
		dump_presp(rp);
}

void
dump_presp(rp)
	register struct presp *rp;
{
	logdebug("pending response (%lx):", (u_long)rp);
	logdebug("\tanycast=%lx", (u_long)rp->which);
	logdebug("\ttimeout=%lx", (u_long)(rp->tp));
	logdebug("\tpeer=%s", inet6pr(&rp->peer));
}

void
dump_hagent(hap)
	struct hagent *hap;
{
	logdebug("home agent (%lx):", (u_long)hap);
	logdebug("\taddress=%s", inet6pr(&hap->addr));
	logdebug("\tpref=%d\tlifetime=%u", hap->hapref, hap->halife);
	logdebug("\ttimeout=%lx\tprefix=%lx",
		 (u_long)(hap->tp), (u_long)(hap->prefix));
}

void
dump_interface(ifp)
	register struct interface *ifp;
{
	register struct prefix *pp;
	register struct anycast *ap;

	logdebug("interface (%lx):", (u_long)ifp);
	logdebug("\tname=%s(if#%d)", ifp->name, ifp->index);
	logdebug("\tlink-local address=%s", inet6pr(&ifp->llocal));
	logdebug("\tstate=%d\tadvtosend=%d", ifp->state, ifp->ntosend);
	logdebug("\tlgtoken %d lgmac %d adphy %s", ifp->ltoken, ifp->lmac,
		 ifp->hasphy ? mac2ascii(ifp->adphy, ifp->lmac) : "NONE");
	logdebug("\tMTU=%lu\tmetric=%lu\tiffflags=%x",
		 ifp->mtu, ifp->metric, (u_int)ifp->flags);
	logdebug("\tnortadv/ripin/out=%d/%d/%d splith=%d flags=%x",
		 ifp->opt.no_rtadv, ifp->opt.no_ripin, ifp->opt.no_ripout,
		 ifp->opt.optsplit, ifp->opt.flags);
	logdebug("\tlife=%d reach/retrans=%d/%d",
		 ifp->opt.life, ifp->opt.reach, ifp->opt.retrans);
	logdebug("\trtadvmin/max=%d/%d (ms)",
		 ifp->opt.minrtadv, ifp->opt.maxrtadv);
	logdebug("\trtadv timeout =%lx riptimeout = %lx", ifp->tp, ifp->rtp);
	logdebug("\tfilter in/out=%x/%x", ifp->filtin, ifp->filtout);
	logdebug("\thome agent pref=%d life=%u",
		 ifp->opt.hapref, ifp->opt.halife);
	if (ifp->usefpref) {
		logdebug("announced prefix list:");
		LOOPALLPP(pp = FFIRST(ifp), pp)
			dump_prefix(pp);
		logdebug("configured prefix list:");
	} else
		logdebug("configured and announced prefix list:");
	FORALLPP(pp, ifp)
		dump_prefix(pp);
	logdebug("orphan anycast/proxy list:");
	FORALLOANY(ap, ifp)
		dump_anycast(ap);
}
