#define NDPD_ROUTER
#include "util_ndpd.h"

union datelife {
	u_int32_t life;
	time_t date;
};
struct valpref {
	signed char	val_date;	/* !=0 if real time */
	signed char	pref_date;
	char	imported;
	union datelife val;
	union datelife pref;
};
/* for comparing real time prefix announces */
#define DELTALIFE	60

struct prefix {
	TAILQ_ENTRY(prefix) prf_list;	/* prefix list */
	TAILQ_HEAD(, anycast) anycasts;	/* anycast (and proxy) address list */
	TAILQ_HEAD(, hagent) hagents;	/* home agent list */
	struct interface *ifp;		/* up pointer */
	struct in6_addr pref;		/* prefix itself */
	struct in6_addr pmask;		/* prefix mask */
	struct in6_addr addr;		/* address */
	struct valpref valpref;		/* valid/pref lifetime/date */
	u_char plen;			/* length */
	u_char slen;			/* site Plength */
	u_char flags;			/* flags */
	int hacnt;			/* count of home agents */
};

struct option {
	struct difopt {		/* any change need action */
		u_char d_no_ripin;		/* do not listen to rip */
		u_char d_no_ripout;		/* do not send rip */
		u_char d_no_rtadv;		/* send rtadv */
	} difopt;
#define no_ripin	difopt.d_no_ripin
#define no_ripout	difopt.d_no_ripout
#define no_rtadv	difopt.d_no_rtadv
#define	NO_ANYCAST(ifp)	(ifp->opt.no_rtadv)
	char optsplit;			/* != : split hor <0 : non poison */
	u_char flags;			/* flags M O H */
	char nopref;			/* !=0 no pref announce */
	char extra;			/* extra options */
#define ADVMTU		0x01		/* advertise MTU */
#define ADVMTUIF	0x02		/* advertise ifp->mtu */
#define ADVADVINT	0x04		/* advertise max interval */
#define ADVHAINF	0x08		/* advertise home-agent infos */
	u_int16_t life;			/* router life value */
	u_int32_t reach;		/* reachable time */
	u_int32_t retrans;		/* retrans time */
	u_int minrtadv;			/* rtadv delay */
	u_int maxrtadv;			/* rtadv delay */
	u_int32_t mtuadv;		/* advertised MTU */
	int16_t hapref;			/* home-agent preference */
	u_int16_t halife;		/* home-agent life value */
	struct valpref valpref;
};

struct anycast {
	TAILQ_ENTRY(anycast) any_list;	/* anycast/proxy list */
	TAILQ_HEAD(, presp) presps;	/* pending response list */
	struct in6_addr addr;		/* address */
	struct prefix *prefix;		/* back pointer (prefix) */
	struct interface *ifp;		/* back pointer (interface) */
	u_int flags;			/* flags */
	int type;			/* type */
	struct timeout *tp;		/* when send gratuitous Nd Adv */
	int ntosend;			/* Gratuitous Nd Adv to send */
};
#define FORALLPANY(ap, pp)	\
	for (ap = pp->anycasts.tqh_first; ap; ap = ap->any_list.tqe_next)
#define FORALLOANY(ap, ifp)	\
	for (ap = ifp->oanycasts.tqh_first; ap; ap = ap->any_list.tqe_next)

struct presp {
	TAILQ_ENTRY(presp) prs_list;	/* pending response list */
	struct anycast *which;		/* anycast pointer */
	struct timeout *tp;		/* when send */
	struct in6_addr peer;		/* peer address */
};

struct hagent {
	TAILQ_ENTRY(hagent) ha_list;	/* list of home agents */
	struct in6_addr addr;		/* address */
	struct prefix *prefix;		/* back pointer */
	struct timeout *tp;		/* timeout (for cleanup) */
	int16_t hapref;			/* home agent preference */
	u_int16_t halife;		/* home agent lifetime */
};
#define FORALLHA(hap, pp)	\
	for (hap = pp->hagents.tqh_first; hap; hap = hap->ha_list.tqe_next)

struct interface {
	TAILQ_ENTRY(interface) if_list;	/* list of interfaces */
	TAILQ_HEAD(, prefix) fprefixes;	/* prefixes from gateway file */
	TAILQ_HEAD(, prefix) prefixes;	/* prefixes */
	TAILQ_HEAD(, anycast) oanycasts;/* orphan anycast/proxy list */
	struct in6_addr llocal;		/* link-local address */
	struct option opt;		/* per if options */
	struct difopt oldopt;		/* change need action */
	u_char adphy[MAXTOKEN];		/* MAC address */
	signed char usefpref;		/* !=0 list of prefix is fprefixes */
#define STATENOLL	0
#define STATELESS	1
	char state;			/* state, LL present ? */
	u_char ltoken;			/* token length */
	u_char lmac;			/* MAC length */
	char hasphy;			/* is there a phy addr ? */
	u_long mtu;			/* IPv6 MTU */
	u_long metric;			/* interface metric */
	short flags;			/* copy of kernel flags */
	u_int index;			/* kernel index */
	char name[IFNAMSIZ];		/* name */
	int ntosend;			/* Rt to send */
	struct timeout *tp;		/* when next */
	struct timeval last;		/* when last */
	struct timeout *rtp;		/* when route */
	struct TFilter *filtin;		/* Rip Filter in */
	struct TFilter *filtout;	/* Rip Filter out */
};
#define FFIRST(ifp)	(ifp->fprefixes.tqh_first)
#define UFIRST(ifp)	(ifp->usefpref? FFIRST(ifp) : FIRST(ifp))

 /* Rip info */
#include <protocols/routed.h>	/* most ipv4 const are OK */

#ifndef TIMER_RATE
#define TIMER_RATE	SUPPLY_INTERVAL
#endif
#if (GARBAGE_TIME == EXPIRE_TIME+60)
#undef GARBAGE_TIME
#define	GARBAGE_TIME	(EXPIRE_TIME+120)
#endif
#ifndef RIP_VERSION_1		/* conflict NetBSD/FreeBSD */
#define RIP_VERSION_1	1
#endif
#define	RIPNG_PORT	521
#define	MADDR6_RIP	9
struct netinfov6 {
	struct in6_addr prefix;
	short routetag;
	u_char plen;
	u_char metric;
};
#define rip_netsv6(x)	(*((struct netinfov6*)&((x).ripun)))

/* a chained list for now. A hashed structure should be better !! */
struct Route {
	TAILQ_ENTRY(Route) route_list;	/* route list */
	struct interface *ifp;		/* up pointer */
	struct in6_addr addr;		/* route itself */
	struct in6_addr mask;		/* mask */
	struct in6_addr gateway;	/* gateway */
	u_int refresh;			/* timeout */
	u_char plen;			/* prefix length */
	u_char flags;			/* flags */
#define	RT_CHANGED	0x01
#define	RT_INSTALLED	0x02
#define	RT_SEEN		0x20
#define	RT_MANUAL	0x04
#define	RT_FILE		0x08
#define	RT_DONTINST	0x10
#define	RT_LOCAL	0x40
#define RT_STAT (RT_MANUAL|RT_FILE|RT_DONTINST)
	u_char metric;			/* metric */
	u_char trunc;			/* metric */
};
/* ordered in decreasing prefix len order (a/x after a/y if x<y) */
TAILQ_HEAD(, Route) routes;
TAILQ_HEAD(, Route) Freeroutes;
#define FORALLRP(rp)	\
	for (rp = routes.tqh_first; rp; rp = rp->route_list.tqe_next)

struct Filter {
	struct in6_addr addr;		/* route itself */
	struct in6_addr mask;		/* mask */
	struct in6_addr gateway;	/* gateway */
	u_char plen;			/* prefix length */
	u_char lg;			/* length of prefix to test */
	u_char typlg;
#define LG_EQ	1
#define LG_GE	2
#define LG_LE	3
	u_char action;
#define FACCEPT	255
#define FREJECT	254
/* #define FTRUNC 0-128 */
};
#define FILTALLOC	10
#define NO_GATEWAY	0xff
struct TFilter {
	int nb;
	struct Filter filter[FILTALLOC];
};
struct TFilter *filtin;			/* global RIP Filter in */
struct TFilter *filtout;		/* global RIP Filter out */

#define GATEWAY6	"/etc/gateway6"

TAILQ_HEAD(, interface) interfaces;
TAILQ_HEAD(, timeout) timeouts;

extern struct option defopt, defoptini;
extern u_int MaxHop;

extern int s, r, u, m;			/* Socket file descriptors */
extern int ripchanges;
extern struct in6_addr localH;

extern u_short lo_index;
extern int optg;
extern int optrip;
extern int optripinst;
extern int rip_in;
extern short ripport;
extern char *bigbuffer;
extern int bigbufferlen;
extern char *gateway6;

extern int verbose;
extern int debug;
extern int timestamps;
extern int mobile;

void start_rtadv(struct interface *);
void stop_rtadv(struct interface *);
void send_grat_apadv(struct anycast *);
void recv_icmp(void);
void cancel_anycast(struct presp *);
void recv_route(void);
void init(void), set_verbose(int), dump(int), finish(int);
void update_prefix(struct prefix *);
void update_advprefix(struct prefix *);
void add_anycast_prefix(struct interface *,
		struct prefix *, struct in6_addr *);
void add_anycast(struct anycast *);
void del_anycast_prefix(struct interface *,
		struct prefix *, struct in6_addr *);
void del_anycast(struct anycast *);
void prefix_inval(struct prefix *);
void sort_prefix(struct interface *);
void dump_prefix(struct prefix *);
void dump_anycast(struct anycast *);
void dump_presp(struct presp *);
void dump_hagent(struct hagent *);
void dump_interface(struct interface *);
int is_router(void);
void decl_route_default(int);
void add_route_prefix(struct prefix *, struct interface *, int);
void del_route_prefix(struct prefix *, struct interface *);
void del_route_ifp(struct interface *);
void recv_rippacket(void);
void dump_route(struct Route *);
void init_rip(void);
void start_rip(struct interface *);
void stop_rip(struct interface *);
void update_rip(struct interface *);
void new_gateway(void);
void clear_rip_changes(void);
void if_up(struct interface *);
void if_down(struct interface *);
void if_rtadv(struct interface *, int);
void if_apadv(struct interface *, int);
void anycast_multi(struct anycast *, int);
void if_ripin(struct interface *, int);
void if_ripout(struct interface *, int);
void read_gateway(int);
void recv_one_route(struct in6_addr *, u_int, u_int, u_int,
	struct in6_addr *, struct interface *);
void add_kern_route(struct Route *);
void del_kern_route(struct Route *);
void change_kern_route(struct Route *);
void localroute(struct interface *, struct in6_addr *, int);
void init_err(void);
void dump_err(void);
int convvalpref(char *, char *, union datelife *);

void disable_binding(int);
void disable_homeaddr(int);
void lost_mobile(struct sockaddr_in6 *);
void recv_destopt(void);
void add_anycast_agent(struct prefix *);
void del_anycast_agent(struct prefix *);
void update_hagent(struct prefix *, struct nd_opt_agti *, struct in6_addr *);
void mobility_init(void);
void mobile_booted(void);
void mobility_cleanup(void);
void mobile_zappref(struct prefix *);
void mobile_chgpref(struct interface *);
void dump_mobiles(void);
