#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <strings.h>
#include <ctype.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/sysctl.h>
#include <sys/uio.h>
#include <sys/resource.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <net/route.h>

#include <netinet/in_systm.h>
#define INET6
#include <netinet/in.h>
#include <netinet/ip6.h>
#include <netinet/ip_var.h>
#include <netinet/ip6_var.h>
#include <netinet/ip6_opts.h>

#include <netdb.h>
#include <arpa/inet.h>

#include <string.h>
#include <stdlib.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <protocols/routed.h>	/* most ipv4 const are OK */
#if (GARBAGE_TIME == EXPIRE_TIME+60)
#undef GARBAGE_TIME
#define	GARBAGE_TIME	(EXPIRE_TIME+120)
#endif
#ifndef RIP_VERSION_1		/* conflict NetBSD/FreeBSD */
#define RIP_VERSION_1	1
#endif
#define	RIPNG_PORT	521
#define	MADDR6_RIP	9
struct netinfov6 {
	struct in6_addr prefix;
	short routetag;
	u_char plen;
	u_char metric;
};
#define rip_netsv6(x)	(*((struct netinfov6*)&((x).ripun)))

#define PACKETSIZE 8192

int ripport = RIPNG_PORT;
char	bigbuffer[PACKETSIZE];
int	nflag;
static int	bufspace = 127*1024;	/* max. input buffer size to request */
char	vb[INET6_ADDRSTRLEN];
extern const struct in6_addr in6addr_any;

int	main(int, char *[]);
void	rip_input(struct sockaddr_in6 *, struct rip *, int);

int
main(argc, argv)
	int argc;
	char **argv;
{
	int c, s, cc;
	struct timeval wt;
	int on;
	struct sockaddr_in6 rips;
	struct rip *ripngp;
	struct netinfov6 *rte = &rip_netsv6(*ripngp);
	char *p;
	fd_set fd;
	int ripslen = sizeof rips;
	struct hostent *hp;

	while ((c = getopt(argc, argv, "nu:")) != EOF)
		switch (c) {
		case 'n':
			nflag++;
			break;
		case 'u':
			if (!(ripport = atoi(optarg)))
				goto usage;
			break;
		default:
			goto usage;
		}
	argv += optind;
	argc -= optind;

	if (!argc) {
usage:		printf("usage: query [-n] [-u port] host values...\n");
		exit(1);
	}


	if ((s = socket(AF_INET6, SOCK_DGRAM, 0)) < 0) {
		perror("socket (RIP)");
		exit(2);
	}
#ifdef SO_RCVBUF
	for (on = bufspace; ; on -= 1024) {
		if (setsockopt(s, SOL_SOCKET, SO_RCVBUF,
		    &on, sizeof (on)) == 0)
			break;
		if (on <= 8*1024) {
			perror("setsockopt SO_RCVBUF (RIP)");
			break;
		}
	}
#endif
	bzero(&rips, sizeof rips);
	rips.sin6_family = AF_INET6;
#ifdef SIN6_LEN
	rips.sin6_len = sizeof(rips);
#endif
	rips.sin6_port = htons(ripport);
	rips.sin6_addr = in6addr_any;

	hp = getipnodebyname(*argv, AF_INET6, 0, &cc);
	if (hp == NULL)
		errx(1, "query %s: %s", *argv, hstrerror(cc));
	memcpy(&rips.sin6_addr, hp->h_addr, hp->h_length);
	freehostent(hp);
	argv++; argc--;

	ripngp = (struct rip *) bigbuffer;
	rte = &rip_netsv6(*ripngp);

	bzero((char *)bigbuffer, sizeof bigbuffer);
	ripngp->rip_vers = RIP_VERSION_1;
	ripngp->rip_cmd = RIPCMD_REQUEST;

	rte->metric = HOPCNT_INFINITY;
	if (argc) rte--;

	while (argc > 0) {
		rte++;
		if ((char *)(rte + 1) > ((char *)bigbuffer) + sizeof bigbuffer)
			errx(1, "too many args");
		bzero((char *)rte, sizeof *rte);
		p = index(*argv, '/');
		if (!p)
			goto errp;
		*p++ = 0;
		rte->plen = atoi(p);
		hp = getipnodebyname(*argv, AF_INET6, 0, &cc);
		p[-1] = '/';
		if (!isdigit(*p) || rte->plen > 128 || hp == NULL) {
    errp:
			errx(1, "incorrect/unknown prefix %s", *argv);
		}
		memcpy(&rte->prefix, hp->h_addr, hp->h_length);
		freehostent(hp);
		argv++; argc--;
	}

	sendto(s, (char *)ripngp,
	       (char *)(rte+1) - (char *)ripngp, 0,
	       (struct sockaddr *)&rips, sizeof(rips));

	for (;;) {
		FD_ZERO(&fd);
		FD_SET(s, &fd);
		wt.tv_sec = 1;
		wt.tv_usec = 0;
		errno = 0;
		cc = select(s + 1, &fd, NULL, NULL, &wt);
		if (cc < 0) {
			if (errno == EINTR)
				continue;
			perror("select");
			(void) close(s);
			exit(1);
		}
		if (cc == 0)
			exit(0);
		cc = recvfrom(s, bigbuffer, sizeof (bigbuffer), 0,
			      (struct sockaddr *)&rips, &ripslen);
		if (cc <= 0) {
			perror("recvfrom");
			exit(1);
		}
		rip_input(&rips, (struct rip *) bigbuffer, cc);
	}
}

void
rip_input(s, rip, cc)
	struct sockaddr_in6 *s;
	struct rip *rip;
	int cc;
{
	struct netinfov6 *rte = &rip_netsv6(*rip);

	if (s->sin6_family != AF_INET6) {
		fprintf(stderr, "packet form faily %d ??\n", s->sin6_family);
		return;
	}
	printf("From %s.%d:",
	       inet_ntop(AF_INET6, &s->sin6_addr, vb, sizeof(vb)),
	       s->sin6_port);
	cc -= sizeof *rip - sizeof *rte;
	if (cc < 0) {
		printf(" truncated (%d chars)\n", cc);
		return;
	}
	if (rip->rip_cmd != RIPCMD_RESPONSE ||
	    rip->rip_vers != RIP_VERSION_1) {
		printf(" cmd %d version %d\n", rip->rip_cmd, rip->rip_vers);
		return;
	}
	for (; cc>=sizeof *rte; cc -= sizeof *rte, rte++) {
		if (rte->plen == 0 && rte->metric == 255) {
			printf("\n\tNH=%s", 
			       inet_ntop(AF_INET6, &rte->prefix,
					 vb, sizeof(vb)));
			continue;
		}
		printf("\n\t%s/%d[%d",
		       inet_ntop(AF_INET6, &rte->prefix, vb, sizeof(vb)),
		       rte->plen, rte->metric);
		if (rte->routetag)
			printf(",tag=%d",rte->routetag);
		printf("]");
	}
	printf("\n");
}
