#include "ndpd.h"
#include <netinet/udp.h>

short ripport = RIPNG_PORT;
int ripchanges;
int rip_in;	/* in socket */

static int	bufspace = 127*1024;	/* max. input buffer size to request */

void route_timeout(struct timeout *);
void recv_rip_request(struct interface *, struct sockaddr_in6 *,
		      struct rip *, char *);
void recv_rip_response(struct interface *, struct sockaddr_in6 *,
		       struct rip *, char *);
void send_route_ifp(struct interface *, int, int, struct sockaddr_in6 *);
void send_route_triggered(void *);
void send_route_periodic(struct interface *);
void send_prompt_rip(struct interface *);
static int filter(struct TFilter *, struct in6_addr *, int *,
		  struct in6_addr *);
/* open input rip socket */
void
init_rip()
{
	int on;
	struct sockaddr_in6 ripsin;
	struct timeout * tp;

	if ((rip_in = socket(AF_INET6, SOCK_DGRAM, 0)) < 0) {
		logperror("socket (RIP)");
		exit(2);
	}
	if (!is_router())
		logerr("Warning: RIPng on a non-router node");
	bzero(&ripsin, sizeof ripsin);
	ripsin.sin6_family = AF_INET6;
#ifdef SIN6_LEN
	ripsin.sin6_len = sizeof(ripsin);
#endif
	ripsin.sin6_port = htons(ripport);
	on = 1;
	if (setsockopt(rip_in, IPPROTO_IPV6, IPV6_RECVPKTINFO,
		       &on, sizeof(on)) < 0) {
		logperror("setsockopt (IP_RECVPKTINFO)");
		exit(3);
	}

	if (setsockopt(rip_in, IPPROTO_IPV6, IPV6_RECVHOPLIMIT,
		       &on, sizeof(on)) < 0) {
		logperror("setsockopt (IP_RECVHOPS)");
		exit(3);
	}

	if (setsockopt(rip_in, SOL_SOCKET, SO_REUSEADDR,
		       &on, sizeof(on)) < 0) {
		logperror("setsockopt(RIP)");
		exit(3);
	}

#ifdef SO_RCVBUF
	for (on = bufspace; ; on -= 1024) {
		if (setsockopt(rip_in, SOL_SOCKET, SO_RCVBUF,
			       &on, sizeof (on)) == 0)
			break;
		if (on <= 8*1024) {
			logperror("setsockopt SO_RCVBUF (RIP)");
			break;
		}
	}
	/* should we do it also on the s socket ???  */
#endif

	if (bind(rip_in, (struct sockaddr *)&ripsin, sizeof(ripsin))) {
		logperror("bind (RIP)");
		exit(3);
	}

	(void)gettimeofday(&now, NULL);
	tp = new_timeout(TIMER_RATE, 0, (func_to) route_timeout, 0);
	if (tp == NULL)
		abort();
	tp->refcnt = 1;	/* keep on expire */
	tp->arg = tp;
	send_route_triggered((void *)(-1)); /* suppress intial triggered send */
}

/* for -g option */
struct in6_addr localH = IN6ADDR_ANY_INIT;

void
decl_route_default(f)
	int f;
{
	if (!optrip)
		return;
	recv_one_route(&localH, 0, 1, f|RT_MANUAL|RT_DONTINST, &localH, NULL);
}

/* add interface route (if new prefix detected) */
void
add_route_prefix(pp, ifp, f)
	struct prefix *pp;
	struct interface *ifp;
	int f;
{
	if (!optrip)
		return;
	recv_one_route(&pp->pref, pp->plen, ifp->metric,
		       f|RT_MANUAL|RT_DONTINST, &ifp->llocal, ifp);
}

/* delete interface route (if prefix deleted) */
void
del_route_prefix(pp, ifp)
	struct prefix *pp;
	struct interface *ifp;
{
	recv_one_route(&pp->pref, pp->plen, HOPCNT_INFINITY,
		       RT_MANUAL|RT_DONTINST, &ifp->llocal, ifp);
}

/* input one RIPng packet */
void recv_rippacket()
{
	int cc;
	struct rip *ripngp;
	int c_cmsg[100];
	struct cmsghdr *cmsgp;
	struct interface *ifp;
	struct msghdr msg;
	struct iovec iov;
	struct in6_pktinfo pi;
	int hlim = -2, haspi = 0;
	struct sockaddr_in6 ripsin_in;

	ripngp = (struct rip *)bigbuffer;
	bzero((caddr_t)&msg, sizeof(struct msghdr));
	bzero((caddr_t)&iov, sizeof(struct iovec));
	msg.msg_iov = &iov;
	msg.msg_iov->iov_base = (char *)ripngp;
	msg.msg_iov->iov_len = bigbufferlen;
	msg.msg_iovlen = 1;
	msg.msg_control = (caddr_t)c_cmsg;
	msg.msg_controllen = sizeof(c_cmsg);
	msg.msg_name =(caddr_t)&ripsin_in;
	msg.msg_namelen = sizeof ripsin_in;

	if ((cc = recvmsg(rip_in, &msg, 0)) < 0) {
		if (verbose)
			logperror("read (RIPng)");
		return;
	}
	if (cc == 0) {
		if (verbose)
			logerr("read RIPng return 0");
		return;
	}
	if (msg.msg_flags & (MSG_TRUNC|MSG_CTRUNC)) {
		if (verbose)
			logerr("read RIPng truncated");
		return;
	}
	cmsgp = msg.msg_controllen? CMSG_FIRSTHDR(&msg) : 0;
	for (; cmsgp; cmsgp = CMSG_NXTHDR(&msg, cmsgp)) {
		if ((cmsgp->cmsg_level == IPPROTO_IPV6) &&
		    (cmsgp->cmsg_type == IPV6_PKTINFO) &&
		    (cmsgp->cmsg_len == sizeof(struct cmsghdr) + sizeof(pi))) {
			bcopy(CMSG_DATA(cmsgp), &pi, sizeof(pi));
			haspi = 1;
		}
		if (((cmsgp->cmsg_level == IPPROTO_IPV6)||
		     (cmsgp->cmsg_level == IPPROTO_IP)) &&      /* ???? */
		     (cmsgp->cmsg_type == IPV6_UNICAST_HOPS) &&
		     (cmsgp->cmsg_len ==
			     sizeof(struct cmsghdr) + sizeof(int))) {
			bcopy(CMSG_DATA(cmsgp), &hlim, sizeof(int));
		}
	}
	if (!haspi) {
		if (verbose)
			logerr("read RIPng doesn't give packet info");
		return;
	}
	if (hlim == -2) {
		if (verbose)
			logerr("read RIPng doesn't give hops limit");
		hlim = ND_DEFAULT_HOPLIM;
	}
	FORALLIFP(ifp)
		if (ifp->index == pi.ipi6_ifindex)
		break;
	if (ifp == NULL || !IS_UP(ifp))
		return;
	if (ifp->opt.no_ripin)
		return;

	if (verbose)
		logtrace("RIP receive %d bytes type %d from %s/%s hlim %d",
			 cc, ripngp->rip_cmd,
			 ifp->name, inet6pr(&ripsin_in.sin6_addr),
			 hlim);

	/* check version */
	if (ripngp->rip_vers != RIP_VERSION_1 ||
	    (char *)&rip_netsv6(*ripngp) > (char *)ripngp+cc ) {
    err1:
		logerr("wrong rip packet received vers %d cmd %d len %d from %s/%s",
		       ripngp->rip_vers, ripngp->rip_cmd, cc,
		       ifp->name, inet6pr(&ripsin_in.sin6_addr));
		return;
	}
	if ((hlim != ND_DEFAULT_HOPLIM) &&
	    (ripngp->rip_cmd != RIPCMD_REQUEST) &&
	    IS_MULTIADDR6(pi.ipi6_addr)) {
		logerr("rip packet with wrong hop limit %d from %s/%s",
		       hlim,
		       ifp->name, inet6pr(&ripsin_in.sin6_addr));
	}

	switch (ripngp->rip_cmd) {
	    default:
		goto err1;

	    case RIPCMD_REQUEST:
		recv_rip_request(ifp, &ripsin_in, ripngp, (char *)ripngp + cc);
		break;

	    case RIPCMD_RESPONSE:
		/* check:
		 *  from LL, not llocal, from ripport, multicast => ttl == 255
		 */
		if (!IS_LINKLADDR6(ripsin_in.sin6_addr) ||
		    (htons(ripsin_in.sin6_port) != ripport) ||
		    SAME_ADDR6(ripsin_in.sin6_addr, ifp->llocal)) {
			if (verbose)
				logerr("wrong rip response from %d/%s port %d",
				       ifp->name,
				       inet6pr(&ripsin_in.sin6_addr),
				       htons(ripsin_in.sin6_port));
			return;
		}
		if (!IS_LINKLADDR6(pi.ipi6_addr) &&
		    !IS_MULTIADDR6(pi.ipi6_addr)) {
			if (verbose)
				logerr("wrong rip response for %s/%s",
				       ifp->name, inet6pr(&pi.ipi6_addr));
			return;
		}
		recv_rip_response(ifp, &ripsin_in,
				  ripngp, (char *)ripngp + cc);
	}
}

/* process input RIPng request packet */
void
recv_rip_request(ifp, src, ripngp, end)
	struct interface *ifp;
	struct sockaddr_in6 *src;
	struct rip *ripngp;
	char *end;
{
	struct netinfov6 *rte = &rip_netsv6(*ripngp);
	int rip_out, on;
	struct sockaddr_in6 rips_out;
	struct Route *rp;

	/* create & bind socket - must create to do bind */
	if ((rip_out = socket(AF_INET6, SOCK_DGRAM, 0)) < 0 ) {
		logperror("socket: outgoing RIP socket");
		return;
	}
	bzero(&rips_out, sizeof rips_out);
	rips_out.sin6_family = AF_INET6;
#ifdef SIN6_LEN
	rips_out.sin6_len = sizeof(rips_out);
#endif
	rips_out.sin6_port = htons(ripport);
	on = 1;
	if (setsockopt(rip_out, SOL_SOCKET, SO_REUSEADDR,
		       &on, sizeof(int)) < 0){
		logperror("setsockopt: SOREUSEADDR outgoing RIP socket");
		goto done;
	}
	/* not necessary but the default scope is global yet them... */
	on = ND_DEFAULT_HOPLIM;
	if (setsockopt(rip_out, IPPROTO_IPV6, IPV6_UNICAST_HOPS,
		       &on, sizeof(on)) < 0) {
		logperror("setsockopt: IPV6_UNICAST_HOPS outgoing RIP socket");
		goto done;
	}
	if (src->sin6_port == rips_out.sin6_port) {
		rips_out.sin6_addr = ifp->llocal;
	/* must set out interf, but kernel cannot do it : BUG ?? */
		rips_out.sin6_scope_id = ifp->index;
	}
	if (bind(rip_out, (struct sockaddr *)&rips_out, sizeof(rips_out))) {
		logperror("bind: outgoing RIP socket");
		goto done;
	}
	rips_out.sin6_port = src->sin6_port;
	rips_out.sin6_addr = src->sin6_addr;

	if (end == (char *)(&rte[1]) && rte->metric == HOPCNT_INFINITY &&
	    rte->plen == 0 && IS_ANYADDR6(rte->prefix)) {
		send_route_ifp(ifp, 0, rip_out, &rips_out);
	} else {
		ripngp->rip_cmd = RIPCMD_RESPONSE;
		for (; (char *)(&rte[1]) <= end; rte++) {
			FORALLRP(rp)
				if (rp->plen == rte->plen &&
				    SAME_ADDR6(rp->addr, rte->prefix))
					break;
			if (rp)
				rte->metric = rp->metric;
			else
				rte->metric = HOPCNT_INFINITY;
		}
		if (rte > &rip_netsv6(*ripngp)) {
			sendto(rip_out, (char *)ripngp,
			       (char *)rte - (char *)ripngp, 0,
			       (struct sockaddr *)&rips_out,
			       sizeof(rips_out));
			if (verbose)
			       logtrace("RIPng: %d entries for request on %s",
					rte - &rip_netsv6(*ripngp), ifp->name);
		}
	}
    done:
	close(rip_out);
}

/* send a prompt: a RIPng request packet */
void
send_prompt_rip(ifp)
	struct interface *ifp;
{
	struct rip *ripngp = (struct rip *)bigbuffer;
	struct netinfov6 *rte = &rip_netsv6(*ripngp);
	int rip_out, on;
	struct sockaddr_in6 rips_out;

	bzero((char *)ripngp, sizeof *ripngp);
	ripngp->rip_vers = RIP_VERSION_1;
	ripngp->rip_cmd = RIPCMD_REQUEST;
	rte->metric = HOPCNT_INFINITY;

	/* create & bind socket - must create to do bind */
	if ((rip_out = socket(AF_INET6, SOCK_DGRAM, 0)) < 0 ) {
		logperror("socket: outgoing RIP socket");
		return;
	}
	bzero(&rips_out, sizeof rips_out);
	rips_out.sin6_family = AF_INET6;
#ifdef SIN6_LEN
	rips_out.sin6_len = sizeof(rips_out);
#endif
	rips_out.sin6_port = htons(ripport);
	rips_out.sin6_addr = ifp->llocal;
	rips_out.sin6_scope_id = ifp->index;
	on = ND_DEFAULT_HOPLIM;
	if (setsockopt(rip_out, IPPROTO_IPV6, IPV6_MULTICAST_HOPS,
		       &on, sizeof(on)) < 0) {
		logperror("setsockopt: IPV6_MULTICAST_HOPS outgoing RIP socket");
		goto done;
	}
	on = 0;
	if (setsockopt(rip_out, IPPROTO_IPV6, IPV6_MULTICAST_LOOP,
		       &on, sizeof(on)) < 0) {
		logperror("setsockopt: IPV6_MULTICAST_LOOP outgoing RIP socket");
		goto done;
	}
	on = 1;
	if (setsockopt(rip_out, SOL_SOCKET, SO_REUSEADDR,
		       &on, sizeof(int)) < 0){
		logperror("setsockopt: SOREUSEADDR outgoing RIP socket");
		goto done;
	}
	on = ifp->index;
	if (setsockopt(rip_out, IPPROTO_IPV6, IPV6_MULTICAST_IF,
		       (char *)&on, sizeof(on)) < 0) {
		logperror("setsockopt: IPV6_MULTICAST_IF outgoing RIP socket");
		goto done;
	}
	if (bind(rip_out,(struct sockaddr*)&rips_out,sizeof(rips_out))){
		logperror("bind:outgoing RIP socket");
		goto done;
	}
	bzero(&rips_out.sin6_addr, sizeof(struct in6_addr));
	rips_out.sin6_addr.s6_addr[0]  = 0xff;
	rips_out.sin6_addr.s6_addr[1]  = MADDR6_SCP_LINK;
	rips_out.sin6_addr.s6_addr[15] = MADDR6_RIP;

	sendto(rip_out, (char *)ripngp,
	       (char *)(rte + 1) - (char *)ripngp, 0,
	       (struct sockaddr *)&rips_out,
	       sizeof(rips_out));
    done:
	close(rip_out);
}

/* process input RIPng response packet */
void
recv_rip_response(ifp, src, ripngp, end)
	struct interface *ifp;
	struct sockaddr_in6 *src;
	struct rip *ripngp;
	char *end;
{
	struct in6_addr *gateway;
	struct netinfov6 *rte = &rip_netsv6(*ripngp);

	gateway = &src->sin6_addr;
	for (rte = &rip_netsv6(*ripngp); (char *)rte < end; rte++) {
		if (rte->metric == 0xff) {
			if (IS_LINKLADDR6(rte->prefix)) {
				gateway = &rte->prefix;
				continue;
			}
			gateway = &src->sin6_addr;
			if (!IS_ANYADDR6(rte->prefix))
			    logerr("wrong proxy %s from %s/%s",
				inet6pr(&rte->prefix),
				ifp->name, inet6pr(gateway));
			continue;
		}
		recv_one_route(&rte->prefix, rte->plen,
			       rte->metric, 0, gateway, ifp);
	}
	if (ripchanges)
		send_route_triggered(0);
}

/* process one entry of input RIPng response */
void
recv_one_route(dest, plen, metric, flags, gateway, ifp)
	struct in6_addr *dest, *gateway;
	u_int plen, metric, flags;
	struct interface *ifp;
{
	struct Route *rp, *rp2;
	struct in6_addr mask;
	struct TFilter *pfilt = 0;

	if (debug)
		logdebug("recv_one_route(%s/%d[%d], 0x%x, %s, %s)",
			inet6pr(dest), plen, metric,
			flags,
			inet6pr(gateway),
			ifp ? ifp->name : "none");
	if (plen > 128 || IS_LINKLADDR6(*dest) || IS_MULTIADDR6(*dest) ||
	    (!(flags & RT_MANUAL) &&
	     (metric == 0 || metric > HOPCNT_INFINITY))) {
		logerr("wrong prefix %s/%d[%d] from %s/%s",
		       inet6pr(dest), plen, metric,
		       ifp ? ifp->name : "CONF",
		       inet6pr(gateway));
		return;
	}
	if (ifp)
		pfilt = ifp->filtin;
	if (!pfilt)
		pfilt = filtin;
	if (pfilt && filter(pfilt, dest, &plen, gateway))
		return;

	/* clean up received destination */
	mk_mask(&mask, plen);
	apply_mask(dest, &mask);

	/* add cost to metric */
	metric++;
	if (metric> HOPCNT_INFINITY)
		metric = HOPCNT_INFINITY;

	(void)gettimeofday(&now, NULL);
	/*
	 * Semi sorted list - decreasing prefix : a/x after a/y if x<y
	 */
	rp = routes.tqh_first;
	while (rp && rp->plen > plen)
		rp = rp->route_list.tqe_next;
	while (rp && plen == rp->plen && !SAME_ADDR6(*dest, rp->addr))
		rp = rp->route_list.tqe_next;
	if (rp && rp->plen == plen) {		/* found */
		if (flags & RT_MANUAL) {
			if (((flags & RT_FILE) == 0) && (rp->flags & RT_FILE))
				return;
			rp->flags |= flags & RT_SEEN;
			if (metric == HOPCNT_INFINITY) {
				if (rp->metric != metric) {
					del_kern_route(rp);
					rp->refresh = now.tv_sec;
					rp->metric = metric;
					ripchanges = (rp->flags |= RT_CHANGED);
				}
				return;
			}
			if ((flags & RT_DONTINST) &&
			    (rp->flags & RT_INSTALLED))
				del_kern_route(rp);
			if ((ifp == rp->ifp) &&
			    (metric == rp->metric) &&
			    ((flags & RT_STAT) == (rp->flags & RT_STAT)) &&
			    SAME_ADDR6(*gateway, rp->gateway)) {
				return;
			}
			if ((ifp != rp->ifp) || (metric != rp->metric))
				ripchanges = (rp->flags |= RT_CHANGED);
			rp->ifp = ifp;
			rp->gateway = *gateway;
			rp->metric = metric;
			rp->flags &= ~RT_STAT;
			rp->flags |= flags & RT_STAT;
			if (rp->flags & RT_INSTALLED)
				change_kern_route(rp);
			else
				add_kern_route(rp);
			return;
		}
		if (rp->flags & RT_MANUAL)
			return;
		/* we've found this route in the routing table */
		if (ifp == rp->ifp &&
		    SAME_ADDR6(*gateway, rp->gateway)) {
			/* same interface and same gateway */
			if (debug)
				logdebug("same gw, metric %d->%d",
					 rp->metric, metric);
			if (rp->metric != metric) {
				if (metric == HOPCNT_INFINITY)
					del_kern_route(rp);
				else if (rp->metric == HOPCNT_INFINITY)
					add_kern_route(rp);
				ripchanges = rp->flags |= RT_CHANGED;
				rp->refresh = now.tv_sec;
				rp->metric = metric;
			} else if (metric < HOPCNT_INFINITY)
				rp->refresh = now.tv_sec;
		} else if (metric < rp->metric ||
			   ((metric == rp->metric) &&
			    (metric < HOPCNT_INFINITY) &&
			    (now.tv_sec - rp->refresh > EXPIRE_TIME/2))) {
			/* new route to a known destination */
			if (debug)
				logdebug("gw %s->, metric %d->%d",
					 inet6pr(&rp->gateway),
					 rp->metric, metric);
			if ((ifp != rp->ifp) || (metric != rp->metric))
				ripchanges = (rp->flags |= RT_CHANGED);
			rp->ifp = ifp;
			rp->gateway = *gateway;
			rp->metric = metric;
			if (rp->metric == HOPCNT_INFINITY)
				add_kern_route(rp);
			else
				change_kern_route(rp);
			rp->refresh = now.tv_sec;
		} else
			return;
	} else if (metric < HOPCNT_INFINITY) {
		if (Calloc(rp2, Route) == NULL) {
			if (verbose)
				logerr("out of memory");
			return;
		}
		if (debug)
			logdebug("adding route");
		if (rp == NULL)
			TAILQ_INSERT_TAIL(&routes, rp2, route_list);
		else
			TAILQ_INSERT_BEFORE(rp, rp2, route_list);
		rp2->addr    = *dest;
		rp2->mask    = mask;
		rp2->plen    = plen;
		rp2->gateway = *gateway;
		rp2->ifp     = ifp;
		rp2->metric  = metric;
		ripchanges = rp2->flags  = flags | RT_CHANGED;
		rp2->refresh = now.tv_sec;
		add_kern_route(rp2);
	} else
		return;
}

/* timout process on rip tables */
void
route_timeout(tp)
	struct timeout *tp;
{
	register struct Route *rp, *rpn;

	(void)gettimeofday(&now, NULL);

	tp->tv.tv_sec += TIMER_RATE;
	add_timeout(tp);
	rpn = routes.tqh_first;
	while(rpn) {
		rp = rpn;
		rpn = rp->route_list.tqe_next;
		if (rp->metric == HOPCNT_INFINITY) {
			if (now.tv_sec - rp->refresh >
				(GARBAGE_TIME - EXPIRE_TIME)) {
				if (debug)
					logdebug("deleting route %s/%d",
					         inet6pr(&rp->addr),
						 rp->plen);
				TAILQ_REMOVE(&routes, rp, route_list);
				free(rp);
			}
			continue;
		}
		if (rp->flags & RT_MANUAL)
			continue;
		if (now.tv_sec - rp->refresh > EXPIRE_TIME) {
			if (debug)
				logdebug("expiring route %s",
					 inet6pr(&rp->gateway));
			del_kern_route(rp);
			rp->metric = HOPCNT_INFINITY;
			rp->refresh = now.tv_sec;
			ripchanges = rp->flags |= RT_CHANGED;
		}
	}
	if (ripchanges)
		send_route_triggered(0);
}

/* send a rip response for ifp */
void
send_route_ifp(ifp, trig, rip_out, rips_out)
	struct interface *ifp;
	int trig, rip_out;
	struct sockaddr_in6 *rips_out;
{
	struct rip *ripngp;
	struct netinfov6 *rte, *rteb;
	int on, metric, maxpacket;
	struct sockaddr_in6 rips_out_x;
	struct Route *rp, *rp2;
	struct TFilter *pfilt;
	struct in6_addr mask, addr;
	u_int plen;

	if (!rips_out && (optrip <= 0 || ifp->opt.no_ripout || !IS_UP(ifp)))
		return;
	if (verbose) {
		logtrace("Sending rip info for %s on %s type %d",
			 inet6pr(&ifp->llocal),
			 ifp->name, trig);
	}
	if (!rips_out) {
		rips_out = &rips_out_x;
		/* create & bind socket - must create to do bind */
		if ((rip_out = socket(AF_INET6, SOCK_DGRAM, 0)) < 0 ) {
			logperror("socket: outgoing RIP socket");
			return;
		}
		bzero(rips_out, sizeof *rips_out);
		rips_out->sin6_family = AF_INET6;
#ifdef SIN6_LEN
		rips_out->sin6_len = sizeof(rips_out);
#endif
		rips_out->sin6_port = htons(ripport);
		rips_out->sin6_addr = ifp->llocal;
		rips_out->sin6_scope_id = ifp->index;
		on = ND_DEFAULT_HOPLIM;
		if (setsockopt(rip_out, IPPROTO_IPV6, IPV6_MULTICAST_HOPS,
			       &on, sizeof(on)) < 0) {
			logperror("setsockopt: IPV6_MULTICAST_HOPS outgoing RIP socket");
			goto done;
		}
		on = 0;
		if (setsockopt(rip_out, IPPROTO_IPV6, IPV6_MULTICAST_LOOP,
			       &on, sizeof(on)) < 0) {
			logperror("setsockopt: IPV6_MULTICAST_LOOP outgoing RIP socket");
			goto done;
		}
		on = 1;
		if (setsockopt(rip_out, SOL_SOCKET, SO_REUSEADDR,
			       &on, sizeof(int)) < 0){
			logperror("setsockopt: SOREUSEADDR outgoing RIP socket");
			exit(3);
		}
		on = ifp->index;
		if (setsockopt(rip_out, IPPROTO_IPV6, IPV6_MULTICAST_IF,
			       (char *)&on, sizeof(on)) < 0) {
			logperror("setsockopt: IPV6_MULTICAST_IF outgoing RIP socket");
			goto done;
		}
		if (bind(rip_out,
			 (struct sockaddr*)rips_out,
			 sizeof(*rips_out))) {
			logperror("bind:outgoing RIP socket");
			goto done;
		}
		bzero(&rips_out->sin6_addr, sizeof(struct in6_addr));
		rips_out->sin6_addr.s6_addr[0]  = 0xff;
		rips_out->sin6_addr.s6_addr[1]  = MADDR6_SCP_LINK;
		rips_out->sin6_addr.s6_addr[15] = MADDR6_RIP;
	}

	pfilt = ifp->filtout;
	if (!pfilt)
		pfilt = filtout;

	if (bigbufferlen < ifp->mtu) {
		bigbufferlen = ifp->mtu;
		if (bigbuffer)
			free(bigbuffer);
		if (!(bigbuffer = malloc(bigbufferlen)) ) {
			logerr("out of memory");
			abort();
		}
	}
	/* packet is at most mtu - sizeof(headers) long */
	maxpacket =
		ifp->mtu - (sizeof (struct ip6_hdr) + sizeof (struct udphdr));
	rp = routes.tqh_first;
	ripngp = (struct rip *)bigbuffer;
	bzero(ripngp, sizeof(*ripngp));
	ripngp->rip_vers = RIP_VERSION_1;
	ripngp->rip_cmd = RIPCMD_RESPONSE;
	rte = rteb = &rip_netsv6(*ripngp);

	rp = routes.tqh_first;
	while (rp) {
		if (!rp->metric) 	/* only added by trunc */
			goto sendtrunc;
		metric = rp->metric;
		if (ifp->opt.optsplit && rp->ifp == ifp &&
		    !(rp->flags & RT_MANUAL)) {
			/* split horizon only for received routes */
			if (ifp->opt.optsplit < 0) /* split horizon only */
				goto reject;
			metric = HOPCNT_INFINITY; /* poisoned reverse */
		}
		if (trig && !(rp->flags & RT_CHANGED))
			goto reject;	/* unchanged-> not in triggered */
		if (!pfilt)
			goto sendit;
		plen = rp->plen;
		if (filter(pfilt, &rp->addr, &plen, NULL))
			goto reject;	/* reject */
		if (plen == rp->plen) {			/* accept */
    accept:
			if (rp->trunc && metric > rp->trunc)	/* also trunc */
				metric = rp->trunc;
			goto sendit;
		}
		/* truncate */
		if (metric != rp->metric)
			goto reject;	/* do not truncate split horizon */
		addr = rp->addr;
		mk_mask(&mask, plen);
		apply_mask(&addr, &mask);
		rp2 = rp;
		while (rp2 && rp2->plen > plen)
			rp2 = rp2->route_list.tqe_next;
		while (rp2 && plen == rp2->plen && !SAME_ADDR6(addr, rp2->addr))
			rp2 = rp2->route_list.tqe_next;
		if (rp2 && rp2->plen == plen) {		/* found */
			if (rp2->trunc == 0 || rp2->trunc > metric)
				rp2->trunc = metric;
		} else {	/* not found */
			struct Route *rp3 = Freeroutes.tqh_first;

			if (rp3)
				TAILQ_REMOVE(&Freeroutes, rp3, route_list);
			else if (Calloc(rp3, Route) == NULL) {
					if (verbose)
						logerr("out of memory");
					goto accept;
			}
			rp3->addr    = addr;
			rp3->mask    = mask;
			rp3->plen    = plen;
			rp3->trunc   = metric;
			if (rp2 == NULL)
				TAILQ_INSERT_TAIL(&routes, rp3, route_list);
			else
				TAILQ_INSERT_BEFORE(rp2, rp3, route_list);
		}
    reject:
		if (!rp->trunc)
			goto freeit;
    sendtrunc:
		metric = rp->trunc;		/* also trunc */

    sendit:
		if ((char *)(rte + 1) > (bigbuffer + maxpacket)) {
			/* send current packet and... */
			sendto(rip_out, (char *)ripngp,
			       (char *)rte - (char *)ripngp, 0,
			       (struct sockaddr *)rips_out,
			       sizeof(*rips_out));
			if (verbose)
				logtrace("RIPng: %d entries on %s",
					 rte-rteb, ifp->name);
			rte = rteb;
		}
		rte->prefix = rp->addr;
		rte->routetag = 0;
		rte->plen = rp->plen;
		rte->metric = metric;
		if (debug)
			logdebug("RIPng set entry %s/%d[%d]",
				 inet6pr(&rte->prefix),
				 rte->plen, rte->metric);
		rte++;
    freeit:
		rp->trunc = 0;
		if (!rp->metric) {	/* trunc only */
			rp2 = rp;
			rp = rp->route_list.tqe_next;
			TAILQ_REMOVE(&routes, rp2, route_list);
			TAILQ_INSERT_TAIL(&Freeroutes, rp2, route_list);
		} else
			rp = rp->route_list.tqe_next;
	}

	/* send last packet if not empty */
	if (rte != rteb) {
		sendto(rip_out, (char *)ripngp,
		       (char *)rte - (char *)ripngp, 0,
		       (struct sockaddr *)rips_out,
		       sizeof(*rips_out));
		if (verbose)
			logtrace("RIPng: %d entries on %s",
				 rte-rteb, ifp->name);
	}
    done:
	close(rip_out);
}

/* send a RIPng response every 30 s */
void
send_route_periodic(ifp)
	struct interface *ifp;
{
	int delay;

	if (ifp->rtp) {
		kill_timeout(ifp->rtp);
		ifp->rtp = NULL;
	}
	read_gateway(0);
	if (optrip <= 0 || !IS_UP(ifp) || ifp->opt.no_ripout)
		return;			/* stop the announces */
/*	if (is_router()) */
		send_route_ifp(ifp, 0, 0, 0);

	(void) gettimeofday(&now, NULL);
#define	MS_SUPPLY_INTERVAL (SUPPLY_INTERVAL*M_SECOND)
	delay = MS_SUPPLY_INTERVAL/2 + random() % (MS_SUPPLY_INTERVAL);
	ifp->rtp = new_timeout(MS_TO_SEC(delay), MS_TO_USEC(delay),
			       (func_to) send_route_periodic, ifp);
	if (!ifp->rtp)
		abort();
}

/*
 * send immediate updates, wait if too ealy foram last update
 * mode: 0 normal, -1 block for a time, other: called from delay
 */
void
send_route_triggered(mode)
	void *mode;
{
	int delay;
	static struct timeout *next_running = 0;
	struct interface *ifp;

	if (mode == (void *)(-1))
		goto block;
	if (mode == 0 && next_running)	/* too early, wait */
		return;
	next_running = 0;
	read_gateway(0);
	if (!ripchanges)
		return;
	ripchanges = 0;
	if (optrip <= 0)
		return;
/*	if (!is_router())
		return;		*/

    block:
	(void) gettimeofday(&now, NULL);
#define	MS_MAX_WAITTIME (MAX_WAITTIME*M_SECOND)
#define	MS_MIN_WAITTIME (1*M_SECOND)
	delay = MS_MIN_WAITTIME + random() % (MS_MAX_WAITTIME-MS_MIN_WAITTIME);
	next_running = new_timeout(MS_TO_SEC(delay), MS_TO_USEC(delay),
				   send_route_triggered, &next_running);
	if (!next_running)
		abort();
	if (mode == (void *)(-1))
		return;

	FORALLIFP(ifp) {
		if (!IS_UP(ifp))
			continue;
		if (ifp->rtp && AFTER(next_running->tv, ifp->rtp->tv))
			send_route_periodic(ifp); /* send next full now */
		else
			send_route_ifp(ifp, 1, 0, 0);	/* send diffs */
	}
	clear_rip_changes();
}

void
clear_rip_changes()
{
	register struct Route *rp;

	FORALLRP(rp)
		rp->flags &= ~(RT_CHANGED | RT_SEEN);
	ripchanges = 0;
}


/* (Re)start sending RIPng periodically */
void
start_rip(ifp)
	struct interface *ifp;
{
	if (!optrip || ifp->opt.no_ripout || !IS_UP(ifp))
		return;
	send_prompt_rip(ifp);
	if (optrip <= 0)
		return;
	send_route_periodic(ifp);
}

void
del_route_ifp(ifp)
	struct interface *ifp;
{
	struct Route *rp;

	(void)gettimeofday(&now, NULL);
	/* delete routes through interface */
	FORALLRP(rp) {
		if (rp->ifp != ifp || rp->flags & RT_MANUAL ||
		    rp->metric == HOPCNT_INFINITY)
			continue;
		if (debug)
			logdebug("deleting route %s/%d through if down %s",
				 inet6pr(&rp->addr), rp->plen, ifp->name);
		del_kern_route(rp);
		ripchanges = (rp->flags |= RT_CHANGED);
		rp->refresh = now.tv_sec;
		rp->metric = HOPCNT_INFINITY;
	}
	if (ripchanges)
		send_route_triggered(0);
}

void
update_rip(ifp)
	struct interface *ifp;
{
	struct Route *rp;

	/* interface change, reinstall routes */
	if (!optripinst)
		return;
	FORALLRP(rp) {
		if (rp->ifp != ifp || !(rp->flags & RT_INSTALLED))
			continue;
		change_kern_route(rp);
	}
}

void
stop_rip(ifp)
	struct interface *ifp;
{
	/* 	TO BE WRITTEN. Should send rip packet ?? */
}

void
dump_route(rp)
	register struct Route *rp;
{
	logdebug("route (%lx):", (u_long)rp);
	logdebug("\tif=%s\tgateway=%s",
		 rp->ifp? rp->ifp->name : "<NONE>",
		 inet6pr(&rp->gateway));
	logdebug("\tdest=%s with len=%d and mask=%s",
		 inet6pr(&rp->addr), rp->plen, inet6pr(&rp->mask));
	logdebug("\tflags=%x metric=%d", rp->flags, rp->metric);
}

void
del_kern_route(rp)
	register struct Route *rp;
{
	register char *cp = m_rtmsg.m_space;
	struct sockaddr_in6 *sin;
	int len;

	if (debug)
		logdebug("del_kern_route %s/%d[%d] gw %s ifp %s fl 0x%x",
			 inet6pr(&rp->addr),
			 rp->plen, rp->metric,
			 inet6pr(&rp->gateway),
			 rp->ifp ? rp->ifp->name : "none",
			 rp->flags);
	if (!optripinst || ((rp->flags & RT_INSTALLED) == 0))
		return;
	bzero((caddr_t)&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = RTM_DELETE;
	if (rp->flags & RT_LOCAL)
		rtm.rtm_flags = RTF_UP | RTF_GATEWAY | RTF_LOCAL;
	else
		rtm.rtm_flags = RTF_UP | RTF_GATEWAY ;
	rtm.rtm_addrs = RTA_DST | RTA_NETMASK;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;
	rtm.rtm_index = rp->ifp->index;

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(rp->addr, sin->sin6_addr);
	ADVANCE(cp, sin);

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_UNSPEC;
	COPY_ADDR6(rp->mask, sin->sin6_addr);
	ADVANCE(cp, sin);

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

#ifdef TRACE
	DUMP("route del",(char *)&m_rtmsg, len);
#endif
	if (write(r, (char *)&m_rtmsg, len) < len) {
		if (verbose)
			logperror("write (route delete)");
		return;
	} else if (verbose)
		logtrace("delete route %s/%d",
			 inet6pr(&rp->addr), rp->plen);
	rp->flags &= ~RT_INSTALLED;
}

void
change_kern_route(rp)
	register struct Route *rp;
{
	register char *cp = m_rtmsg.m_space;
	struct sockaddr_in6 *sin;
	struct sockaddr_dl *sdl;
	int len;

	if (debug)
		logdebug("change_kern_route %s/%d[%d] gw %s ifp %s fl 0x%x",
			 inet6pr(&rp->addr),
			 rp->plen, rp->metric,
			 inet6pr(&rp->gateway),
			 rp->ifp ? rp->ifp->name : "none",
			 rp->flags);
	if (!optripinst || !(rp->flags & RT_INSTALLED))
		return;
	if (!rp->ifp) {
		if (verbose)
			logperror("no ifp for route");
		return;
	}
	bzero((caddr_t)&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = RTM_CHANGE;
	if (rp->flags & RT_LOCAL)
		rtm.rtm_flags = RTF_UP | RTF_GATEWAY | RTF_LOCAL;
	else
		rtm.rtm_flags = RTF_UP | RTF_GATEWAY;
	rtm.rtm_addrs = RTA_DST | RTA_GATEWAY | RTA_NETMASK | RTA_IFP;
	if (FIRST(rp->ifp))
		rtm.rtm_addrs |= RTA_IFA;
	else if (verbose)
		logerr("change kern route without ifa for %s", rp->ifp->name);
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;
	rtm.rtm_index = rp->ifp->index;

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(rp->addr, sin->sin6_addr);
	ADVANCE(cp, sin);

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(rp->gateway, sin->sin6_addr);
	ADVANCE(cp, sin);

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_UNSPEC;
	COPY_ADDR6(rp->mask, sin->sin6_addr);
	ADVANCE(cp, sin);

	sdl = (struct sockaddr_dl *)cp;
	sdl->sdl_family = AF_LINK;
	sdl->sdl_len = sizeof(struct sockaddr_dl);
	sdl->sdl_index = rp->ifp->index;
	ADVANCE(cp, sdl);

	if (FIRST(rp->ifp)) {
		sin = (struct sockaddr_in6 *)cp;
		sin->sin6_len = sizeof(struct sockaddr_in6);
		sin->sin6_family = AF_INET6;
		COPY_ADDR6(FIRST(rp->ifp)->addr, sin->sin6_addr);
		ADVANCE(cp, sin);
	}

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

	/* should exist, ESRCH is not a correct answer */
#ifdef TRACE
	DUMP("route chg",(char *)&m_rtmsg, len);
#endif
	if (write(r, (char *)&m_rtmsg, len) < len) {
		if (verbose)
			logperror("write (route change)");
		return;
	} else if (verbose)
		logtrace("change route  %s/%d (link#%d)",
			 inet6pr(&rp->addr),
			 rp->plen,
			 rp->ifp->index);
}

void
add_kern_route(rp)
	register struct Route *rp;
{
	register char *cp;
	struct sockaddr_in6 *sin;
	struct sockaddr_dl *sdl;
	int len;
	int already = 0;

	if (!optripinst || rp->flags & (RT_DONTINST | RT_INSTALLED))
		return;
	if (!rp->ifp) {
		if (verbose)
			logperror("no ifp for route");
		return;
	}
    retry:
	cp = m_rtmsg.m_space;
	bzero((caddr_t)&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = RTM_ADD;
	if (rp->flags & RT_LOCAL)
		rtm.rtm_flags = RTF_UP | RTF_GATEWAY | RTF_LOCAL;
	else
		rtm.rtm_flags = RTF_UP | RTF_GATEWAY;
	rtm.rtm_addrs = RTA_DST | RTA_GATEWAY | RTA_NETMASK | RTA_IFP;
	if (FIRST(rp->ifp))
		rtm.rtm_addrs |= RTA_IFA;
	else if (verbose && !already)
		logerr("change kern route without ifa for %s", rp->ifp->name);
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;
	rtm.rtm_index = rp->ifp->index;

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(rp->addr, sin->sin6_addr);
	ADVANCE(cp, sin);

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	COPY_ADDR6(rp->gateway, sin->sin6_addr);
	ADVANCE(cp, sin);

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_UNSPEC;
	COPY_ADDR6(rp->mask, sin->sin6_addr);
	ADVANCE(cp, sin);

	sdl = (struct sockaddr_dl *)cp;
	sdl->sdl_family = AF_LINK;
	sdl->sdl_len = sizeof(struct sockaddr_dl);
	sdl->sdl_index = rp->ifp->index;
	ADVANCE(cp, sdl);

	if (FIRST(rp->ifp)) {
		sin = (struct sockaddr_in6 *)cp;
		sin->sin6_len = sizeof(struct sockaddr_in6);
		sin->sin6_family = AF_INET6;
		COPY_ADDR6(FIRST(rp->ifp)->addr, sin->sin6_addr);
		ADVANCE(cp, sin);
	}

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

	/*
	 * If exists do a delete/add.
	 * Change is not enough if the old route is without gateway
	 */
#ifdef TRACE
	DUMP("route add", (char *)&m_rtmsg, len);
#endif
	if (write(r, (char *)&m_rtmsg, len) < len) {
		if (errno == EEXIST && !already) {
			rp->flags |= RT_INSTALLED;
			del_kern_route(rp);
			already = 1;
			goto retry;
		} else if (verbose)
			logperror("write (route add)");
		return;
	} else if (verbose)
		logtrace("set route %s/%d (link#%d)",
			 inet6pr(&rp->addr), rp->plen, rp->ifp->index);
	/* set ifa */
	rp->flags |= RT_INSTALLED;
}

void
new_gateway() {
	read_gateway(0);
	if (ripchanges)
		send_route_triggered(0);
}

static int
filter(pfilt, addr, aplen, gateway)
	struct TFilter *pfilt;
	struct in6_addr *addr, *gateway;
	int *aplen;
{
	int i;
	struct Filter *f, *fend = &pfilt->filter[pfilt->nb];

	for (f = &pfilt->filter[0]; f < fend; f++) {
		if (f->plen > *aplen)
			continue;
		switch (f->typlg) {
		    case LG_EQ:
			if (*aplen != f->lg)
				continue;
			break;
		    case LG_GE:
			if (*aplen < f->lg)
				continue;
			break;
		    case LG_LE:
			if (*aplen > f->lg)
				continue;
			break;
		}
		if (f->gateway.s6_addr8[0] != NO_GATEWAY)
			if (!SAME_ADDR6(f->gateway, *gateway))
				continue;
		if (f->plen)
			for (i = 0; i < sizeof(f->mask); i++)
				if (f->addr.s6_addr[i] !=
				    (addr->s6_addr[i]&f->mask.s6_addr[i]))
					goto nxt;
		switch (f->action) {
		    case FACCEPT:
			if (debug)
				logdebug("accept");
			return 0;
		    case FREJECT:
			if (debug)
				logdebug("reject");
			return 1;
		    default:
			if (*aplen < f->action)
				continue;
			*aplen = f->action;
			if (debug)
				logdebug("truncate to %d", *aplen);
			return 0;
		}
    nxt:
	}
	return 0;
}
