#include "ndpd.h"

/* add/modify a interface prefix */
/* TODO : detect reordering of prefix */
void
update_prefix(pp)
	register struct prefix *pp;
{

}

/* Modify advertised prefix list */
void
update_advprefix(pp)
	register struct prefix *pp;
{
	read_gateway(1);	/* recompute everything !! */
}

/* add anycast entries for a prefix or an interface */
void
add_anycast_prefix(ifp, pp, addr)
	struct interface *ifp;
	struct prefix *pp;
	struct in6_addr *addr;
{
	register struct anycast *ap;

	if (NO_ANYCAST(ifp))
		return;
	/* verify automatic cases */
	if (pp && (addr == NULL) &&
	    ((pp->plen > (128 - 8 * ifp->ltoken)) || (pp->plen == 0)))
		return;
	if ((pp == NULL) && (addr == NULL) && IS_ANYADDR6(ifp->llocal))
		return;

	if (Calloc(ap, anycast) == NULL) {
		if (verbose)
			logerr("out of memory");
		return;
	}

	TAILQ_INIT(&ap->presps);
	ap->prefix = pp;
	ap->ifp = ifp;
	ap->type = IP6ANY_ANYCAST;
	ap->ntosend = 1;
	if (addr)
		COPY_ADDR6(*addr, ap->addr);
	else if (pp)
		COPY_ADDR6(pp->pref, ap->addr);
	else {
		ap->addr.s6_addr[0] = 0xfe;
		ap->addr.s6_addr[1] = 0x80;
	}
	if (pp)
		TAILQ_INSERT_TAIL(&pp->anycasts, ap, any_list);
	else
		TAILQ_INSERT_TAIL(&ifp->oanycasts, ap, any_list);

	add_anycast(ap);

	if (pp && (ifp->ltoken == 8) && (pp->hacnt == -1) &&
	    (ifp->opt.flags & ND_RA_FLAG_HAGENT))
		add_anycast_agent(pp);
}

/* add an anycast or a proxy */
void
add_anycast(ap)
	struct anycast *ap;
{
	struct if_laddrreq ifr;
	struct sockaddr_in6 *sin;

	bzero((caddr_t)&ifr, sizeof(struct if_laddrreq));
	strncpy(ifr.iflr_name, ap->ifp->name, IFNAMSIZ);
	sin = (struct sockaddr_in6 *)&ifr.addr;
	sin->sin6_family = AF_INET6;
	sin->sin6_len = sizeof(*sin);
	ifr.flags = ap->type;
	COPY_ADDR6(ap->addr, sin->sin6_addr);

	if (ioctl(u, SIOCADDANY, (caddr_t)&ifr) < 0) {
		if (verbose)
			logperror("ioctl (SIOCADDANY)");
		return;
	}
	anycast_multi(ap, 1);
	send_grat_apadv(ap);
}

/* delete anycast entries for a prefix or an interface */
void
del_anycast_prefix(ifp, pp, addr)
	struct interface *ifp;
	struct prefix *pp;
	struct in6_addr *addr;
{
	register struct anycast *ap;
	struct in6_addr aa;

	if (NO_ANYCAST(ifp))
		return;
	/* verify automatic cases */
	if (pp && (addr == NULL) &&
	    ((pp->plen > (128 - 8 * ifp->ltoken)) || (pp->plen == 0)))
		return;
	if ((pp == NULL) && (addr == NULL) && IS_ANYADDR6(ifp->llocal))
		return;

	if (addr)
		COPY_ADDR6(*addr, aa);
	else if (pp)
		COPY_ADDR6(pp->pref, aa);
	else {
		bzero((caddr_t)&aa, sizeof(struct in6_addr));
		aa.s6_addr[0] = 0xfe;
		aa.s6_addr[1] = 0x80;
	}

	if (pp) {
		FORALLPANY(ap, pp)
			if (SAME_ADDR6(ap->addr, aa))
				break;
	} else FORALLOANY(ap, ifp)
		if (SAME_ADDR6(ap->addr, aa))
			break;

	if (ap)
		del_anycast(ap);
	else if (verbose)
		logerr("can't find the anycast to delete");

	if (pp && (ifp->ltoken == 8) && (pp->hacnt != -1) &&
	    (ifp->opt.flags & ND_RA_FLAG_HAGENT))
		del_anycast_agent(pp);
}

/* delete an anycast entry */
void
del_anycast(ap)
	register struct anycast *ap;
{
	register struct presp *rp;
	struct prefix *pp = ap->prefix;
	struct interface *ifp = ap->ifp;
	struct if_laddrreq ifr;
	struct sockaddr_in6 *sin;

	anycast_multi(ap, 0);

	while ((rp = ap->presps.tqh_first) != NULL)
		cancel_anycast(rp);

	if (ap->tp) {
		kill_timeout(ap->tp);
		ap->tp = NULL;
	}

	if (pp)
		TAILQ_REMOVE(&pp->anycasts, ap, any_list);
	else
		TAILQ_REMOVE(&ifp->oanycasts, ap, any_list);

	bzero((caddr_t)&ifr, sizeof(struct if_laddrreq));
	strncpy(ifr.iflr_name, ifp->name, IFNAMSIZ);
	sin = (struct sockaddr_in6 *)&ifr.addr;
	sin->sin6_family = AF_INET6;
	sin->sin6_len = sizeof(*sin);
	COPY_ADDR6(ap->addr, sin->sin6_addr);

	if (ioctl(u, SIOCDELANY, (caddr_t)&ifr) < 0) {
		if (verbose)
			logperror("ioctl (SIOCDELANY)");
	}
	free(ap);
}

/* delete an invalid prefix */
void
prefix_inval(pp)
	register struct prefix *pp;
{
	register struct anycast *ap;
	register struct prefix *pp1, *pp2;

	while ((ap = pp->anycasts.tqh_first) != NULL)
		del_anycast(ap);

	if (pp->ifp->usefpref)
	    for (pp1 = FFIRST(pp->ifp); pp1; pp1 = pp2) {
		pp2 = pp1->prf_list.tqe_next;
		if (pp1->valpref.imported &&
		    SAME_ADDR6(pp->addr, pp1->addr))
			TAILQ_REMOVE(&pp->ifp->fprefixes, pp1, prf_list);
	    }
	TAILQ_REMOVE(&pp->ifp->prefixes, pp, prf_list);

	free(pp);
}
