#include <util_ndpd.h>
#include <hw.h>

struct neighbor {
	TAILQ_ENTRY(neighbor) ngb_list;	/* neighbor list */
	struct in6_addr addr;		/* IPv6 address */
	struct sockaddr_dl sdl;		/* data-link address */
	TIMEOUT *tp;			/* timeout */
	u_char flags;			/* flags */
#define ND6F_IS_ROUTER		0x01
#define ND6F_HAS_ROUTE		0x02
#define ND6F_HAS_LLADDR		0x04
#define ND6F_IS_SELF		0x08
	char state;			/* state */
#define ND6S_REJECT		-1
#define ND6S_INCOMPLETE		0
#define ND6S_REACHABLE		1
#define ND6S_STALE		2
#define ND6S_DELAY		3
#define ND6S_PROBE		4
#define ND6S_PENDING		5
#define ND6S_LOCKED		6
	char pstate;			/* pending next state */
	u_char asked;			/* solicitation counter */
	struct nd_opt_hdr *llinfo;	/* link-layer info */
	void *circuit;			/* pointer to circuit info */
};
TAILQ_HEAD(, neighbor) neighbors;
#define FORALLND(ndp)	\
	for (ndp = neighbors.tqh_first; ndp; ndp = ndp->ngb_list.tqe_next)

#define IN6_IS_ADDR_SOLICITED(a)	\
	(((a)->s6_addr[0] == 0xff) &&	\
	 ((a)->s6_addr[1] == 0x02) &&	\
	 ((a)->s6_addr[2] == 0x00) &&	\
	 ((a)->s6_addr[3] == 0x00) &&	\
	 ((a)->s6_addr[4] == 0x00) &&	\
	 ((a)->s6_addr[5] == 0x00) &&	\
	 ((a)->s6_addr[6] == 0x00) &&	\
	 ((a)->s6_addr[7] == 0x00) &&	\
	 ((a)->s6_addr[8] == 0x00) &&	\
	 ((a)->s6_addr[9] == 0x00) &&	\
	 ((a)->s6_addr[10] == 0x00) &&	\
	 ((a)->s6_addr[11] == 0x01) &&	\
	 ((a)->s6_addr[12] == 0xff))

struct prefix {
	TAILQ_ENTRY(prefix) prf_list;	/* prefix list */
	struct in6_addr pref;		/* prefix itself */
	struct in6_addr pmask;		/* prefix mask */
	struct in6_addr addr;		/* address */
	struct in6_addr amask;		/* address mask */
	struct in6_addr dest;		/* destination address */
	int plen;			/* length */
	struct neighbor self;		/* prefix as a neighbor */
};
TAILQ_HEAD(, prefix) prefixes;
#define FORALLPP(pp)	\
	for (pp = prefixes.tqh_first; pp; pp = pp->prf_list.tqe_next)

struct interface {
	struct in6_addr llocal;		/* link-local address */
	char name[IFNAMSIZ];		/* name */
	u_char ltoken;			/* token length */
	u_char lmac;			/* MAC length */
	short flags;			/* copy of kernel flags */
	u_long mtu;			/* IPv6 MTU */
	u_int index;			/* kernel index */
	struct nd_opt_hdr *llinfo;	/* link-layer info */
	/* link-layer info functions */
	struct nd_opt_hdr * (*lladdr_get) (struct nd_opt_hdr *);
	int (*lladdr_cmp) (struct nd_opt_hdr *, struct nd_opt_hdr *);
	void (*lladdr_updated) (struct neighbor *);
	void (*circuit_close) (struct neighbor *);
	char * (*lladdr_pr) (struct nd_opt_hdr *);
	char * (*sdl_pr) (struct sockaddr_dl *);
};
struct interface itf;

extern int icmp_sd, route_sd;		/* Socket file descriptors */

extern int verbose;
extern int debug;
extern int timestamps;

/* Protocol Parameters */

extern u_int BaseReachableTime;
extern u_int ReachableTime;
extern u_int InactiveTime;
extern u_int HoldDownTime;
extern u_int RetransTimer;
extern u_int DelayTimer;

extern TIMEOUT *randpar;

/* send_ndsol kinds */
#define NS6K_DAD	0			/* DAD operation */
#define NS6K_SOL	1			/* use solicited multicast */
#define NS6K_DIR	2			/* use neighbor address */

void enter_pending(struct neighbor *, char);
void leave_pending(struct neighbor *);
void enter_state(struct neighbor *, char);
void retransmit(struct neighbor *);
void do_nud(struct neighbor *);
void probe(struct neighbor *);
void del_neighbor(struct neighbor *);
void update(struct neighbor *, struct nd_opt_hdr *, int, int, int);

void recv_icmp(void);
void recv_rtsol(struct ip6_hdr *, int);
void recv_rtadv(struct ip6_hdr *, int);
void recv_ndsol(struct ip6_hdr *, int);
void recv_ndadv(struct ip6_hdr *, int);
void recv_redirect(struct ip6_hdr *, int);
struct neighbor *ndplookup(struct in6_addr *, int);
int lladdr_overwrite(struct neighbor *, struct nd_opt_hdr *, char *);

int send_rtsol(void);
int send_ndsol(struct neighbor *, int);
int send_ndadv(struct neighbor *, struct prefix *, int);

int init(char *);
void init_param(void);
void setReachableTime(u_int);
void setRetransTimer(u_int);
void recv_route(void);
void recv_newaddr(struct ifa_msghdr *);
void recv_deladdr(struct ifa_msghdr *);
void recv_resolve(struct rt_msghdr *);
int is_router(void);
int is_apaddr(struct in6_addr *);

void ndpd_init(void);
void set_verbose(int);
void finish(int);
void dump(int);
void dump_interface(struct interface *);
void dump_prefix(struct prefix *);
void dump_neighbor(struct neighbor *);

int del_route(struct neighbor *);
int change_route(struct neighbor *);
int reject_route(struct neighbor *);
int local_route(struct neighbor *);
