/* Neighbor discovery specific routines for PPP */

#include <ndpd.h>

/*
 * Hardware dependent interface initialisations
 */
int
new_ppp_if(ifm, first)
	struct if_msghdr *ifm;
	int first;
{
	struct sockaddr_dl *sdl;

	/* first time? */

	if (first != 1)
		return;

	/* addresses should be interface only */

	if (ifm->ifm_addrs != RTA_IFP) {
		logerr("bad ifinfo (no ifp#%d)", ifm->ifm_index);
		return 0;
	}

	/* get and check the data-link address */

	sdl = (struct sockaddr_dl *)((char *)ifm + sizeof(struct if_msghdr));
	if ((sdl->sdl_family != AF_LINK) || (sdl->sdl_nlen == 0)) {
		if (verbose)
			logerr("bad sdl address");
		return 0;
	}

	/* verify flags and MTU */

	if ((ifm->ifm_flags & IFF_LOOPBACK) ||
	    (ifm->ifm_flags & IFF_BROADCAST) ||
	    ((ifm->ifm_flags & IFF_POINTOPOINT) != IFF_POINTOPOINT) ||
	    ((ifm->ifm_flags & IFF_MULTICAST) != IFF_MULTICAST) ||
	    (ifm->ifm_data.ifi_mtu < IP6_MMTU)) {
		logerr("incompatible flags 0x%x for interface %.*s",
		       (unsigned)ifm->ifm_flags,
		       sdl->sdl_nlen, sdl->sdl_data);
		return 0;
	}

	/* verify type */

	if (ifm->ifm_data.ifi_type != IFT_PPP) {
		logerr("incompatible type %d for interface %.*s",
		       ifm->ifm_data.ifi_type,
		       sdl->sdl_nlen, sdl->sdl_data);
		return 0;
	}

	/* PPP has no data-link addresses */

	if ((ifm->ifm_data.ifi_addrlen != 0) || (sdl->sdl_alen != 0)) {
		logerr("bad data-link address length (%d) for interface %.*s",
		       ifm->ifm_data.ifi_addrlen,
		       sdl->sdl_nlen, sdl->sdl_data);
		return 0;
	}

	/* fill the interface structure */

	itf.mtu = ifm->ifm_data.ifi_mtu;
	itf.flags = ifm->ifm_flags;
	itf.ltoken = PPPTOKEN;
	bcopy(sdl->sdl_data, itf.name, sdl->sdl_nlen);
#ifdef notyet
	itf.ntosend = MAX_RTR_SOLICITATIONS;
#endif

	/* fill methods */

	itf.lladdr_get = ppp_addr_get;
	itf.lladdr_cmp = ppp_addr_cmp;
	itf.lladdr_updated = ppp_addr_updated;
	itf.lladdr_pr = ppp_addr_pr;
	itf.sdl_pr = ppp_sdl_pr;

	/* success */

	if (verbose)
		logtrace("got interface %s#%d", itf.name, itf.index);
	return 1;
}

/*
 * Transform a PPP link-layer info in an usable stuff
 * (link-layer info should not be used)
 */
struct nd_opt_hdr *
ppp_addr_get(lp)
	struct nd_opt_hdr *lp;
{
	if (verbose)
		logerr("ppp_addr_get called?");
	return lp;
}

/*
 * Compare two PPP link-layer infos (always return 0 ie equal)
 */
int
ppp_addr_cmp(l1, l2)
	struct nd_opt_hdr *l1, *l2;
{
	return 0;
}

/*
 * Update a data-link address from a PPP link-layer info
 */
void
ppp_addr_updated(ndp)
	struct neighbor *ndp;
{
	/* nothing to do */

	ndp->flags |= ND6F_HAS_LLADDR;
}

/*
 * CLose a circuit (none for PPP)
 */
void
ppp_circuit_close(ndp)
	struct neighbor *ndp;
{
	if (debug && (ndp->circuit != NULL))
		logerr("%s as a circuit 0x%x on PPP",
		       inet6pr(&ndp->addr), (u_long)ndp->circuit);
}

/*
 * Print a PPP link-layer info
 */
char *
ppp_addr_pr(lp)
	struct nd_opt_hdr *lp;
{
	return "-";
}

/*
 * Print a PPP data-link address
 */
char *
ppp_sdl_pr(sdl)
	struct sockaddr_dl *sdl;
{
	return "[]";
}
