/* Neighbor discovery route message emission */

#include <ndpd.h>

/*
 * Delete a route for a neighbor
 */
int
del_route(ndp)
	struct neighbor *ndp;
{
	char *cp = m_rtmsg.m_space;
	struct sockaddr_in6 *sin;
	int len;

	if (verbose)
		logtrace("delete route for %s", inet6pr(&ndp->addr));

	/* fill the route message header */

	bzero(&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = RTM_DELETE;
	rtm.rtm_flags = RTF_UP | RTF_HOST;
	rtm.rtm_addrs = RTA_DST;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;
	rtm.rtm_index = itf.index;

	/* fill the destination */

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	bcopy(&ndp->addr, &sin->sin6_addr, sizeof(struct in6_addr));
	ADVANCE(cp, sin);

	/* send the message */

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

	DUMPR("delete route");

	if (write(route_sd, (char *)&m_rtmsg, len) < len) {
		if (verbose)
			logperror("write (delete route)");
		return 0;
	}

	/* success */

	ndp->flags &= ~ND6F_HAS_ROUTE;
	return 1;
}

/*
 * Update a route for a neighbor with a known data-link address
 */
int
change_route(ndp)
	struct neighbor *ndp;
{
	char *cp = m_rtmsg.m_space;
	struct sockaddr_in6 *sin;
	struct sockaddr_dl *sdl;
	int len;

	if (verbose)
		logtrace("change route to %s with %s",
			 inet6pr(&ndp->addr),
			 itf.sdl_pr(&ndp->sdl));

	/* fill the route message header */

	bzero(&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = RTM_CHANGE;
	rtm.rtm_flags = RTF_UP | RTF_HOST;
	rtm.rtm_addrs = RTA_DST | RTA_GATEWAY | RTA_IFP | RTA_IFA;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;
	rtm.rtm_index = itf.index;

	/* fill the destination */

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	bcopy(&ndp->addr, &sin->sin6_addr, sizeof(struct in6_addr));
	ADVANCE(cp, sin);

	/* fill the gateway (data-link address) */

	sdl = (struct sockaddr_dl *)cp;
	sdl->sdl_family = AF_LINK;
	sdl->sdl_len = sizeof(struct sockaddr_dl);
	sdl->sdl_index = itf.index;
	sdl->sdl_alen = ndp->sdl.sdl_alen;
	bcopy(LLADDR(&ndp->sdl), LLADDR(sdl), sdl->sdl_alen);
	ADVANCE(cp, sdl);

	/* fill the interface */

	sdl = (struct sockaddr_dl *)cp;
	sdl->sdl_family = AF_LINK;
	sdl->sdl_len = sizeof(struct sockaddr_dl);
	sdl->sdl_index = itf.index;
	ADVANCE(cp, sdl);

	/* fill the interface address: hack... */

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	bcopy(&ndp->addr, &sin->sin6_addr, sizeof(struct in6_addr));
	ADVANCE(cp, sin);

	/* send the message */

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

	DUMPR("change route");

	if (write(route_sd, (char *)&m_rtmsg, len) < len) {
		if (verbose)
			logperror("write (change route)");
		return 0;
	}

	/* success */

	return 1;
}

/*
 * Mark a route as reject (neighbor unreachable, need a kernel change)
 */
int
reject_route(ndp)
	struct neighbor *ndp;
{
	char *cp = m_rtmsg.m_space;
	struct sockaddr_in6 *sin;
	int len;

	if (verbose)
		logtrace("reject route for %s", inet6pr(&ndp->addr));

	/* fill the route message header */

	bzero(&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = RTM_CHANGE;
	rtm.rtm_flags = RTF_UP | RTF_HOST | RTF_REJECT;
	rtm.rtm_addrs = RTA_DST;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;
	rtm.rtm_index = itf.index;

	/* fill the destination */

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	bcopy(&ndp->addr, &sin->sin6_addr, sizeof(struct in6_addr));
	ADVANCE(cp, sin);

	/* send the message */

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

	DUMPR("reject route");

	if (write(route_sd, (char *)&m_rtmsg, len) < len) {
		if (verbose)
			logperror("write (reject route)");
		return 0;
	}

	/* success */

	return 1;
}

/*
 * Add a local route for myself
 */
int
local_route(ndp)
	struct neighbor *ndp;
{
	char *cp = m_rtmsg.m_space;
	struct sockaddr_in6 *sin;
	struct sockaddr_dl *sdl;
	int len, type = RTM_ADD;

	if (verbose)
		logtrace("add local route for %s", inet6pr(&ndp->addr));

	/* come back here if route doesn't exist */

    retry:

	/* fill the route message header */

	bzero(&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = type;
	rtm.rtm_flags = RTF_UP | RTF_HOST | RTF_LOCAL;
	rtm.rtm_addrs = RTA_DST | RTA_GATEWAY | RTA_IFP | RTA_IFA;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;
	rtm.rtm_index = itf.index;

	/* fill the destination */

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	bcopy(&ndp->addr, &sin->sin6_addr, sizeof(struct in6_addr));
	ADVANCE(cp, sin);

	/* fill the gateway (data-link address) */

	sdl = (struct sockaddr_dl *)cp;
	sdl->sdl_family = AF_LINK;
	sdl->sdl_len = sizeof(struct sockaddr_dl);
	sdl->sdl_index = itf.index;
	sdl->sdl_alen = ndp->sdl.sdl_alen;
	bcopy(LLADDR(&ndp->sdl), LLADDR(sdl), sdl->sdl_alen);
	ADVANCE(cp, sdl);

	/* fill the interface */

	sdl = (struct sockaddr_dl *)cp;
	sdl->sdl_family = AF_LINK;
	sdl->sdl_len = sizeof(struct sockaddr_dl);
	sdl->sdl_index = itf.index;
	ADVANCE(cp, sdl);

	/* fill the interface address */

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	sin->sin6_family = AF_INET6;
	bcopy(&ndp->addr, &sin->sin6_addr, sizeof(struct in6_addr));
	ADVANCE(cp, sin);

	/* send the message */

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;

	DUMPR("add local route");

	if (write(route_sd, (char *)&m_rtmsg, len) < len) {
		if ((type == RTM_ADD) && (errno == EEXIST)) {
			/* route already exists: retry with change */
			type = RTM_CHANGE;
			goto retry;
		}
		if (verbose)
			logperror("write (add local route)");
		return 0;
	}

	/* success */

	ndp->flags |= ND6F_HAS_LLADDR | ND6F_HAS_ROUTE;
	return 1;
}
