/* Neighbor discovery regular timeout routines */

#include <ndpd.h>

#if __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif

void
add_timeout(tp)
	register struct timeout *tp;
{
	register struct timeout *tp2;

	tp->status = 1;
	if (timeouts.tqh_first == NULL) {
		TAILQ_INSERT_HEAD(&timeouts, tp, tm_list);
		return;
	}
	for (tp2 = timeouts.tqh_first; tp2; tp2 = tp2->tm_list.tqe_next)
		if (!AFTER(tp->tv, tp2->tv))
			break;

	if (tp2 == NULL)
		TAILQ_INSERT_TAIL(&timeouts, tp, tm_list);
	else
		TAILQ_INSERT_BEFORE(tp2, tp, tm_list);
}

void
set_timeout(tp, sec, usec, func, arg)
	register struct timeout *tp;
	int sec, usec;
	func_to func;
	void * arg;
{
	tp->tv.tv_sec = now.tv_sec + sec;
	tp->tv.tv_usec = now.tv_usec + usec;
	NORMTIME(tp->tv);
	tp->func = func;
	tp->arg = arg;
	add_timeout(tp);
}

struct timeout *
new_timeout(sec, usec, func, arg)
	int sec, usec;
	func_to func;
	void * arg;
{
	register struct timeout *tp;

	if (Calloc(tp, timeout) == NULL) {
		logerr("out of memory");
		return 0;
	}
	set_timeout(tp, sec, usec, func, arg);
	return tp;
}

void
kill_timeout(tp)
	register struct timeout *tp;
{
	if (tp->status > 0) {
		TAILQ_REMOVE(&timeouts, tp, tm_list);
		tp->status = 0;
	}
	if (tp->status >= 0 && tp->refcnt == 0)
		free(tp);
}

void
run_timeouts()
{
	register struct timeout *tp;

	for(;;) {
		(void) gettimeofday(&now, NULL);
		tp = timeouts.tqh_first;
		if (tp == 0 || !AFTER(now, tp->tv))
			return;
		TAILQ_REMOVE(&timeouts, tp, tm_list);
		tp->status = -1;
		(*(tp->func)) (tp->arg);
		if (tp->status < 0)	/* not reused */
			if (tp->refcnt == 0)
				free(tp);
			else
				tp->status = 0;
	}
}

void
get_nexttime(tvp)
	struct timeval *tvp;
{
	struct timeout *tp = timeouts.tqh_first;

	if (tp == 0) {
		tvp->tv_sec = 1;
		tvp->tv_usec = 0;
		return;
	}

	tvp->tv_usec = tp->tv.tv_usec + 1000000 - now.tv_usec;
	tvp->tv_sec = tp->tv.tv_sec - now.tv_sec;
	NORMTIME(*tvp);
	if (tvp->tv_sec > 0)
		tvp->tv_sec -= 1;
	else
		tvp->tv_usec = 10000;
}

void
dump_timeout(tp)
	register struct timeout *tp;
{
	struct timeval delta;
	int sign = 0;

	logdebug("timeout (%lx):", (u_long)tp);
	if (AFTER(tp->tv, now)) {
		sign = +1;
		delta.tv_sec = tp->tv.tv_sec - now.tv_sec - 1;
		delta.tv_usec = tp->tv.tv_usec + 1000000 - now.tv_usec;
		NORMTIME(delta);
	} else {
		sign = -1;
		delta.tv_sec =  now.tv_sec - tp->tv.tv_sec - 1;
		delta.tv_usec = now.tv_usec + 1000000 - tp->tv.tv_usec;
		NORMTIME(delta);
	}
	logdebug("\texpire=%+ld.%06lu", sign * delta.tv_sec, delta.tv_usec);
	logdebug("\tstatus=%d\trefcnt=%d", tp->status, tp->refcnt);
	logdebug("\tfunction=%p\targument=%p", tp->func, tp->arg);
}
