/*
 * Regular timeout definitions
 */

typedef void (*func_to) (void *);
struct timeout {
	TAILQ_ENTRY(timeout) tm_list;	/* timeout list */
	struct timeval tv;		/* when */
	func_to func;			/* function */
	void *arg;			/* argument */
	int status;			/* 0: inactive, >0: in timeouts list */
					/* <0: being expired (do not free!) */
	int refcnt;			/* refcount: if 0, expire=>free */
};
TAILQ_HEAD(, timeout) timeouts;

#define DO_TIMEOUTS_INIT

#define FORALLTP(tp)	\
	for (tp = timeouts.tqh_first; tp; tp = tp->tm_list.tqe_next)

#define NORMTIME(tv) \
	while ((tv).tv_usec >= 1000000) { \
		(tv).tv_sec += 1; \
		(tv).tv_usec -= 1000000; \
	}

/* tv1 >= tv2 ? */
#define AFTER(tv1, tv2) \
	(((tv1).tv_sec > (tv2).tv_sec) || \
	 (((tv1).tv_sec == (tv2).tv_sec) && ((tv1).tv_usec >= (tv2).tv_usec)))

struct timeout *new_timeout(int, int, func_to, void *);
void set_timeout(struct timeout *, int, int, func_to, void *);
void add_timeout(struct timeout *);
void dump_timeout(struct timeout *);
void kill_timeout(struct timeout *);
void run_timeouts(void);
void get_nexttime(struct timeval *);

#define TIMEOUT		struct timeout
#define NEW_TIMEOUT(sec, usec, func, arg)	\
	new_timeout((sec), (usec), (func_to) (func), (arg))
#define SET_TIMEOUT	set_timeout
#define ADD_TIMEOUT	add_timeout
#define DUMP_TIMEOUT	dump_timeout
#define KILL_TIMEOUT	kill_timeout
#define RUN_TIMEOUTS	run_timeouts
#define GET_NEXTTIME	get_nexttime
