/*-
 * Copyright (c) 1990 The Regents of the University of California.
 * All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Van Jacobson.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
char copyright[] =
"@(#) Copyright (c) 1990 The Regents of the University of California.\n\
 All rights reserved.\n";
#endif /* not lint */

#ifndef lint
static char sccsid[] = "@(#)traceroute.c	5.4 (Berkeley) 5/15/90";
#endif /* not lint */

/*
 * traceroute host  - trace the route ip packets follow going to "host".
 *
 * Attempt to trace the route an ip packet would follow to some
 * internet host.  We find out intermediate hops by launching probe
 * packets with a small ttl (time to live) then listening for an
 * icmp "time exceeded" reply from a gateway.  We start our probes
 * with a ttl of one and increase by one until we get an icmp "port
 * unreachable" (which means we got to "host") or hit a max (which
 * defaults to 30 hops & can be changed with the -m flag).  Three
 * probes (change with -q flag) are sent at each ttl setting and a
 * line is printed showing the ttl, address of the gateway and
 * round trip time of each probe.  If the probe answers come from
 * different gateways, the address of each responding system will
 * be printed.  If there is no response within a 5 sec. timeout
 * interval (changed with the -w flag), a "*" is printed for that
 * probe.
 *
 * Probe packets are UDP format.  We don't want the destination
 * host to process them so the destination port is set to an
 * unlikely value (if some clod on the destination is using that
 * value, it can be changed with the -p flag).
 *
 * A sample use might be:
 *
 *     [yak 71]% traceroute nis.nsf.net.
 *     traceroute to nis.nsf.net (35.1.1.48), 30 hops max, 56 byte packet
 *      1  helios.ee.lbl.gov (128.3.112.1)  19 ms  19 ms  0 ms
 *      2  lilac-dmc.Berkeley.EDU (128.32.216.1)  39 ms  39 ms  19 ms
 *      3  lilac-dmc.Berkeley.EDU (128.32.216.1)  39 ms  39 ms  19 ms
 *      4  ccngw-ner-cc.Berkeley.EDU (128.32.136.23)  39 ms  40 ms  39 ms
 *      5  ccn-nerif22.Berkeley.EDU (128.32.168.22)  39 ms  39 ms  39 ms
 *      6  128.32.197.4 (128.32.197.4)  40 ms  59 ms  59 ms
 *      7  131.119.2.5 (131.119.2.5)  59 ms  59 ms  59 ms
 *      8  129.140.70.13 (129.140.70.13)  99 ms  99 ms  80 ms
 *      9  129.140.71.6 (129.140.71.6)  139 ms  239 ms  319 ms
 *     10  129.140.81.7 (129.140.81.7)  220 ms  199 ms  199 ms
 *     11  nic.merit.edu (35.1.1.48)  239 ms  239 ms  239 ms
 *
 * Note that lines 2 & 3 are the same.  This is due to a buggy
 * kernel on the 2nd hop system -- lbl-csam.arpa -- that forwards
 * packets with a zero ttl.
 *
 * A more interesting example is:
 *
 *     [yak 72]% traceroute allspice.lcs.mit.edu.
 *     traceroute to allspice.lcs.mit.edu (18.26.0.115), 30 hops max
 *      1  helios.ee.lbl.gov (128.3.112.1)  0 ms  0 ms  0 ms
 *      2  lilac-dmc.Berkeley.EDU (128.32.216.1)  19 ms  19 ms  19 ms
 *      3  lilac-dmc.Berkeley.EDU (128.32.216.1)  39 ms  19 ms  19 ms
 *      4  ccngw-ner-cc.Berkeley.EDU (128.32.136.23)  19 ms  39 ms  39 ms
 *      5  ccn-nerif22.Berkeley.EDU (128.32.168.22)  20 ms  39 ms  39 ms
 *      6  128.32.197.4 (128.32.197.4)  59 ms  119 ms  39 ms
 *      7  131.119.2.5 (131.119.2.5)  59 ms  59 ms  39 ms
 *      8  129.140.70.13 (129.140.70.13)  80 ms  79 ms  99 ms
 *      9  129.140.71.6 (129.140.71.6)  139 ms  139 ms  159 ms
 *     10  129.140.81.7 (129.140.81.7)  199 ms  180 ms  300 ms
 *     11  129.140.72.17 (129.140.72.17)  300 ms  239 ms  239 ms
 *     12  * * *
 *     13  128.121.54.72 (128.121.54.72)  259 ms  499 ms  279 ms
 *     14  * * *
 *     15  * * *
 *     16  * * *
 *     17  * * *
 *     18  ALLSPICE.LCS.MIT.EDU (18.26.0.115)  339 ms  279 ms  279 ms
 *
 * (I start to see why I'm having so much trouble with mail to
 * MIT.)  Note that the gateways 12, 14, 15, 16 & 17 hops away
 * either don't send ICMP "time exceeded" messages or send them
 * with a ttl too small to reach us.  14 - 17 are running the
 * MIT C Gateway code that doesn't send "time exceeded"s.  God
 * only knows what's going on with 12.
 *
 * The silent gateway 12 in the above may be the result of a bug in
 * the 4.[23]BSD network code (and its derivatives):  4.x (x <= 3)
 * sends an unreachable message using whatever ttl remains in the
 * original datagram.  Since, for gateways, the remaining ttl is
 * zero, the icmp "time exceeded" is guaranteed to not make it back
 * to us.  The behavior of this bug is slightly more interesting
 * when it appears on the destination system:
 *
 *      1  helios.ee.lbl.gov (128.3.112.1)  0 ms  0 ms  0 ms
 *      2  lilac-dmc.Berkeley.EDU (128.32.216.1)  39 ms  19 ms  39 ms
 *      3  lilac-dmc.Berkeley.EDU (128.32.216.1)  19 ms  39 ms  19 ms
 *      4  ccngw-ner-cc.Berkeley.EDU (128.32.136.23)  39 ms  40 ms  19 ms
 *      5  ccn-nerif35.Berkeley.EDU (128.32.168.35)  39 ms  39 ms  39 ms
 *      6  csgw.Berkeley.EDU (128.32.133.254)  39 ms  59 ms  39 ms
 *      7  * * *
 *      8  * * *
 *      9  * * *
 *     10  * * *
 *     11  * * *
 *     12  * * *
 *     13  rip.Berkeley.EDU (128.32.131.22)  59 ms !  39 ms !  39 ms !
 *
 * Notice that there are 12 "gateways" (13 is the final
 * destination) and exactly the last half of them are "missing".
 * What's really happening is that rip (a Sun-3 running Sun OS3.5)
 * is using the ttl from our arriving datagram as the ttl in its
 * icmp reply.  So, the reply will time out on the return path
 * (with no notice sent to anyone since icmp's aren't sent for
 * icmp's) until we probe with a ttl that's at least twice the path
 * length.  I.e., rip is really only 7 hops away.  A reply that
 * returns with a ttl of 1 is a clue this problem exists.
 * Traceroute prints a "!" after the time if the ttl is <= 1.
 * Since vendors ship a lot of obsolete (DEC's Ultrix, Sun 3.x) or
 * non-standard (HPUX) software, expect to see this problem
 * frequently and/or take care picking the target host of your
 * probes.
 *
 * Other possible annotations after the time are !H, !N, !P (got a host,
 * network or protocol unreachable, respectively), !S or !F (source
 * route failed or fragmentation needed -- neither of these should
 * ever occur and the associated gateway is busted if you see one).  If
 * almost all the probes result in some kind of unreachable, traceroute
 * will give up and exit.
 *
 * Notes
 * -----
 * This program must be run by root or be setuid.  (I suggest that
 * you *don't* make it setuid -- casual use could result in a lot
 * of unnecessary traffic on our poor, congested nets.)
 *
 * This program requires a kernel mod that does not appear in any
 * system available from Berkeley:  A raw ip socket using proto
 * IPPROTO_RAW must interpret the data sent as an ip datagram (as
 * opposed to data to be wrapped in a ip datagram).  See the README
 * file that came with the source to this program for a description
 * of the mods I made to /sys/netinet/raw_ip.c.  Your mileage may
 * vary.  But, again, ANY 4.x (x < 4) BSD KERNEL WILL HAVE TO BE
 * MODIFIED TO RUN THIS PROGRAM.
 *
 * The udp port usage may appear bizarre (well, ok, it is bizarre).
 * The problem is that an icmp message only contains 8 bytes of
 * data from the original datagram.  8 bytes is the size of a udp
 * header so, if we want to associate replies with the original
 * datagram, the necessary information must be encoded into the
 * udp header (the ip id could be used but there's no way to
 * interlock with the kernel's assignment of ip id's and, anyway,
 * it would have taken a lot more kernel hacking to allow this
 * code to set the ip id).  So, to allow two or more users to
 * use traceroute simultaneously, we use this task's pid as the
 * source port (the high bit is set to move the port number out
 * of the "likely" range).  To keep track of which probe is being
 * replied to (so times and/or hop counts don't get confused by a
 * reply that was delayed in transit), we increment the destination
 * port number before each probe.
 *
 * Don't use this as a coding example.  I was trying to find a
 * routing problem and this code sort-of popped out after 48 hours
 * without sleep.  I was amazed it ever compiled, much less ran.
 *
 * I stole the idea for this program from Steve Deering.  Since
 * the first release, I've learned that had I attended the right
 * IETF working group meetings, I also could have stolen it from Guy
 * Almes or Matt Mathis.  I don't know (or care) who came up with
 * the idea first.  I envy the originators' perspicacity and I'm
 * glad they didn't keep the idea a secret.
 *
 * Tim Seaver, Ken Adelman and C. Philip Wood provided bug fixes and/or
 * enhancements to the original distribution.
 *
 * I've hacked up a round-trip-route version of this that works by
 * sending a loose-source-routed udp datagram through the destination
 * back to yourself.  Unfortunately, SO many gateways botch source
 * routing, the thing is almost worthless.  Maybe one day...
 *
 *  -- Van Jacobson (van@helios.ee.lbl.gov)
 *     Tue Dec 20 03:50:13 PST 1988
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/queue.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip6.h>
#include <netinet/icmp6.h>
#include <netinet/ip_var.h>
#include <netinet/ip6_var.h>
#include <netinet/udp.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <netdb.h>
#include <ctype.h>

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define	MAXPACKET	65535	/* max IPv6 packet size */
#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN	64
#endif

#ifndef FD_SET
#define NFDBITS         (8*sizeof(fd_set))
#define FD_SETSIZE      NFDBITS
#define FD_SET(n, p)    ((p)->fds_bits[(n)/NFDBITS] |= (1 << ((n) % NFDBITS)))
#define FD_CLR(n, p)    ((p)->fds_bits[(n)/NFDBITS] &= ~(1 << ((n) % NFDBITS)))
#define FD_ISSET(n, p)  ((p)->fds_bits[(n)/NFDBITS] & (1 << ((n) % NFDBITS)))
#define FD_ZERO(p)      bzero((char *)(p), sizeof(*(p)))
#endif

#define Fprintf (void)fprintf
#define Sprintf (void)sprintf
#define Printf (void)printf

#define MAX_RTHD	IP6_RT_MAX

/*
 * format of a (udp) probe packet.
 */
struct opacket {
	struct ip6_hdr ip;
	struct udphdr udp;
	u_char seq;		/* sequence number of this packet */
	u_char hlim;		/* hlim packet left with */
	struct timeval tv;	/* time packet left */
};

struct in6_addr gateway[MAX_RTHD + 1];

struct rthd {
	struct ipv6_rthdr rh;
	struct in6_addr ra[MAX_RTHD];
} rthd;
int rthda, rthdlen = 0;

u_char	packet[1024];		/* last inbound (ICMPv6) packet */
struct opacket	*outpacket;	/* last output (UDP) packet */
char *inetname();

int s;				/* receive (ICMPv6) socket file descriptor */
int sndsock;			/* send (UDP) socket file descriptor */
struct timezone tz;		/* leftover */

struct sockaddr_in6 whereto;	/* Who to try to reach */
struct sockaddr_in6 fromwhere;	/* Which source address to use */
struct sockaddr_in6 sinifaddr;	/* interface to use */
#define IFADDR	(&sinifaddr.sin6_addr)

int datalen;		/* How much data */
int mtus[] = {
	32768,
	32748,
	32728,
	17914,
	17894,
	17874,
	16384,
	16364,
	16344,
	 8166,
	 8146,
	 8126,
	 4464,
	 4444,
	 4424,
	 4352,
	 4332,
	 4312,
	 2048,
	 2028,
	 2008,
	 2002,
	 1982,
	 1962,
	 1536,
	 1516,
	 1500,
	 1496,
	 1492,
	 1480,
	 1472,
	 1460,
	 1452,
	 1280,
	    0,
	    0
};
int *mtuptr = &mtus[0];
/* start at 16 Mb token ring, since not all systems like SO_SNDBUF of 65535 */

char source[MAXHOSTNAMELEN + 1];
char hostname[MAXHOSTNAMELEN + 1];
char temp[MAXHOSTNAMELEN + 1];

int nprobes = 3;
int max_hlim = 30;
u_short ident;
u_short port = 32768+666;	/* start udp dest port # for probe packets */
int options;			/* socket options */
int verbose;
int waittime = 5;		/* time to wait for response (in seconds) */
int nflag;			/* print addresses numerically */
int setsndif;			/* set sending interface */

void getsrcfor();

main(argc, argv)
	char *argv[];
{
	extern char *optarg;
	extern int optind;
	struct hostent *hp;
	struct sockaddr_in6 *from, *to;
	struct in6_addr gw;
	int ch, i, on, probe, seq, dsb, flow, hlim, hlim_flag;
 
	on = 1;
	seq = dsb = flow = rthda = 0;
	hlim_flag = 0;
	to = &whereto;
	from = &fromwhere;
	while ((ch = getopt(argc, argv, "dg:i:lm:np:q:rs:t:f:w:v")) != EOF)
		switch(ch) {
		case 'd':
			options |= SO_DEBUG;
			break;
		case 'g':
			if (rthda >= MAX_RTHD)
				errx(1, "too many gateways; max %d", MAX_RTHD);
			hp = getipnodebyname(optarg, AF_INET6, 0, &i);
			if (hp == 0)
				errx(1, "unknown host %s", optarg);
			memcpy(&gw, hp->h_addr, hp->h_length);
			freehostent(hp);
			if (IN6_IS_ADDR_V4MAPPED(&gw))
				errx(1, "no IPv6 address for %s", optarg);
			COPY_ADDR6(gw, gateway[rthda]);
			rthda++;
			if (rthda == 1)
				rthdlen = sizeof (struct ipv6_rthdr);
			rthdlen += sizeof (struct in6_addr);
			break;
		case 'i':
			(void) bzero((char *)&sinifaddr,
				     sizeof(struct sockaddr_in6));
			sinifaddr.sin6_len = sizeof(struct sockaddr_in6);
			sinifaddr.sin6_family = AF_INET6;
			hp = getipnodebyname(optarg, AF_INET6, 0, &i);
			if (hp == 0)
				errx(1, "unknown interface %s", optarg);
			bcopy(hp->h_addr, (caddr_t)IFADDR, hp->h_length);
			freehostent(hp);
			setsndif = 1;
			break;
		case 'l':
			hlim_flag = 1;
			break;
		case 'm':
			max_hlim = atoi(optarg);
			if (max_hlim <= 1) {
				Fprintf(stderr,
				    "traceroute6: max hlim must be >1.\n");
				exit(1);
			}
			break;
		case 'n':
			nflag++;
			break;
		case 'p':
			port = atoi(optarg);
			if (port < 1) {
				Fprintf(stderr,
				    "traceroute6: port must be >0.\n");
				exit(1);
			}
			break;
		case 'q':
			nprobes = atoi(optarg);
			if (nprobes < 1) {
				Fprintf(stderr,
				    "traceroute6: nprobes must be >0.\n");
				exit(1);
			}
			break;
		case 'r':
			options |= SO_DONTROUTE;
			break;
		case 's':
			/*
			 * set the IPv6 source address of the outbound
			 * probe (e.g., on a multi-homed host).
			 */
			strncpy(source, optarg, MAXHOSTNAMELEN);
			break;
		case 't':
			dsb = atoi(optarg);
			if (dsb < 0 || dsb > 255) {
				Fprintf(stderr,
				"traceroute6: DS byte must be 0 to 255.\n");
				exit(1);
			}
			dsb = IPV6_SET_CLASS(dsb);
			break;
		case 'f':
			flow = atoi(optarg);
			if (flow < 0 || flow > 1048575) {
				Fprintf(stderr,
				"traceroute6: flow must be 0 to 1048575.\n");
				exit(1);
			}
			flow = IPV6_SET_FLOWLABEL(flow);
			break;
		case 'v':
			verbose++;
			break;
		case 'w':
			waittime = atoi(optarg);
			if (waittime <= 1) {
				Fprintf(stderr,
				    "traceroute6: wait must be >1 sec.\n");
				exit(1);
			}
			break;
		default:
			usage();
		}
	argc -= optind;
	argv += optind;

	if (argc < 1) 
		usage();

	setlinebuf (stdout);

	if (*source == '\0') {
		if (rthdlen)
			getsrcfor(inet_ntop(AF_INET6, &gateway[0],
					    temp, MAXHOSTNAMELEN));
		else
			getsrcfor(*argv);
	}

	(void) bzero((char *)&fromwhere, sizeof(struct sockaddr_in6));
#ifdef SIN6_LEN
	from->sin6_len = sizeof(struct sockaddr_in6);
#endif
	hp = getipnodebyname(source, AF_INET6, 0, &i);
	if (hp == 0)
		errx(1, "unknown source %s", source);
	from->sin6_family = hp->h_addrtype;
	bcopy(hp->h_addr, (caddr_t)&from->sin6_addr, hp->h_length);
	strncpy(source, hp->h_name, MAXHOSTNAMELEN);
	freehostent(hp);
	if ((from->sin6_family != AF_INET6) ||
	    IN6_IS_ADDR_V4MAPPED(&from->sin6_addr))
		errx(1, "no IPv6 address for source %s", source);

	(void) bzero((char *)&whereto, sizeof(struct sockaddr_in6));
#ifdef SIN6_LEN
	to->sin6_len = sizeof(struct sockaddr_in6);
#endif
	hp = getipnodebyname(*argv, AF_INET6, 0, &i);
	if (hp == 0)
		errx(1, "unknown host %s", *argv);
	to->sin6_family = hp->h_addrtype;
	bcopy(hp->h_addr, (caddr_t)&to->sin6_addr, hp->h_length);
	strncpy(hostname, hp->h_name, MAXHOSTNAMELEN);
	freehostent(hp);
	if ((to->sin6_family != AF_INET6) ||
	    IN6_IS_ADDR_V4MAPPED(&to->sin6_addr))
		errx(1, "no IPv6 address for host %s", *argv);

	if (*++argv) {
		datalen = atoi(*argv);
		while(datalen + sizeof(struct opacket) <= *mtuptr)
			mtuptr++;
	} else
		datalen = (*mtuptr++) - sizeof(struct opacket);

	if (datalen < 0 || datalen >= MAXPACKET - sizeof(struct opacket)) {
		Fprintf(stderr,
		    "traceroute6: packet size must be 0 <= s < %ld.\n",
		    MAXPACKET - sizeof(struct opacket));
		exit(1);
	}
	datalen += sizeof(struct opacket);
	outpacket = (struct opacket *)malloc((unsigned)datalen);
	if (! outpacket) {
		perror("traceroute6: malloc");
		exit(1);
	}
	(void) bzero((char *)outpacket, datalen);
	outpacket->ip.ip6_flow = flow | dsb | IPV6_VERSION;
	outpacket->ip.ip6_nxt = rthdlen ? IP6_NHDR_RT : IPPROTO_UDP;
	COPY_ADDR6(from->sin6_addr, outpacket->ip.ip6_src);
	COPY_ADDR6(to->sin6_addr, outpacket->ip.ip6_dst);

	ident = (getpid() & 0xffff) | 0x8000;

	if (rthdlen != 0) {
		memset(&rthd, 0, sizeof(rthd));
		rthd.rh.ir6_nh = IPPROTO_UDP;
		rthd.rh.ir6_hlen = rthda << 1;
		rthd.rh.ir6_type = IP6_LSRRT;
		rthd.rh.ir6_sglt = rthda;
		for (i = 1; i <= rthda; i++)
			COPY_ADDR6(gateway[i], rthd.ra[i - 1]);
		COPY_ADDR6(whereto.sin6_addr, rthd.ra[rthda - 1]);
	}

	if ((s = socket(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6)) < 0) {
		perror("traceroute6: ICMPv6 socket");
		exit(5);
	}
	if (options & SO_DEBUG)
		(void) setsockopt(s, SOL_SOCKET, SO_DEBUG,
				  (char *)&on, sizeof(on));
	if (options & SO_DONTROUTE)
		(void) setsockopt(s, SOL_SOCKET, SO_DONTROUTE,
				  (char *)&on, sizeof(on));

	if ((sndsock = socket(AF_INET6, SOCK_RAW, IPPROTO_RAW)) < 0) {
		perror("traceroute6: raw socket");
		exit(5);
	}
#ifdef SO_SNDBUF
	if (setsockopt(sndsock, SOL_SOCKET, SO_SNDBUF, (char *)&datalen,
		       sizeof(datalen)) < 0) {
		perror("traceroute6: SO_SNDBUF");
		exit(6);
	}
#endif SO_SNDBUF
	if (setsockopt(sndsock, IPPROTO_IP, IP_HDRINCL, (char *)&on,
		       sizeof(on)) < 0) {
		perror("traceroute6: IP_HDRINCL");
		exit(6);
	}
	if (options & SO_DEBUG)
		(void) setsockopt(sndsock, SOL_SOCKET, SO_DEBUG,
				  (char *)&on, sizeof(on));
	if (options & SO_DONTROUTE)
		(void) setsockopt(sndsock, SOL_SOCKET, SO_DONTROUTE,
				  (char *)&on, sizeof(on));
	if (setsndif) {
		if ((to->sin6_addr.s6_addr[0] == 0xff) &&
		    (setsockopt(sndsock, IPPROTO_IPV6,
				IPV6_MULTICAST_IF, IFADDR,
				sizeof(struct in6_addr)) < 0)) {
			perror("traceroute6: IPV6_MULTICAST_IF");
			exit(6);
		} else if (setsockopt(sndsock, IPPROTO_IPV6,
				      IPV6_SENDIF, &sinifaddr,
				      sinifaddr.sin6_len) < 0) {
			perror("traceroute6: IPV6_SENDIF");
			exit(6);
		}
	}

	Fprintf(stderr, "traceroute6 to %s (%s)", hostname,
		inet_ntop(AF_INET6, &to->sin6_addr, temp, MAXHOSTNAMELEN));
	Fprintf(stderr, " from %s (%s)", source,
		inet_ntop(AF_INET6, &from->sin6_addr, temp, MAXHOSTNAMELEN));
	Fprintf(stderr, ", %d hops max\n", max_hlim);
	(void) fflush(stderr);

	for (hlim = 1; hlim <= max_hlim; ++hlim) {
		struct in6_addr lastaddr;
		int got_there = 0;
		int unreachable = 0;

		CLR_ADDR6(lastaddr);
	again:
		Printf("%2d ", hlim);
		for (probe = 0; probe < nprobes; ++probe) {
			int cc;
			struct timeval tv;
			struct ip6_hdr *ip;

			(void) gettimeofday(&tv, &tz);
			send_probe(++seq, hlim);
			while (cc = wait_for_reply(s, from, &tv)) {
				if ((i = packet_ok(packet, cc, from, seq))) {
					int dt = deltaT(&tv);
					if (!SAME_ADDR6(from->sin6_addr, lastaddr)) {
						print(packet, cc, from);
						COPY_ADDR6(from->sin6_addr, lastaddr);
					}
					ip = (struct ip6_hdr *)packet;
					Printf("  %d ms", dt);
					if (hlim_flag)
						Printf(" (ttl = %d)",
						       ip->ip6_hlim);
					switch(i) {
#define UNREACH_NOROUTE	((ICMP6_DST_UNREACH << 8) | ICMP6_DST_UNREACH_NOROUTE)
					case UNREACH_NOROUTE:
						++unreachable;
						Printf(" !N");
						break;
#define UNREACH_ADMIN	((ICMP6_DST_UNREACH << 8) | ICMP6_DST_UNREACH_ADMIN)
					case UNREACH_ADMIN:
						++unreachable;
						Printf(" !A");
						break;
#define UNREACH_BYDSCP	((ICMP6_DST_UNREACH << 8) | ICMP6_DST_UNREACH_BEYONDSCOPE)
					case UNREACH_BYDSCP:
						++unreachable;
						Printf(" !S");
						break;
#define UNREACH_ADDRESS	((ICMP6_DST_UNREACH << 8) | ICMP6_DST_UNREACH_ADDR)
					case UNREACH_ADDRESS:
						++unreachable;
						Printf(" !H");
						break;
#define UNREACH_PORT	((ICMP6_DST_UNREACH << 8) | ICMP6_DST_UNREACH_NOPORT)
					case UNREACH_PORT:
						if (ip->ip6_hlim <= 1)
							Printf(" !");
						++got_there;
						break;
#define PKTTOOBIG	(ICMP6_PACKET_TOO_BIG << 8)
					case PKTTOOBIG:
						frag_err(&from);
						goto again;
#define PARAMPROB_NH	((ICMP6_PARAM_PROB << 8) | ICMP6_PARAMPROB_NEXTHEADER)
					case PARAMPROB_NH:
						++got_there;
						Printf(" !P");
						break;
					}
					break;
				}
			}
			if (cc == 0)
				Printf(" *");
			(void) fflush(stdout);
		}
		putchar('\n');
		if (got_there || unreachable >= nprobes-1)
			exit(0);
	}
}

wait_for_reply(sock, from, tp)
	int sock;
	struct sockaddr_in6 *from;
	struct timeval *tp;
{
	fd_set fds;
	struct timeval now, wait;
	int cc = 0;
	int fromlen = sizeof (*from);

	FD_ZERO(&fds);
	FD_SET(sock, &fds);
	wait.tv_sec = tp->tv_sec + waittime;
	wait.tv_usec = tp->tv_usec;
	(void)gettimeofday(&now, &tz);
	tvsub(&wait, &now);
	if (wait.tv_sec < 0) {
		wait.tv_sec = 0;
		wait.tv_usec = 1;
	}

	if (select(sock+1, &fds, (fd_set *)0, (fd_set *)0, &wait) > 0) {
		if (verbose)
			(void) bzero((char *)packet, sizeof(packet));
		cc = recvfrom(s, (char *)packet, sizeof(packet), 0,
			      (struct sockaddr *)from, &fromlen);
	}
	return(cc);
}

/*
 * Received ICMPv6 packet too big.
 */

long nextmtu;	/* from ICMPv6 error, set by packet_ok(), might be 0 */

frag_err(from)
struct sockaddr_in6 *from;
{
	int i;

	if (nextmtu > 0) {
		Printf("\nICMPv6 packet too big, next hop MTU = %d\n",
		       nextmtu);
		datalen = nextmtu;
		for (i = 0; mtus[i] > 0; i++) {
			if (mtus[i] == nextmtu) {
				mtuptr = &mtus[i+1];	/* next one to try */
				return;
			}
		}
		Printf("(MTU not in table!)\n");
		
	} else {
		datalen = *mtuptr++;
		Printf("\nfragmentation required, trying new MTU = %d\n",
			datalen);
	}
	if (*mtuptr == 0) {
		Fprintf(stderr, "\ngets zero MTU ?!\n");
		exit(1);
	}
}

send_probe(seq, hlim)
{
	struct opacket *op = outpacket;
	struct ip6_hdr *ip = &op->ip;
	struct udphdr *up = &op->udp;
	int i;
	struct ip6ovck save_ipc, *ipc;
	static int firsttime = 1;

again:
	ip->ip6_plen = htons((u_short)(datalen - sizeof(struct ip6_hdr)));
	ip->ip6_hlim = hlim;
	up->uh_sport = htons(ident);
	up->uh_dport = htons(port+seq);
	up->uh_ulen = htons(ntohs(ip->ip6_plen) - rthdlen);
	op->seq = seq;
	op->hlim = hlim;
	(void) gettimeofday(&op->tv, &tz);

	ipc = (struct ip6ovck *)ip;
	save_ipc = *ipc;
	ipc->ih6_wrd1 = ipc->ih6_wrd0 = 0;
	ipc->ih6_pr = IPPROTO_UDP;
	ipc->ih6_len = up->uh_ulen;
	if (rthdlen)
		COPY_ADDR6(whereto.sin6_addr, ip->ip6_dst);
	up->uh_sum = 0;
	up->uh_sum = in_cksum((u_short *)op, datalen);
	if (up->uh_sum == 0)
		up->uh_sum = 0xffff;
	*ipc = save_ipc;

	if (rthdlen) {
		memmove((char *)&op->udp + rthdlen,
			&op->udp,
			datalen - rthdlen - sizeof(struct ip6_hdr));
		memcpy(&op->udp, &rthd, rthdlen);
		COPY_ADDR6(gateway[rthda - 1], ip->ip6_dst);
	}

	i = sendto(sndsock, (char *)outpacket, datalen, 0,
		   (struct sockaddr *)&whereto, sizeof(struct sockaddr_in6));
	if (i < 0 || i != datalen)  {
		if (i < 0) {
			/*
			 * An errno of EMSGSIZE means we're writing too big a
			 * datagram for the interface.  We have to just decrease
			 * the packet size until we find one that works.
			 */
			if (errno == EMSGSIZE) {
				if (verbose)
				   Fprintf(stderr, "EMSGSIZE for %d\n", datalen);
				datalen = *mtuptr++;
				if (*mtuptr == 0) {
					Fprintf(stderr, "gets zero MTU ?!\n");
					exit(1);
				}
				goto again;
			}
			perror("sendto");
		}
		Printf("traceroute6: wrote %s %d chars, ret=%d\n", hostname,
			datalen, i);
		(void) fflush(stdout);
	}
	if (firsttime) {
		Fprintf(stderr, "outgoing MTU = %d\n", datalen);
		firsttime = 0;
	}
}


deltaT(tp)
	struct timeval *tp;
{
	struct timeval tv;

	(void) gettimeofday(&tv, &tz);
	tvsub(&tv, tp);
	return (tv.tv_sec*1000 + (tv.tv_usec + 500)/1000);
}


/*
 * Convert an ICMPv6 "type" field to a printable string.
 */
char *
pr_type(t)
	u_char t;
{
	static char *ttab[] = {
	"ICMP 0",	"Unreachable",	"Too Big",	"Time Exceeded",
	"Param Problem","IMCP 5",	"ICMP 6",	"ICMP 7",
	"ICMP 8",	"ICMP 9",	"ICMP 10",	"ICMP 11",
	"ICMP 12",	"ICMP 13",	"ICMP 14",	"ICMP 15",
	"ICMP 16",	"ICMP 17",	"ICMP 18",	"ICMP 19",
	"ICMP 20",	"ICMP 21",	"ICMP 22",	"ICMP 23",
	"ICMP 24",	"ICMP 25",	"ICMP 26",	"ICMP 27",
	"ICMP 28",	"ICMP 29",	"ICMP 30",	"ICMP 31",
	"ICMP 32",	"ICMP 33",	"ICMP 34",	"ICMP 35",
	"ICMP 36",	"ICMP 37",	"ICMP 38",	"ICMP 39",
	"ICMP 40",	"ICMP 41.",	"ICMP 42",	"ICMP 43",
	"ICMP 44",	"ICMP 45",	"ICMP 46",	"ICMP 47",
	"ICMP 48",	"ICMP 49",	"ICMP 50",	"ICMP 51",
	"ICMP 52",	"ICMP 53",	"ICMP 54",	"ICMP 55",
	"ICMP 56",	"ICMP 57",	"ICMP 58",	"ICMP 59",
	"ICMP 60",	"ICMP 61",	"ICMP 62",	"ICMP 63",
	"ICMP 64",	"ICMP 65",	"ICMP 66",	"ICMP 67",
	"ICMP 68",	"ICMP 69",	"ICMP 70",	"ICMP 71",
	"ICMP 72",	"ICMP 73",	"ICMP 74",	"ICMP 75",
	"ICMP 76",	"ICMP 77",	"ICMP 78",	"ICMP 79",
	"ICMP 80",	"ICMP 81",	"ICMP 82",	"ICMP 83",
	"ICMP 84",	"ICMP 85",	"ICMP 86",	"ICMP 87",
	"ICMP 88",	"ICMP 89",	"ICMP 90",	"ICMP 91",
	"ICMP 92",	"ICMP 93",	"ICMP 94",	"ICMP 95",
	"ICMP 96",	"ICMP 97",	"ICMP 98",	"ICMP 99",
	"ICMP 100",	"ICMP 101",	"ICMP 102",	"ICMP 103",
	"ICMP 104",	"ICMP 105",	"ICMP 106",	"ICMP 107",
	"ICMP 108",	"ICMP 109",	"ICMP 110",	"ICMP 111",
	"ICMP 112",	"ICMP 113",	"ICMP 114",	"ICMP 115",
	"ICMP 116",	"ICMP 117",	"ICMP 118",	"ICMP 119",
	"ICMP 120",	"ICMP 121",	"ICMP 122",	"ICMP 123",
	"ICMP 124",	"ICMP 125",	"ICMP 126",	"ICMP 127",
	"Echo Request",	"Echo Reply",	"Group Query",	"Group Report",
	"Group Term",	"Router Sol",	"Router Adv",	"Neighbor Sol",
	"Neighbor Adv",	"Redirect",	"Router Renum",	"ICMP 139",
	"ICMP 140",	"ICMP 141",	"ICMP 142",	"ICMP 143",
	"ICMP 144",	"ICMP 145",	"ICMP 146",	"ICMP 147",
	"ICMP 148",	"ICMP 149",	"ICMP 150",	"ICMP 151",
	"ICMP 152",	"ICMP 153",	"ICMP 154",	"ICMP 155",
	"ICMP 156",	"ICMP 157",	"ICMP 158",	"ICMP 159",
	"ICMP 160",	"ICMP 161",	"ICMP 162",	"ICMP 163",
	"ICMP 164",	"ICMP 165",	"ICMP 166",	"ICMP 167",
	"ICMP 168",	"ICMP 169",	"ICMP 170",	"ICMP 171",
	"ICMP 172",	"ICMP 173",	"ICMP 174",	"ICMP 175",
	"ICMP 176",	"ICMP 177",	"ICMP 178",	"ICMP 179",
	"ICMP 180",	"ICMP 181",	"ICMP 182",	"ICMP 183",
	"ICMP 184",	"ICMP 185",	"ICMP 186",	"ICMP 187",
	"ICMP 188",	"ICMP 189",	"ICMP 190",	"ICMP 191",
	"ICMP 192",	"ICMP 193",	"ICMP 194",	"ICMP 195",
	"ICMP 196",	"ICMP 197",	"ICMP 198",	"ICMP 199",
	"ICMP 200",	"ICMP 201",	"ICMP 202",	"ICMP 203",
	"ICMP 204",	"ICMP 205",	"ICMP 206",	"ICMP 207",
	"ICMP 208",	"ICMP 209",	"ICMP 210",	"ICMP 211",
	"ICMP 212",	"ICMP 213",	"ICMP 214",	"ICMP 215",
	"ICMP 216",	"ICMP 217",	"ICMP 218",	"ICMP 219",
	"ICMP 220",	"ICMP 221",	"ICMP 222",	"ICMP 223",
	"ICMP 224",	"ICMP 225",	"ICMP 226",	"ICMP 227",
	"ICMP 228",	"ICMP 229",	"ICMP 230",	"ICMP 231",
	"ICMP 232",	"ICMP 233",	"ICMP 234",	"ICMP 235",
	"ICMP 236",	"ICMP 237",	"ICMP 238",	"ICMP 239",
	"ICMP 240",	"ICMP 241",	"ICMP 242",	"ICMP 243",
	"ICMP 244",	"ICMP 245",	"ICMP 246",	"ICMP 247",
	"ICMP 248",	"ICMP 249",	"ICMP 250",	"ICMP 251",
	"ICMP 252",	"ICMP 253",	"ICMP 254",	"ICMP 255",
	};
	return(ttab[t]);
}


packet_ok(buf, cc, from, seq)
	u_char *buf;
	int cc;
	struct sockaddr_in6 *from;
	int seq;
{
	register struct icmp6_hdr *icp;
	u_char type, code;
	int hlen;
	struct ip6_hdr *ip;

	ip = (struct ip6_hdr *) buf;
	hlen = sizeof(struct ip6_hdr);
	if (cc < hlen + sizeof(*icp)) {
		if (verbose)
			Printf("packet too short (%d bytes) from %s\n", cc,
			       inet_ntop(AF_INET6, &from->sin6_addr,
					 temp, MAXHOSTNAMELEN));
		return (0);
	}
	cc -= hlen;
	icp = (struct icmp6_hdr *)(buf + hlen);
	type = icp->icmp6_type; code = icp->icmp6_code;
	if ((type == ICMP6_TIME_EXCEEDED &&
	     code == ICMP6_TIME_EXCEED_TRANSIT) ||
	    (type == ICMP6_PARAM_PROB && code == ICMP6_PARAMPROB_NEXTHEADER) ||
	    type == ICMP6_DST_UNREACH || type == ICMP6_PACKET_TOO_BIG) {
		struct ip6_hdr *hip;
		struct udphdr *up;

		nextmtu = htonl(icp->icmp6_mtu); /* for frag_err() */
		hip = (struct ip6_hdr *)(icp + 1);
		hlen = sizeof(struct ip6_hdr);
		if (rthdlen && hip->ip6_nxt == IP6_NHDR_RT)
			hlen += rthdlen;
		up = (struct udphdr *)((u_char *)hip + hlen);
		if (hlen + 12 <= cc &&
		    (hip->ip6_nxt == IPPROTO_UDP ||
		     *((u_char *)hip + sizeof(struct ip6_hdr)) ==
			IPPROTO_UDP) &&
		    up->uh_sport == htons(ident) &&
		    up->uh_dport == htons(port+seq)) {
			if (type == ICMP6_TIME_EXCEEDED)
				return (-1);
			else
				return ((type << 8) | code);
		}
	}
	if (verbose) {
		int i;
		u_long *lp = (u_long *)(icp + 1);

		Printf("\n%d bytes from %s to ", cc,
		       inet_ntop(AF_INET6, &from->sin6_addr,
				 temp, MAXHOSTNAMELEN));
		Printf("%s: ICMPv6 type %d (%s) code %d\n", 
		       inet_ntop(AF_INET6, &ip->ip6_dst,
				 temp, MAXHOSTNAMELEN),
		       type, pr_type(type), icp->icmp6_code);
		cc -= sizeof(*icp);
		for (i = 0; i < cc ; i += sizeof(long))
			Printf("%2d: 0x%8.8lx\n", i, *lp++);
	}
	return(0);
}


print(buf, cc, from)
	u_char *buf;
	int cc;
	struct sockaddr_in6 *from;
{
	struct ip6_hdr *ip;
	int hlen;

	ip = (struct ip6_hdr *) buf;
	hlen = sizeof(struct ip6_hdr);
	cc -= hlen;

	if (nflag)
		Printf(" %s",
		       inet_ntop(AF_INET6, &from->sin6_addr,
				 temp, MAXHOSTNAMELEN));
	else
		Printf(" %s (%s)", inetname(&from->sin6_addr),
		       inet_ntop(AF_INET6, &from->sin6_addr,
				 temp, MAXHOSTNAMELEN));

	if (verbose)
		Printf(" %d bytes to %s", cc,
		       inet_ntop(AF_INET6, &ip->ip6_dst,
				 temp, MAXHOSTNAMELEN));
}


/*
 * Checksum routine for Internet Protocol family headers (C Version)
 */
in_cksum(addr, len)
u_short *addr;
int len;
{
	register int nleft = len;
	register u_short *w = addr;
	register u_short answer;
	register int sum = 0;

	/*
	 *  Our algorithm is simple, using a 32 bit accumulator (sum),
	 *  we add sequential 16 bit words to it, and at the end, fold
	 *  back all the carry bits from the top 16 bits into the lower
	 *  16 bits.
	 */
	while (nleft > 1)  {
		sum += *w++;
		nleft -= 2;
	}

	/* mop up an odd byte, if necessary */
	if (nleft == 1)
		sum += *(u_char *)w;

	/*
	 * add back carry outs from top 16 bits to low 16 bits
	 */
	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum;				/* truncate to 16 bits */
	return (answer);
}

/*
 * Subtract 2 timeval structs:  out = out - in.
 * Out is assumed to be >= in.
 */
tvsub(out, in)
register struct timeval *out, *in;
{
	if ((out->tv_usec -= in->tv_usec) < 0)   {
		out->tv_sec--;
		out->tv_usec += 1000000;
	}
	out->tv_sec -= in->tv_sec;
}


/*
 * Construct an IPv6 address representation.
 * If the nflag has been supplied, give 
 * numeric value, otherwise try for symbolic name.
 */
char *
inetname(in)
	struct in6_addr *in;
{
	register char *cp;
	static char line[50];
	struct hostent *hp;
	static char domain[MAXHOSTNAMELEN + 1];
	static int first = 1;
	int er;

	if (first && !nflag) {
		first = 0;
		if (gethostname(domain, MAXHOSTNAMELEN) == 0 &&
		    (cp = index(domain, '.'))) {
			(void) strncpy(domain, cp + 1, sizeof(domain));
			domain[sizeof(domain) - 1] = '\0';
		} else
			domain[0] = 0;
	}
	hp = 0;
	if (!nflag && !IS_ANYADDR6(*in)) {
		hp = getipnodebyaddr(in, sizeof (*in), AF_INET6, &er);
		if (hp) {
			if ((cp = index(hp->h_name, '.')) &&
			    !strcmp(cp + 1, domain))
				*cp = 0;
		}
	}
	if (hp) {
		(void) strncpy(line, hp->h_name, sizeof(line));
		line[sizeof(line) - 1] = '\0';
		freehostent(hp);
	} else
		(void) inet_ntop(AF_INET6, in, line, sizeof(line));
	return (line);
}

void
getsrcfor(destname)
	char *destname;
{
	struct sockaddr_in6 sin;
	int er, len = sizeof(sin);
	register struct hostent *hp;

	if ((s = socket(AF_INET6, SOCK_DGRAM, 0)) < 0)
		err(1, "UDP socket");

	memset(&sin, 0, sizeof(sin));
#ifdef SIN6_LEN
	sin.sin6_len = sizeof(struct sockaddr_in6);
#endif
	hp = getipnodebyname(destname, AF_INET6, 0, &er);
	if (hp == 0)
		errx(1, "unknown host: %s", destname);
	sin.sin6_family = hp->h_addrtype;
	memcpy(&sin.sin6_addr, hp->h_addr, hp->h_length);
	sin.sin6_port = htons(NAMESERVER_PORT);
	freehostent(hp);

	if ((sin.sin6_family != AF_INET6) ||
	    IN6_IS_ADDR_V4MAPPED(&sin.sin6_addr))
		errx(1, "no IPv6 address for: %s", destname);

	printf("trying to get source for %s\n", destname);
	if (connect(s, (struct sockaddr *)&sin, len) < 0)
		err(1, "UDP connect");
	memset(&sin, 0, sizeof(sin));
	if (getsockname(s, (struct sockaddr *)&sin, &len) < 0)
		err(1, "getsockname");
	(void)close(s);
	if ((sin.sin6_family != AF_INET6) ||
	    (len != sizeof(struct sockaddr_in6)))
		err(1, "getsockname returns garbage");
	(void) inet_ntop(AF_INET6, &sin.sin6_addr, source, MAXHOSTNAMELEN);
	printf("source should be %s\n", source);
}

usage()
{
	(void)fprintf(stderr, 
"usage: traceroute6 [-dnrv] [-g gateway_addr] ... [-m max_hlim] [-p port#]\n\t\
[-q nqueries] [-i if_addr] [-s src_addr] [-t pri] [-f flow]\n\t[-w wait] host [data size]\n");
	exit(1);
}
