/* Send arbritary ICMPv6 packets */

#include <sys/param.h>
#include <sys/queue.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/time.h>
#include <signal.h>

#include <net/if.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip6.h>
#include <netinet/icmp6.h>
#include <netinet/ip_var.h>
#include <netinet/ip6_var.h>
#include <netinet/ip6_opts.h>
#include <netinet/if_ether.h>
#if !defined(__NetBSD__)
#include <netinet/if_ether6.h>
#endif
#include <netinet/if_ndp6.h>
#include <netinet/udp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <sys/sysctl.h>
#include <sys/socket.h>
#include <net/route.h>
#include <net/if_dl.h>


char packet[8192];
char option[2048];
char token[128];
char *cp, *op;
int len, olen, droplen;
struct sockaddr_in6 dst = { sizeof(dst), AF_INET6 };
int s;
struct ip6ovck save_ip;
u_int indif = 0;
int proto = -1;
int isnh = -1;

struct ip6_hdr *ip;
struct ip6ovck *iop;
struct icmp6_hdr *icmp;
struct udphdr *udp;
struct ipv6_h2hhdr *hhp;
struct ipv6_rthdr *rtp;
struct ipv6_fraghdr *frp;
struct ipv6_authhdr *ahp;
struct ipv6_dopthdr *dop;
struct nd_opt_hdr *lap;
struct nd_opt_prefix_info *pfp;
struct nd_opt_mtu *mtup;
struct nd_opt_advit *aitp;
struct nd_opt_agti *haip;

#define TOKEN_IS(x)	(strcasecmp(token, x) == 0)

int main __P((int, char *[]));
int getindex __P((char *));
void dump __P((u_char *, int));
int Mgetchar __P((void));
int getoken __P((int));
u_int getint __P((void));
u_int getudpport __P((void));
void gethexa __P((int));
int parse_escape  __P((void));
void getipv6 __P((void));
void geth2h __P((void));
void getrt __P((void));
void getfrg __P((void));
void getdopt __P((void));
void getah __P((void));
void getnone __P((void));
void geticmpv6 __P((void));
void getudp __P((void));
void getlladdr __P((int));
void getprefinfo __P((void));
void getmtu __P((void));
void getadvit __P((void));
void getagti __P((void));
void getpacket __P((void));
int in_cksum __P((u_short *, int));

#if ! BYTE_ORDER
Must know byte order
#endif
#if BYTE_ORDER == BIG_ENDIAN
u_int htoll __P((u_int));
u_short htols __P((u_short));
inline u_int htoll(x)
register u_int x;
{
	return (x>>24)|((x&0xff)<<24)|((x&0xff0000)>>8)|((x&0xff00)<<8);
}
inline u_short htols(x)
register u_short x;
{
	return (x>>8)|((x&0xff)<<8);
}
#endif
#if BYTE_ORDER == LITTLE_ENDIAN
#define htoll(x)        (x)
#define htols(x)        (x)
#endif

int
Mgetchar()
{
	int c = getchar();

	if (c == '#') {
		while (((c = getchar()) != EOF) &&
		       (c != '\n') &&
		       (c != '\r'))
			;
	}
	return c;
}

int
getoken(term)
	int term;
{
	register int ch, state = 0;
	register char *p;

	bzero(token, 128);
	p = token;
	for (;;) switch (ch = Mgetchar()) {
	    case EOF:
		if (state != 0) {
			fprintf(stderr, "%sEOF?\n", token);
			exit(1);
		}
		return ch;

	    case ' ':
	    case '\t':
	    case '\n':
	    case '\r':
		if (state == 1) {
			if (term == 0)
				return ch;
			else
				state = 2;
		}
		break;
				
	    case '=':
		if ((term == 0) || ((state != 1) && (state != 2))) {
			fprintf(stderr, "bad '=' in %s\n", token);
			exit(1);
		}
		if (term == 1)
			return ch;
		else
			state = 3;
		break;

	    case '{':
		if ((term != 2) || ((state != 2) && (state != 3))) {
			 fprintf(stderr, "bad '{' in %s\n", token);
			 exit(1);
		}
		return ch;
		break;

	    case '}':
		if (state != 0) {
			fprintf(stderr, "bad '}' in %s\n", token);
			exit(1);
		}
		*p++ = ch;
		return ch;

	    case '-': case '.': case ':':
	    case '0': case '1': case '2': case '3': case '4': case '5':
	    case '6': case '7': case '8': case '9':
	    case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
	    case 'G': case 'H': case 'I': case 'J': case 'K': case 'L':
	    case 'M': case 'N': case 'O': case 'P': case 'Q': case 'R':
	    case 'S': case 'T': case 'U': case 'V': case 'W': case 'X':
	    case 'Y': case 'Z':
	    case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
	    case 'g': case 'h': case 'i': case 'j': case 'k': case 'l':
	    case 'm': case 'n': case 'o': case 'p': case 'q': case 'r':
	    case 's': case 't': case 'u': case 'v': case 'w': case 'x':
	    case 'y': case 'z':
		if (state > 1) {
			fprintf(stderr, "bad character %s??%c\n", token, ch);
			exit(1);
		}
		state = 1;
		*p++ = ch;
		break;

	    default:
		fprintf(stderr, "bad character %s??%c\n", token, ch);
		exit(1);
	}
}

u_int
getint()
{
	u_int v;
	char *tokend;

	if (getoken(0) == EOF) {
		fprintf(stderr, "EOF in integer\n");
		exit(1);
	}
	v = strtoul(token, &tokend, 0);
	if ((tokend > token) &&
	    (tokend < (token + 128)) &&
	    (*tokend == '\0'))
		return v;
	fprintf(stderr, "bad number %s\n", token);
	exit(1);
}

u_int
getudpport()
{
	u_int v;
	char *tokend;
	struct servent *sp ;

	if (getoken(0) == EOF) {
		fprintf(stderr, "EOF in port\n");
		exit(1);
	}
	sp = getservbyname(token, "udp") ;     /* find daemon's port */
	if (sp)
		return sp->s_port;
	v = strtoul(token, &tokend, 0);
	if ((tokend > token) &&
	    (tokend < (token + 128)) &&
	    (v <= 0xffff) &&
	    (*tokend == '\0'))
		return v;
	fprintf(stderr, "bad udp port %s\n", token);
	exit(1);
}

void
gethexa(skip)
{
	register int ch, valid = 0;
	register u_int v = 0;
	register int base = 16, size = 1, machorder = 0;
	register struct hostent *hp;
	int er;

	for (;;) switch (ch = Mgetchar()) {
	    case EOF:
		fprintf(stderr, "EOF?\n");
		exit(1);

	    case '{':
		if (skip) {
			skip = 0;
			break;
		}
		goto err;

	    case '}':
	    case ' ':
	    case '\t':
	    case '\n':
	    case '\r':
		if (valid) {
			if (size == 1)
				*cp++ = v;
			else if (size == 2) {
				u_short vv = v;
				char *p = (char *)&vv;

				if (machorder == 0)
					vv = htons(vv);
				else if (machorder == 2)
					vv = htols(vv);
				*cp++ = *p++;
				*cp++ = *p++;
			} else {
				u_int vv = v;
				char *p = (char *)&vv;

				if (machorder == 0)
					vv = htonl(vv);
				else if (machorder == 2)
					vv = htoll(vv);
				*cp++ = *p++;
				*cp++ = *p++;
				*cp++ = *p++;
				*cp++ = *p++;
			}
		}
		if (ch == '}')
			return;
		valid = 0;
		v = 0;
		base = 16;
		size = 1;
		machorder = 0;
		break;

	    case '0':
	    case '1':
	    case '2':
	    case '3':
	    case '4':
	    case '5':
	    case '6':
	    case '7':
	    case '8':
	    case '9':
		ch -= '0';
		goto mul;

	    case 'A':
	    case 'B':
	    case 'C':
	    case 'D':
	    case 'E':
	    case 'F':
		ch -= 'A' - 10;
		goto mul;

	    case 'a':
	    case 'b':
	    case 'c':
	    case 'd':
	    case 'e':
	    case 'f':
		ch -= 'a' - 10;
	mul:
		if (valid > 1 || ch >= base)
			goto err;
		valid = 1;
		v = v * base + ch;
		break;

	    case 'o':
	    case 'O':
		base = 8;
		goto setbase;

	    case 'x':
	    case 'X':
		base = 16;
		goto setbase;

	    case 't':
	    case 'T':
		base = 10;
	setbase:
		if (valid != 1 || v)
			goto err;
		break;
		
	    case 'l':
	    case 'L':
		size = 4;
		goto setsize;

	    case 's':
	    case 'S':
		size = 2;
	setsize:
		if (valid != 1)
			goto err;
		valid = 2;
		break;

	    case 'h':
	    case 'H':
		machorder = 1;
		goto setorder;

	    case 'r':
	    case 'R':
		machorder = 2;
	setorder:
		if (valid != 2)
			goto err;
		valid = 3;
		break;

	    case '\'':
	    case '"':
		v = ch;
		while (1) {
			ch = getchar();
			if (ch == '\n' || ch == EOF)
				goto err;
			else if (ch == '\\')
				ch = parse_escape();
			else if (ch == v)
				break;
			*cp++ = ch;
		}
		break;

	    case '@':
		if (getoken(0) == EOF) {
			fprintf(stderr, "EOF in hexa field\n");
			exit(1);
		}
		hp = getipnodebyname(token, AF_INET6, 0, &er);
		if (hp == NULL) {
			fprintf(stderr, "bad address %s\n", token);
			exit(1);
		}
		bcopy(hp->h_addr, cp, hp->h_length);
		cp += hp->h_length;
		freehostent(hp);
		break;

	    default:
	    err:
		fprintf(stderr, "bad character %c (%d)\n", ch, (u_int)ch);
		exit(1);
	}
}

int
parse_escape ()
{
	register int c = getchar();

	switch (c) {
	    case 'a':
		return '\a';
	    case 'b':
		return '\b';
	    case 'e':
	    case 'E':
		return '\e';
	    case 'f':
		return '\f';
	    case 'n':
		return '\n';
	    case 'r':
		return '\r';
	    case 't':
		return '\t';
	    case 'v':
		return '\v';
	    case '\n':
	    case EOF:
		fprintf(stderr, "bad character %c (%d)\n", c, (u_int)c);
		exit(1);

	    case '0':
	    case '1':
	    case '2':
	    case '3':
	    case '4':
	    case '5':
	    case '6':
	    case '7':
		{
		register int i = c - '0';
		register int count = 0;

		while (++count < 3) {
			c = getchar();
			if (c >= '0' && c <= '7')
				i = (i << 3) + c - '0';
			else {
				ungetc(c, stdin);
				break;
			}
		}
		return i & 0xff;
		}
	    default:
	      return c;
	}
}

void
getipv6()
{
	register struct hostent *hp;
	int dsbyte = 0, flow = 0;
	int er;

#define HEAD()	(IPV6_VERSION | \
		 IPV6_SET_CLASS(dsbyte) | IPV6_SET_FLOWLABEL(flow))

	ip = (struct ip6_hdr *)cp;
	cp += sizeof(struct ip6_hdr);
	ip->ip6_nxt = IPPROTO_ICMPV6;
	ip->ip6_hlim = 1;

	for (;;) {
		if (getoken(1) == EOF) {
			fprintf(stderr, "EOF in IPv6 header\n");
			exit(1);
		}
		if (TOKEN_IS("}"))
			break;
		else if (TOKEN_IS("dsbyte") || TOKEN_IS("class"))
			dsbyte = getint();
		else if (TOKEN_IS("flow") || TOKEN_IS("flowlabel"))
			flow = getint();
		else if (TOKEN_IS("lengh"))
			ip->ip6_plen = htons(getint());
		else if (TOKEN_IS("next-header") || TOKEN_IS("protocol")) {
			isnh = 1;
			ip->ip6_nxt = getint();
		} else if (TOKEN_IS("hop-limit") || TOKEN_IS("ttl"))
			ip->ip6_hlim = getint();
		else if (TOKEN_IS("interface") || TOKEN_IS("if")) {
			(void) getoken(0);
			indif = getindex(token);
			if (indif == 0) {
				fprintf(stderr, "bad if %s\n", token);
				exit(1);
			}
		} else if (TOKEN_IS("source")) {
			(void) getoken(0);
			hp = getipnodebyname(token, AF_INET6, 0, &er);
			if (hp == NULL) {
				fprintf(stderr, "bad address %s\n", token);
				exit(1);
			}
			bcopy(hp->h_addr, &ip->ip6_src, hp->h_length);
			freehostent(hp);
		} else if (TOKEN_IS("dest") || TOKEN_IS("destination")) {
			(void) getoken(0);
			hp = getipnodebyname(token, AF_INET6, 0, &er);
			if (hp == NULL) {
				fprintf(stderr, "bad address %s\n", token);
				exit(1);
			}
			bcopy(hp->h_addr, &ip->ip6_dst, hp->h_length);
			freehostent(hp);
		} else {
			fprintf(stderr, "unknown token %s\n", token);
			exit(1);
		}
	}
	ip->ip6_flow = HEAD();
}

void
geth2h()
{
	char *oldcp = cp;

	hhp = (struct ipv6_h2hhdr *)op;
	op += 2 * sizeof(u_int8_t);
	len = 0;

	for (;;) {
		if (getoken(1) == EOF) {
			fprintf(stderr, "EOF in hop-by-hop header\n");
			exit(1);
		}
		if (TOKEN_IS("}"))
			break;
		else if (TOKEN_IS("next-header") || TOKEN_IS("protocol"))
				hhp->ih6_nh = getint();
		else if (TOKEN_IS("length"))
			len = getint();
		else if (TOKEN_IS("pad")) {
			len = getint();
			op += len - 2;
		} else if (TOKEN_IS("hexa") || TOKEN_IS("hexadecimal") ||
			 TOKEN_IS("data"))
			gethexa(1);
		else {
			fprintf(stderr, "unknown token %s\n", token);
			exit(1);
		}
	}
	if (cp != oldcp) {
		bcopy(oldcp, op, cp - oldcp);
		op += cp - oldcp;
		if (len == 0)
			len = (cp - oldcp) + 2;
		cp = oldcp;
	}
	if (len < 8) {
		fprintf(stderr, "hop-by-hop too short (%d)\n", len);
		exit(1);
	}
	if (len & 7) {
		fprintf(stderr, "hop-by-hop bad align (%d)\n", len);
		exit(1);
	}
	hhp->ih6_hlen = (len >> 3) - 1;
}

void
getrt()
{
	register struct hostent *hp;
	int er, numa = 0;

	rtp = (struct ipv6_rthdr *)op;
	op += sizeof(struct ipv6_rthdr);
	rtp->ir6_type = IP6_LSRRT;
	len = 0;

	for (;;) {
		if (getoken(1) == EOF) {
			fprintf(stderr, "EOF in srcrt header\n");
			exit(1);
		}
		if (TOKEN_IS("}"))
			break;
		else if (TOKEN_IS("next-header") || TOKEN_IS("protocol"))
			rtp->ir6_nh = getint();
		else if (TOKEN_IS("length"))
			len = getint();
		else if (TOKEN_IS("type"))
			rtp->ir6_type = getint();
		else if (TOKEN_IS("sglt") || TOKEN_IS("segment-left"))
			rtp->ir6_sglt = getint();
		else if (TOKEN_IS("slmsk") || TOKEN_IS("mask"))
			rtp->ir6_slmsk = getint();
		else if (TOKEN_IS("hop")) {
			(void) getoken(0);
			hp = getipnodebyname(token, AF_INET6, 0, &er);
			if (hp == NULL) {
				fprintf(stderr, "bad hop %s\n", token);
				exit(1);
			}
			bcopy(hp->h_addr, op, hp->h_length);
			freehostent(hp);
			op += sizeof(struct in6_addr);
			numa++;
		} else {
			fprintf(stderr, "unknown token %s\n", token);
			exit(1);
		}
	}
	if (len == 0) {
		len = sizeof(struct ipv6_rthdr);
		len += numa * sizeof(struct in6_addr);
	}
	if (len < 8) {
		fprintf(stderr, "srcrt too short (%d)\n", len);
		exit(1);
	}
	if (len & 7) {
		fprintf(stderr, "srcrt bad align (%d)\n", len);
		exit(1);
	}
	rtp->ir6_hlen = (len >> 3) - 1;
}

void
getfrg()
{
	frp = (struct ipv6_fraghdr *)op;
	op += sizeof(struct ipv6_fraghdr);
	len = 0;

	for (;;) {
		if (getoken(1) == EOF) {
			fprintf(stderr, "EOF in fragment header\n");
			 exit(1);
		}
		if (TOKEN_IS("}"))
			break;
		else if (TOKEN_IS("next-header") || TOKEN_IS("protocol"))
			frp->if6_nh = getint();
		else if (TOKEN_IS("drop"))
			droplen = getint();
		else if (TOKEN_IS("length"))
			len = getint();
		else if (TOKEN_IS("offset"))
			frp->if6_off = getint();
		else if (TOKEN_IS("more")) {
			if (getint())
				frp->if6_off |= IP6_MF;
			else
				frp->if6_off &= ~IP6_MF;
		} else if (TOKEN_IS("id") || TOKEN_IS("ident"))
			frp->if6_id = htonl(getint());
		else {
			fprintf(stderr, "unknown token %s\n", token);
			exit(1);
		}
	}
	if (len == 0)
		len = sizeof(struct ipv6_fraghdr);
	if (len < 8) {
		fprintf(stderr, "fragment too short (%d)\n", len);
		exit(1);
	}
	if (len & 7) {
		fprintf(stderr, "fragment bad align (%d)\n", len);
		exit(1);
	}
	frp->if6_res = (len >> 3) - 1;
	frp->if6_off = htons(frp->if6_off);
}

void
getah()
{
	char *oldcp = cp;
	int klen = 0;

	ahp = (struct ipv6_authhdr *)op;
	op += sizeof(struct ipv6_authhdr);
	len = 0;

	for (;;) {
		if (getoken(1) == EOF) {
			fprintf(stderr, "EOF in AH header\n");
			exit(1);
		}
		if (TOKEN_IS("}"))
			break;
		else if (TOKEN_IS("next-header") || TOKEN_IS("protocol"))
			ahp->ah6_nh = getint();
		else if (TOKEN_IS("length"))
			len = getint();
		else if (TOKEN_IS("iclen"))
			klen = getint();
		else if (TOKEN_IS("pad"))
			ahp->ah6_pad = htons(getint());
		else if (TOKEN_IS("spi"))
			ahp->ah6_spi = htonl(getint());
		else if (TOKEN_IS("seq") || TOKEN_IS("sequence"))
			ahp->ah6_seq = htonl(getint());
		else if (TOKEN_IS("icv"))
			gethexa(1);
		else {
			fprintf(stderr, "unknown token %s\n", token);
			exit(1);
		}
	}
	if (cp != oldcp) {
		bcopy(oldcp, op, cp - oldcp);
		op += cp - oldcp;
		if (klen == 0)
			klen = cp - oldcp;
		cp = oldcp;
	}
	if (klen & 7)
		klen = (klen + 7) & ~7;
	if (len == 0)
		len = sizeof(struct ipv6_authhdr) + klen;
	if (len < 8) {
		fprintf(stderr, "AH too short (%d)\n", len);
		exit(1);
	}
	if (len & 7) {
		fprintf(stderr, "AH bad align (%d)\n", len);
		exit(1);
	}
	ahp->ah6_hlen = (len >> 2) - 1;
}

void
getdopt()
{
	char *oldcp = cp;

	dop = (struct ipv6_dopthdr *)op;
	op += 2 * sizeof(u_int8_t);
	len = 0;

	for (;;) {
		if (getoken(1) == EOF) {
			fprintf(stderr, "EOF in dest-opt header\n");
			exit(1);
		}
		if (TOKEN_IS("}"))
			break;
		else if (TOKEN_IS("next-header") || TOKEN_IS("protocol"))
			dop->io6_nh = getint();
		else if (TOKEN_IS("length"))
			len = getint();
		else if (TOKEN_IS("pad"))
			op += len - 2;
		else if (TOKEN_IS("hexa") || TOKEN_IS("hexadecimal") ||
			 TOKEN_IS("data"))
			gethexa(1);
		else {
			fprintf(stderr, "unknown token %s\n", token);
			exit(1);
		}
	}
	if (cp != oldcp) {
		bcopy(oldcp, op, cp - oldcp);
		op += cp - oldcp;
		if (len == 0)
			len = (cp - oldcp) + 2;
		cp = oldcp;
	}
	if (len < 8) {
		fprintf(stderr, "dest-opt too short (%d)\n", len);
		exit(1);
	}
	if (len & 7) {
		fprintf(stderr, "dest-opt bad align (%d)\n", len);
		exit(1);
	}
	dop->io6_hlen = (len >> 3) - 1;
}

void
getnone()
{
	if (proto < 0)
		proto = IP6_NHDR_NONH;
	gethexa(0);
}

void
geticmpv6()
{
	register struct hostent *hp;
	struct nd_router_advert *rtadv;
	struct nd_redirect *redir;
	int er;

	if (proto < 0)
		proto = IPPROTO_ICMPV6;
	icmp = (struct icmp6_hdr *)cp;
	rtadv = (struct nd_router_advert *)icmp;
	redir = (struct nd_redirect *)icmp;
	len = 0;

	for (;;) {
		if (getoken(1) == EOF) {
			fprintf(stderr, "EOF in ICMPv6 header\n");
			exit(1);
		}
		if (TOKEN_IS("}"))
			break;
		else if (TOKEN_IS("length"))
			len = getint();
		else if (TOKEN_IS("type"))
			icmp->icmp6_type = getint();
		else if (TOKEN_IS("code"))
			icmp->icmp6_code = getint();
		else if (TOKEN_IS("path-mtu"))
			icmp->icmp6_mtu = htonl(getint());
		else if (TOKEN_IS("pointer") || TOKEN_IS("parameter-pointer"))
			icmp->icmp6_pptr = htonl(getint());
		else if (TOKEN_IS("flags"))
			icmp->icmp6_data32[0] = htonl(getint());
		else if (TOKEN_IS("id") || TOKEN_IS("ident"))
			icmp->icmp6_id = htons(getint());
		else if (TOKEN_IS("seq") || TOKEN_IS("sequence"))
			icmp->icmp6_seq = htons(getint());
		else if (TOKEN_IS("mrd") || TOKEN_IS("multicast-delay"))
			icmp->icmp6_maxdelay = htons(getint());
		else if (TOKEN_IS("mhlim") || TOKEN_IS("max-hop-limit") ||
			 TOKEN_IS("hop-limit"))
			rtadv->nd_ra_curhoplimit = getint();
		else if (TOKEN_IS("aflags") || TOKEN_IS("adv-flags") ||
			 TOKEN_IS("ra-flags"))
			rtadv->nd_ra_flags_reserved = getint();
		else if (TOKEN_IS("life") || TOKEN_IS("lifetime"))
			rtadv->nd_ra_router_lifetime = htons(getint());
		else if (TOKEN_IS("reachable") || TOKEN_IS("reachable-time"))
			rtadv->nd_ra_reachable = htonl(getint());
		else if (TOKEN_IS("retrans") || TOKEN_IS("retrans-timer"))
			rtadv->nd_ra_retransmit = htonl(getint());
		else if (TOKEN_IS("target") || TOKEN_IS("group")) {
			(void) getoken(0);
			hp = getipnodebyname(token, AF_INET6, 0, &er);
			if (hp == NULL) {
				fprintf(stderr, "bad address %s\n", token);
				exit(1);
			}
			bcopy(hp->h_addr, &redir->nd_rd_target, hp->h_length);
			freehostent(hp);
		} else if (TOKEN_IS("remote")) {
			(void) getoken(0);
			hp = getipnodebyname(token, AF_INET6, 0, &er);
			if (hp == NULL) {
				fprintf(stderr, "bad address %s\n", token);
				exit(1);
			}
			bcopy(hp->h_addr, &redir->nd_rd_dst, hp->h_length);
			freehostent(hp);
		} else {
			fprintf(stderr, "unknown token %s\n", token);
			exit(1);
		}
	}
	if (len) {
		cp += len;
		return;
	}
	switch (icmp->icmp6_type) {
	    case ICMP6_MEMBERSHIP_QUERY:
	    case ICMP6_MEMBERSHIP_REPORT:
	    case ICMP6_MEMBERSHIP_REDUCTION:
		len = sizeof (struct mld6_hdr);
		break;

	    case ND_ROUTER_SOLICIT:
		len = sizeof (struct nd_router_solicit);
		break;

	    case ND_ROUTER_ADVERT:
		len = sizeof (struct nd_router_advert);
		break;

	    case ND_NEIGHBOR_SOLICIT:
		len = sizeof (struct nd_neighbor_solicit);
		break;

	    case ND_NEIGHBOR_ADVERT:
		len = sizeof (struct nd_neighbor_advert);
		break;

	    case ND_REDIRECT:
		len = sizeof (struct nd_redirect);
		break;

	    default:
		len = sizeof (struct icmp6_hdr);
	}
	cp += len;
}

void
getudp()
{
	if (proto < 0)
		proto = IPPROTO_UDP;
	udp = (struct udphdr *)cp;
	len = 0;

	for (;;) {
		if (getoken(1) == EOF) {
			fprintf(stderr, "EOF in UDP6 header\n");
			exit(1);
		}
		if (TOKEN_IS("}"))
			break;
		else if (TOKEN_IS("length"))
			len = getint();
		else if (TOKEN_IS("ulen"))
			udp->uh_ulen = htons(getint());
		else if (TOKEN_IS("sport") || TOKEN_IS("src"))
			udp->uh_sport = htons(getudpport());
		else if (TOKEN_IS("dport") || TOKEN_IS("dest"))
			udp->uh_dport = htons(getudpport());
		else {
			fprintf(stderr, "unknown token %s\n", token);
			exit(1);
		}
	}
	if (len) {
		cp += len;
		return;
	}
	len = 8;;
	cp += len;
}

void
getlladdr(type)
	int type;
{
	struct ether_addr *ep;
#if !defined(__NetBSD__)
	extern struct ether_addr *ether_aton(char *);
#endif

	lap = (struct nd_opt_hdr *)cp;
	lap->nd_opt_type = type;
	lap->nd_opt_len = 1;
	len = 0;

	for (;;) {
		if (getoken(1) == EOF) {
			fprintf(stderr, "EOF in Link-Layer Address option\n");
			exit(1);
		}
		if (TOKEN_IS("}"))
			break;
		else if (TOKEN_IS("type"))
			lap->nd_opt_type = getint();
		else if (TOKEN_IS("length"))
			lap->nd_opt_len = getint();
		else if (TOKEN_IS("real-length"))
			len = getint();
		else if (TOKEN_IS("address") || TOKEN_IS("ethernet-address")) {
			(void) getoken(0);
			ep = ether_aton(token);
			if (ep == NULL) {
				fprintf(stderr, "bad address %s\n", token);
				exit(1);
			}
			bcopy(ep, lap + 1, ETHER_ADDR_LEN);
		} else {
			fprintf(stderr, "unknown token %s\n", token);
			exit(1);
		}
	}
	if (len == 0)
		len = lap->nd_opt_len << 3;
	cp += len;
}

void
getmtu()
{
	mtup = (struct nd_opt_mtu *)cp;
	mtup->nd_opt_mtu_type = ND_OPT_MTU;
	mtup->nd_opt_mtu_len = 1;
	len = 0;

	for (;;) {
		if (getoken(1) == EOF) {
			fprintf(stderr, "EOF in MTU option\n");
			exit(1);
		}
		if (TOKEN_IS("}"))
			break;
		else if (TOKEN_IS("type"))
			mtup->nd_opt_mtu_type = getint();
		else if (TOKEN_IS("length"))
			mtup->nd_opt_mtu_len = getint();
		else if (TOKEN_IS("real-length"))
			len = getint();
		else if (TOKEN_IS("reserved"))
			mtup->nd_opt_mtu_reserved = htons(getint());
		else if (TOKEN_IS("mtu"))
			mtup->nd_opt_mtu_mtu = htonl(getint());
		else {
			fprintf(stderr, "unknown token %s\n", token);
			exit(1);
		}
	}
	if (len == 0)
		len = mtup->nd_opt_mtu_len << 3;
	cp += len;
}

void
getadvit()
{
	aitp = (struct nd_opt_advit *)cp;
	aitp->nd_opt_ait_type = ND_OPT_MOB_ADVERT_INTERVAL;
	aitp->nd_opt_ait_len = 1;
	len = 0;

	for (;;) {
		if (getoken(1) == EOF) {
			fprintf(stderr,
				"EOF in Advertisement Interval option\n");
			exit(1);
		}
		if (TOKEN_IS("}"))
			break;
		else if (TOKEN_IS("type"))
			aitp->nd_opt_ait_type = getint();
		else if (TOKEN_IS("length"))
			aitp->nd_opt_ait_len = getint();
		else if (TOKEN_IS("real-length"))
			len = getint();
		else if (TOKEN_IS("reserved"))
			aitp->nd_opt_ait_reserved = htons(getint());
		else if (TOKEN_IS("it") || TOKEN_IS("interval"))
			aitp->nd_opt_ait_it = htonl(getint());
		else {
			fprintf(stderr, "unknown token %s\n", token);
			exit(1);
		}
	}
	if (len == 0)
		len = aitp->nd_opt_ait_len << 3;
	cp += len;
}

void
getagti()
{
	haip = (struct nd_opt_agti *)cp;
	haip->nd_opt_hai_type = ND_OPT_MOB_AGENT_INFORMATION;
	haip->nd_opt_hai_len = 1;
	len = 0;

	for (;;) {
		if (getoken(1) == EOF) {
			fprintf(stderr,
				"EOF in Home-Agent Information option\n");
			exit(1);
		}
		if (TOKEN_IS("}"))
			break;
		else if (TOKEN_IS("type"))
			haip->nd_opt_hai_type = getint();
		else if (TOKEN_IS("length"))
			haip->nd_opt_hai_len = getint();
		else if (TOKEN_IS("real-length"))
			len = getint();
		else if (TOKEN_IS("reserved"))
			haip->nd_opt_hai_reserved = htons(getint());
		else if (TOKEN_IS("pref") || TOKEN_IS("preference"))
			haip->nd_opt_hai_pref = htons(getint());
		else if (TOKEN_IS("lifetime"))
			haip->nd_opt_hai_life = htons(getint());
		else {
			fprintf(stderr, "unknown token %s\n", token);
			exit(1);
		}
	}
	if (len == 0)
		len = haip->nd_opt_hai_len << 3;
	cp += len;
}

void
getprefinfo()
{
	register struct hostent *hp;
	int er;

	pfp = (struct nd_opt_prefix_info *)cp;
	pfp->nd_opt_pi_type = ND_OPT_PREFIX_INFORMATION;
	pfp->nd_opt_pi_len = 4;
	len = 0;

	for (;;) {
		if (getoken(1) == EOF) {
			fprintf(stderr, "EOF in Prefix Info option\n");
			exit(1);
		}
		if (TOKEN_IS("}"))
			break;
		else if (TOKEN_IS("type"))
			pfp->nd_opt_pi_type = getint();
		else if (TOKEN_IS("length"))
			pfp->nd_opt_pi_len = getint();
		else if (TOKEN_IS("real-length"))
			len = getint();
		else if (TOKEN_IS("prefix-length"))
			pfp->nd_opt_pi_prefix_len = getint();
		else if (TOKEN_IS("flags"))
			pfp->nd_opt_pi_flags_reserved = getint();
		else if (TOKEN_IS("invlife") ||
			 TOKEN_IS("invalidation") ||
			 TOKEN_IS("invalidation-lifetime"))
			pfp->nd_opt_pi_valid_time = htonl(getint());
		else if (TOKEN_IS("deplife") ||
			 TOKEN_IS("deprecation") ||
			 TOKEN_IS("deprecation-lifetime"))
			pfp->nd_opt_pi_preferred_time = htonl(getint());
		else if (TOKEN_IS("reserved") || TOKEN_IS("reserved2"))
			pfp->nd_opt_pi_reserved_site_plen = htonl(getint());
		else if (TOKEN_IS("prefix")) {
			(void) getoken(0);
			hp = getipnodebyname(token, AF_INET6, 0, &er);
			if (hp == NULL) {
				fprintf(stderr, "bad address %s\n", token);
				exit(1);
			}
			bcopy(hp->h_addr,
			      &pfp->nd_opt_pi_prefix,
			      hp->h_length);
			freehostent(hp);
		} else {
			fprintf(stderr, "unknown token %s\n", token);
			exit(1);
		}
	}
	if (len == 0)
		len = pfp->nd_opt_pi_len << 3;
	cp += len;
}

void
getpacket()
{
	while (getoken(2) != EOF) {
		if (TOKEN_IS("hexa") || TOKEN_IS("hexadecimal") ||
		    TOKEN_IS("data"))
			gethexa(0);
		else if (TOKEN_IS("ip") || TOKEN_IS("ipv6"))
			getipv6();
		else if (TOKEN_IS("h2h") || TOKEN_IS("hop-by-hop"))
			geth2h();
		else if (TOKEN_IS("rt") || TOKEN_IS("srcrt") ||
			 TOKEN_IS("route") || TOKEN_IS("source-route"))
			getrt();
		else if (TOKEN_IS("frg") || TOKEN_IS("fragment"))
			getfrg();
		else if (TOKEN_IS("dopt") || TOKEN_IS("dest-opt") ||
			 TOKEN_IS("destination-options"))
			getdopt();
		else if (TOKEN_IS("ah"))
			getah();
		else if (TOKEN_IS("none"))
			getnone();
		else if (TOKEN_IS("icmp") || TOKEN_IS("icmpv6"))
			geticmpv6();
		else if (TOKEN_IS("udp") || TOKEN_IS("udpv6"))
			getudp();
		else if (TOKEN_IS("source-link-layer"))
			getlladdr(ND_OPT_SOURCE_LINKADDR);
		else if (TOKEN_IS("target-link-layer"))
			getlladdr(ND_OPT_TARGET_LINKADDR);
		else if (TOKEN_IS("prefix-info"))
			getprefinfo();
		else if (TOKEN_IS("mtu"))
			getmtu();
		else if (TOKEN_IS("advertisement-interval"))
			getadvit();
		else if (TOKEN_IS("home-agent-information"))
			getagti();
		else {
			fprintf(stderr, "unknown token %s\n", token);
			exit(1);
		}
	}
}

/* Internet checksum */
int
in_cksum(addr, len)
	u_short *addr;
	int len;
{
	register int nleft = len;
	register u_short *w = addr;
	register u_short answer;
	u_short odd_byte = 0;
	register int sum = 0;

	/*
	 *  Our algorithm is simple, using a 32 bit accumulator (sum),
	 *  we add sequential 16 bit words to it, and at the end, fold
	 *  back all the carry bits from the top 16 bits into the lower
	 *  16 bits.
	 */
	while( nleft > 1 )  {
		sum += *w++;
		nleft -= 2;
	}

	/* mop up an odd byte, if necessary */
	if( nleft == 1 ) {
		*(u_char *)(&odd_byte) = *(u_char *)w;
		sum += odd_byte;
	}

	/*
	 * add back carry outs from top 16 bits to low 16 bits
	 */
	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum;				/* truncate to 16 bits */
	return (answer);
}

int
main(argc, argv)
	int argc;
	char **argv;
{
	int on, cc;
	extern char *optarg;
	extern int optind;
	int xflag = 0, nflag = 0, ch;

	while ((ch = getopt(argc, argv, "xni:")) != EOF) {
 		switch(ch) {
		case 'x':
			xflag++;
			break;
		case 'n':
			nflag++;
			break;
		case 'i':
			indif = getindex(optarg);
			if (!indif)
				goto err;
			break;
		default:
		err:
			fprintf(stderr, "Use : zing [-i if] [-x] [file]\n");
			exit(1);
		}
	}

	argc -= optind;
	argv += optind;
	if (argc > 1)
		goto err;
	if (argc == 1) {
		if (freopen(*argv, "r", stdin) == NULL) {
			fprintf(stderr, "%s: ", *argv);
			perror("cannot open : ");
			exit(1);
		}
	}

	cp = packet;
	bzero(cp, 8192);
	op = option;
	bzero(op, 2048);

	getpacket();
	len = cp - packet;
	olen = op - option;

	ip = (struct ip6_hdr *)packet;
	iop = (struct ip6ovck *)packet;
	icmp = (struct icmp6_hdr *)(ip + 1);
	udp = (struct udphdr *)(ip + 1);
	if (proto < 0)
		proto = IPPROTO_ICMPV6;
	if (isnh < 0)
		ip->ip6_nxt = proto;
	if ((proto == IPPROTO_ICMPV6) && (icmp->icmp6_cksum == 0)) {
		save_ip = *iop;
		iop->ih6_wrd0 = iop->ih6_wrd1 = 0;
		iop->ih6_pr = proto;
		iop->ih6_len = htons(len - sizeof(struct ip6_hdr));
		icmp->icmp6_cksum = in_cksum((u_short *)packet, len);
		*iop = save_ip;
	} else if ((proto == IPPROTO_UDP) && (udp->uh_sum == 0)) {
		save_ip = *iop;
		iop->ih6_wrd0 = iop->ih6_wrd1 = 0;
		iop->ih6_pr = proto;
		iop->ih6_len = htons(len - sizeof(struct ip6_hdr));
		if (udp->uh_ulen == 0)
			udp->uh_ulen = htons(cp - (char *)udp);
		udp->uh_sum = in_cksum((u_short *)packet, len);
		*iop = save_ip;
	}

	len -= sizeof(struct ip6_hdr);
	if (olen) {
		cp = (char *)(ip + 1);
		if (droplen == 0)
			bcopy(cp, cp + olen, len);
		else if (droplen > 0) {
			if (len < droplen) {
				fprintf(stderr, "drop too large\n");
				exit(1);
			}
			len -= droplen;
			bzero(cp + len, droplen);
			bcopy(cp, cp + olen, len);
		} else if (droplen < 0) {
			droplen = -droplen;
			if (len < droplen) {
				fprintf(stderr, "drop too large\n");
				exit(1);
			}
			len -= droplen;
			bzero(cp, droplen);
			bcopy(cp + droplen, cp + olen, len);
		}
		bcopy(option, cp, olen);
		len += olen;
	}

	COPY_ADDR6(ip->ip6_dst, dst.sin6_addr);
	if (ip->ip6_plen == 0)
		ip->ip6_plen = htons(len);
	len +=  sizeof(struct ip6_hdr);

	if (xflag)
		dump(packet, len);
	if (nflag)
		exit(0);

	if ((s = socket(PF_INET6, SOCK_RAW, IPPROTO_RAW)) < 0) {
		perror("socket");
		exit(1);
	}
	on = 1;
	if (setsockopt(s, IPPROTO_IPV6, IP_HDRINCL,
		       (char *)&on, sizeof(on)) < 0) {
		perror("setsockopt(IP_HDRINCL)");
		exit(1);
	}
	if (IS_MULTIADDR6(ip->ip6_dst)) {
		u_int hops;

		hops = ip->ip6_hlim;
		if (setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_HOPS,
			       (char *)&hops, sizeof(hops)) < 0) {
			perror("setsockopt(IPV6_MULTICAST_HOPS)");
			exit(1);
		}
		if (indif) {
		     if (setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_IF,
				    (char *)&indif, sizeof indif) < 0) {
				perror("setsockopt(IPV6_MULTICAST_IF)");       
				exit(1);
			}
		}
	} else if (IS_LINKLADDR6(dst.sin6_addr))
		dst.sin6_scope_id = indif;
	cc = sendto(s, packet, len, 0, (struct sockaddr *)&dst, dst.sin6_len);
	if (cc < 0) {
		perror("sendto");
		exit(1);
	} else if (cc != len) {
		fprintf(stderr, "sendto return %d != %d\n", cc, len);
		exit(1);
	}
	exit(0);
}

int
getindex(s)
char *s;
{
	static int mib[] = { CTL_NET, PF_ROUTE, 0, 0, NET_RT_IFLIST, 0 };
	register struct if_msghdr *ifm;
	struct sockaddr_dl *sdl;
	char *buf, *cp, *ap;
	int r = 0;
	unsigned bufsize;

	if (sysctl(mib, 6, NULL, &bufsize, NULL, 0) < 0) {
		perror("sysctl0");
		exit(4);
	}
	buf = (char *)malloc(bufsize);
	if (buf == NULL) {
		fprintf(stderr, "out of memory\n");
		exit(4);
	}
	if (sysctl(mib, 6, buf, &bufsize, NULL, 0) < 0) {
		perror("sysctl");
		exit(4);
	}

	for (cp = buf; cp < buf + bufsize; cp += ifm->ifm_msglen) {
		ifm = (struct if_msghdr *)cp;

		if (ifm->ifm_type == RTM_IFINFO) {
			ap = cp + sizeof(struct if_msghdr);
			sdl = (struct sockaddr_dl *)ap;
			if ((sdl->sdl_family == AF_LINK) &&
			    (sdl->sdl_nlen != 0) &&
			    (bcmp(sdl->sdl_data, s, sdl->sdl_nlen) != 0)) {
				continue;
			}
			r = ifm->ifm_index;
			break;
		}
	}
	free(buf);
	return(r);
}

void
dump(s, l)
u_char *s;
int l;
{
	int i = 0;
	while (l > 0) {
		printf("%02x", *s++);
		l--; i++;
		if (l) {
			if (i < 16)
				printf(" ");
			else {
				i = 0;
				printf("\n");
			}
		}

	}
	printf("\n");
}
